// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import BackgroundTasks
import AuthenticationServices
import Logging
import CocoaLumberjackSwift
import CocoaLumberjackSwiftLogBackend
#if DEBUG
import Pulse
#endif
#if canImport(Sentry)
import Sentry
#endif
#if DEBUG
import Atlantis
#endif

@UIApplicationMain
class AppDelegate: UIResponder, UIApplicationDelegate {

    var window: UIWindow?
	let app = App.shared

    func application(_ application: UIApplication,
                     didFinishLaunchingWithOptions launchOptions: [UIApplication.LaunchOptionsKey: Any]?) -> Bool {

        // Swift Log Setup
//        let fileLogger = DDFileLogger()
//        fileLogger.logFileManager.maximumNumberOfLogFiles = 0
//        DDLog.sharedInstance.add(fileLogger, with: .debug)

        #if DEBUG
        let osLogger = DDOSLogger()
        osLogger.logFormatter = OSLogFormatter()
        DDLog.sharedInstance.add(osLogger, with: .all)
        #endif

        // Sentry Breadcrumb logging if sentry is available
        #if canImport(Sentry)
        let breadcrumbLogger = SentryBreadcrumbLogger()
        DDLog.sharedInstance.add(breadcrumbLogger, with: .debug)
        #endif

        let logHandlerFactory = DDLogHandler.handlerFactory(
            for: .sharedInstance,
            defaultLogLevel: .debug,
            loggingSynchronousAsOf: .info,
            synchronousLoggingMetadataKey: DDLogHandler.defaultSynchronousLoggingMetadataKey
        )
        #if DEBUG
//        LoggingSystem.bootstrap({
//            MultiplexLogHandler([logHandlerFactory($0), PersistentLogHandler(label: $0)])
//        })
        LoggingSystem.bootstrap(logHandlerFactory)
        #else
        LoggingSystem.bootstrap(logHandlerFactory)
        #endif

		// Setup main window
		let window = UIWindow(frame: UIScreen.main.bounds)
		self.window = window
        // Start app coordinator on window
		app.startCoordinator(with: window)
		window.makeKeyAndVisible()

        // Start Sentry if available
        #if canImport(Sentry) && !DEBUG
        let crashOptInSetting = App.shared.settings[SettingsKeys.sentryCrashLogsOptIn] as? Bool ?? true

//        let logFileAttachment: Attachment? = {
//            if let logFileInfo = fileLogger.currentLogFileInfo {
//                return Attachment(path: logFileInfo.filePath, filename: logFileInfo.fileName)
//            }
//            return nil
//        }()

        if crashOptInSetting {
            // Create a Sentry client and start crash handler if user didnt opt out
            let bundlePlist = Bundle.main.infoDictionary
            let release = bundlePlist?["CFBundleShortVersionString"] as? String ?? "(unknown version)"
            let dist = bundlePlist?[kCFBundleVersionKey as String] as? String ?? "(unknown build)"

            SentrySDK.configureScope { (scope) in
                scope.setExtra(value: App.shared.deviceIdentifier.uuidString,
                               key: HiddenSettings.deviceIdentifier)

//                if let attachment = logFileAttachment {
//                    scope.add(attachment)
//                }
            }
            SentrySDK.start { options in
                options.dsn = "https://fd78b2f7b267467497caeb8cc37b2c56@sentry.tumexam.de/2"
                options.enableAutoSessionTracking = true
                options.integrations = Sentry.Options.defaultIntegrations().filter { (name) -> Bool in
                    // This will disable  SentryAutoBreadcrumbTrackingIntegration
                    return name != "SentryAutoBreadcrumbTrackingIntegration"
                }
                options.maxBreadcrumbs = 1024
                options.releaseName = release
                options.dist = dist
                options.add(inAppInclude: "TUMexam")
                #if DEBUG
                options.debug = true
                options.environment = "debug"
                #else
                options.debug = false
                options.environment = "release"
                options.beforeSend = { event in
                    if event.level == .fatal {
                        DispatchQueue.main.async {
                            Logger.appDelegate.info("app crashed during last session, asking user for feedback")
                            let host = SentryFeedbackViewController(eventId: event.eventId.sentryIdString)
                            host.modalPresentationStyle = .formSheet
                            UIApplication.topmostViewController()?.present(host, animated: true)
                        }
                    }
                    return event
                }
                #endif
            }
            Logger.appDelegate.info("Enabled Sentry")
        }
        #endif

        // Register Background task for locks
        let bundleId = Bundle.main.bundleIdentifier!
        BGTaskScheduler.shared.register(forTaskWithIdentifier: "\(bundleId).lock-keepalive",
                                        using: nil) { task in
            App.shared.kit?.locks.handleAppRefresh(task)
        }

        #if DEBUG
        // Enable Atlantis for Proxyman debugging if compiled for debug mode
        Atlantis.start()
        Logger.appDelegate.critical("Enabled Atlantis")
        #endif

        BreadcrumbManager.addLifecycle("application did finish launching")

        return true
    }

    func applicationWillResignActive(_ application: UIApplication) {
        BreadcrumbManager.addLifecycle("application will resign active")
        // Save settings before entering background
        SettingsStore.shared.scheduleSave()
    }

    func applicationDidEnterBackground(_ application: UIApplication) {
        BreadcrumbManager.addLifecycle("application did enter background")
        // Schedule Lock Keepalive tasks when entering background
        App.shared.kit?.locks.stopTimers()
        let bundleId = Bundle.main.bundleIdentifier!
        let request = BGAppRefreshTaskRequest(identifier: "\(bundleId).lock-keepalive")
        // Wait 20 min
        request.earliestBeginDate = Date(timeIntervalSinceNow: 20 * 60)
        do {
            try BGTaskScheduler.shared.submit(request)
            Logger.appDelegate.info("Registered BGAppRefreshTask for Keepalive Locks")
        } catch {
            Logger.appDelegate.info("Could not register BGAppRefreshTask to keepalive locks: \(error)")
        }
    }

    func applicationWillEnterForeground(_ application: UIApplication) {
        BreadcrumbManager.addLifecycle("application will enter foreground")
        // Cancel backround lock keepalive tasks when entering foreground
        BGTaskScheduler.shared.cancelAllTaskRequests()
        Logger.appDelegate.info("Cancelled all Background Refresh Tasks")
        // Re-add normal in-app lock timers
        App.shared.kit?.locks.startTimers()
    }

    func applicationDidReceiveMemoryWarning(_ application: UIApplication) {
        BreadcrumbManager.addLifecycle("application did receive memory warning")
        // Just log this, critical components like caches in the app receive their own notifications for this
        Logger.appDelegate.critical("TUMexam received Memory Warning!")
    }

    func applicationDidBecomeActive(_ application: UIApplication) {
        BreadcrumbManager.addLifecycle("application did become active")
        // Check for updates when returning from background
        App.shared.checkForUpdates(userInitiated: false)
        App.shared.coordinator?.showDeprecationAlertIfNecessary()
    }

    func applicationWillTerminate(_ application: UIApplication) {
        BreadcrumbManager.addLifecycle("application will terminate")
        // Write Settings to disk before exiting
        SettingsStore.shared.scheduleSave()
        UserDefaults.standard.synchronize()
    }
}
