// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit
import Combine
import Version
import TUMexamKit
import BackgroundTasks
#if canImport(Sentry)
import Sentry
#endif

/// This class handles general App lifecycle functions.
///
/// It keeps track of the apps state and holds the following references:
///  - The App's kit object, which comes is the Entrypoint to TUMexamKit and thus, the backend logic of the app
///  - The ``AppCoordinator`` which handles view controller transitions between main parts of the app (e.g. when transitioning from the setup to the main app or when starting a correction)
///  - The ``SettingsStore`` of the App
final class App {
	/// Singleton
	static let shared: App = App()

    /// App State enum
	enum State: Equatable {
		/// Not yet checked
		case uninitialized
		/// Nothing is setup yet
		case backendSetupRequired
        /// Backend was chosen, but the current exam still needs to be selected
		case examSetupRequired(URL)
		/// All systems go
		case nominal(URL, Int)
	}

    /// The current app state
	var state: State = .uninitialized

    /// The App settings are stored here
	let settings: SettingsStore
    /// Holds a list of previously connected instances
    let connectionHistory: ConnectionHistoryStore

    /// Drawing options used by canvas views throughout the app
	let globalDrawingOptions = GlobalDrawingOptions()

    /// Flag that indicates if we should make use of animation or not
	var shouldAnimate: Bool {
		!(UIAccessibility.isVoiceOverRunning
		|| UIAccessibility.isReduceMotionEnabled)
	}

	/// Unique Device / Install Identification for locks
	var deviceIdentifier: UUID

    /// The backend url to use when making network requests
	private(set) var backendURL: URL?

    /// The id of the currently selected exam to work with
    ///
    /// Name of this exam is displayed in the sidebar in the app
	private var selectedExamId: Int?

    /// Reference to the update check result
    ///
    /// The Update check checks against the TUMexam website if any new versions are available
    private var updateCheckCancellable: AnyCancellable?
    /// Reference to the update enforcement result
    ///
    /// The Update enforcement checks against the selected backend if it requires a new version
    private var updateEnforcementCancellable: AnyCancellable?

    /// Reference to the global kit object
    ///
    /// The kit is initialized (and reinitialized) using the ``ensureKit(_:)`` function.
    /// The app only ever has one kit object. When the backend url changes, the old kit is destroyed and a new kit is created.
    /// See the ``reset()`` function for details.
	var kit: Kit?

    /// The currently selected exam object
	@Published var currentExam: ExamType?

    /// Handles view controller lifecycle of the app, virtually only exists inside the ``App`` Singleton
	var coordinator: AppCoordinator?

    /// This should never be called directly. Instead, use ``App.shared``
	init() {
		let store = SettingsStore.shared
        let history = ConnectionHistoryStore.shared
		self.settings = store
        self.connectionHistory = history
		self.deviceIdentifier = store.deviceIdentifier

        self.settings.registerDefaults([
            SettingsKeys.sentryCrashLogsOptIn: true,
            SettingsKeys.allowModificationsToAllProblems: false,
            SettingsKeys.applePencilDoubleTapGestureEnabled: true,
            SettingsKeys.askBeforeDismissingCorrectionSession: true,
            SettingsKeys.enhancedDarkMode: false,
            SettingsKeys.sampleSolutionOpacity: 1.0,
            SettingsKeys.hideOverlays: false,
            SettingsKeys.retainPageSelection: false
        ])

        // Check previously installed version
        guard let versionStr = settings[HiddenSettings.installedVersion] as? String else {
            // Either we weren't installed before or app is so old that it doesnt have the key
            self.reset()
            setupAppState()
            return
        }

        if let dictVersion = Version(versionStr),
           let cfBundleVersionStr = Bundle.main.object(forInfoDictionaryKey: "CFBundleShortVersionString") as? String,
           let version = Version(cfBundleVersionStr),
           version != dictVersion {
            // Installed version differs from previously installed version
            // -> Purge all settings in order to remove incompabilities
            self.reset()
        }

		setupAppState()
	}

    /// Resets the App to a clean state, only preserves connection history
	func reset() {
        BreadcrumbManager.addLifecycle("App - resetting app")
		currentExam = nil
        Kit.resetKeychain()
        kit?.clearCache()
		kit = nil
		selectedExamId = nil
		backendURL = nil
        settings.reset()
        self.deviceIdentifier = settings.deviceIdentifier
		coordinator?.showWelcomeScreen(animated: true)
	}

    /// Assures that the ``state`` variable hold the correct app state based on the available information
	func setupAppState() {
		assert(state == .uninitialized)

		self.backendURL = settings.backendURL
		self.selectedExamId = settings.selectedExamId

		if let url = self.backendURL {
			if let exam = self.selectedExamId {
				state = .nominal(url, exam)
			} else {
				state = .examSetupRequired(url)
			}
		} else {
			state = .backendSetupRequired
		}
	}

    /// Persists the given backend url in the settings and tries to setup the kit object with this url
	func setBackendURL(_ url: URL) {
		self.backendURL = url
		settings.backendURL = url
        connectionHistory.add(url)

		_ = ensureKit(url: url)
	}

    /// Updates the currently selected exam
	func setCurrentExam(_ exam: ExamType) {
		settings.selectedExamId = exam.identifier
		currentExam = exam
	}

    /// Clears the authentication information in the keychain persisted by the kit object
	func resetKeychain() {
		Kit.resetKeychain()
	}

    /// Sets up the app coordinator object on the given window.
    ///
    /// This is called from the AppDelegate
	func startCoordinator(with window: UIWindow) {
		let coordinator = AppCoordinator(window: window)
		self.coordinator = coordinator

		coordinator.setupAppearance()
        #if DEBUG
        coordinator.setupDebugGestureRecognizer()
        #endif

		switch state {
		case .uninitialized:
			fatalError("trying to start coordinator without a state")
		case .backendSetupRequired:
			coordinator.showWelcomeScreen()
		case let .examSetupRequired(url):
			_ = ensureKit(url: url)
			coordinator.showExamSetup()
		case let .nominal(url, exam):
			let kit = ensureKit(url: url)
			currentExam = kit.exam(withId: exam)
			coordinator.showMain()
		}
        checkForUpdates(userInitiated: false)
        loadInstanceMetadata()
	}

    /// Tries to create the Kit object if not already present and start authentication
	func ensureKit(url: URL) -> Kit {
        #if canImport(Sentry)
        SentrySDK.configureScope { (scope) in
            scope.setTag(value: url.absoluteString, key: HiddenSettings.backendBaseURL)
        }
        #endif
		if let kit = kit {
			if kit.backendURL == url {
				return kit
			}
			// destroy old kit
			self.kit = nil
		}

		let kit = Kit(backendURL: url, useKeychain: true, uuid: deviceIdentifier)
		self.kit = kit
		kit.shibbolethAuthentication.delegate = coordinator
        checkMinimumAppVersion()
		return kit
	}

    /// Checks if the app version as required by the backend matches the currently installed app version
    func checkMinimumAppVersion() {
        guard let plistVersion = Bundle.main.object(forInfoDictionaryKey: "CFBundleShortVersionString") as? String,
              let currentVersion = Version(plistVersion),
              let updateEnforcement = kit?.updateEnforcement else {
            return
        }

        let cancellable = updateEnforcement.state.receiveOnMain().sink(receiveValue: { [weak self] state in
            guard state == .loaded else {
                return
            }
            defer {
                self?.updateEnforcementCancellable = nil
            }
            if updateEnforcement.updateRequired(currentVersion: currentVersion) {
                // Notify the user
                let alert = UIAlertController(
                    title: L10n.UpdateRequired.title,
                    message: L10n.UpdateRequired.message(updateEnforcement.requiredVersion ?? "(nil)", currentVersion),
                    preferredStyle: .alert
                )
                alert.addAction(UIAlertAction(title: L10n.cancel, style: .cancel, handler: { _ in
                    App.shared.reset()
                }))
                alert.addAction(UIAlertAction(title: L10n.install, style: .default, handler: { _ in
                    UIApplication.shared.open(URL(string: "https://tumexam.de/app.html")!,
                                              options: [:],
                                              completionHandler: nil)
                    App.shared.reset()
                }))
                alert.show(animated: true)
            }
        })

        updateEnforcement.load()
        updateEnforcementCancellable = cancellable
    }

    /// Checks the TUMexam website for a newer version of the iPad app and, if so, notifies the user.
    /// - Parameter userInitiated: If this request was made by the user manually, notify him if no newer version was found
    func checkForUpdates(userInitiated: Bool) {
        let updateInfo = kit?.updateInfo
        let cancellable = updateInfo?.version.receiveOnMain().sink(receiveValue: { [weak self] version in
            defer {
                self?.updateCheckCancellable = nil
            }
            guard let plistVersion = Bundle.main.object(forInfoDictionaryKey: "CFBundleShortVersionString") as? String,
                  let currentVersion = Version(plistVersion),
                  let updateVersion = version else {
                return
            }

            guard updateVersion > currentVersion else {
                if userInitiated {
                    let alert = UIAlertController(title: L10n.Update.Alert.NotFound.title,
                                                  message: L10n.Update.Alert.NotFound.message,
                                                  preferredStyle: .alert)
                    let okay = UIAlertAction(title: L10n.ok, style: .default, handler: nil)
                    alert.addAction(okay)
                    alert.show(animated: true)
                }

                return
            }

            let alert = UIAlertController(title: L10n.Update.Alert.Found.title,
                                          message: L10n.Update.Alert.Found.message(currentVersion.description,
                                                                                   updateVersion.description),
                                          preferredStyle: .alert)
            let cancel = UIAlertAction(title: L10n.cancel, style: .cancel, handler: nil)
            let install = UIAlertAction(title: L10n.install, style: .default) { _ in
                UIApplication.shared.open(URL(string: "https://tumexam.de/app.html")!,
                                          options: [:],
                                          completionHandler: nil)
            }
            alert.addAction(cancel)
            alert.addAction(install)
            alert.show(animated: true)
        })
        updateCheckCancellable = cancellable
        kit?.updateInfo.load()
    }

    /// Loads the metadata object (only used for the instance name atm)
    func loadInstanceMetadata() {
        kit?.metadata.load()
    }
}

extension App {
    /// Generates the App version string displayed at the top of the screen for debugging purposes
    static func versionDescription() -> String {
        let bundlePlist = Bundle.main.infoDictionary
        let version = bundlePlist?["CFBundleShortVersionString"] ?? "(unknown version)"
        let buildId = bundlePlist?[kCFBundleVersionKey as String] ?? "(unknown build)"
        #if os(iOS)
        let osVersion = "iOS \(ProcessInfo.processInfo.operatingSystemVersionString)"
        #elseif os(macOS)
        let osVersion = "macOS \(ProcessInfo.processInfo.operatingSystemVersionString)"
        #endif
        return "TUMexam \(version) (build: \(buildId); \(osVersion))"
    }
}
