// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import SwiftUI
import AuthenticationServices
import Logging
import TUMexamKit
#if DEBUG
import PulseUI
#endif

/// This class handles presenting and transitioning between view controllers inside the app
class AppCoordinator: NSObject {
    /// the apps main window
	let window: UIWindow

    /// the label that shows the version information at the top of the screen
    private let debugLabel = UILabel().autolayout()
    /// session for the shibboleth authentication
    private var webAuthenticationSession: ASWebAuthenticationSession?
    /// modal transitioning controller for custom modal vc transitions
    private var modalTransitionController: UIViewControllerAnimatedTransitioning?
    #if DEBUG
    /// Flag that indicates whether the PulseUI is currently visible
    private var isShowingPulse = false
    #endif

    /// Initializer called from the singleton App object
	init(window: UIWindow) {
		self.window = window
	}

    /// Displays the start of the apps setup tour
    /// - Parameter animated: flag that indicates whether this trnasition shoud be animated
	func showWelcomeScreen(animated: Bool = false) {
        BreadcrumbManager.addNavigation("[Coordinator] Showing welcome screen")
        let setup = UIHostingController(rootView: SetupWelcomeView(history: App.shared.connectionHistory.history))
        let navigation = UINavigationController(rootViewController: setup)
        navigation.setNavigationBarHidden(true, animated: false)
		navigation.makeBarTransparent()
		if animated {
			UIView.transition(with: window, duration: 0.5, options: .transitionFlipFromLeft, animations: {
				self.window.rootViewController = navigation
			}, completion: { _ in
                self.window.bringSubviewToFront(self.debugLabel)
            })
		} else {
			window.rootViewController = navigation
		}
	}

    /// Shows the exam selection screen.
    ///
    /// This is necessary because the exam setup is a view controller and not written in SwiftUI
	func showExamSetup() {
        BreadcrumbManager.addNavigation("[Coordinator] Showing exam setup")
		let navigation: UINavigationController
		if let root = window.rootViewController as? UINavigationController {
			navigation = root
		} else {
			showWelcomeScreen()
			// swiftlint:disable force_cast
			navigation = window.rootViewController as! UINavigationController
		}

        if #available(iOS 14.0, *) {
            let examSetup = ExamSetupViewController()
            navigation.pushViewController(examSetup, animated: true)
        } else {
            let examSetup = LegacyExamSetupViewController()
            navigation.pushViewController(examSetup, animated: true)
        }
	}

    /// Sets up UIAppearance of certain UIControls to match the TUMexam corporate design
    ///
    /// Also adds the debug label to the view hierarchy
	func setupAppearance() {
		let navigationBar = UINavigationBar.appearance()
		navigationBar.tintColor = Asset.barButtonColor.color
		navigationBar.titleTextAttributes = [.foregroundColor: Asset.barButtonColor.color]

		let tabBar = UITabBar.appearance()
		tabBar.tintColor = Asset.barButtonColor.color

        // Show always on label with version info
        debugLabel.text = App.versionDescription()
        debugLabel.font = .systemFont(ofSize: 12, weight: .light)
        debugLabel.textColor = .secondaryLabel
        debugLabel.textAlignment = .center
        debugLabel.numberOfLines = 1
        debugLabel.minimumScaleFactor = 0.6
        debugLabel.adjustsFontSizeToFitWidth = true

        DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) {
            self.window.addSubview(self.debugLabel)
            NSLayoutConstraint.activate([
                self.debugLabel.leadingAnchor.constraint(equalTo: self.window.safeAreaLayoutGuide.leadingAnchor,
                                                         constant: 130),
                self.debugLabel.trailingAnchor.constraint(equalTo: self.window.safeAreaLayoutGuide.trailingAnchor,
                                                          constant: -130),
                self.debugLabel.topAnchor.constraint(equalTo: self.window.topAnchor, constant: 2),
                self.debugLabel.addHeight(15)
            ])
        }
	}

    #if DEBUG
    /// Sets up a gesture recognizer for the PulseUI component for runtime debugging without having to attach a debugger on a mac
    func setupDebugGestureRecognizer() {
        let gesture = UITapGestureRecognizer(target: self, action: #selector(debugGesture))
        gesture.numberOfTapsRequired = 2
        gesture.numberOfTouchesRequired = 3
        window.addGestureRecognizer(gesture)
    }

    /// Handler for the debug gesture recognizer
    @objc private func debugGesture() {
        if isShowingPulse {
            window.visibleViewController?.dismiss(animated: true, completion: nil)
            isShowingPulse = false
        } else {
            let pulse = PulseUI.MainViewController()
            pulse.modalPresentationStyle = .fullScreen
            guard let safeTarget = window.visibleViewController else {
                return
            }
            safeTarget.present(pulse, animated: true) {
                self.window.bringSubviewToFront(self.debugLabel)
            }
            isShowingPulse = true
        }
    }
    #endif

    /// Transitions the App the the main menu (logged in state) of the app
	private func showMainController() {
        BreadcrumbManager.addNavigation("[Coordinator] Showing main tabbar controller")
        // UISplitViewController is only available on iOS 14 or newer
        let main = MainSplitViewController()
        window.rootViewController = main
        self.window.bringSubviewToFront(self.debugLabel)

	}

    /// Warns the user if their device is running iOS 13 or lower
    func showDeprecationAlertIfNecessary() {
        let version = (UIDevice.current.systemVersion as NSString).floatValue
        if version < 14 {
            let alert = UIAlertController(title: L10n.DeprecationWarning.title,
                                          message: L10n.DeprecationWarning.message,
                                          preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: nil))
            alert.show(animated: true)
        }
    }
}

extension AppCoordinator {
    /// Helper function to show the main controller AND prefetch the sample solution
	func showMain() {
		showMainController()
		// We assume an exam was set and the auth done.
		App.shared.currentExam?.solution.prepare()
	}

    /// Shows detailed correction status by subproblem
    /// - Parameters:
    ///   - status: correction status object containing the status info
    ///   - target: the view controller to present the modal view controller in
    ///   - source: the source view to be flipped in the trnasition (usually the table view/collection view cell the user tapped on)
    func showSubproblemStatus(_ status: CorrectionProblemStatusType, target: UIViewController? = nil, source: UIView) {
        let statusView = SubproblemStatusView(dismissHandler: { [weak self] in
            guard let safeTarget = target ?? self?.window.visibleViewController else { return }
            safeTarget.presentedViewController?.dismiss(animated: true, completion: nil)
        }, status: status)
        BreadcrumbManager.addNavigation("[Coordinator] Showing subproblem status controller for problem \(status.problem)")
        // The view is written in SwiftUI, need a hosting controller
        let hosting = UIHostingController(rootView: statusView)
        let nav = UINavigationController(rootViewController: hosting)

        nav.view.cornerRadius = 10
        nav.view.clipsToBounds = true
        nav.modalPresentationStyle = .pageSheet
        // Custom Flip transition
        let transition = FlipTransition(source)
        nav.transitioningDelegate = transition
        self.modalTransitionController = transition

        guard let safeTarget = target ?? window.visibleViewController else { return }

        safeTarget.present(nav, animated: true) {
            self.window.bringSubviewToFront(self.debugLabel)
        }
    }

    /// Start normal correction using an interactive correction controller
    /// - Parameters:
    ///   - controller: an Interactive promise based correction controller
    ///   - atIndex: the index to start the correction on (0 is default)
    ///   - target: the view controller to modally transition on (default is the current topmost viewcontroller of the window)
	func startCorrection(
		using controller: InteractiveCorrectionPromiseControllerType,
		atIndex: Int = 0,
		target: UIViewController? = nil
	) {
        BreadcrumbManager.addNavigation("[Coordinator] Starting correction with controller \(controller) at index \(atIndex)")
		guard let safeTarget = target ?? window.visibleViewController else { return }

        let dataSource = InteractivePageDataSource(correctionController: controller)

        let viewController = OuterCorrectionPageViewController(dataSource: dataSource)
        let wrapper = EditorWrapperViewController(
            editorContent: viewController,
            context: viewController.correctionContext
        )
		let navigation = UINavigationController(rootViewController: wrapper)
		navigation.modalPresentationStyle = .fullScreen

        safeTarget.present(navigation, animated: true) {
            self.window.bringSubviewToFront(self.debugLabel)
        }

		navigation.view.backgroundColor = Asset.correctionBackground.color
	}

    /// Start Correction using a static page data source
    ///
    /// Mainly used for starting correction from "Manual Search" view controller
    /// - Parameters:
    ///   - controller: the correction controller to use as a data source
    ///   - atIndex: the index to start correction on
    ///   - target: the target view controller to present modally from
    func startCorrection(
        using controller: CorrectionControllerType,
        atIndex: Int = 0,
        target: UIViewController? = nil
    ) {
        BreadcrumbManager.addNavigation("[Coordinator] Starting correction with controller \(controller) at index \(atIndex)")
        guard let safeTarget = target ?? window.visibleViewController else { return }

        let dataSource = StaticPageDataSource(correctionController: controller, startIndex: atIndex)

        let viewController = OuterCorrectionPageViewController(dataSource: dataSource)
        let wrapper = EditorWrapperViewController(
            editorContent: viewController,
            context: viewController.correctionContext
        )
        let navigation = UINavigationController(rootViewController: wrapper)
        navigation.modalPresentationStyle = .fullScreen

        safeTarget.present(navigation, animated: true) {
            self.window.bringSubviewToFront(self.debugLabel)
        }

        navigation.view.backgroundColor = Asset.correctionBackground.color
    }

    /// Starts subproblem correction using an existing page data source, a target cutout rect and a page number of the exam. Used when using a StaticCorrectionDataSource, aka when coming from Manual Search
    ///
    /// This is called from the OuterPageViewController when selecting a subproblem to correct
    /// - Parameters:
    ///   - dataSource: an existing outer page data source in use by the current correction stack
    ///   - rect: a rect on the page to cutout for correction
    ///   - page: the page number to start subproblem correction on
    ///   - target: the target view controller to present the subproblem correction from (default = topmost navigation controller)
    func startSubproblemCorrection(
        using dataSource: OuterCorrectionDataSource,
        rect: RectOnPage,
        page: Int,
        startIndex: Int = 0,
        target: UIViewController? = nil) {
        BreadcrumbManager.addNavigation("[Coordinator] Starting subproblem correction with rect \(rect) on page \(page)")

        guard let safeTarget = target ?? window.visibleViewController else { return }
        guard let navigation = safeTarget.navigationController else { return }

        let subproblemDataSource = dataSource.createSubproblemDataSource(rect: rect, page: page)
        let viewController = SubproblemPageViewController(dataSource: subproblemDataSource, startIndex: startIndex)
        let wrapper = EditorWrapperViewController(
            editorContent: viewController,
            context: viewController.correctionContext
        )
        navigation.pushViewController(wrapper, animated: true)
    }

    /// Starts subproblem correction using an existing outer correction data source for a given problem(s) and pass combination
    ///
    /// The key difference to the other function above is, that here, a new correction controller is created to fetch uncorrected subproblems using `problems` and `pass`
    /// - Parameters:
    ///   - dataSource: an existing out page data source in use by the current correction stack
    ///   - rect: a rect on the page to cutout for correction
    ///   - page: the page number to start subproblem correction on
    ///   - problems: an array of problem identifiers
    ///   - pass: the pass to correct
    ///   - target: the target view controller to present the subproblem correction from (default = topmost navigation controller)
    func startSubproblemCorrection(
        using dataSource: OuterCorrectionDataSource,
        rect: RectOnPage,
        page: Int,
        problems: [ProblemIdentifier],
        pass: Int,
        target: UIViewController? = nil) {
        BreadcrumbManager.addNavigation("[Coordinator] Starting subproblem correction with rect \(rect) on page \(page)")

        guard let safeTarget = target ?? window.visibleViewController else { return }
        guard let navigation = safeTarget.navigationController else { return }

        assert(problems.count > 0)
        let problem = problems.first!.problem
        guard problems.allSatisfy({ $0.problem == problem }) else {
            fatalError("ProblemIdentifiers given to InteractiveLoadingController must be from the same problem")
        }
        var controller: InteractiveCorrectionPromiseControllerType!

        controller = App.shared.currentExam?.interactiveCorrectionPromiseController(
            forProblem: problem,
            subproblems: problems.compactMap({ $0.subproblem }),
            pass: pass,
            filterLocked: false
        )

        guard let controllerUnwrapped = controller else { return }

        let subproblemDataSource = SubproblemInteractiveDataSource(rect: rect,
                                                                   page: page,
                                                                   correctionController: controllerUnwrapped)
        let viewController = SubproblemPageViewController(dataSource: subproblemDataSource)
        let wrapper = EditorWrapperViewController(
            editorContent: viewController,
            context: viewController.correctionContext
        )
        navigation.pushViewController(wrapper, animated: true)
    }

    /// Pushes a single correction stack without any data source attached to the topmost navigation controller of target
    ///
    /// Used when double tapping a subproblem during subproblem correction
    /// - Parameters:
    ///   - correction: a correction instance to display
    ///   - target: the target view controller to present from (default = topmost view controller)
    ///   - pageIndex: the page to display
    func showCorrection(_ correction: CorrectionType, target: UIViewController? = nil, pageIndex: Int? = nil) {
        BreadcrumbManager.addNavigation("[Coordinator] Showing correction\(correction)")

		guard let safeTarget = target ?? window.visibleViewController else { return }
		guard let navigation = safeTarget.navigationController else { return }

		let viewController = SingleCorrectionViewController(correction: correction)
        if let index = pageIndex {
            viewController.context.currentPageIndex = index
        }
        let editor = EditorWrapperViewController(
            editorContent: viewController,
            context: viewController.context
        )
		navigation.pushViewController(editor, animated: true)
	}

    /// Shows a single correction stack for the given correction object modally
    /// - Parameters:
    ///   - correction: a correction object
    ///   - target: the view controller to modally push onto
    func showCorrectionModally(_ correction: CorrectionType, target: UIViewController? = nil) {
        BreadcrumbManager.addNavigation("[Coordinator] Showing correction\(correction)")

        let correctionController = SingleCorrectionViewController(correction: correction)
        correctionController.isModal = true
        let viewController = EditorWrapperViewController(
            editorContent: correctionController,
            context: correctionController.context
        )
        guard let safeTarget = target ?? window.visibleViewController else { return }
        let navigation = UINavigationController(rootViewController: viewController)
        navigation.modalPresentationStyle = .fullScreen

        safeTarget.present(navigation, animated: true) {
            self.window.bringSubviewToFront(self.debugLabel)
        }
    }

    /// Pushes the sample solution editor onto the current navigation stack
    ///
    /// Used when tapping "Edit Solution" during correction
    /// - Parameters:
    ///   - page:The initial page to display
    ///   - target: the target vc to present from
    ///   - context: existing correction context, if any
	func startSampleSolutionEditing(atPage page: PageType,
									target: UIViewController? = nil,
									context: CorrectionContext? = nil) {
        BreadcrumbManager.addNavigation("[Coordinator] Start Sample Solution editing for page \(page)")

		guard let safeTarget = target ?? window.visibleViewController,
			  let navigation = safeTarget.navigationController,
			  let context = context else { return }
		let tempFirstVC = CorrectionViewController(page: page, context: context)
		let vc = SolutionPageViewController(vc: tempFirstVC,
											context: context)
        let wrapper = EditorWrapperViewController(
            editorContent: vc,
            context: vc.correctionContext
        )
		navigation.pushViewController(wrapper, animated: true)
	}

    /// Starts diffing for solution editor using a rejection model from the server
    /// - Parameters:
    ///   - pages: the pages which have conflicts
    ///   - target: the targer vc to present the diffing controller from (default=topmost view controller)
	func startDiffOfPages(_ pages: [SolutionRejectionModel],
						  target: UIViewController? = nil) {
        BreadcrumbManager.addNavigation("[Coordinator] Starting diffing for pages \(pages)")

		guard let safeTarget = target ?? window.visibleViewController else { return }
		let container = DiffingContainerViewController(pages: pages)
		container.modalPresentationStyle = .fullScreen
		safeTarget.present(container, animated: true)
	}

    /// Shows the settings view controller, either modally or in a popover from a bar item if supplied
    /// - Parameters:
    ///   - controller: the view controller to present from
    ///   - barButtonItem: the bar button item for popover presentation (if any)
	func showSettings(using controller: UIViewController, barButtonItem: UIBarButtonItem? = nil) {
        BreadcrumbManager.addNavigation("[Coordinator] showing settings popover")

		let viewController = SettingsViewController()
        let nav = UINavigationController(rootViewController: viewController)

        if let item = barButtonItem {
            nav.modalPresentationStyle = .popover
            guard let popover = nav.popoverPresentationController else {
                fatalError("couldn't open settings as popover!")
            }
            popover.barButtonItem = item
        } else {
            nav.modalPresentationStyle = .formSheet
        }

		controller.present(nav, animated: true, completion: nil)
	}
}

extension AppCoordinator: ShibbolethAuthenticationDelegate {
    /// Called by the ASWebAuthenticationSession once we have been redirected to a `tumexam://` URL
    private func webAuthenticationSessionCallback(
        auth: ShibbolethAuthentication,
        originalURL: URL,
        _ callbackURL: URL?,
        _ error: Error?
    ) {
        guard error == nil, let callbackURL = callbackURL else {
            Logger.shibboleth.error("Got Error from auth session: \(error?.localizedDescription ?? "nil")")

            if let aserr = error as? ASWebAuthenticationSessionError,
               aserr.errorCode == ASWebAuthenticationSessionError.canceledLogin.rawValue {
                // Notify auth object that auth was canceled by user
                self.webAuthenticationSession = nil
                auth.cancel(url: originalURL, byUser: true)
                return
            }

            auth.failure(url: originalURL, reason: .unknown)
            return
        }

        // Parse authorization info from callback
        // We expect something like this:
        // tumexam://callback/success?token=bla&user=blubb&exp=dategoeshere
        guard callbackURL.lastPathComponent == "success" else {
            auth.failure(url: originalURL, reason: .unauthorized)
            return
        }

        let components = URLComponents(url: callbackURL, resolvingAgainstBaseURL: true)
        guard let queryItems = components?.queryItems, queryItems.count == 3 else {
            auth.failure(url: originalURL, reason: .unknown)
            return
        }

        // Transform query items to dictionary for named access
        let queryDict = queryItems.reduce(into: [String: String]()) {
            $0[$1.name] = $1.value
        }

        guard let userId = queryDict["user"],
              let token = queryDict["token"],
              let tokenExpirationDate = queryDict["exp"]
        else {
            // This should never happen, would be a backend issue
            auth.failure(url: originalURL, reason: .unknown)
            return
        }

        // Auth was successful
        _ = auth.authorizeUsing(userId: userId, token: token, tokenExpiration: tokenExpirationDate)
    }

    func shibboleth(_ auth: ShibbolethAuthentication, requiresAuthenticationUsingURL url: URL, refreshExisting: Bool) {
        guard self.webAuthenticationSession == nil else {
            Logger.shibboleth.info("tried to present shibboleth auth, but an authentication session is already running \(webAuthenticationSession!.debugDescription)")
            return
        }
        // Create URL for Authentication
        let components = URLComponents(url: url, resolvingAgainstBaseURL: true)
        let host = components?.host
        guard let instanceName = host?.split(separator: ".").first else { return } // e.g. 2021ss-in-grnvs

        var authURLComponents = URLComponents()
        authURLComponents.scheme = "https"
        authURLComponents.host = "hq.tumexam.de"
        authURLComponents.path = "/ipadlogin/\(instanceName)"
        guard let authURL = authURLComponents.url else { return } // e.g. https://hq.tumexam.de/ipadlogin/2021ss-in-grnvs

        // This create the authentication session, waiting for a redirect to a url with scheme tumexam://...
        let scheme = "tumexam"
        let authSession = ASWebAuthenticationSession(
            url: authURL,
            callbackURLScheme: scheme
        ) { [unowned self] (callbackURLOpt, error) in
            self.webAuthenticationSessionCallback(auth: auth, originalURL: url, callbackURLOpt, error)
        }

        // We don't want to share cookies with Safari
        // If we don't set this flag, we wont be able to sign in as a different user until our Shibboleth cookie expires
        authSession.prefersEphemeralWebBrowserSession = true

        // Update session property
        self.webAuthenticationSession = authSession

        // Check if we are trying to refresh an existing session, for UI/UX purposes
        if refreshExisting {
            // Show an alert thate tells the user the session expired before starting the auth session
            let alert = UIAlertController(title: L10n.Shibboleth.Refresh.Alert.title,
                                          message: L10n.Shibboleth.Refresh.Alert.message,
                                          preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: L10n.Shibboleth.Refresh.Alert.disconnect,
                                          style: .destructive,
                                          handler: { _ in
                                            // Abort and disconnect
                                            self.webAuthenticationSession = nil
                                            App.shared.reset()
            }))
            alert.addAction(UIAlertAction(title: L10n.ok,
                                          style: .default,
                                          handler: { _ in
                                            // Start Authentication
                                            // swiftlint:disable:next line_length
                                            BreadcrumbManager.addNavigation("[Coordinator] Showing ASWebAuthenticationSession to the user")
                                            authSession.presentationContextProvider = self
                                            authSession.start()
            }))
            alert.show(animated: true)
        } else {
            // Show an alert to the user that an initial authentication has to happen before starting auth
            let alert = UIAlertController(title: L10n.Shibboleth.Initial.Alert.title,
                                          message: L10n.Shibboleth.Initial.Alert.message,
                                          preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: L10n.ok,
                                          style: .default,
                                          handler: { _ in
                                            // Start Authentication
                                            // swiftlint:disable:next line_length
                                            BreadcrumbManager.addNavigation("[Coordinator] Showing ASWebAuthenticationSession to the user")
                                            authSession.presentationContextProvider = self
                                            authSession.start()
            }))
            alert.show(animated: true)
        }
	}

	func shibbolethDidAuthenticate(_ auth: ShibbolethAuthentication) {
        self.webAuthenticationSession = nil
		App.shared.currentExam?.solution.prepare()
	}

    func shibbolethAuthenticationFailed(_ auth: ShibbolethAuthentication,
                                        reason: ShibbolethAuthenticationFailureReason) {
        self.webAuthenticationSession = nil
        switch reason {
        case .unauthorized, .unknown:
            let alert = UIAlertController(title: L10n.Shibboleth.Failed.Alert.title,
                                          message: L10n.Shibboleth.Failed.Alert.message,
                                          preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: { _ in
                // This checks if we are in the initial setup process of the app
                // Might break in future iOS versions due to SwiftUI/UIHostingController changes
                if let navcon = self.window.rootViewController as? UINavigationController,
                   (navcon.viewControllers.first as? UIHostingController<SetupWelcomeView>) != nil {
                    // We are in the initial exam setup
                    // Pop the exam setup screen
                    navcon.popViewController(animated: true)
                }
            }))
            // This delay is necessary, otherwise it happens while the web authentication session is still displayed
            // The alert would not be visible then
            DispatchQueue.main.asyncAfter(deadline: DispatchTime.now().advanced(by: .milliseconds(750))) {
                alert.show(animated: true)
            }
        case .canceled:
            // This checks if we are in the initial setup process of the app
            // Might break in future iOS versions due to SwiftUI/UIHostingController changes
            if let navcon = self.window.rootViewController as? UINavigationController,
               (navcon.viewControllers.first as? UIHostingController<SetupWelcomeView>) != nil {
                // We are in the initial exam setup
                // Pop the exam setup screen
                navcon.popViewController(animated: true)
            }
        case .maintenance:
            // Instance is under maintenance, tell that to the user and log them out
            let alert = UIAlertController(title: L10n.Shibboleth.Maintenance.Alert.title,
                                          message: L10n.Shibboleth.Maintenance.Alert.message,
                                          preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: L10n.Shibboleth.Refresh.Alert.disconnect,
                                          style: .destructive,
                                          handler: { _ in
                // Logout / Reset app to setup
                App.shared.reset()
            }))
            DispatchQueue.main.asyncAfter(deadline: DispatchTime.now().advanced(by: .milliseconds(750))) {
                alert.show(animated: true)
            }
        }
    }
}

extension AppCoordinator: ASWebAuthenticationPresentationContextProviding {
    func presentationAnchor(for session: ASWebAuthenticationSession) -> ASPresentationAnchor {
        return self.window
    }
}
