// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import TUMexamKit

/// A CorrectionContext object holds information about current correction canvas settings, such as the current drawing tool or if the solution is currently shown
/// It is used to keep state between different pages of a correction (= different CorrectionViewControllers) and to make that state observable using Combine's `@Published`
class CorrectionContext {
    /// Current Mode, either ``.normal`` or ``.subproblemSelection``
	enum Mode {
		case normal
		case subproblemSelection
	}

    /// Index of the currently displayed page
	@Published var currentPageIndex: Int?
    /// Flag indicating whether the currently displayed page is flagged as empty using the "Is Empty" toolbar button
	@Published var currentPageIsEmpty: Bool = false
    /// Flag indicating whether or not the solution is being displayed
	@Published var showSolution: Bool = false
    /// Flag indicating whether or not thepersonalized solution is being displayed
    @Published var showsPersonalizedSolution: Bool = false
    /// Flag indicating whether or not the raw scan is being displayed
    @Published var showsRaw: Bool = false
    /// Flag indicating whether or not correction overlays such as the flag buttons and the cross buttons are being displayed
    @Published var showOverlay: Bool = !(App.shared.settings[SettingsKeys.hideOverlays] as? Bool ?? false)
    /// The current correction ``Mode``
	@Published private(set) var mode: Mode = .normal

    /// The current undo managers used by the active correction view controller
	@Published var currentUndoManager: ContextUndoManager?
    /// Flag indicating whether a shared undo manager is used
    var useSharedUndoManager: Bool = true

    /// The currently used drawing tool
	@Published var drawingTool: Tool = .pen

    /// Flag indicating whether the lasso tool is enabled for the currently displayed correction view controller
	@Published var enableLassoTool: Bool = false

    /// Function to change the current correction mode
    /// - Parameter mode: the new correction mode
	func setMode(_ mode: Mode) {
		guard mode != self.mode else { return }
		self.mode = mode
	}
}
