// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import UIKit
import TUMexamKit
import Logging

extension CorrectionCanvasViewController: CorrectionViewSelectionElementDelegate {
	func selection(_ element: CorrectionViewSelectionElement, atPath normalizedPath: CGPath) {
		guard let pass = page.focusedPass,
			  let layer = page.layer(pass: pass),
			  let drawing = layer.drawing.value else {
			return
		}

		// try getting image
		guard let cgImage = drawing.getImage() else { return }
		let image = UIImage(cgImage: cgImage)

		// try cutout
		guard let cutout = image.cutout(normalizedPath: normalizedPath) else {
			return
		}
		guard let (cropped, hasAlpha) = cutout.cropUsingAlphaChannel(extend: 5) else {
			return
		}
		guard hasAlpha else {
            Logger.correctionUI.error("[CanvasVC|\(page.erid ?? "nil")|\(page.pageNumber)] selection has no content")
			return
		}
		sharedCanvasContext?.startSelectionInsertion(cropped)
	}
}

extension CorrectionCanvasViewController: LassoSelectionDelegate {
	func showSelectionInsertion(_ element: CorrectionViewSelectionInsertionElement,
								at point: CGPoint) {

		element.becomeFirstResponder()
		let menuController = UIMenuController.shared
		let pasteItem = UIMenuItem(title: L10n.paste, action: #selector(pasteSelection))
		menuController.menuItems = [pasteItem]
		let menuSize = CGSize(width: 100, height: 50)
		menuController.showMenu(from: view,
								rect: CGRect(center: point, size: menuSize))
	}

	func showSelectionPlacement(_ element: CorrectionViewSelectionPlacementElement,
								didSelectRect rect: CGRect,
								forSelection selection: Image) {
		let element = SelectionElement(selection: selection, normalizedRect: rect)
		self.drawing?.canvasView?.drawing.append(element: .selection(element))

		sharedCanvasContext?.endSelectionPlacement()
	}

	@objc
	func pasteSelection() {
		guard let selection = selectionInsertionElement?.selection else {
			preconditionFailure("No element to insert where one should be.")
		}
		let elem = CorrectionViewSelectionPlacementElement(context: canvasContext)
		elem.selection = selection
		selectionPlacementElement = addElement(elem)
		selectionPlacementElement?.placementDelegate = self
		selectionInsertionElement?.removeFromSuperview()
		selectionInsertionElement = nil
	}

	func enableSelectionInsertion(_ isEnabled: Bool) {
		if isEnabled, let sharedCtx = sharedCanvasContext {
			let insertion = CorrectionViewSelectionInsertionElement(context: canvasContext)
			insertion.selection = sharedCtx.selection
			addElement(insertion)
			insertion.delegate = self
			self.selectionInsertionElement = insertion
		} else {
			self.selectionPlacementElement?.removeFromSuperview()
			self.selectionInsertionElement?.removeFromSuperview()
			self.selectionPlacementElement = nil
			self.selectionInsertionElement = nil
		}
	}
}
