// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import UIKit
import TUMexamKit
import Logging

class CorrectionCanvasViewController: UIViewController {

    let correctionContext: CorrectionContext?
    let canvasContext: CorrectionCanvasContext
    let page: PageType

    var state: PageLoadingState = .idle

    var lockReference: LockReferenceType? {
        didSet {
            lockReferenceDidChange()
        }
    }
    var lockObserver: AnyCancellable?

    private var elements: [CorrectionViewElement] = []

    private var scan: CorrectionViewScanElement?
    private var raw: CorrectionViewRawElement?
    // TODO: Refactor this to be private again
    var boxes: CorrectionViewBoxesElement?
    private var emptyBox: CorrectionViewEmptyBoxElement?
    private var solution: CorrectionViewSolutionElement?
    private var personalizedSolution: CorrectionViewPersonalizedSolutionElement?

    private var correctionLayers: [CorrectionViewLayerElement] = []
    var drawing: CorrectionViewDrawingElement?
    private var subproblemSelection: CorrectionViewSubproblemSelectionElement?
    private var snippetCreation: CorrectionViewSnippetCreationElement?
    private var snippetPlacement: CorrectionViewSnippetPlacementElement?
    private var empty: CorrectionViewEmptyElement?
    private var shadow: CorrectionViewShadowElement?
    private var lockState: CorrectionViewLockStateElement?
    private var lassoSelection: CorrectionViewSelectionElement?

    var selectionInsertionElement: CorrectionViewSelectionInsertionElement?
    var selectionPlacementElement: CorrectionViewSelectionPlacementElement?

    var overlayObserver: AnyCancellable?
    var solutionObserver: AnyCancellable?
    var personalizedSolutionObserver: AnyCancellable?
    var rawObserver: AnyCancellable?
    var modeObserver: AnyCancellable?
    var canvasModeObserver: AnyCancellable?
    var toolObserver: AnyCancellable?
    var selectionToolObserver: AnyCancellable?
    var emptyBoxObserver: AnyCancellable?
    var emptyBoxSelectedObserver: AnyCancellable?

    var snippetToast: Toast?
    var sharedCanvasContextObserver: AnyCancellable?

    /// keeps track whether this view was already visible at some point (used in order to check for changes when returning to a visible state)
    var isFirstAppearance: Bool = true

    var resetUndoManagerWhenAppearing: Bool = false

    var showsRaw: Bool = false {
        didSet {
            raw?.isHidden = !showsRaw
        }
    }

    var sharedCanvasContext: CanvasContextBridge? {
        didSet {
            if sharedCanvasContext != nil {
                observeSharedCanvasContext()
            } else {
                sharedCanvasContextObserver = nil
            }
        }
    }

    init(page: PageType,
         normalizedPageRect: CGRect? = nil,
         correctionContext: CorrectionContext? = nil,
         targetUndoManager: ContextUndoManager? = nil,
         lockReference: LockReferenceType? = nil) {
        self.correctionContext = correctionContext
        self.page = page
        self.lockReference = lockReference
        canvasContext = CorrectionCanvasContext(page: page)

        // either use the given one, or create a new one
        if let undoManager = targetUndoManager {
            canvasContext.undoManager = undoManager
        } else {
            canvasContext.undoManager = ContextUndoManager()
        }

        if let rect = normalizedPageRect {
            canvasContext.normalizedPageRect = rect
        }

        super.init(nibName: nil, bundle: nil)
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        Logger.correctionUI.debug("[CanvasVC|\(page.containingCorrection?.identifier.stringIdentifier ?? "nil")|\(page.pageNumber)|\(pointerString(self))] viewDidLoad")

        view.backgroundColor = .white
        view.clipsToBounds = false

        if let correctionContext = correctionContext {
            self.setupObservers(correctionContext)
        }

        lockReferenceDidChange()

        setupDrop()
    }

    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        isFirstAppearance = false
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)

        if resetUndoManagerWhenAppearing {
            // reset undo manager
            canvasContext.undoManager?.reset()
            resetUndoManagerWhenAppearing = false
        }

        // lock the correction
        if lockReference == nil {
            Logger.correctionUI.debug("[CanvasVC|\(page.containingCorrection?.identifier.stringIdentifier ?? "nil")|\(page.pageNumber)|\(pointerString(self))] added lock req")
            lockReference = page.containingCorrection?.addLockRequirement()

        }
    }

    override func didMove(toParent parent: UIViewController?) {
        super.didMove(toParent: parent)
        updateUndoManager()
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        updateUndoManager()
    }

    func lockReferenceDidChange() {
        guard isViewLoaded else { return }
        if let ref = lockReference {
            lockObserver = ref.state.receiveOnMain().sink { [weak self] (state) in
                Logger.correctionUI.debug("[CanvasVC|\(self?.page.containingCorrection?.identifier.stringIdentifier ?? "nil")|\(self?.page.pageNumber ?? -1)|\(pointerString(self))] lock state changed to \(state)")
                self?.updateEditableState(state.editable)
            }
        } else {
            lockObserver = nil
        }
    }

    func updateEditableState(_ editable: Bool) {
        boxes?.isUserInteractionEnabled = editable
        drawing?.isUserInteractionEnabled = editable
        lockState?.isHidden = editable
    }

    func retryLock() {
        let newLock = page.containingCorrection?.addLockRequirement()
        lockReference = newLock
    }

    func updateUndoManager() {
        // update undo manager (if enabled)
        if let correctionContext = correctionContext {
            if correctionContext.useSharedUndoManager == true,
               let parent = self.parent(conformingTo: ContextUndoManagerProvider.self) {
                canvasContext.undoManager = parent.contextUndoManager
            } else if let parent = self.parent(conformingTo: ContextUndoManagerCoordinator.self) {
                if !parent.shouldChangeUndoManager(to: self) {
                    Logger.undoRedo.debug("[CanvasVC|\(page.containingCorrection?.identifier.stringIdentifier ?? "nil")|\(page.pageNumber)] not switching undo context due to parent")
                    return
                }
            } else {
                return
            }

            Logger.undoRedo.debug("[CanvasVC|\(page.containingCorrection?.identifier.stringIdentifier ?? "nil")|\(page.pageNumber)] switching to undo context \(pointerString(canvasContext.undoManager))")
            correctionContext.currentUndoManager = canvasContext.undoManager
        }
    }

    override func viewDidDisappear(_ animated: Bool) {
        super.viewDidDisappear(animated)

        // remove lock
        lockReference = nil
        Logger.undoRedo.debug("[CanvasVC|\(page.containingCorrection?.identifier.stringIdentifier ?? "nil")|\(page.pageNumber)|\(pointerString(self))] lock ref removed bc viewDidDisappear")
    }

    func vacateLock() {
        lockReference = nil
    }

    // MARK: - Observer Setup

    func setupObservers(_ context: CorrectionContext) {
        overlayObserver = context.$showOverlay.receiveOnMain().sink { [weak self] (show) in
            self?.boxes?.overlayMode = show
        }

        solutionObserver = context.$showSolution.receiveOnMain().sink { [weak self] (show) in
            if show {
                self?.solution?.loadDrawings()
            }
            self?.solution?.isHidden = !show
        }

        personalizedSolutionObserver = context.$showsPersonalizedSolution.receiveOnMain().sink { [weak self] show in
            self?.personalizedSolution?.isHidden = !show
        }

        rawObserver = context.$showsRaw.receiveOnMain().sink(receiveValue: { [weak self] (show) in
            if show {
                self?.raw?.loadableImage.load()
            }
            self?.raw?.isHidden = !show
        })

        modeObserver = context.$mode.receiveOnMain().sink { [weak self] mode in
            self?.updateMode(mode)
        }

        toolObserver = context.$drawingTool.receiveOnMain().sink { [weak self] tool in
            self?.drawing?.tool = tool
        }

        canvasModeObserver = canvasContext.$mode.receiveOnMain().sink { [weak self] mode in
            self?.showSnippetCreationElement(mode == .snippetCreation)
            self?.showSnippetPlacementElement(mode == .snippetPlacement)
            self?.enableSelectionInsertion(mode == .selectionInsertion)
        }

        selectionToolObserver = context.$enableLassoTool.receiveOnMain().sink { [weak self] isEnabled in
            self?.enableLassoTool(isEnabled)
        }

        emptyBoxObserver = page.emptyBox.receiveOnMain().sink { [weak self] emptyBox in
            guard let box = emptyBox else {
                return
            }
            self?.emptyBoxSelectedObserver = box.selected.receiveOnMain()
                .sink { [weak self] _ in
                    self?.updateEmptyElement()
            }
            self?.emptyBoxObserver = nil
        }
    }

    private func observeSharedCanvasContext() {
        guard let sharedContext = sharedCanvasContext else { return }
        sharedCanvasContextObserver = sharedContext.$mode
            .receiveOnMain()
            .sink { [weak self] mode in
                self?.canvasContext.setMode(mode)
        }
    }

    // MARK: - Toolbar Element Setup

    func addDefaultElements() {
        /* Order (back to front)
         ... Compact Elements
         - (Subproblem Selection)
         - Solution
         - Personalized Solution
         - Raw Scan
         */
        addCompactElements()

        subproblemSelection = addElement(CorrectionViewSubproblemSelectionElement(context: canvasContext))
        subproblemSelection?.selectionDelegate = self
        subproblemSelection?.isHidden = true

        // only add solution if a solution is available
        if page.solution != nil {
            solution = addElement(CorrectionViewSolutionElement(context: canvasContext))
            var alpha = App.shared.settings[SettingsKeys.sampleSolutionOpacity] as? Double ?? 1.0
            if alpha < 1.0 {
                alpha = 0.8 * alpha + 0.1
            }
            solution?.alpha = CGFloat(alpha)
            solution?.isUserInteractionEnabled = !(alpha < 1.0)
        }

        if page.personalizedSolution != nil {
            personalizedSolution = addElement(CorrectionViewPersonalizedSolutionElement(context: canvasContext))
            var alpha = App.shared.settings[SettingsKeys.sampleSolutionOpacity] as? Double ?? 1.0
            if alpha < 1.0 {
                alpha = 0.8 * alpha + 0.1
            }
            personalizedSolution?.alpha = CGFloat(alpha)
            personalizedSolution?.isUserInteractionEnabled = !(alpha < 1.0)
        }

        raw = addElement(CorrectionViewRawElement(context: canvasContext))
        raw?.isHidden = !showsRaw
    }

    func addCompactElements() {
        /* Order (back to front)
         - Shadow Element
         - Scan Image
         - Layer (pass 0)
         - ...
         - Layer (pass n)
         - Drawing Canvas
         - Boxes
         - Empty Element
         - Empty Box Element (if empty box exists)
         - Lock State
         */

        shadow = addElement(CorrectionViewShadowElement(context: canvasContext))

        scan = addElement(CorrectionViewScanElement(context: canvasContext))

        for layer in page.layers.value {
            if let focusedPass = page.focusedPass, layer.pass >= focusedPass {
                continue
            }
            correctionLayers.append(addElement(CorrectionViewLayerElement(layer: layer, in: canvasContext)))
        }

        drawing = addElement(CorrectionViewDrawingElement(context: canvasContext))
        if let tool = correctionContext?.drawingTool {
            drawing?.tool = tool
        }

        boxes = addElement(CorrectionViewBoxesElement(context: canvasContext))
        boxes?.boxesDelegate = self

        empty = addElement(CorrectionViewEmptyElement(context: canvasContext))
        empty?.isHidden = true

        if page.emptyBox.value != nil {
            self.emptyBox = addElement(CorrectionViewEmptyBoxElement(context: canvasContext))
        }

        lockState = addElement(CorrectionViewLockStateElement(context: canvasContext))
        lockState?.retryHandler = { [weak self] in
            self?.retryLock()
        }
        lockState?.isHidden = true
    }

    func addERIDElement() {
        addElement(CorrectionViewERIDElement(context: canvasContext))
    }

    func updateEmptyElement() {
        if page.empty {
//            correctionContext?.currentPageIsEmpty = true
            empty?.activatedByCrossbox = false
            empty?.isHidden = false
            return
        }
        if let box = page.emptyBox.value, box.selected.value {
            empty?.activatedByCrossbox = true
            empty?.isHidden = false
            correctionContext?.currentPageIsEmpty = false
            return
        }
        empty?.activatedByCrossbox = false
        empty?.isHidden = true
    }

    @discardableResult
    func addElement<V: CorrectionViewElement>(_ element: V) -> V {
        view.addSubview(element)
        element.pinToSuperview()
        elements.append(element)
        return element
    }

    var canUndo: Bool {
        canvasContext.undoManager?.canUndo ?? false
    }

    var canRedo: Bool {
        canvasContext.undoManager?.canRedo ?? false
    }

    func undo() {
        canvasContext.undoManager?.undo()
    }

    func redo() {
        canvasContext.undoManager?.redo()
    }

    func updateMode(_ mode: CorrectionContext.Mode) {
        self.showSubproblemSelection(mode == .subproblemSelection)
    }

    func showSubproblemSelection(_ mode: Bool) {
        subproblemSelection?.isHidden = !mode
        if mode == true {
            subproblemSelection?.startSegmentation()
        }
    }

    // MARK: - Lasso

    func enableLassoTool(_ isEnabled: Bool) {
        if lassoSelection == nil, isEnabled {
            lassoSelection = addElement(CorrectionViewSelectionElement(context: canvasContext))
            lassoSelection?.lassoSelectionDelegate = self
        } else {
            lassoSelection?.isHidden = !isEnabled
        }
    }

    // MARK: - Snippets

    func showSnippetCreationElement(_ visible: Bool) {
        if visible {
            if snippetCreation == nil {
                snippetCreation = addElement(CorrectionViewSnippetCreationElement(context: canvasContext))
                snippetCreation?.snippetDelegate = self
            }
            if snippetToast == nil {
                let toast = self.snippetToast ?? createSnippetToast()
                self.snippetToast = toast
            }

            if let target = parent(conformingTo: ToastPresentingViewController.self),
               let toast = snippetToast {
                target.show(toast)
            }
        } else {
            if let toast = snippetToast {
                toast.dismiss()
                snippetToast = nil
            }
        }
        snippetCreation?.isHidden = !visible
    }

    func showSnippetPlacementElement(_ visible: Bool) {
        if visible && snippetPlacement == nil {
            snippetPlacement = addElement(CorrectionViewSnippetPlacementElement(context: canvasContext))
            snippetPlacement?.snippetDelegate = self
        }
        snippetPlacement?.isHidden = !visible
        snippetPlacement?.snippet = canvasContext.snippetToPlace
    }

    func showSnippets(forProblem problem: ProblemType, source: UIView, focus: SnippetType? = nil) {
        let viewController = SnippetsTableViewController(problem: problem, focus: focus)
        viewController.delegate = self
        let navigation = UINavigationController(rootViewController: viewController)
        navigation.modalPresentationStyle = .popover
        guard let popover = navigation.popoverPresentationController else {
            fatalError("couldn't open snippets as popover!")
        }
        popover.sourceView = source
        popover.sourceRect = source.bounds

        self.present(navigation, animated: true, completion: nil)
    }

    private func createSnippetToast() -> Toast {
        Toast(
            .info,
            title: L10n.Toast.SnippetMode.title,
            subtitle: L10n.Toast.SnippetMode.subtitle,
            action: .init(title: L10n.Toast.SnippetMode.action, handler: { [weak self] _ in
                self?.canvasContext.setMode(.normal)
            }),
            dismissable: false
        )
    }

    func showFlags(forProblem problem: ProblemType, source: UIView) {
        guard let erid = self.canvasContext.page.erid else {
            return
        }

        let viewController = FlagSelectionCollectionViewController(
            problem: problem,
            erid: erid,
            correctionPass: self.page.containingCorrection?.focusedPass
        )
        // viewController.delegate = self
        let navigation = UINavigationController(rootViewController: viewController)
        navigation.modalPresentationStyle = .popover
        guard let popover = navigation.popoverPresentationController else {
            fatalError("couldn't open flags as popover!")
        }
        popover.sourceView = source
        popover.sourceRect = source.bounds

        self.present(navigation, animated: true, completion: nil)
    }
}

// MARK: - Subproblem Selection Delegate

extension CorrectionCanvasViewController: CorrectionViewSubproblemSelectionElementDelegate {
    func subproblemSelection(_ element: CorrectionViewSubproblemSelectionElement,
                             shouldStartUsing rect: RectOnPage,
                             page: Int,
                             problems: [ProblemIdentifier]?) {
        if let parent = self.parent?.parent(conformingTo: CorrectionViewSubproblemSelectionElementDelegate.self) {
            parent.subproblemSelection(element, shouldStartUsing: rect, page: page, problems: problems)

            resetUndoManagerWhenAppearing = true
        } else {
            Logger.correctionUI.error("[CanvasVC|\(self.page.containingCorrection?.identifier.stringIdentifier ?? "nil")|\(self.page.pageNumber)] parent can't handle subproblem selection: \(self.parent?.debugDescription ?? "no description available")")
        }
    }
}

extension CorrectionCanvasViewController: CorrectionViewBoxesElementDelegate {
    func boxes(
        _ element: CorrectionViewBoxesElement,
        didTapActionButton button: UIButton,
        forProblem problem: ProblemType) {
        showSnippets(forProblem: problem, source: button)
    }

    func boxes(_ element: CorrectionViewBoxesElement,
               didTapFlagButton button: UIButton,
               forProblem problem: ProblemType) {
        showFlags(forProblem: problem, source: button)
    }
}
