// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import UIKit
import TUMexamKit
import Logging

final class CorrectionViewController: UIViewController, IndexRepresentingViewController {
	let page: PageType
	let context: CorrectionContext?

	lazy var scrollView: UIScrollView = CenteringScrollView().autolayout()

	var representingIndexInCollection: Int?

	var didPerformSetup: Bool = false

	var preferredCanvasSize: CGSize = .zero

	let loadingViewController: LoadingWrapperViewController<PageType, CorrectionCanvasViewController>
    private var widthConstraint: NSLayoutConstraint?
    private var heightConstraint: NSLayoutConstraint?

    // MARK: UIViewController Lifecycle

	init(page: PageType,
		 context: CorrectionContext? = nil,
		 sharedContext: CanvasContextBridge? = nil) {

		self.page = page
		self.context = context
		self.loadingViewController = LoadingWrapperViewController(
			PageLoadingResource(page: page),
			block: { page -> CorrectionCanvasViewController in
				let canvas = CorrectionCanvasViewController(page: page, correctionContext: context)
				canvas.addDefaultElements()
				canvas.sharedCanvasContext = sharedContext
				return canvas
		})

		super.init(nibName: nil, bundle: nil)
        Logger.correctionUI.debug("[CorrectionVC] - Page[\(page.pageIdentifier.stringIdentifier)] init")
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

    deinit {
        vacateLock()
    }

    override func viewDidLoad() {
        super.viewDidLoad()

		edgesForExtendedLayout = []

		scrollView.delegate = self

		// Add Container View, constraint to max size for screen
		loadingViewController.view.translatesAutoresizingMaskIntoConstraints = false

//        scrollView.maximumZoomScale = 1.0 / self.traitCollection.displayScale
        scrollView.maximumZoomScale = 2.0

        scrollView.panGestureRecognizer.allowedTouchTypes = [UITouch.TouchType.direct.rawValue as NSNumber]
        scrollView.pinchGestureRecognizer?.allowedTouchTypes = [UITouch.TouchType.direct.rawValue as NSNumber]

		addChild(loadingViewController)
		scrollView.addSubview(loadingViewController.view)
		loadingViewController.didMove(toParent: self)

		// setup the view
		// view = UIView()
		view.addSubview(scrollView)
		scrollView.pinToSuperview()
        loadingViewController.view.pinToSuperview()
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        // swiftlint:disable:next line_length
        BreadcrumbManager.addNavigation("[CorrectionVC] viewDidAppear - \(page.erid ?? "nil") page \(page.pageNumber) pass \(page.focusedPass ?? -1)")
    }

    override func viewDidDisappear(_ animated: Bool) {
        super.viewDidDisappear(animated)
        // swiftlint:disable:next line_length
        BreadcrumbManager.addNavigation("[CorrectionVC] - viewDidDisappear - \(page.erid ?? "nil") page \(page.pageNumber) pass \(page.focusedPass ?? -1)")
    }

	func preferredContentSize(for containerSize: CGSize) -> CGSize {
//        if let size = page.scan.image?.size {
//            return size
//        }
		let pageSize: CGSize = page.pageSize

		let pageRatio: CGFloat = pageSize.aspectRatio
		let containerRatio = containerSize.aspectRatio

		var factor: CGFloat = 0.0
		if pageRatio >= containerRatio {
			factor = containerSize.width / pageSize.width
		} else {
			factor = containerSize.height / pageSize.height
		}

		let size = CGSize(width: floor(pageSize.width * factor), height: floor(pageSize.height * factor))

        // swiftlint:disable:next line_length
        Logger.correctionUI.debug("[CorrectionVC] - Page[\(page.pageNumber)] preferred size for \(containerSize.debugDescription): \(size.debugDescription)")
		return size
	}

    override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()

        preferredCanvasSize = preferredContentSize(for: self.view.bounds.size)
        let width = widthConstraint ?? {
            let const = loadingViewController.view.widthAnchor.constraint(equalToConstant: preferredCanvasSize.width)
            const.isActive = true
            return const
        }()
        let height = heightConstraint ?? {
            let const = loadingViewController.view.heightAnchor.constraint(equalToConstant: preferredCanvasSize.height)
            const.isActive = true
            return const
        }()
        width.constant = preferredCanvasSize.width
        height.constant = preferredCanvasSize.height
        self.widthConstraint = width
        self.heightConstraint = height
        scrollView.setNeedsLayout()

        updateScrollViewMinScale()
        self.scrollView.setZoomScale(self.scrollView.minimumZoomScale, animated: false)
    }

	func updateScrollViewMinScale() {
		guard preferredCanvasSize != .zero, !scrollView.bounds.isEmpty else {
//			fatalError()
			return
		}

		let canvasSize = preferredCanvasSize

		let scaleX = scrollView.bounds.size.width / canvasSize.width
		let scaleY = scrollView.bounds.size.height / canvasSize.height
		let scale = min(scaleX, scaleY)

		scrollView.minimumZoomScale = scale
		if scrollView.zoomScale < scale {
			scrollView.setZoomScale(scale, animated: true)
		}
    }

    func vacateLock() {
        guard let content = loadingViewController.contentViewController else { return }
        content.vacateLock()
    }

	var canUndo: Bool {
		loadingViewController.contentViewController?.canUndo ?? false
	}

	var canRedo: Bool {
		loadingViewController.contentViewController?.canRedo ?? false
	}

	func undo() {
		loadingViewController.contentViewController?.undo()
	}

	func redo() {
		loadingViewController.contentViewController?.redo()
	}

	var empty: Bool {
		get {
			page.empty
		}
		set {
			page.empty = newValue
			if let content = loadingViewController.contentViewController {
				content.updateEmptyElement()
			} else {
				// back to false if not yet loaded
				context?.currentPageIsEmpty = false
			}
		}
	}
}
