// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import UIKit
import TUMexamKit
import os.log

class CorrectionViewController: UIViewController, IndexRepresentingViewController {
	let page: PageType
	let context: CorrectionContext?

	lazy var scrollView: UIScrollView = CenteringScrollView().autolayout()

	var representingIndexInCollection: Int?

	var didPerformSetup: Bool = false

	var preferredCanvasSize: CGSize = .zero

	let loadingWrapper: CorrectionPageLoadingWrapperDataSource
	lazy var loadingWrapperViewController = LoadingWrapperViewController(dataSource: self.loadingWrapper)

	// Views and Settings

//	lazy var canvasViewController: CorrectionCanvasViewController
//		= CorrectionCanvasViewController(page: self.page, correctionContext: self.context)

    var pencilDrawGestureRecognizer: DrawGestureRecognizer!

    // MARK: UIViewController Lifecycle

	init(page: PageType, context: CorrectionContext? = nil) {
		self.page = page
		self.context = context
		self.loadingWrapper = CorrectionPageLoadingWrapperDataSource(page: page, context: context)

		super.init(nibName: nil, bundle: nil)
		os_log(.debug, log: .correctionUI, "Page[%@] init", page.pageNumber as NSNumber)
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

    override func viewDidLoad() {
        super.viewDidLoad()

		edgesForExtendedLayout = []

		scrollView.delegate = self

		// Add Container View, constraint to max size for screen
		let size = preferredContentSize(for: UIScreen.main.bounds.size)
		preferredCanvasSize = size
		loadingWrapperViewController.view.translatesAutoresizingMaskIntoConstraints = false
		NSLayoutConstraint.activate([
			loadingWrapperViewController.view.widthAnchor.constraint(equalToConstant: size.width),
			loadingWrapperViewController.view.heightAnchor.constraint(equalToConstant: size.height)
		])

        scrollView.maximumZoomScale = 2.0

        scrollView.panGestureRecognizer.allowedTouchTypes = [UITouch.TouchType.direct.rawValue as NSNumber]
        scrollView.pinchGestureRecognizer?.allowedTouchTypes = [UITouch.TouchType.direct.rawValue as NSNumber]

		addChild(loadingWrapperViewController)
		scrollView.addSubview(loadingWrapperViewController.view)
		loadingWrapperViewController.didMove(toParent: self)

		// setup the view
		view = UIView()
		view.addSubview(scrollView)
		scrollView.pinToSuperview()

		os_log(.debug, log: .correctionUI, "Page[%@] viewDidLoad", page.pageNumber as NSNumber)
    }

	func preferredContentSize(for containerSize: CGSize) -> CGSize {
		let pageSize: CGSize = page.pageSize

		let pageRatio: CGFloat = pageSize.aspectRatio
		let containerRatio = containerSize.aspectRatio

		var factor: CGFloat = 0.0
		if pageRatio >= containerRatio {
			factor = containerSize.width / pageSize.width
		} else {
			factor = containerSize.height / pageSize.height
		}

		let size =  CGSize(width: pageSize.width * factor, height: pageSize.height * factor)
		return size
	}

	override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()

        updateScrollViewMinScale()
        self.scrollView.setZoomScale(self.scrollView.minimumZoomScale, animated: false)
    }

	func updateScrollViewMinScale() {
		guard preferredCanvasSize != .zero, !scrollView.bounds.isEmpty else {
			return
		}

		let canvasSize = preferredCanvasSize

		let scaleX = scrollView.bounds.size.width / canvasSize.width
		let scaleY = scrollView.bounds.size.height / canvasSize.height
		let scale = min(scaleX, scaleY)

		scrollView.minimumZoomScale = scale
		if scrollView.zoomScale < scale {
			scrollView.setZoomScale(scale, animated: true)
		}
    }

	func prepareForSave() {
		loadingWrapper.contentViewController?.prepareForSave()
	}

	var empty: Bool {
		get {
			page.empty
		}
		set {
			page.empty = newValue
			if let content = loadingWrapper.contentViewController {
				content.showEmptyElement(page.empty)
			} else {
				// back to false if not yet loaded
				context?.currentPageIsEmpty = false
			}
		}
	}
}

extension CorrectionViewController: CorrectionViewSubproblemSelectionElementDelegate {
	func subproblemSelection(_ element: CorrectionViewSubproblemSelectionElement, shouldStartUsing rect: RectOnPage) {
		if let parent = parent as? CorrectionViewSubproblemSelectionElementDelegate {
			parent.subproblemSelection(element, shouldStartUsing: rect)
		} else {
			os_log(.error, log: .correctionUI, "Parent can't handle subproblem selection %@", self)
		}
	}
}
