// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import os.log

class CorrectionViewProblemPassBoxesView: BaseView {
	let problemPass: ProblemPassType
	var buttons: [CorrectionBoxButton] = []
	var overlayMode: Bool = false {
		didSet {
			guard oldValue != overlayMode else { return }
			for button in buttons {
				button.overlayMode = overlayMode
			}
		}
	}
    let overrideModeActive: Bool

    init(problemPass: ProblemPassType, overrideModeActive: Bool = false) {
		self.problemPass = problemPass
        self.overrideModeActive = overrideModeActive

		super.init()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override func setup() {
		super.setup()

		for box in problemPass.boxes {
			let button = addButton(for: box)
			buttons.append(button)
		}
	}

	override func layoutSubviews() {
		super.layoutSubviews()

		for button in buttons {
			button.frame = button.normalizedFrame.mappedTo(size: bounds.size)
		}
	}

	func addButton(for box: CorrectionBoxType) -> CorrectionBoxButton {
		// create button using a default rect
		let emptyFrame = CGRect(origin: .zero, size: CGSize(width: 10.0, height: 10.0))
		let button = CorrectionBoxButton(frame: emptyFrame, box: box, overrideModeActive: overrideModeActive)

		button.normalizedFrame = box.normalizedFrame.normalizedIn(rect: problemPass.normalizedFrame)

		// Change button cross color if pass != 1 (red is default)
		if problemPass.pass == 2 {
			button.crossColor = Asset.correctionPass2.color
		} else if problemPass.pass == 3 {
			button.crossColor = Asset.correctionPass3.color
		}

		// Add CorrectionBoxButton to view and keep a reference
		self.addSubview(button)
		return button
	}
}

class CorrectionViewProblemBoxesView: BaseView {
	let problem: ProblemType
	var passViews: [CorrectionViewProblemPassBoxesView] = []
    var actionStackView: UIStackView?
	var actionButton: UIButton?
    var flagButton: UIButton?
	var hasActionButton: Bool { return actionButton != nil }
	var onAction: ((UIButton, ProblemType) -> Void)?
    var onFlag: ((UIButton, ProblemType) -> Void)?
    let activePass: Int

	var overlayMode: Bool = false {
		didSet {
			guard oldValue != overlayMode else { return }
			for pass in passViews {
				pass.overlayMode = overlayMode
			}
            actionStackView?.isHidden = !overlayMode
		}
	}

    init(problem: ProblemType, activePass: Int) {
		self.problem = problem
        self.activePass = activePass

		super.init()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override func setup() {
		super.setup()
        self.clipsToBounds = false
		for pass in problem.passes {
            // swiftlint:disable:next line_length
            let view = CorrectionViewProblemPassBoxesView(problemPass: pass, overrideModeActive: activePass == pass.pass)
			addSubview(view)
			passViews.append(view)
		}

        let actionStack = UIStackView()
        self.actionStackView = actionStack
        actionStack.alignment = .center
        actionStack.distribution = .fillProportionally
        actionStack.axis = .vertical

        addSubview(actionStack)

        // Flag Button
        let flagButton = UIButton(type: .custom).autolayout()
        self.flagButton = flagButton
        flagButton.setImage(UIImage(systemName: "flag"), for: .normal)
        flagButton.tintColor = Asset.tuMexamBlue.color
        flagButton.addTarget(self, action: #selector(didTapFlag(_:)), for: .touchUpInside)
        flagButton.imageView?.contentMode = .scaleAspectFit

        actionStack.addArrangedSubview(flagButton)

		// problem can be edited, and problem is credits
		// -> provide action button
		if problem.editable, problem.evaluationType == .credits {
			let button = UIButton(type: .custom)
			self.actionButton = button

			button.setImage(UIImage(systemName: "pencil.and.ellipsis.rectangle"), for: .normal)
			button.tintColor = Asset.tuMexamBlue.color
			button.addTarget(self, action: #selector(didTapActionButton(_:)), for: .touchUpInside)
            button.imageView?.contentMode = .scaleAspectFit

            actionStack.addArrangedSubview(button)
		}
	}

	override func layoutSubviews() {
		super.layoutSubviews()
        guard let superWidth = self.superview?.bounds.size.width else {
            return
        }

        if let actionStack = actionStackView {
            var stackFrame = bounds
            let isLeft = self.frame.minX < superWidth / 2
            let minWidth = 0.05 * superWidth
            stackFrame.size.width = max(bounds.size.width, minWidth)
            stackFrame.origin.x =  isLeft ?
                -stackFrame.size.width : bounds.size.width
            stackFrame.origin.y = 0
            stackFrame.size.height = bounds.size.height
            actionStack.frame = stackFrame
        }

		for pass in passViews {
			pass.frame = pass.problemPass.normalizedFrame
				.normalizedIn(rect: self.problem.normalizedFrame)
                .mappedTo(size: bounds.size)
		}
	}

	@objc func didTapActionButton(_ sender: UIButton) {
		onAction?(sender, problem)
	}

    @objc func didTapFlag(_ sender: UIButton) {
        onFlag?(sender, problem)
    }

    override func point(inside point: CGPoint, with event: UIEvent?) -> Bool {
        guard let stackView = actionStackView else {
            return super.point(inside: point, with: event)
        }
        if stackView.bounds.contains(convert(point, to: stackView)) {
            return true
        }
        return super.point(inside: point, with: event)
    }
}

protocol CorrectionViewBoxesElementDelegate: AnyObject {
	func boxes(
		_ element: CorrectionViewBoxesElement,
		didTapActionButton button: UIButton,
		forProblem problem: ProblemType
	)

    func boxes(
        _ element: CorrectionViewBoxesElement,
        didTapFlagButton button: UIButton,
        forProblem problem: ProblemType
    )
}

class CorrectionViewBoxesElement: CorrectionViewElement {
	var boxGroups: [CorrectionViewProblemBoxesView] = []
	var overlayMode: Bool = false {
		didSet {
			guard oldValue != overlayMode else { return }
			for box in boxGroups {
				box.overlayMode = overlayMode
			}
		}
	}

	weak var boxesDelegate: CorrectionViewBoxesElementDelegate?

	override func setup() {
		super.setup()

		for problem in context.page.problems.value {
            let boxGroupView = CorrectionViewProblemBoxesView(problem: problem, activePass: context.page.containingCorrection?.focusedPass ?? 1)
			boxGroupView.onAction = { [weak self] button, problem in
				guard let self = self else { return }
				self.boxesDelegate?.boxes(self, didTapActionButton: button, forProblem: problem)
			}
            boxGroupView.onFlag = { [weak self] button, problem in
                guard let self = self else { return }
                self.boxesDelegate?.boxes(self, didTapFlagButton: button, forProblem: problem)
            }

			boxGroups.append(boxGroupView)
			addSubview(boxGroupView)
		}
	}

	override func layoutSubviews() {
		super.layoutSubviews()

		for boxGroup in boxGroups {
			let boxFrame = boxGroup.problem.normalizedFrame
				.normalizedIn(rect: context.normalizedPageRect)
				.mappedTo(size: self.bounds.size)
			boxGroup.frame = boxFrame
			boxGroup.isHidden = !boxGroup.frame.intersects(bounds)
		}
	}

	override func point(inside point: CGPoint, with event: UIEvent?) -> Bool {
		return boxGroups.contains { $0.point(inside: convert(point, to: $0), with: event) }
	}

	func boxGroup(forProblem problem: ProblemIdentifier) -> CorrectionViewProblemBoxesView? {
		boxGroups.first(problem, for: \.problem.identifier)
	}
}
