// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import Combine
import Vision
import TUMexamKit

class CorrectionViewSubproblemIndicatorView: BaseView {
	var normalizedRect: CGRect = .zero
	let label: UILabel = UILabel()
    let problem: ProblemIdentifier

    init(_ problem: ProblemIdentifier) {
        self.problem = problem
        super.init()
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

	override func setup() {
		super.setup()
		self.backgroundColor = UIColor.systemBlue.withAlphaComponent(0.35)
		self.layer.cornerRadius = 10.0

		label.font = UIFont.preferredFont(forTextStyle: .title1)
		label.textAlignment = .center
		addSubview(label)
	}

	override func layoutSubviews() {
		super.layoutSubviews()

		label.frame = self.bounds
	}
}

// Taken from the apple documentation
extension CGImagePropertyOrientation {
	init(_ uiOrientation: UIImage.Orientation) {
		switch uiOrientation {
		case .up: self = .up
		case .upMirrored: self = .upMirrored
		case .down: self = .down
		case .downMirrored: self = .downMirrored
		case .left: self = .left
		case .leftMirrored: self = .leftMirrored
		case .right: self = .right
		case .rightMirrored: self = .rightMirrored
		default: self = .up
		}
    }
}

protocol CorrectionViewSubproblemSelectionElementDelegate: AnyObject {
    func subproblemSelection(_ element: CorrectionViewSubproblemSelectionElement,
                             shouldStartUsing rect: RectOnPage,
                             page: Int,
                             problems: [ProblemIdentifier]?)
}

class CorrectionViewSubproblemSelectionElement: CorrectionViewElement {
	var indicatorViews: [CorrectionViewSubproblemIndicatorView] = []
	let rectangleSelectionView = RectangleSelectionView()

	var imageRequestHandler: VNImageRequestHandler?
	weak var segmenter: PageSegmentationType?

	var tapGestureRecognizer: UITapGestureRecognizer?

	private var segmentationSubscriber: AnyCancellable?

	weak var selectionDelegate: CorrectionViewSubproblemSelectionElementDelegate?

	override func setup() {
		super.setup()

		let tap = UITapGestureRecognizer(target: self, action: #selector(didTap(_:)))
		addGestureRecognizer(tap)

		rectangleSelectionView.onTapInside = { [weak self] in
			self?.rectangleSelectionDidFinish()
		}
		rectangleSelectionView.onDidBegin = { [weak self] in
			self?.setIndicatorViews(hidden: true)
		}
		rectangleSelectionView.onDidEnd = { [weak self] in
			self?.setIndicatorViews(hidden: false)
		}
		addSubview(rectangleSelectionView)
		rectangleSelectionView.pinToSuperview()
	}

	@objc func didTap(_ recognizer: UITapGestureRecognizer) {
		guard rectangleSelectionView.rectangle == nil else { return }

		let location = recognizer.location(in: self)
		for indicator in indicatorViews {
			if indicator.frame.contains(location) {
				let normalizedRect = indicator.frame.normalizedUsing(size: self.bounds.size)
				let pageSize = self.context.page.pageSize
				let rectOnPage = RectOnPage(normalizedRect: normalizedRect, pageSize: pageSize)

                let page = context.page.pageNumber
                selectionDelegate?.subproblemSelection(
                    self,
                    shouldStartUsing: rectOnPage,
                    page: page,
                    problems: [indicator.problem]
                )
				return
			}
		}
	}

	func rectangleSelectionDidFinish() {
		guard let rect = rectangleSelectionView.rectangle?.rect else { return }
		let normalizedRect = rect.normalizedUsing(size: rectangleSelectionView.bounds.size)
		let pageSize = self.context.page.pageSize
		let rectOnPage = RectOnPage(normalizedRect: normalizedRect, pageSize: pageSize)
        let pageNumber = context.page.pageNumber

        let intersectingProblems = context.page.problems.value.filter {
            $0.normalizedFrame.intersects(normalizedRect)
        }.sorted {
            $0.identifier.problem <= $1.identifier.problem && $0.identifier.subproblem ?? "" < $1.identifier.subproblem ?? ""
        }

        let initial = intersectingProblems.first?.identifier.problem
        if intersectingProblems.allSatisfy({ $0.identifier.problem == initial }) {
            selectionDelegate?.subproblemSelection(
                self,
                shouldStartUsing: rectOnPage,
                page: pageNumber,
                problems: intersectingProblems.count > 0 ? intersectingProblems.map({ $0.identifier }) : nil
            )
        } else {
            guard let first = intersectingProblems.first else {
                selectionDelegate?.subproblemSelection(
                    self,
                    shouldStartUsing: rectOnPage,
                    page: pageNumber,
                    problems: nil
                )
                return
            }
            selectionDelegate?.subproblemSelection(
                self,
                shouldStartUsing: rectOnPage,
                page: pageNumber,
                problems: [first.identifier]
            )
        }
	}

	func setIndicatorViews(hidden: Bool) {
		UIView.animate(withDuration: 0.25) {
			for indicator in self.indicatorViews {
				indicator.alpha = (hidden ? 0.0 : 1.0)
			}
		}

	}

	func startSegmentation() {
        self.segmenter = context.page.segmentation
		guard let segmenter = self.segmenter else { return }
		guard segmenter.state.value != .loaded else { return }

		segmenter.load()
		segmentationSubscriber = segmenter.segments.receiveOnMain().sink { [weak self] segments in
			guard segments.count > 0 else { return }
			self?.segmentationFinished(segments)
		}
	}

	func segmentationFinished(_ segments: [Segment]) {
		for seg in segments {
			switch seg {
			case .textAnswer(let textAnswer):
				self.addTextAnswerIndicator(textAnswer)
			default:
				break
			}
		}
	}

	func addTextAnswerIndicator(_ textAnswer: TextAnswerSegment) {
        let newView = CorrectionViewSubproblemIndicatorView(textAnswer.problem)

		let range = textAnswer.range
		newView.label.text = textAnswer.problem.shortName
		var rect = CGRect()
		rect.origin.x = 0.0
		rect.size.width = 1.0
		rect.origin.y = range.lowerBound
		rect.size.height = range.upperBound - range.lowerBound

		newView.normalizedRect = rect

		addSubview(newView)
		indicatorViews.append(newView)
	}

	override func layoutSubviews() {
		super.layoutSubviews()

		for indicator in indicatorViews {
			indicator.frame = indicator.normalizedRect.mappedTo(size: self.bounds.size)
		}
	}

}
