// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

protocol CorrectionViewSelectionElementDelegate: AnyObject {
	func selection(_ element: CorrectionViewSelectionElement, atPath normalizedPath: CGPath)
}

final class CorrectionViewSelectionElement: CorrectionViewElement {
	var strokeGestureRecognizer: SimpleStrokeGestureRecognizer?
	let lassoLayer: CAShapeLayer = CAShapeLayer()
	let confirmButton: UIButton = UIButton(type: .custom)
	var currentPath: CGPath?

	weak var lassoSelectionDelegate: CorrectionViewSelectionElementDelegate?

	override func setup() {
		super.setup()

		let stroke = SimpleStrokeGestureRecognizer(target: self, action: #selector(didStroke(_:)))
		stroke.delegate = self
		self.strokeGestureRecognizer = stroke
		self.addGestureRecognizer(stroke)

		lassoLayer.fillColor = UIColor.systemBlue.withAlphaComponent(0.5).cgColor

		confirmButton.setTitle(L10n.CorrectionController.Canvas.SelectionConfirm.button, for: .normal)
		confirmButton.titleLabel?.font = .preferredFont(forTextStyle: .headline)

		confirmButton.backgroundColor = .systemBlue

		confirmButton.setTitleColor(.white, for: .normal)
		confirmButton.contentEdgeInsets = UIEdgeInsets(top: 10, left: 20, bottom: 10, right: 20)
		confirmButton.sizeToFit()
		confirmButton.layer.cornerRadius = confirmButton.bounds.size.height * 0.5

		confirmButton.addTarget(self, action: #selector(confirmSelection(_:)), for: .touchUpInside)

		self.layer.addSublayer(lassoLayer)
		addSubview(confirmButton)
		confirmButton.isHidden = true
	}

	override func layoutSubviews() {
		super.layoutSubviews()
		lassoLayer.frame = layer.bounds
	}

	func deselect() {
		lassoLayer.path = nil
		confirmButton.isHidden = true
	}

	@objc
	func didStroke(_ recognizer: SimpleStrokeGestureRecognizer) {
		let path = convertToPath(recognizer.samples)
		lassoLayer.path = path
		currentPath = path

		switch recognizer.state {
		case .began:
			confirmButton.isHidden = true
		case .ended:
			// check if selection is at least 40x40
			let pathFrame = path.boundingBoxOfPath.integral
			if pathFrame.size.height < 40.0 || pathFrame.size.width < 40.0 {
				deselect()
				return
			}

			// show button over the selection
			confirmButton.isHidden = false
			var buttonFrame = confirmButton.frame
			buttonFrame.origin.y = max(10.0, pathFrame.origin.y - buttonFrame.size.height - 10.0)
			buttonFrame.origin.x = (pathFrame.origin.x + (pathFrame.size.width * 0.5)) - (buttonFrame.size.width * 0.5)
			confirmButton.frame = buttonFrame
		default:
			break
		}
	}

	@objc func confirmSelection(_ sender: UIButton) {
		guard let path = currentPath?.normalizedCopy(using: bounds.size) else { return }
		lassoSelectionDelegate?.selection(self, atPath: path)
		deselect()
	}

	func convertToPath(_ samples: [SimpleStrokeGestureRecognizer.StrokeSample]) -> CGPath {
		let path = CGMutablePath()
		guard samples.count > 1 else { return path }
		path.move(to: samples[0].location)

		for sample in samples[1..<samples.count] {
			path.addLine(to: sample.location)
		}

		path.closeSubpath()

		return path
	}
}

extension CorrectionViewSelectionElement: UIGestureRecognizerDelegate {
	func gestureRecognizer(_ gestureRecognizer: UIGestureRecognizer, shouldReceive touch: UITouch) -> Bool {
		if !confirmButton.isHidden && confirmButton.frame.contains(touch.location(in: self)) {
			return false
		}
		return true
	}
}
