// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Logging

protocol CorrectionViewSnippetPlacementElementDelegate: AnyObject {
    func snippetPlacement(_ element: CorrectionViewSnippetPlacementElement,
                          didSelectRect rect: CGRect,
                          forSnippet snippet: SnippetType)
}

class CorrectionViewSnippetPlacementElement: CorrectionViewElement {
    var panGestureRecognizer: UIPanGestureRecognizer?
    var tapGestureRecognizer: UITapGestureRecognizer?

    var imageLayer: CALayer = CALayer()
    var normalizedImageRect: CGRect = .zero
    var originalNormalizedImageRect: CGRect = .zero

    var snippet: SnippetType? {
        didSet {
            updateSnippet()
        }
    }

    var isInside: Bool = true

    weak var snippetDelegate: CorrectionViewSnippetPlacementElementDelegate?

    override func setup() {
        super.setup()

        imageLayer.disableAnimations()
        imageLayer.borderWidth = 2.0
        imageLayer.borderColor = Asset.tuMexamBlue.color.cgColor
        layer.addSublayer(imageLayer)

        let pan = UIPanGestureRecognizer(target: self, action: #selector(didPan(_:)))
        addGestureRecognizer(pan)
        self.panGestureRecognizer = pan

        let tap = UITapGestureRecognizer(target: self, action: #selector(didTap(_:)))
        addGestureRecognizer(tap)
        self.tapGestureRecognizer = tap
    }

    @objc func didPan(_ recognizer: UIPanGestureRecognizer) {
        switch recognizer.state {
        case .began:
            originalNormalizedImageRect = normalizedImageRect
            let hitTest = recognizer.location(in: self)
            isInside = imageLayer.frame.contains(hitTest)
            fallthrough
        case .changed, .ended:
            guard isInside else { return }
            let location = recognizer.location(in: self).normalized(using: bounds.size)
            normalizedImageRect = CGRect(center: location, size: originalNormalizedImageRect.size)
            setNeedsLayout()
        case .cancelled:
            guard isInside else { return }
            normalizedImageRect = originalNormalizedImageRect
            setNeedsLayout()
        default:
            break
        }
    }

    @objc func didTap(_ recognizer: UITapGestureRecognizer) {
        guard let snippet = self.snippet else { return }
        snippetDelegate?.snippetPlacement(self, didSelectRect: normalizedImageRect, forSnippet: snippet)
    }

    func updateSnippet() {
        guard let snippet = self.snippet, let image = snippet.image else {
            imageLayer.contents = nil
            self.isUserInteractionEnabled = false
            return
        }

        guard let scanSize = context.page.scan.image.value?.size else {
            Logger.correctionUI.error("[SnipPlacementElem] Can't position snippet without knowing the image resolution")
            return
        }

        isUserInteractionEnabled = true

        let imageSize = image.size

        let normSize = imageSize.normalizedIn(size: scanSize)
        var normalized = context.normalizedPageRect
            .centeredRect(size: normSize)
            .convertTo(normalizedSubRect: context.normalizedPageRect)

        if let initial = context.snippetInitialLocation {
            normalized = CGRect(center: initial, size: normalized.size)
        }

        normalizedImageRect = normalized

        imageLayer.contents = image.cgImage
        imageLayer.frame = normalizedImageRect.mappedTo(size: bounds.size)
    }

    override func layoutSubviews() {
        super.layoutSubviews()

        imageLayer.frame = normalizedImageRect.mappedTo(size: bounds.size)
    }
}
