// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
class DrawingOptionsViewController: UIViewController {
	// MARK: Properties

	var drawingOptions: GlobalDrawingOptions = App.shared.globalDrawingOptions
	var toolOptions: GlobalDrawingOptions.ToolOptions? {
		didSet {
			updateLineWidthBounds()
			updateLineWidth()
		}
	}

    // MARK: View Elements

	let lineWidthSlider = UISlider().autolayout()

	let pressureLabel = UILabel().autolayout()
	let pressureSwitch = UISwitch().autolayout()

    // MARK: - UIViewController Lifecycle

	init() {
		super.init(nibName: nil, bundle: nil)
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

    override func viewDidLoad() {
        super.viewDidLoad()

		preferredContentSize = CGSize(width: 310, height: 164)

		// setup slider action
		lineWidthSlider.addTarget(self, action: #selector(lineWidthSliderValueChanged(_:)), for: .valueChanged)

		// setup label and switch for pressure sensitivity option
		pressureLabel.text = L10n.DrawingOptions.PressureSensitivity.title
		pressureSwitch.addTarget(self, action: #selector(pressureSensitivityChanged(_:)), for: .valueChanged)
		let pressureStack = UIStackView(arrangedSubviews: [pressureLabel, pressureSwitch]).autolayout()
		pressureStack.axis = .horizontal

		// setup vertical main stack
		let stack = UIStackView(arrangedSubviews: [lineWidthSlider, pressureStack]).autolayout()
		stack.axis = .vertical
		stack.spacing = 16.0
		view.addSubview(stack)
		stack.pinToSuperview(UIEdgeInsets(top: 16, left: 16, bottom: 16, right: 16))

		// update slider bounds
		updateLineWidthBounds()
    }

	override func viewWillAppear(_ animated: Bool) {
		super.viewWillAppear(animated)

        if let options = toolOptions, options.supportsPressureSensitiveLineWidth {
            // just to be sure, load the current settings from the shared object
            pressureSwitch.isOn = drawingOptions.pressureSensitiveLineWidth
            pressureSwitch.isEnabled = true
            pressureLabel.textColor = .label
        } else {
            pressureSwitch.isEnabled = false
            pressureSwitch.isOn = false
            pressureLabel.textColor = .secondaryLabel
        }
        updateLineWidth()
	}

	func updateLineWidthBounds() {
		guard let toolOptions = toolOptions else { return }
		lineWidthSlider.minimumValue = Float(toolOptions.lineWidthMin)
		lineWidthSlider.maximumValue = Float(toolOptions.lineWidthMax)
	}

	func updateLineWidth() {
		guard let toolOptions = toolOptions else { return }
		lineWidthSlider.value = Float(toolOptions.lineWidth)
	}

    // MARK: IBActions
    @IBAction func lineWidthSliderValueChanged(_ sender: Any) {
		guard let toolOptions = toolOptions else { return }
		toolOptions.lineWidth = CGFloat(lineWidthSlider.value)
    }

    @IBAction func pressureSensitivityChanged(_ sender: Any) {
		drawingOptions.pressureSensitiveLineWidth = pressureSwitch.isOn
    }

}
