// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit

extension EditorWrapperViewController.NavigationTitle {
	var title: String {
		switch self {
		case let .title(string):
			return string
		case let .subtitle(string, _):
			return string
		}
	}
	func withSubtitle(_ subtitle: String) -> EditorWrapperViewController.NavigationTitle {
		let title = self.title
		return .subtitle(title, subtitle)
	}
}

protocol CorrectionTitleTarget: AnyObject {
    func updateTitle()
}

extension EditorWrapperViewController: CorrectionTitleTarget {
    var currentCorrectionViewController: CorrectionRepresentingViewController? {
        guard let content = self.editorContent as? UIPageViewController,
              let current = content.viewControllers?.first
        else { return nil }
        if let current = current as? CorrectionRepresentingViewController {
            return current
        }
        return nil
    }

	func updateTitle() {
        guard let current = self.currentCorrectionViewController else {

            guard let title = self.editorContent.navigationTitle else { return }
            navigationTitle = .title(title)

            return
        }
        guard let correction = current.representingCorrection else { return }

        let erid = correction.erid
		let problem = correction.focusedProblem
		let pass = correction.focusedPass
		let problemPass = L10n.CorrectionController.ProblemPass.title("\(problem)", "\(pass)")

		navigationTitle = .title("\(erid) | \(problemPass)")
		// if second pass -> load name of corrector for first pass
		if pass == 2 {
			firstPassCorrector = correction.corrector(forProblem: problem, pass: 1)
			firstPassCorrector?.load()
		} else {
			firstPassCorrector = nil
		}
	}

	func updateNavigationItem() {
		switch navigationTitle {
		case let .title(title):
			navigationSubtitleView.set(title: title, subtitle: nil)
		case let .subtitle(title, subtitle):
			navigationSubtitleView.set(title: title, subtitle: subtitle)
		}
	}

	func firstPassCorrectorDidChange() {
		guard let corrector = firstPassCorrector else { return }
		firstPassCorrectorNameObserver = corrector.name.receiveOnMain().sink { [unowned self] name in
			guard let name = name else { return }
			let formatted = L10n.CorrectionController.ProblemPass.subtitle(name)
			self.navigationTitle = self.navigationTitle.withSubtitle(formatted)
		}
	}

    @objc func subtitleButtonPressed() {
        BreadcrumbManager.addInteraction("Toolbar - tapped title, showing corrector list")
        guard let current = self.currentCorrectionViewController else { return }
        guard let correction = current.representingCorrection else { return }

        let problem = correction.focusedProblem
        let correctorListController = CorrectorListTableViewController()
        correctorListController.correctorData = correction.allCorrectors(forProblem: problem)

        let navigation = UINavigationController(rootViewController: correctorListController)
        navigation.modalPresentationStyle = .popover
        let popover = navigation.popoverPresentationController
        popover?.sourceView = navigationSubtitleView

        self.present(navigation, animated: true, completion: nil)
    }
}
