// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import TUMexamKit
import UIKit

extension EditorWrapperViewController {
	func setupToolbarManager() {
		let manager = ToolbarManager(
            navigationItem: navigationItem,
            context: correctionContext,
            toolbar: drawingToolbar
        )
		manager.delegate = self
		self.toolbarManager = manager

		// observe changes to undo manager
		undoManagerObserver = correctionContext
			.$currentUndoManager
			.receiveOnMain()
			.assignNoRetain(to: \.currentUndoManager, on: manager)

        switch editorContent.leftToolbarMode {
        case .subproblem:
            manager.setupLeftItemsForSuproblemCorrection()
            navigationItem.leftItemsSupplementBackButton = true
            manager.setupDrawingItems(
                editorToolbarMode: editorContent.leftToolbarMode,
                includeSubproblem: editorContent.supportsSubproblemCorrection
            )
        case .solution:
            manager.setupDrawingItems(
                editorToolbarMode: editorContent.leftToolbarMode,
                includeSubproblem: editorContent.supportsSubproblemCorrection
            )
            navigationItem.leftItemsSupplementBackButton = true
        case .single:
            manager.setupDrawingItems(
                editorToolbarMode: editorContent.leftToolbarMode,
                includeSubproblem: editorContent.supportsSubproblemCorrection
            )
            manager.setupLeftItems(
                includeDismiss: false,
                includeSubproblem: false
            )
            navigationItem.leftItemsSupplementBackButton = true
        case .singleModal:
            manager.setupDrawingItems(
                editorToolbarMode: editorContent.leftToolbarMode,
                includeSubproblem: editorContent.supportsSubproblemCorrection
            )
            manager.setupLeftItems(
                includeDismiss: true,
                includeSubproblem: false
            )
        default:
            manager.setupLeftItems(
                includeDismiss: true,
                includeSubproblem: editorContent.supportsSubproblemCorrection
            )
            manager.setupDrawingItems(
                editorToolbarMode: editorContent.leftToolbarMode,
                includeSubproblem: editorContent.supportsSubproblemCorrection
            )
        }

		manager.setupRightItems()

        let app = UINavigationBarAppearance()
        app.configureWithTransparentBackground()
        app.backgroundColor = .secondarySystemBackground
        app.shadowColor = .clear
        navigationController?.navigationBar.standardAppearance = app
        navigationController?.navigationBar.scrollEdgeAppearance = app
	}
}

extension EditorWrapperViewController: ToolbarManagerDelegate {
    func setEmpty(_ empty: Bool) {
        guard let content = editorContent as? ToolbarManagerDelegate else {
            return
        }
        content.setEmpty(empty)
    }

    func save() {
        guard let content = editorContent as? ToolbarManagerDelegate else {
            return
        }
        content.save()
    }

    func toolbarWantsToDismiss() {
        guard let content = editorContent as? ToolbarManagerDelegate else {
            return
        }
        content.toolbarWantsToDismiss()
    }

    func presentDrawingOptions(_ viewController: UIViewController) {
        self.present(viewController, animated: true, completion: nil)
    }

    func presentToolbarSettings(_ viewController: UIViewController) {
        self.present(viewController, animated: true, completion: nil)
    }

    func editSampleSolution() {
        guard let content = editorContent as? ToolbarManagerDelegate else {
            return
        }
        content.editSampleSolution()
    }

    func expandParentView() {
    }

    func showPreview() {
    }

    func changeDrawingToolbarPosition(_ newPosition: DrawingToolbarPosition) {
        UIView.animate(withDuration: 0.3, delay: 0, options: .curveEaseInOut) {
            self.setDrawingToolbarPosition(position: newPosition)
            if newPosition == .top {
                let items = self.drawingToolbar.items
                if var items = items, let first = items.first, first.tag != 10 {
                    let item = UIBarButtonItem(systemItem: .flexibleSpace, primaryAction: nil, menu: nil)
                    item.tag = 10
                    items.insert(item, at: 0)
                    self.drawingToolbar.items = items
                }
            } else {
                let items = self.drawingToolbar.items
                if var items = items, let first = items.first, first.tag == 10 {
                    items.removeFirst()
                    self.drawingToolbar.items = items
                }
            }
            self.view.layoutIfNeeded()
        }
    }
}

extension OuterCorrectionPageViewController: ToolbarManagerDelegate {
	func setEmpty(_ empty: Bool) {
        guard let viewController = correctionDataSource.currentPageViewController else { return }
        viewController.setCurrentPageEmpty(empty)
	}

	func save() {
        guard let viewController = correctionDataSource.currentPageViewController else { return }
        self.save(correction: viewController.correction)
	}

	func editSampleSolution() {
		guard let viewController = correctionDataSource.currentPageViewController else { return }
		guard let first = viewController.viewControllers?.first as? CorrectionViewController,
			  let context = first.context else {
			return
		}
		if let solutionPage = first.page as? CorrectionPageType {
			App.shared.coordinator?.startSampleSolutionEditing(
                atPage: solutionPage.solutionPage.value,
				context: context)
		}
	}

	func toolbarWantsToDismiss() {
		// if confirm dialog disabled -> dismiss right away
		guard App.shared.settings.askBeforeDismissingCorrectionSession else {
			self.dismiss(animated: true, completion: nil)
			return
		}

		// get unsaved corrections
        let unsaved = correctionDataSource.getChangedERIDs()

		// if nothing unsaved -> dismiss right away
		guard !unsaved.isEmpty else {
			self.dismiss(animated: true, completion: nil)
			return
		}

		// get the corrections in a human readable form
		let readableCorrections = unsaved.joined(separator: ", ")

		// create the alert
		let alert = UIAlertController(
			title: L10n.CorrectionController.Alert.DismissChanges.title,
			message: L10n.CorrectionController.Alert.DismissChanges.message(readableCorrections),
			preferredStyle: .alert
		)

		// add 'cancel' option -> back to correction
		alert.addAction(
			UIAlertAction(
				title: L10n.CorrectionController.Alert.DismissChanges.cancel,
				style: .default,
				handler: { _ in
					// actually do nothing...
				})
		)

		// add 'ok' option -> discard
		alert.addAction(
			UIAlertAction(
				title: L10n.CorrectionController.Alert.DismissChanges.ok,
				style: .destructive,
				handler: { [weak self] _ in
					self?.dismiss(animated: true, completion: nil)
				})
		)

		// present the alert
		self.present(alert, animated: true, completion: nil)
	}

	func presentDrawingOptions(_ viewController: UIViewController) {
		self.present(viewController, animated: true, completion: nil)
	}

	func expandParentView() {}
	func showPreview() {}

    func changeDrawingToolbarPosition(_ newPosition: DrawingToolbarPosition) { }
}

extension OuterCorrectionPageViewController: UIToolbarDelegate {
    func position(for bar: UIBarPositioning) -> UIBarPosition {
        return .topAttached
    }
}
