// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit
import Combine
import Instructions
import TUMexamKit

enum EditorToolbarMode {
    case standard
    case subproblem
    case solution
    case diffing
    case single
    case singleModal
}

protocol EditorContentViewController: UIViewController {
    var supportsSubproblemCorrection: Bool { get }
    var leftToolbarMode: EditorToolbarMode { get }
    var navigationTitle: String? { get }
}

class EditorWrapperViewController: UIViewController {
    enum NavigationTitle: Equatable {
        case title(String)
        case subtitle(String, String)
    }

    var correctionContext: CorrectionContext
    var toolbarManager: ToolbarManager?
    var navigationTitle: NavigationTitle = .title("") {
        didSet {
            if oldValue != navigationTitle {
                updateNavigationItem()
            }
        }
    }
    var firstPassCorrector: ProblemPassCorrectorType? { didSet { firstPassCorrectorDidChange() }}

    var modeObserver: AnyCancellable?
    var firstPassCorrectorNameObserver: AnyCancellable?
    var undoManagerObserver: AnyCancellable?

    let toolbarContainer = UIView().autolayout()
    let drawingToolbar = UIToolbar(
        frame: CGRect(
            origin: CGPoint(x: 50, y: 200),
            size: CGSize(width: 300, height: 40)
        )
    ).autolayout()

    var toolbarConstraints: [NSLayoutConstraint]?

    let horizontalHairlineView = UIView().autolayout()
    let verticalHairlineView = UIView().autolayout()
    let editorContent: EditorContentViewController
    lazy var navigationSubtitleView: NavigationItemTitleSubtitleView = NavigationItemTitleSubtitleView()

    /// For Tutorial on first use
    let coachMarksController = CoachMarksController()
    var coachContent = [(String, UIView?, CGPoint?, CutoutPathMaker?)]()

    init(editorContent: EditorContentViewController, context: CorrectionContext) {
        self.editorContent = editorContent
        self.correctionContext = context
        super.init(nibName: nil, bundle: nil)
        self.modalPresentationStyle = .overFullScreen
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        setupToolbarManager()
        setupObservers()

        self.edgesForExtendedLayout = []
        self.navigationItem.titleView = navigationSubtitleView

        let navigationSubtitleGestureRecognizer = UITapGestureRecognizer(target: self,
                                                                         action: #selector(subtitleButtonPressed))
        self.navigationSubtitleView.addGestureRecognizer(navigationSubtitleGestureRecognizer)

        view.addSubview(toolbarContainer)
        toolbarContainer.addSubview(drawingToolbar)
        drawingToolbar
            .center()
        drawingToolbar.delegate = self

        view.addSubview(horizontalHairlineView)
        horizontalHairlineView.backgroundColor = .systemGray4
        horizontalHairlineView
            .height(1)

        view.addSubview(verticalHairlineView)
        verticalHairlineView.backgroundColor = .systemGray4
        verticalHairlineView
            .top(to: \.topAnchor)
            .width(1)
            .bottom(to: \.bottomAnchor)

        editorContent.willMove(toParent: self)
        addChild(editorContent)
        view.addSubview(editorContent.view)
        editorContent.view.translatesAutoresizingMaskIntoConstraints = false

        setDrawingToolbarPosition(position: App.shared.settings.drawingToolbarPosition)
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        if !((App.shared.settings[HiddenSettings.didShowPagingCorrectionTutorial] as? Bool) ?? false) {
            BreadcrumbManager.addNavigation("[OuterCorrectionPageVC] show tutorial")
            setupCoachMarks()
            coachMarksController.dataSource = self
            coachMarksController.delegate  = self
            let skipView = CoachMarkSkipDefaultView()
            skipView.setTitle("Skip Tutorial", for: .normal)
            coachMarksController.skipView = skipView
            coachMarksController.start(in: .window(over: self))
        }
    }

    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        self.coachMarksController.stop(immediately: true)
    }

    func setupObservers() {
        modeObserver = correctionContext
            .$mode
            .receiveOnMain()
            .sink { [weak self] mode in
                self?.updateMode(mode)
        }
    }

    func updateMode(_ mode: CorrectionContext.Mode) {
        self.toolbarManager?.updateSubproblemCorrectionMode(mode == .subproblemSelection)
    }

    func setDrawingToolbarPosition(position: DrawingToolbarPosition) {
        App.shared.settings.drawingToolbarPosition = position
        toolbarConstraints?.forEach { $0.deactivate() }
        var constraints = [NSLayoutConstraint]()

        switch position {
        case .left:
            self.verticalHairlineView.isHidden = false
            constraints.append(contentsOf: [
                self.drawingToolbar.addWidth(to: \.heightAnchor),
                self.drawingToolbar.addHeight(to: \.widthAnchor),
                // Container
                self.toolbarContainer.addTopConstraint(to: \.topAnchor, constant: -20),
                self.toolbarContainer.addLeadingConstraint(to: \.leadingAnchor),
                self.toolbarContainer.addBottomConstraint(to: \.bottomAnchor),
                self.toolbarContainer.addWidth(44),
                // Editor
                self.editorContent.view.addTopConstraint(to: \.topAnchor),
                self.editorContent.view.addLeadingConstraint(to: \.trailingAnchor, of: self.toolbarContainer),
                self.editorContent.view.addTrailingConstraint(to: \.trailingAnchor),
                self.editorContent.view.addBottomConstraint(to: \.bottomAnchor),
                // Horizontal Hairline
                self.horizontalHairlineView.addTopConstraint(to: \.topAnchor),
                self.horizontalHairlineView.addLeadingConstraint(to: \.trailingAnchor, of: self.toolbarContainer),
                self.horizontalHairlineView.addTrailingConstraint(to: \.trailingAnchor),
                // Vertical Hairline
                self.verticalHairlineView.addLeadingConstraint(to: \.trailingAnchor, of: self.toolbarContainer)
            ])
            self.updateDrawingItemsRotation()

        case .right:
            self.verticalHairlineView.isHidden = false
            constraints.append(contentsOf: [
                self.drawingToolbar.addWidth(to: \.heightAnchor),
                self.drawingToolbar.addHeight(to: \.widthAnchor),
                // Container
                self.toolbarContainer.addTopConstraint(to: \.topAnchor, constant: -20),
                self.toolbarContainer.addTrailingConstraint(to: \.trailingAnchor),
                self.toolbarContainer.addBottomConstraint(to: \.bottomAnchor),
                self.toolbarContainer.addWidth(44),
                // Editor
                self.editorContent.view.addTopConstraint(to: \.topAnchor),
                self.editorContent.view.addLeadingConstraint(to: \.leadingAnchor),
                self.editorContent.view.addTrailingConstraint(to: \.leadingAnchor, of: self.toolbarContainer),
                self.editorContent.view.addBottomConstraint(to: \.bottomAnchor),
                // Horizontal Hairline
                self.horizontalHairlineView.addTopConstraint(to: \.topAnchor),
                self.horizontalHairlineView.addLeadingConstraint(to: \.leadingAnchor),
                self.horizontalHairlineView.addTrailingConstraint(to: \.leadingAnchor, of: self.toolbarContainer),
                // Vertical Hairline
                self.verticalHairlineView.addTrailingConstraint(to: \.leadingAnchor, of: self.toolbarContainer)
            ])
            self.updateDrawingItemsRotation()
        case .top:
            self.verticalHairlineView.isHidden = true
            constraints.append(contentsOf: [
                self.drawingToolbar.addWidth(to: \.widthAnchor),
                self.drawingToolbar.addHeight(to: \.heightAnchor),
                // Container
                self.toolbarContainer.addTopConstraint(to: \.topAnchor),
                self.toolbarContainer.addLeadingConstraint(to: \.leadingAnchor),
                self.toolbarContainer.addTrailingConstraint(to: \.trailingAnchor),
                self.toolbarContainer.addHeight(44),
                // Editor
                self.editorContent.view.addTopConstraint(to: \.bottomAnchor, of: self.toolbarContainer),
                self.editorContent.view.addLeadingConstraint(to: \.leadingAnchor),
                self.editorContent.view.addTrailingConstraint(to: \.trailingAnchor),
                self.editorContent.view.addBottomConstraint(to: \.bottomAnchor),
                // Horizontal Hairline
                self.horizontalHairlineView.addTopConstraint(to: \.bottomAnchor, of: self.toolbarContainer),
                self.horizontalHairlineView.addLeadingConstraint(to: \.leadingAnchor),
                self.horizontalHairlineView.addTrailingConstraint(to: \.trailingAnchor)
            ])
            self.updateDrawingItemsRotation()
        }
        self.toolbarConstraints = constraints
    }

    override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()
        updateDrawingItemsRotation()
    }

    private func updateDrawingItemsRotation() {
        if toolbarManager?.drawingToolbarPosition == .top {
            drawingToolbar.transform = .identity
            drawingToolbar.subviews(conformingTo: SelectableBarButton.self).forEach { view in
                view.transform = .identity
            }
        } else {
            drawingToolbar.transform = .init(rotationAngle: .pi / 2)
            drawingToolbar.subviews(conformingTo: SelectableBarButton.self).forEach { view in
                view.transform = .init(rotationAngle: -.pi / 2)
            }
        }
    }
}

extension EditorWrapperViewController: UIToolbarDelegate {
    func position(for bar: UIBarPositioning) -> UIBarPosition {
        return .top
    }
}
