// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import Combine

class LoadingItem {
    enum State {
        case waiting
        case loading
        case loaded
        case error
    }
    let title: String

    @Published var state: State = .waiting

    init(title: String) {
        self.title = title
    }
}

class LoadingItemView: BaseView {
    private let label: UILabel = UILabel().autolayout()
    private let stateContainer = UIView().autolayout()
    private let stack = UIStackView().autolayout()

    private var stateObserver: AnyCancellable?

    let item: LoadingItem

    init(_ item: LoadingItem) {
        self.item = item
        super.init()
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func setup() {
        super.setup()

        stack.spacing = 5.0
        stack.axis = .horizontal
        stack.alignment = .center

        stack.addArrangedSubview(stateContainer)
        stack.addArrangedSubview(label)

        stateContainer.widthAnchor.constraint(equalTo: stateContainer.heightAnchor).isActive = true
        stateContainer.widthAnchor.constraint(equalToConstant: 20.0).isActive = true

        addSubview(stack)
        stack.pinToSuperview()

        label.text = item.title

        stateObserver = item.$state.receiveOnMain().removeDuplicates().sink { [weak self] state in
            self?.stateChanged(state)
        }
    }

    func stateChanged(_ state: LoadingItem.State) {
        // due to .removeDuplicates, this is only called when the state actually changes (i hope)
        // -> remove all subviews in container
        let containerSubviews = stateContainer.subviews
        for subview in containerSubviews {
            subview.removeFromSuperview()
        }

        switch state {
        case .waiting:
            label.textColor = .secondaryLabel
            let stateView = UIActivityIndicatorView(style: .medium).autolayout()
            stateContainer.addSubview(stateView)
            stateView.pinToSuperview()
            stateView.startAnimating()
        case .loading:
            label.textColor = .label
            let stateView = UIActivityIndicatorView(style: .medium).autolayout()
            stateContainer.addSubview(stateView)
            stateView.pinToSuperview()
            stateView.startAnimating()
        case .loaded:
            label.textColor = .label
            let stateView = UIImageView(image: UIImage(systemName: "checkmark.circle.fill")).autolayout()
            stateView.tintColor = .systemGreen
            stateContainer.addSubview(stateView)
            stateView.pinToSuperview()
        case .error:
            label.textColor = .label
            let stateView = UIImageView(image: UIImage(systemName: "xmark.circle.fill")).autolayout()
            stateView.tintColor = .systemRed
            stateContainer.addSubview(stateView)
            stateView.pinToSuperview()
        }
    }
}

class DetailedLoadingViewController: UIViewController {
    let label: UILabel = UILabel().autolayout()
    let stack = UIStackView().autolayout()
    weak var provider: LoadingItemProvider?

    init(provider: LoadingItemProvider) {
        self.provider = provider
        super.init(nibName: nil, bundle: nil)
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        guard let provider = provider else { return }

        provider.onChange = { [weak self] in
            self?.recreate()
        }

        label.text = provider.loadingTitle

        // Do any additional setup after loading the view.
        stack.axis = .vertical
        stack.alignment = .fill
        stack.spacing = 5.0
        stack.addArrangedSubview(label)

        addItems(provider.loadingItems)

        view.addSubview(stack)

        stack.centerInSuperview()
    }

    func recreate() {
        let oldItems = stack.arrangedSubviews
        for item in oldItems {
            if item === label {
                continue
            }

            item.removeFromSuperview()
        }

        guard let provider = provider else { return }
        label.text = provider.loadingTitle
        addItems(provider.loadingItems)
    }

    func addItems(_ items: [LoadingItem]) {
        let itemViews = items.map { LoadingItemView($0) }
        for itemView in itemViews {
            stack.addArrangedSubview(itemView)
        }
    }
}
