// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import Combine
import TUMexamKit
import Logging

class LoadingWrapperViewController<ResourceType, ContentVCType: UIViewController>: UIViewController {
	typealias ContentCreationBlock = (ResourceType) -> ContentVCType
	let loadingResource: LoadingResource<ResourceType>
	let block: ContentCreationBlock

	private var currentViewController: UIViewController?
	private var stateObserver: AnyCancellable?

	var contentViewController: ContentVCType?

	init(_ loadingResource: LoadingResource<ResourceType>, block: @escaping ContentCreationBlock) {
		self.loadingResource = loadingResource
		self.block = block

		super.init(nibName: nil, bundle: nil)
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

    override func viewDidLoad() {
        super.viewDidLoad()

        // Do any additional setup after loading the view.
		stateObserver = loadingResource.$state.receiveOnMain().removeDuplicates().sink { [weak self] state in
			self?.stateChanged(state)
		}

        if loadingResource.state == .idle || loadingResource.state == .reset {
            Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))] calling load on resource from viewDidLoad because resource state is \(loadingResource.state)")
            loadingResource.load()
        }
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        if loadingResource.state == .idle || loadingResource.state == .reset {
            Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))] calling load() on resource from viewWillAppear because resource state is \(loadingResource.state)")
            loadingResource.load()
        } else {
            Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))] NOT calling load() on resource from viewWillAppear because resource state is \(loadingResource.state)")
        }
    }

	func stateChanged(_ state: LoadingResource<ResourceType>.State) {
        Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))] resource \(self.loadingResource) state changed to \(state)")
		switch state {
        case .loading, .reset:
			showLoadingViewController()
		case let .error(error):
			showErrorViewController(error)
		case let .content(content):
			showContentViewController(content)
		case .idle:
            showLoadingViewController()
            Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))]: state changed to .idle, calling load() on resource")
            loadingResource.load()
		}
	}

	func showLoadingViewController() {
        Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))]: Switching to LoadingVC with resource \(loadingResource)")

        // if the loading resource provides items...
        if let itemProvider = loadingResource as? LoadingItemProvider {
            // ... use them to show the detailed loading vc
            let loading = DetailedLoadingViewController(provider: itemProvider)
            showViewController(loading)
        } else {
            // ... otherwise just the standard activity indicator
			let loading = LoadingViewController()
			showViewController(loading)
        }
	}

	func showContentViewController(_ content: ResourceType) {
        Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))]: Switching to content for loading resource \(loadingResource)")
		let contentVC = block(content)
        contentViewController = contentVC
		showViewController(contentVC)
	}

	func showErrorViewController(_ error: Error) {
        Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))]: Switching to Error for resource \(loadingResource)")

        if case let .reachedEnd(pass) = error as? PromiseLoadingResourceError {
            // Handle end case specially
            // Check if we are in subproblem mode, then we DO NOT want to display the message inside the individual controllers
            if (self.parent as? SubproblemPageContentViewController) == nil {
                let endVC = EndViewController(pass ?? 1)
                showViewController(endVC)
            } else {
                // Show empty view controller
                showViewController(UIViewController())
            }
        } else {
            let errorVC = ErrorViewController(nibName: nil, bundle: nil)
            errorVC.errors = [error]
            showViewController(errorVC)
        }
	}

	func showViewController(_ viewController: UIViewController) {
		guard currentViewController != viewController else {
			let currentPointer = pointerString(currentViewController)
			let viewPointer = pointerString(viewController)
            Logger.loadingUI.notice("LoadingWrapper[\(pointerString(self))]: Already showing \(currentPointer) [wants to show \(viewPointer)]")
			return
		}

        Logger.loadingUI.info("LoadingWrapper[\(pointerString(self))] showViewController \(pointerString(viewController)) for loading resource \(loadingResource)")

		// delete old one if required
		if let current = self.currentViewController {
			current.willMove(toParent: nil)
			current.view.removeFromSuperview()
			current.removeFromParent()
		}

		// add new one
		addChild(viewController) { (view) in
			view.translatesAutoresizingMaskIntoConstraints = false
			self.view.addSubview(view)
			view.pinToSuperview()
		}
        self.currentViewController = viewController
	}
}

protocol CorrectionRepresentingViewController: UIViewController {
    var representingCorrection: CorrectionType? { get }
}

class CorrectionLoadingWrapperViewController: LoadingWrapperViewController<
                                                CorrectionPromiseType, CorrectionPageViewController> {

}

extension CorrectionLoadingWrapperViewController: CorrectionRepresentingViewController {
    var representingCorrection: CorrectionType? {
        if let content = contentViewController {
            return content.representingCorrection
        }
        return nil
    }
}
