// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import os.log

/// a loading resource that bundles multiple SubloadingResources together and waits for them all to finish before providing the content
class CompoundLoadingResource<T>: LoadingResource<T>, LoadingItemProvider {
    private var subloadingResources: [SubloadingResource] = []

    // MARK: - Loading Item Provider
    var loadingTitle: String {
        fatalError("override in subclass")
    }

    var loadingItems: [LoadingItem] {
        subloadingResources.map { $0.loadingItem }
    }

    var onChange: (() -> Void)?

    // MARK: - Initialization

    override init() {
        super.init()

        subloadingResources = createSubloadingResources()
        // now set the callbacks
        for sub in subloadingResources {
            sub.onError = { [weak self] (resource, error) in
                self?.subresource(resource, failed: error)
            }
            sub.onCompletion = { [weak self] (resource) in
                self?.subresourceCompleted(resource)
            }
        }
        checkAll()
    }

    // MARK: - Override in Subclass

    /// override and provide the subloading resources required
    /// - Returns: the subloading resources
    func createSubloadingResources() -> [SubloadingResource] {
        fatalError("override in subclass")
    }

    /// override and return the final value
    /// - Returns: the value this resource has as content
    func didFinish() -> T {
        fatalError("override in subclass")
    }

    // MARK: - Internal Logic

    override func load() {
        guard state == .idle else { return }

        for sub in subloadingResources {
            sub.load()
        }
    }

    func subresource(_ resource: SubloadingResource, failed error: Error) {
        guard resource.isMandatory else {
            return
        }
        state = .error(error)
    }

    func subresourceCompleted(_ resource: SubloadingResource) {
        // no more completion if already encountered an error
        if case .error = state {
            return
        }
        checkAll()
    }

    func checkAll() {
        let currentState = checkState()
        state = currentState
    }

    func checkState() -> State {
        // first check if everything's loaded
        if subloadingResources.all(.loaded, for: \.loadingItem.state) {
            let result = didFinish()
            return .content(result)
        }

        return .loading
    }
}
