// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import TUMexamKit
import Logging

class PageLoadingResource: CompoundLoadingResource<PageType> {
    let page: PageType

    var debugIdentifier: String {
        "\(page.erid ?? "").\(page.pageNumber)"
    }

    private var subloadingResources: [StatefulLoadableSubloadingResource] = []

    override var loadingTitle: String {
        L10n.Loading.Page.title
    }

    init(page: PageType) {
        self.page = page
        super.init()
        Logger.loadingUI.debug("[PageLoadingResource|\(debugIdentifier)] created loading resource")
    }

    override func createSubloadingResources() -> [SubloadingResource] {
        var resources: [StatefulLoadableSubloadingResource] = []
        // scan
        resources.append(StatefulLoadableSubloadingResource(page.scan, title: L10n.Loading.Page.scan))
        // solution
        if let solution = page.solution {
            resources.append(StatefulLoadableSubloadingResource(solution,
                                                                          title: L10n.Loading.Page.solution,
                                                                          isMandatory: false))
        }
        if let personalizedSolution = page.personalizedSolution {
            resources.append(StatefulLoadableSubloadingResource(
                                personalizedSolution,
                                title: L10n.Loading.Page.personalizedSolution,
                                isMandatory: false
            ))
        }
        // all layers
        for layer in page.layers.value {
            if let focusedPass = page.focusedPass, layer.pass > focusedPass {
                continue
            }
            let title = L10n.Loading.Page.layer("\(layer.pass)")
            resources.append(StatefulLoadableSubloadingResource(layer.image, title: title, isMandatory: false))
        }

        return resources
    }

    override func didFinish() -> PageType {
        return page
    }

    override var description: String {
        "PageLoadingResource[\(pointerString(self))|\(page.erid ?? "")-\(page.pageNumber)]"
    }
}
