// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import TUMexamKit
import Logging

// typealias PromiseResult = (CorrectionType, LockReferenceType?)

//    class PromiseResult {
//        let correction: CorrectionType
//        var lockReference: LockReferenceType?
//
//        init(correction: CorrectionType, lockReference: LockReferenceType? = nil) {
//            self.correction = correction
//            self.lockReference = lockReference
//        }
//    }

class PromiseLoadingResource: LoadingResource<CorrectionPromiseType>, LoadingItemProvider {
    let promise: CorrectionPromiseType

    private var promiseObserver: AnyCancellable?

    private let lockLoadingItem = LoadingItem(title: L10n.Loading.Correction.lock)

    var loadingTitle: String {
        L10n.Loading.Correction.title
    }

    var loadingItems: [LoadingItem] {
        [lockLoadingItem]
    }

    var onChange: (() -> Void)?

    init(promise: CorrectionPromiseType) {
        self.promise = promise
    }

    override func load() {
        if promiseObserver == nil {
            promiseObserver = promise.state.sink { [weak self] (promiseState) in
                self?.promiseStateChanged(promiseState)
            }
        }
        Logger.loadingUI.info("\(self) load() called")
        promise.load()
    }

    private func promiseStateChanged(_ promiseState: CorrectionPromiseState) {
        Logger.loadingUI.info("\(self) internal promise state changed to \(promiseState)")
        switch promiseState {
        case .loading:
            self.state = .loading
            self.lockLoadingItem.state = .loading
        case .ready:
            self.lockLoadingItem.state = .loaded
            self.state = .content(promise)
        case let .error(error):
            self.state = .error(error)
        case let .reachedEnd(pass):
            self.state = .error(PromiseLoadingResourceError.reachedEnd(pass: pass))
        case .expired:
            self.state = .reset
            self.lockLoadingItem.state = .waiting
        }
    }

    override var description: String {
        return "PromiseLoadingResource[\(promise.debugDescription)]"
    }
}
