// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import TUMexamKit
import Combine
import Logging

typealias PromisePageResult = (PageType, CorrectionPromiseType?)

class PromisePageLoadingResource: LoadingResource<PromisePageResult>, LoadingItemProvider {
    private let promise: CorrectionPromiseType
    private let pageNumber: Int
    private let rect: RectOnPage

    private var promiseLoading: PromiseLoadingResource
    private var pageLoading: PageLoadingResource?
    private var promiseObserver: AnyCancellable?
    private var pageObserver: AnyCancellable?

    var correction: CorrectionType? { promise.state.value.correction }

    var loadingTitle: String {
        if let page = pageLoading {
            return page.loadingTitle
        }
        return promiseLoading.loadingTitle
    }

    var loadingItems: [LoadingItem] {
        if let page = pageLoading {
            return page.loadingItems
        }
        return promiseLoading.loadingItems
    }

    var onChange: (() -> Void)?

    init(promise: CorrectionPromiseType, pageNumber: Int, rect: RectOnPage) {
        self.promise = promise
        self.pageNumber = pageNumber
        self.rect = rect
        self.promiseLoading = PromiseLoadingResource(promise: promise)
    }

    override func load() {
        guard promiseObserver == nil else { return }
        Logger.loadingUI.info("[PromisePageLoadingResource|\(pointerString(self))|\(promise.debugDescription)] load() called")
        promiseObserver = promiseLoading.$state
            .receiveOnMain()
            .debounce(for: 0.1, scheduler: RunLoop.main)
            .sink { [weak self] state in
                self?.promiseStateChanged(state)
        }
        promiseLoading.load()
    }

    func promiseStateChanged(_ state: PromiseLoadingResource.State) {
        Logger.loadingUI.info("[PromisePageLoadingResource|\(pointerString(self))|\(promise.debugDescription)] promise state changed to \(state)")
        switch state {
        case .idle:
            break
        case .loading:
            self.state = .loading
        case let .content(content):
            let promise = content
            guard let correction = promise.state.value.correction else {
                Logger.correctionUI.error("trying to access correction of before loaded")
                // self.state = .error(NetworkError.lockTaken)
                return
            }
            // try getting the page
            if let page = correction.page(withNumber: pageNumber) {
                loadPage(page)
            } else {
                // no page -> error
                self.state = .error(SubproblemPageResourceError.pageNotFound)
            }
        case let .error(err):
            self.state = .error(err)
        case .reset:
            // reset page loading
            // self.promiseLoading = PromiseLoadingResource(promise: promise)
            self.pageLoading = nil
            self.promiseObserver = nil
            self.pageObserver = nil
            self.state = .reset
        }
    }

    private func loadPage(_ page: PageType) {
        Logger.loadingUI.info("[PromisePageLoadingResource|\(pointerString(self))|\(promise.debugDescription)] loadPage() called")
        let pageLoading = PageLoadingResource(page: page)
        self.pageLoading = pageLoading
        // inform about new loading items
        onChange?()

        pageObserver = pageLoading.$state.receiveOnMain().sink { [weak self] state in
            guard let self = self else { return }
            switch state {
            case .idle, .reset:
                self.state = .idle
            case .loading:
                self.state = .loading
            case let .content(page):
                self.state = .content((page, self.promise))
            case let .error(err):
                self.state = .error(err)
            }
        }

        pageLoading.load()
    }

    override var description: String {
        "PromisePageLoadingResource[\(promise.debugDescription)]"
    }
}
