// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import UIKit
import TUMexamKit
import Logging

enum PromiseLoadingResourceError: Error, LocalizedError {
    case reachedEnd(pass: Int?)

    var errorDescription: String? {
        switch self {
        case .reachedEnd:
            return ""
        }
    }
}

enum SubproblemPageResourceError: Error {
	case pageNotFound
}

class SubproblemPageLoadingResource: LoadingResource<PromisePageResult>, LoadingItemProvider {
    let correction: CorrectionType
    let pageNumber: Int
    let rect: RectOnPage

    var pageLoading: PageLoadingResource?
    var observer: AnyCancellable?
    var lockReference: LockReferenceType?

    private let correctionLoadingItem = LoadingItem(title: L10n.Loading.Correction.page)

    var loadingTitle: String {
        if let page = pageLoading {
            return page.loadingTitle
        }
        return L10n.Loading.Correction.title
    }

    var loadingItems: [LoadingItem] {
        if let page = pageLoading {
            return page.loadingItems
        }
        return [correctionLoadingItem]
    }

    var onChange: (() -> Void)?

    init(correction: CorrectionType, pageNumber: Int, rect: RectOnPage) {
        self.correction = correction
        self.pageNumber = pageNumber
        self.rect = rect
    }

    override func load() {
        Logger.loadingUI.debug("[SubPageLoadingResource|\(correction.erid)|\(pointerString(self))] load()")
        guard state == .idle else {
            Logger.loadingUI.debug("[SubPageLoadingResource|\(correction.erid)|\(pointerString(self))] load() called, but state is \(state)")
            return
        }
        if let page = correction.page(withNumber: pageNumber) {
            loadPage(page)
            return
        }

        Logger.loadingUI.debug("[SubPageLoadingResource|\(correction.erid)|\(pointerString(self))] add lock requirement")
        lockReference = correction.addLockRequirement()

        observer = correction.state.removeDuplicates().receiveOnMain().sink { [weak self] state in
            self?.correctionStateChanged(state)
        }
    }

    func correctionStateChanged(_ resourceState: ResourceObjectState) {
        Logger.loadingUI.debug("[SubPageLoadingResource|\(correction.erid)|\(pointerString(self))] correction state changed to \(resourceState)")
        switch resourceState {
        case .idle:
            // do nothing
            correctionLoadingItem.state = .waiting
        case .loading:
            state = .loading
            correctionLoadingItem.state = .loading
        case .loaded:
            correctionLoadingItem.state = .loaded
            if let page = correction.page(withNumber: pageNumber) {
                loadPage(page)
            } else {
                state = .error(SubproblemPageResourceError.pageNotFound)
            }
        case let .error(error):
            state = .error(error)
        }
    }

    private func loadPage(_ page: PageType) {
        Logger.loadingUI.debug("[SubPageLoadingResource|\(correction.erid)|\(pointerString(self))] loadPage()")
        let pageLoading = PageLoadingResource(page: page)
        self.pageLoading = pageLoading

        // inform that loading items changed
        onChange?()

        observer = pageLoading.$state.removeDuplicates().receiveOnMain().sink { [weak self] state in
            Logger.loadingUI.debug("[SubPageLoadingResource|\(self?.correction.erid ?? "nil")|\(pointerString(self ?? nil))] pageLoading state changed to \(state)")
            switch state {
            case .idle, .reset:
                self?.state = .idle
            case .loading:
                self?.state = .loading
            case let .content(page):
                self?.state = .content((page, nil))
            case let .error(err):
                self?.state = .error(err)
            }
        }
        pageLoading.load()
    }

    override var description: String {
        "<SubproblemPageLoadingResource: correction = \(self.correction.identifier.stringIdentifier), page = \(self.pageNumber), lockref = \(self.lockReference.debugDescription)>"
    }
}
