// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Combine
import Logging

class CorrectionPageViewController: UIPageViewController, IndexRepresentingViewController {
    let correction: CorrectionType
    var representingIndexInCollection: Int?

    var context: CorrectionContext?

    var stateObserver: AnyCancellable?
    var modeObserver: AnyCancellable?

//    var arrayDataSource: ArrayDataSource?
    var cache: PageViewControllerCache<CorrectionViewController>

    var lockReference: LockReferenceType?
    var lockStateObserver: AnyCancellable?
    var lockedToast: Toast?
    var lockLoadingToast: Toast?
    var lockStartedLoad = false

    var isCurrentlyAppearing: Bool = false
    var switchToPageWhenAppearing: Bool = false
    var isPresentingSubproblemCorrection: Bool = false

    init(correction: CorrectionType,
         cache: PageViewControllerCache<CorrectionViewController>,
         context: CorrectionContext? = nil,
         lockReference: LockReferenceType? = nil) {
        self.correction = correction
        self.cache = cache
        self.context = context
        self.lockReference = lockReference

        super.init(
            transitionStyle: .scroll,
            navigationOrientation: .horizontal,
            options: [.interPageSpacing: 32.0 as NSNumber]
        )

        self.title = correction.erid
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        disablePencilForScrolling()
        observeState(correction)

        modeObserver = context?.$mode.receiveOnMain().sink { [weak self] mode in
            self?.pageScrollView?.isScrollEnabled = mode == .normal
        }

        // create a lock reference, and observe the status
        let lockReference = self.lockReference ?? correction.addLockRequirement()

        lockStateObserver = lockReference.state.receiveOnMain().sink { [weak self] state in
            self?.lockStateChanged(state)
        }
        self.lockReference = lockReference
        Logger.correctionUI.debug("PageVC added lock req (\(self.correction.identifier.stringIdentifier)) \(pointerString(lockReference as AnyObject)))")
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)

        if let current = context?.currentPageIndex, switchToPageWhenAppearing {
            // this means we're already loaded
            _ = showPage(at: current)
            switchToPageWhenAppearing = false
        }
    }

    override func viewDidAppear(_ animated: Bool) {
        BreadcrumbManager.addNavigation("[CorrectionPageVC] viewDidAppear - \(correction.identifier.stringIdentifier)")
        super.viewDidAppear(animated)
        isCurrentlyAppearing = true
        restoreLockToast()

        if lockReference == nil {
            self.lockReference = correction.addLockRequirement()
        }

        if let target = parent(conformingTo: CorrectionTitleTarget.self) {
            target.updateTitle()
        }

        guard let current = self.viewControllers?.first as? CorrectionViewController else {
            return
        }
        context?.currentPageIsEmpty = current.empty
    }

    override func viewDidDisappear(_ animated: Bool) {
        BreadcrumbManager.addNavigation("[CorrectionPageVC] viewDidDisappear - \(correction.identifier.stringIdentifier)")
        super.viewDidDisappear(animated)
        isCurrentlyAppearing = false
        switchToPageWhenAppearing = !isPresentingSubproblemCorrection
        isPresentingSubproblemCorrection = false
        hideLockToast()
    }

    func createDataSource() {
        var currentPageIndex = context?.currentPageIndex ?? 0

        self.dataSource = self
        self.delegate = self

        // show the current page (based on context) or fallback to 0 (if no context available)
        if !showPage(at: currentPageIndex) {
            if let focusedPage = correction.pageForFocusedSubproblem() {
                currentPageIndex = focusedPage.pageNumber - 1
            } else {
                currentPageIndex = 0
            }
            context?.currentPageIndex = currentPageIndex
            showPage(at: currentPageIndex)
        }
    }

    /// Called when the Page View Controlelr first appears
    @discardableResult
    func showPage(at index: Int) -> Bool {
        if let page = correction.pages.value[safeAccess: index], let current = cache.viewController(for: page) {
            self.setViewControllers([current], direction: .forward, animated: false, completion: nil)
            precachePage(at: index - 1)
            precachePage(at: index + 1)
            return true
        } else {
            return false
        }
    }

    @discardableResult
    func precachePage(at index: Int) -> Bool {
        if let page = correction.pages.value[safeAccess: index], cache.viewController(for: page) != nil {
            return true
        } else {
            return false
        }
    }

    func page(at index: Int) -> PageType? {
        let pages = correction.pages.value
        guard index >= 0, index < pages.count else { return nil }
        return pages[index]
    }

    func switchToContent() {
        // if no current page set yet, try to infer it from the correction
        if context?.currentPageIndex == nil {
            if let focusedPage = correction.pageForFocusedSubproblem() {
                context?.currentPageIndex = focusedPage.pageNumber - 1
            } else {
                context?.currentPageIndex = 0
            }
        }

        createDataSource()
    }

    func vacateLock() {
        Logger.correctionUI.debug("PageVC is vacating Lock (\(self.correction.identifier.stringIdentifier)) \(pointerString(lockReference as AnyObject)))")
        lockReference = nil

        for page in cache.currentlyHeldControllers() {
            page.vacateLock()
        }
    }

    func setCurrentPageEmpty(_ empty: Bool) {
        guard let current = self.viewControllers?.first as? CorrectionViewController else {
            return
        }

        current.empty = empty
    }

    var canUndo: Bool {
        guard let current = self.viewControllers?.first as? CorrectionViewController else { return false }
        return current.canUndo
    }

    var canRedo: Bool {
        guard let current = self.viewControllers?.first as? CorrectionViewController else { return false }
        return current.canRedo
    }

    func undo() {
        guard let current = self.viewControllers?.first as? CorrectionViewController else {
            return
        }

        current.undo()
    }

    func redo() {
        guard let current = self.viewControllers?.first as? CorrectionViewController else {
            return
        }

        current.redo()
    }
}

extension CorrectionPageViewController: StateObserving {
    func stateChanged(_ state: ResourceObjectState) {
        switch state {
        case .idle:
            // do nothing
            break
        case .loading:
            break
        case .loaded:
            switchToContent()
        case let .error(error):
            Logger.correctionUI.error("PageVC: Error on state change: \(String(describing: error))")
        }
    }
}

extension CorrectionPageViewController: UIPageViewControllerDelegate {
    func pageViewController(
        _ pageViewController: UIPageViewController,
        didFinishAnimating finished: Bool,
        previousViewControllers: [UIViewController],
        transitionCompleted completed: Bool) {
        guard let current = self.viewControllers?.first as? CorrectionViewController else {
            return
        }

        if App.shared.settings[SettingsKeys.retainPageSelection] as? Bool ?? false {
            let pageNumber = current.page.pageNumber
            context?.currentPageIndex = pageNumber - 1
        }

        context?.currentPageIsEmpty = current.empty
    }
}

extension CorrectionPageViewController: CorrectionViewSubproblemSelectionElementDelegate {
    func subproblemSelection(_ element: CorrectionViewSubproblemSelectionElement,
                             shouldStartUsing rect: RectOnPage,
                             page: Int,
                             problems: [ProblemIdentifier]?) {
        if let parent = parent(conformingTo: CorrectionViewSubproblemSelectionElementDelegate.self) {
            isPresentingSubproblemCorrection = true
            parent.subproblemSelection(element, shouldStartUsing: rect, page: page, problems: problems)
        } else {
            Logger.correctionUI.error("PageVC (\(pointerString(self)): Parent can't handle subproblem selection")
        }
    }
}

extension CorrectionPageViewController: CorrectionRepresentingViewController {
    var representingCorrection: CorrectionType? {
        correction
    }
}

extension CorrectionPageViewController: ContextUndoManagerCoordinator {
    func shouldChangeUndoManager(to canvasView: CorrectionCanvasViewController) -> Bool {
        guard let currentCorrection = self.viewControllers?.first as? CorrectionViewController,
              let currentCanvas = currentCorrection.loadingViewController.contentViewController else { return false }

        guard canvasView === currentCanvas else { return false }

        // We also need to check if we are the current page view controller being displayed

        // This is old code. When using a StaticPageDataSource, there are no LoadingWarppers betwee the Outer and Inner PageViewControllers
//        guard let parent = self.parent as? CorrectionLoadingWrapperViewController,
//              let outer = parent.parent as? OuterCorrectionPageViewController,
//              let currentPageWrap = outer.viewControllers?.first as? CorrectionLoadingWrapperViewController,
//              let currentPage = currentPageWrap.contentViewController else { return false }

        if let parent = self.parent as? CorrectionLoadingWrapperViewController,
           let outer = parent.parent as? OuterCorrectionPageViewController,
           let currentPageWrap = outer.viewControllers?.first as? CorrectionLoadingWrapperViewController,
           let currentPage = currentPageWrap.contentViewController {

            return currentPage === self

        } else if let outer = self.parent as? OuterCorrectionPageViewController,
                  let currentPage = outer.currentCorrectionViewController {

            return currentPage === self

        } else if let single = self.parent as? SingleCorrectionViewController {
            return single.correctionPageViewController === self
        } else {
            return false
        }
    }

    func becomeUndoMager() {
        guard let currentCorrection = self.viewControllers?.first as? CorrectionViewController,
              let currentCanvas = currentCorrection.loadingViewController.contentViewController else { return }

        currentCanvas.updateUndoManager()
    }
}
