// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Logging

class InteractivePageDataSource: NSObject, UIPageViewControllerDataSource {
    typealias LoadingVC = CorrectionLoadingWrapperViewController
    let correctionController: InteractiveCorrectionPromiseControllerType
    var correctionContext: CorrectionContext?

    weak var target: UIPageViewController?

    let pageContainerCache: CorrectionPromiseViewControllerCache<LoadingVC>
    let pageCache: PageViewControllerCache<CorrectionViewController>

    let prelockCount: Int = 2
    let postlockCount: Int = 1

    private var maxIndex: Int = 0 {
        didSet {
            guard maxIndex > oldValue else { return }
            prefetchIfNecessary()
        }
    }
    var currentIndex: Int

    init(correctionController: InteractiveCorrectionPromiseControllerType) {
        self.correctionController = correctionController

        self.pageContainerCache = CorrectionPromiseViewControllerCache()
        self.pageCache = PageViewControllerCache()

        currentIndex = 0

        super.init()

        pageContainerCache.creation = { [weak self] promise in
            guard let self = self else { return nil }
            let resource = PromiseLoadingResource(promise: promise)
            let cache = self.pageCache
            let context = self.correctionContext
            return LoadingVC(resource) { result -> CorrectionPageViewController in
                CorrectionPageViewController(
                    correction: result.state.value.correction!,
                    cache: cache,
                    context: context,
                    lockReference: result.takeLockOwnership()
                )
            }
        }

        pageCache.creation = { [weak self] page in
            guard let self = self else { return nil }
            let pageVC = CorrectionViewController(page: page, context: self.correctionContext)
            page.prefetch()
            return pageVC
        }
    }

    func start(on pageViewController: UIPageViewController) {
        guard let promise = correctionController.getNextCorrection() else {
            fatalError()
        }
        guard let viewController = pageContainerCache.viewController(for: promise) else {
            fatalError()
        }
        pageViewController.setViewControllers([viewController], direction: .forward, animated: false, completion: nil)
        pageViewController.dataSource = self
        self.target = pageViewController

        prefetchIfNecessary()
    }

    private func indexOf(promise: CorrectionPromiseType) -> Int? {
        return correctionController.correctionPromises.firstIndex { $0 === promise }
    }

    private func indexOf(viewController: UIViewController) -> Int? {
        guard let viewController = viewController as? LoadingVC else { return nil }
        guard let promiseResource = viewController.loadingResource as? PromiseLoadingResource else { return nil }
        return indexOf(promise: promiseResource.promise)
    }

    func pageViewController(_ pageViewController: UIPageViewController,
                            viewControllerBefore viewController: UIViewController) -> UIViewController? {
        guard let index = indexOf(viewController: viewController) else { return nil }
        guard index > 0 else { return nil }

        let before = correctionController.correctionPromises[index - 1]
        return pageContainerCache.viewController(for: before)
    }

    func pageViewController(_ pageViewController: UIPageViewController,
                            viewControllerAfter viewController: UIViewController) -> UIViewController? {
        guard var index = indexOf(viewController: viewController) else {
            fatalError()
        }
        index += 1
        if index < correctionController.correctionPromises.count {
            let after = correctionController.correctionPromises[index]
            Logger.correctionUI.debug("Interactive got (index=\(index)): \(pointerString(after as AnyObject))")
            after.prefetch(problem: correctionController.problem)
            return pageContainerCache.viewController(for: after)
        } else {
            guard let promise = correctionController.getNextCorrection() else { return nil }
            Logger.correctionUI.debug("Interactive got (new): \(pointerString(promise as AnyObject))")
            promise.prefetch(problem: correctionController.problem)
            return pageContainerCache.viewController(for: promise)
        }
    }

    func viewControllerDidChange(to viewController: UIViewController) {
        guard let index = indexOf(viewController: viewController) else {
            fatalError()
        }

        currentIndex = index

        let toUnlock = index - (postlockCount + 1)
        if toUnlock >= 0 {
            let promise = correctionController.correctionPromises[toUnlock]
            if let viewController = pageContainerCache.viewController(for: promise, createIfNecessary: false) {
                viewController.contentViewController?.vacateLock()
            }
        }

        maxIndex = max(index, maxIndex)
    }

    func prefetchIfNecessary() {
        // iterate until the prelockCount has been reached
        while maxIndex + prelockCount > correctionController.correctionPromises.count {
            // if no more promises -> return
            guard let promise = correctionController.getNextCorrection() else { return }
            promise.prefetch(problem: correctionController.problem)
            Logger.correctionUI.debug("prelocked \(pointerString(promise as AnyObject))")
        }
    }
}

extension InteractivePageDataSource: OuterCorrectionDataSource {
    var supportsSubproblemSelection: Bool {
        true // always supporting suproblem mode
    }

    var currentPageViewController: CorrectionPageViewController? {
        if let viewController = target?.viewControllers?.first as? LoadingVC {
            return viewController.contentViewController
        }
        return nil
    }

    func getChangedERIDs() -> [String] {
        // get unsaved corrections
        let unsaved = correctionController.correctionPromises
            .compactMap { $0.state.value.correction }
            .filter(true, for: \.hasChanges)
            .map { $0.erid }

        return unsaved
    }

    func didReceiveMemoryWarning() {
        Logger.correctionUI.info("clearing page cache and page container cache due to memory warning")
        pageCache.removeAllObjects()
        pageContainerCache.removeAllObjects()
    }

    func createSubproblemDataSource(rect: RectOnPage, page: Int) -> SubproblemDataSource {
        SubproblemInteractiveDataSource(rect: rect, page: page, correctionController: correctionController)
    }
}
