// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Combine
import Instructions

protocol OuterCorrectionDataSource: AnyObject {
    var correctionContext: CorrectionContext? { get set }

    var currentPageViewController: CorrectionPageViewController? { get }

    var supportsSubproblemSelection: Bool { get }

    var currentIndex: Int { get }

    func getChangedERIDs() -> [String]

    func start(on pageViewController: UIPageViewController)

    func viewControllerDidChange(to viewController: UIViewController)

    func didReceiveMemoryWarning()

    func createSubproblemDataSource(rect: RectOnPage, page: Int) -> SubproblemDataSource
}

class OuterCorrectionPageViewController: UIPageViewController {
//	enum NavigationTitle: Equatable {
//		case title(String)
//		case subtitle(String, String)
//	}

	let correctionContext: CorrectionContext

	let startIndex: Int

    let correctionDataSource: OuterCorrectionDataSource

//	var indexedDataSource: IndexedPageViewDataSource<CorrectionType, CorrectionPageViewController>?

	var stateObserver: AnyCancellable?
	var modeObserver: AnyCancellable?
	var undoManagerObserver: AnyCancellable?

	var pencilInteraction: UIPencilInteraction?

	var toastViewController: ToastViewController?

	var subproblemToast: Toast?

    var activeSavingHandlers: [SubproblemSavingHandler] = []

    // Toast Bubbles
    let toastBubble: ToastBubbleView = ToastBubbleView()

    var currentCorrectionViewController: CorrectionRepresentingViewController? {
        guard let current = self.viewControllers?.first
        else { return nil }
        if let current = current as? CorrectionRepresentingViewController {
            return current
        }
        return nil
    }

    init(dataSource: OuterCorrectionDataSource, startIndex: Int = 0) {
		self.correctionContext = CorrectionContext()
		self.startIndex = startIndex

        self.correctionDataSource = dataSource

		super.init(
			transitionStyle: .scroll,
			navigationOrientation: .vertical,
			options: [.interPageSpacing: 32.0 as NSNumber]
		)

        self.correctionContext.useSharedUndoManager = false

        self.correctionDataSource.correctionContext = self.correctionContext
		self.modalPresentationStyle = .overFullScreen
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override func viewDidLoad() {
		super.viewDidLoad()

		edgesForExtendedLayout = []

		// Do any additional setup after loading the view.
		self.view.backgroundColor = Asset.correctionBackground.color
		self.delegate = self

		disablePencilForScrolling()

		setupObservers()
		setupPencilInteraction()

        correctionDataSource.start(on: self)
	}

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        correctionDataSource.didReceiveMemoryWarning()
    }

	func setupObservers() {
		modeObserver = correctionContext
			.$mode
			.receiveOnMain()
			.sink { [weak self] mode in
				self?.updateMode(mode)
		}
	}

	// MARK: - Private functions

	private func createSubproblemToast() -> Toast {
		Toast(.info, title: L10n.Toast.SubproblemMode.title, subtitle: L10n.Toast.SubproblemMode.subtitle)
	}

	func updateMode(_ mode: CorrectionContext.Mode) {
		self.pageScrollView?.isScrollEnabled = (mode == .normal)

		// Subproblem Correction Mode?
//		self.toolbarManager?.updateSubproblemCorrectionMode(mode == .subproblemSelection)
		if mode == .subproblemSelection {
			let toast = self.subproblemToast ?? createSubproblemToast()
			self.subproblemToast = toast
			show(toast)
		} else {
			if let toast = self.subproblemToast, toast.isBeingPresented {
				toast.dismiss()
			}
		}
	}
}

extension OuterCorrectionPageViewController: CorrectionViewSubproblemSelectionElementDelegate {
    func subproblemSelection(_ element: CorrectionViewSubproblemSelectionElement,
                             shouldStartUsing rect: RectOnPage,
                             page: Int,
                             problems: [ProblemIdentifier]?) {
        if let problems = problems,
           let dataSource = self.correctionDataSource as? InteractivePageDataSource {
            App.shared.coordinator?.startSubproblemCorrection(
                using: correctionDataSource,
                rect: rect,
                page: page,
                problems: problems,
                pass: dataSource.correctionController.pass,
                target: self
            )
        } else {
            App.shared.coordinator?.startSubproblemCorrection(
                using: correctionDataSource,
                rect: rect,
                page: page,
                startIndex: correctionDataSource.currentIndex,
                target: self)
        }
		// disable subproblem selection mode
		correctionContext.setMode(.normal)
	}
}

extension OuterCorrectionPageViewController: DefaultToastPresentingViewController {
    func show(_ toast: Toast) {
        switch toast.style {
        case .error:
            let viewController = ensureToastViewController()
            viewController.add(toast)
        case .info:
            guard toast.accessory == .activityIndicator else {
                let viewController = ensureToastViewController()
                viewController.add(toast)
                return
            }
            // Activity
            toastBubble.add(toast)
            guard let nav = parent(conformingTo: UINavigationController.self) else {
                return
            }
            toastBubble.present(in: nav.view)
        default:
            return
        }
    }
}

extension OuterCorrectionPageViewController: DefaultSavingManager { }

extension OuterCorrectionPageViewController: EditorContentViewController {
    var leftToolbarMode: EditorToolbarMode {
        .standard
    }

    var supportsSubproblemCorrection: Bool {
        correctionDataSource.supportsSubproblemSelection
    }

    var navigationTitle: String? { nil }
}
