// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Combine

class StaticPageDataSource {
    var correctionContext: CorrectionContext?
    let correctionController: CorrectionControllerType
    var stateObserver: AnyCancellable?

    let startIndex: Int
    var currentIndex: Int

    let postlockCount: Int = 1

    let pageContainerCache: CorrectionViewControllerCache<CorrectionPageViewController>
    let pageCache: PageViewControllerCache<CorrectionViewController>

    var indexedDataSource: IndexedPageViewDataSource<CorrectionType, CorrectionPageViewController>?

    private var isInEditMode: Bool = false

    weak var target: UIPageViewController?

    init(correctionController: CorrectionControllerType, startIndex: Int = 0) {
        self.correctionController = correctionController
        self.startIndex = startIndex
        self.currentIndex = startIndex

        self.pageContainerCache = CorrectionViewControllerCache()
        self.pageCache = PageViewControllerCache()

        pageContainerCache.creation = { [weak self] correction in
            guard let self = self else { return nil }
            return CorrectionPageViewController(
                correction: correction,
                cache: self.pageCache,
                context: self.correctionContext
            )
        }
        pageCache.creation = { [weak self] page in
            guard let self = self else { return nil }
            let pageVC = CorrectionViewController(page: page, context: self.correctionContext)
            page.prefetch()
            return pageVC
        }
    }

    func switchToEditMode() {
        guard let target = target else { return }
        isInEditMode = true
        if correctionController.corrections.count > 0 {
            // create new indexed data source with indexed access to the corrections
            // and a simple creation block for the view controller
            let indexedDataSource = IndexedPageViewDataSource(
                for: target,
                access: { [weak self] index in self?.correction(at: index) },
                creation: { [weak self] in
                    return self?.pageContainerCache.viewController(for: $0)
                }
            )
            self.indexedDataSource = indexedDataSource
            target.dataSource = indexedDataSource

            indexedDataSource.show(index: startIndex, animated: false)
        } else {
//            show(Toast(.error, title: "No Corrections found"))
        }
    }

    func correction(at index: Int) -> CorrectionType? {
        let corrections = correctionController.corrections
        guard index >= 0, index < corrections.count else { return nil }
        return corrections[index]
    }
}

extension StaticPageDataSource: OuterCorrectionDataSource {
    var supportsSubproblemSelection: Bool {
        correctionController.filter.problem != nil && correctionController.filter.pass != nil
    }

    var currentPageViewController: CorrectionPageViewController? {
        target?.viewControllers?.first as? CorrectionPageViewController
    }

    func getChangedERIDs() -> [String] {
        let unsaved = correctionController.corrections.filter(true, for: \.hasChanges)
        return unsaved.map { $0.erid }
    }

    func start(on pageViewController: UIPageViewController) {
        self.target = pageViewController
        observeState(correctionController)
        correctionController.load()
    }

    func viewControllerDidChange(to viewController: UIViewController) {
        guard let viewController = viewController as? IndexRepresentingViewController else { return }
        guard let index = viewController.representingIndexInCollection else { return }
        self.currentIndex = index
        let currentCount = correctionController.corrections.count
        if index >= currentCount - 1 {
            correctionController.loadMore()
        }

        let toUnlock = index - (postlockCount + 1)
        if toUnlock >= 0 {
            let correction = correctionController.corrections[toUnlock]
            if let viewController = pageContainerCache.viewController(for: correction, createIfNecessary: false) {
                viewController.vacateLock()
            }
        }
    }

    func didReceiveMemoryWarning() {
        pageCache.removeAllObjects()
        pageContainerCache.removeAllObjects()
    }

    func createSubproblemDataSource(rect: RectOnPage, page: Int) -> SubproblemDataSource {
        SubproblemStaticDataSource(rect: rect, page: page, correctionController: correctionController)
    }
}

extension StaticPageDataSource: StateObserving {
    func stateChanged(_ state: ResourceObjectState) {
        // only listen to state changes if we're not yet in edit mode
        guard !isInEditMode else {
            return
        }

        switch state {
        case .loading:
            let loading = LoadingViewController()
            loading.style = .clear
            target?.setViewControllers([loading], direction: .forward, animated: false, completion: nil)
        case .loaded:
            switchToEditMode()
        default:
            break
        }
    }
}
