// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import TUMexamKit
import Combine
import UIKit

/// this class manages the presentation, dismissal, etc of the toasts and alerts for a save operation of a correction
class SubproblemSavingHandler {
    typealias CompletionBlock = (SubproblemSavingHandler) -> Void
    let correction: CorrectionType
    var lockReference: LockReferenceType?
    weak var target: ToastPresentingViewController?
    var completion: CompletionBlock?

    var observer: AnyCancellable?

    var savingToast: Toast?

    init(correction: CorrectionType, target: ToastPresentingViewController, completion: CompletionBlock?) {
        self.correction = correction
        self.target = target
        self.completion = completion
    }

    func start() {
        self.observer = correction.submissionState.receiveOnMain().sink { [weak self] (state) in
            self?.savingStateChanged(state)
        }

        lockReference = correction.addLockRequirement()
        correction.submissionCompletion = { [weak self] result in
            switch result {
            case .failure(let error):
                DispatchQueue.main.async {
                    self?.savingFailed(error)
                }
            case .success:
                DispatchQueue.main.async {
                    self?.savingSuccessful()
                }
            }
        }
        correction.submitChanges()
    }

    func savingStateChanged(_ state: ResourceObjectState) {
        switch state {
        case .idle:
            savingIdle()
        case .loading:
            showSavingToast()
        case .loaded:
            savingSuccessful()
        case let .error(error):
            savingFailed(error)
        }
    }

    func savingIdle() {
        if let toast = savingToast {
            toast.dismiss()
        }
    }

    func savingSuccessful() {
        savingIdle()

        correction.submissionCompletion = nil

        let toast = Toast(
            .success,
            title: L10n.Toast.SavingDone.title,
            subtitle: correction.erid,
            dismissable: true)
        target?.show(toast)
        toast.dismiss(after: 3.0)

        completion?(self)

        lockReference = nil
    }

    func showSavingToast() {
        guard savingToast == nil else { return }
        let toast = Toast(
            .info,
            title: L10n.Toast.Saving.title,
            subtitle: correction.erid,
            accessory: .activityIndicator,
            dismissable: false)
        savingToast = toast
        DispatchQueue.main.async {
            self.target?.show(toast)
        }
    }

    func savingFailed(_ error: Error) {
        // TODO: special handling when no lock
        // create the alert
        let alert = UIAlertController(
            title: L10n.CorrectionController.Alert.SavingError.title(correction.erid),
            message: error.localizedDescription,
            preferredStyle: .alert
        )

        // add 'cancel' option -> discard
        alert.addAction(
            UIAlertAction(
                title: L10n.CorrectionController.Alert.SavingError.cancel,
                style: .destructive,
                handler: { [weak self] _ in
                    // do nothing (maybe reset correction)
                    self?.cancelSaving()
                })
        )

        // add 'retry' option -> submit changes
        alert.addAction(
            UIAlertAction(
                title: L10n.CorrectionController.Alert.SavingError.retry,
                style: .default,
                handler: { [weak self] _ in
                    self?.correction.submitChanges()
                })
        )

        target?.present(alert, animated: true, completion: nil)
    }

    func cancelSaving() {
        savingToast?.dismiss()
        savingToast = nil
        correction.submissionCompletion = nil
        completion?(self)
        lockReference = nil
    }
}
