// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import TUMexamKit
import UIKit

/// this protocol can be used by sub-viewcontrollers in the hierachy to delegate their saving process to an containing manager.
/// if we'd just manage the saving notifications/alerts/etc in the actual view controller, we wouldn't have any lifetime guarantees
protocol SavingManager: AnyObject {
    /// tells the manager to save the correction
    /// - Parameter correction: the correction to save
    func save(correction: CorrectionType)
}

protocol DefaultSavingManager: ToastPresentingViewController, SavingManager {
    var activeSavingHandlers: [SubproblemSavingHandler] { get set }
}

extension DefaultSavingManager {
    func save(correction: CorrectionType) {
        guard correction.hasChanges else { return }

        // don't add twice
        guard !activeSavingHandlers.contains(where: { $0.correction === correction }) else { return }

        // create and add handler
        let handler = SubproblemSavingHandler(correction: correction, target: self) { [weak self] (savingHandler) in
            // when done -> remove
            self?.savingHandlerDone(savingHandler)
        }
        activeSavingHandlers.append(handler)

        // start saving
        handler.start()
    }

    func savingHandlerDone(_ handler: SubproblemSavingHandler) {
        activeSavingHandlers.removeAll { $0 === handler }
    }

    func notifyNothingSaved(correction: CorrectionType) {
        let toast = Toast(.info,
                          title: L10n.Toast.NothingSaved.title,
                          subtitle: L10n.Toast.NothingSaved.subtitle(correction.erid),
                          accessory: .symbol("exclamationmark.triangle"),
                          action: .none,
                          dismissable: false)
        self.show(toast)
        toast.dismiss(after: 3.0)
    }
}
