// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import Foundation
import os.log
import TUMexamKit
import UIKit

final class SolutionPageViewController: UIPageViewController, IndexRepresentingViewController {

//	var toolbarManager: ToolbarManager?
	var undoManagerObserver: AnyCancellable?
    var pencilInteraction: UIPencilInteraction?

	private var submissionObserver: AnyCancellable?

	// MARK: - DataSource

	var arrayDataSource: ArrayDataSource?
	var representingIndexInCollection: Int?

	// MARK: - Default Saving Manager
	var toastViewController: ToastViewController?
	var savingToast: Toast?

	// MARK: - UIPageView Properties

	var isCurrentlyAppearing: Bool = false
	var switchToPageWhenAppearing: Bool = false

	private var currentCorrectionViewController: CorrectionViewController? {
		viewControllers?.first as? CorrectionViewController
	}

	internal var solution: SolutionType? {
		App.shared.currentExam?.solution
	}

	let correctionContext: CorrectionContext

	private let pageSize: CGSize

	/**
	This initializer is temporary.
	*/
	init(vc: CorrectionViewController, context: CorrectionContext) {
		self.correctionContext = context
		self.pageSize = vc.page.pageSize
		super.init(
			transitionStyle: .scroll,
			navigationOrientation: .horizontal,
			options: [.interPageSpacing: 32.0 as NSNumber]
		)
		createDataSource(startingWith: vc)

		setViewControllers([vc], direction: .forward,
						   animated: false,
						   completion: nil)
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override func viewDidLoad() {
		super.viewDidLoad()
		view.backgroundColor = Asset.correctionBackground.color
		edgesForExtendedLayout = []

		disablePencilForScrolling()
//		setupToolbarManager()

		setCustomBackButton()
        setupPencilInteraction()

        submissionObserver = App.shared.currentExam?.solution.submissionState
			.receive(on: DispatchQueue.main)
			.sink(receiveValue: { [weak self] state in
			switch state {
			case let .error(error):
				if case let NetworkError.rejected(pages) = error {
					let readableString = pages.map { "\($0.page)" }.joined(separator: ", ")

					let msg: String
					if pages.count == 1 {
						msg = L10n.Solution.Submission.Rejected.messageSingle(readableString)
					} else {
						msg = L10n.Solution.Submission.Rejected.message(readableString)
					}
					let title = L10n.Solution.Submission.Rejected.title

					let alert = UIAlertController(title: title,
												  message: msg,
												  preferredStyle: .alert)
					let cancel = UIAlertAction(title: L10n.cancel, style: .cancel) { _ in
						App.shared.currentExam?.solution.resetSubmission()
					}

					let diffTitle = L10n.Solution.Submission.Rejected.inspect
					let diff = UIAlertAction(title: diffTitle, style: .default) { _ in
						App.shared.coordinator?.startDiffOfPages(pages)
					}

					alert.addAction(cancel)
					alert.addAction(diff)
					DispatchQueue.main.async {
						alert.show(animated: true)
					}
					print("The following pages were rejected: \(pages)")
				} else {
					print("error submitting pages \(error)")
				}
				self?.savingToast?.dismiss()
				self?.savingToast = nil
			case .loading:
				self?.showSavingToast()
			case .loaded:
				self?.savingToast?.dismiss()
				self?.savingToast = nil
                self?.navigationController?.popToRootViewController(animated: true)
			default:
				break
			}
		})
	}

	deinit {
		submissionObserver?.cancel()
	}

	private func showSavingToast() {
		guard savingToast == nil else { return }
		guard let subtitleString = self.solution?.pages.value
				.filter(\.hasChanges)
				.map({ "\($0.pageNumber)" })
				.joined(separator: ", ") else { return }

		let toast = Toast(
			.info,
			title: L10n.Toast.Saving.title,
			subtitle: L10n.Toast.Saving.Solution.subtitle(subtitleString),
			accessory: .activityIndicator,
			dismissable: false)
		savingToast = toast
		show(toast)
	}

	private func setCustomBackButton() {
		navigationItem.setHidesBackButton(true, animated: false)

		let button = UIBarButtonItem(title: L10n.CorrectionController.dismiss,
									 style: .done,
									 target: self, action: #selector(onDismiss))
		navigationItem.leftBarButtonItem = button
	}

	@objc
	private func onDismiss() {
		toolbarWantsToDismiss()
	}

	// MARK: - Page Display

	func createDataSource(startingWith vc: CorrectionViewController) {
		let currentPageIndex = correctionContext.currentPageIndex ?? 0

		var allPages = solution?.pages.value.compactMap {
			CorrectionViewController(page: $0, context: correctionContext)
		} ?? []

		guard allPages.count - 1 >= currentPageIndex else { return }

		allPages[currentPageIndex] = vc

		let dataSource = ArrayDataSource(allPages)
		arrayDataSource = dataSource
		self.dataSource = dataSource
		self.delegate = self

		// show the current page (based on context) or fallback to 0 (if no context available)
		showPage(at: currentPageIndex)
	}

	func showPage(at index: Int) {
		if let current = arrayDataSource?.viewController(at: index) {
			self.setViewControllers([current], direction: .forward, animated: false, completion: nil)
		}
	}

	override func viewDidAppear(_ animated: Bool) {
		super.viewDidAppear(animated)

		if let target = parent(conformingTo: CorrectionTitleTarget.self) {
			target.updateTitle()
		}

		guard let current = self.viewControllers?.first as? CorrectionViewController else {
			return
		}
		correctionContext.currentPageIsEmpty = current.empty

	}

	func setCurrentPageEmpty(_ empty: Bool) {
		currentCorrectionViewController?.empty = empty
	}
}

extension SolutionPageViewController: UIPageViewControllerDelegate {
	func pageViewController(
		_ pageViewController: UIPageViewController,
		didFinishAnimating finished: Bool,
		previousViewControllers: [UIViewController],
		transitionCompleted completed: Bool) {
		guard let current = self.viewControllers?.first as? CorrectionViewController else {
			return
		}

		correctionContext.currentPageIsEmpty = current.empty
	}
}

extension SolutionPageViewController: ContextUndoManagerCoordinator {
	func shouldChangeUndoManager(to canvasView: CorrectionCanvasViewController) -> Bool {
		guard let currentCorrection = currentCorrectionViewController,
			  let currentCanvas = currentCorrection.loadingViewController.contentViewController else { return false }
		return canvasView === currentCanvas
	}

	func becomeUndoMager() {}
}

extension SolutionPageViewController: DefaultToastPresentingViewController {}
