// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Logging

protocol SubproblemPageContentViewControllerDelegate: AnyObject {
    func subproblemPageContent(_ viewController: SubproblemPageContentViewController,
                               didChangeCorrectionCount correctionCount: Int)

    func subproblemPageContent(_ viewController: SubproblemPageContentViewController,
                               correctionControllerAtIndex index: Int) -> SubproblemCanvasWrapperViewController?
}

class SubproblemPageContentViewController: UIViewController {
    static let spacing: CGFloat = 20.0
    static let landscapeCompactWidth: CGFloat = 0.5

    let rect: RectOnPage
    let stackView = UIStackView().autolayout()
    var widthConstraint: NSLayoutConstraint?
    var lastBounds: CGRect = .zero

    let topSpacer = UIView().autolayout()
    let bottomSpacer = UIView().autolayout()

    var correctionViewControllers: [SubproblemCanvasWrapperViewController] = []

    var correctionCount: Int {
        correctionViewControllers.count
    }

    let contextUndoManager = ContextUndoManager()
    var resetUndoManagerWhenAppearing: Bool = false

    var isActive: Bool = false

    var isLast: Bool = false
    var possibleCorrectionCount: Int {
        let compactWidth = calculateCompactWidth()
        return calculateCorrectionCount(bounds: view.bounds, widthFactor: compactWidth)
    }

    var startIndex: Int {
        didSet {
            guard oldValue != startIndex else { return }
            Logger.correctionUI.debug("[Sub \(oldValue)] becomes [Sub \(startIndex)]")
            recreateStack()
        }
    }

    private var doubleTapGesture: UITapGestureRecognizer?

    weak var delegate: SubproblemPageContentViewControllerDelegate?

    init(rect: RectOnPage, startIndex: Int = 0) {
        self.rect = rect
        self.startIndex = startIndex

        super.init(nibName: nil, bundle: nil)

        Logger.correctionUI.debug("[Sub \(startIndex)] initialized \(String(describing: self))")
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        stackView.axis = .vertical
        stackView.spacing = Self.spacing
        stackView.alignment = .fill
        stackView.distribution = .equalSpacing

        stackView.addArrangedSubview(topSpacer)
        stackView.addArrangedSubview(bottomSpacer)

        view.addSubview(stackView)

        NSLayoutConstraint.activate([
            stackView.topAnchor.constraint(equalTo: view.topAnchor),
            stackView.bottomAnchor.constraint(equalTo: view.bottomAnchor),
            stackView.centerXAnchor.constraint(equalTo: view.centerXAnchor),

            topSpacer.heightAnchor.constraint(equalToConstant: 0),
            bottomSpacer.heightAnchor.constraint(equalToConstant: 0)
        ])

        let doubleTap = UITapGestureRecognizer(target: self, action: #selector(didDoubleTap(_:)))
        doubleTap.numberOfTapsRequired = 2
        doubleTap.cancelsTouchesInView = false
        doubleTap.swiftAllowedTouchTypes = [.direct]
        view.addGestureRecognizer(doubleTap)
        doubleTapGesture = doubleTap
    }

    override func viewWillLayoutSubviews() {
        super.viewWillLayoutSubviews()

        updateIfNecessary()
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)

        if resetUndoManagerWhenAppearing {
            contextUndoManager.reset()
            resetUndoManagerWhenAppearing = false
        }
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)

        isActive = true
        updateIfNecessary()
    }

    override func viewDidDisappear(_ animated: Bool) {
        super.viewDidDisappear(animated)
        save()
        isActive = false
    }

    @objc func didDoubleTap(_ recognizer: UITapGestureRecognizer) {
        let location = recognizer.location(in: view)

        for correctionViewController in correctionViewControllers {
            if correctionViewController.view.frame.contains(location) {
                guard let correction = correctionViewController.contentViewController?.page.containingCorrection else {
                    return
                }

                // mark undo manager for reset
                resetUndoManagerWhenAppearing = true

                // show full correction
                if let pageNumber = correctionViewController.contentViewController?.page.pageNumber {
                    App.shared.coordinator?.showCorrection(
                        correction,
                        target: self,
                        pageIndex: pageNumber - 1
                    )
                } else {
                    App.shared.coordinator?.showCorrection(
                        correction,
                        target: self
                    )
                }
                return
            }
        }
    }

    func updateWidth(percent: CGFloat = 1.0) {
        if let widthConstraint = widthConstraint {
            widthConstraint.isActive = false
            self.widthConstraint = nil
        }

        let width = stackView.widthAnchor.constraint(equalTo: view.widthAnchor, multiplier: percent)
        widthConstraint = width
        width.isActive = true
    }

    func calculateCompactWidth() -> CGFloat {
        return 0.9
    }

    func calculateCorrectionCount(bounds: CGRect, widthFactor: CGFloat) -> Int {
        let aspectRatio = rect.aspectRatio

        let targetSize = CGSize(width: (bounds.width * widthFactor), height: bounds.size.height)
        let singleCorrectionHeight = targetSize.width / aspectRatio
        let count = (targetSize.height - stackView.spacing) / (singleCorrectionHeight + stackView.spacing)

        let deviceMemory = ProcessInfo.processInfo.physicalMemory
        let maxCountMemoryBased = max(1, (deviceMemory * 10 / 6_442_450_944)) // 6 GB == 10 exams per page

        return min(max(1, Int(floor(count))), Int(maxCountMemoryBased))
    }

    func recreateStack() {
        let all = correctionViewControllers
        for viewController in all {
            removeChild(viewController)
        }
        correctionViewControllers = []

        updateVisibleViews()
    }

    func updateIfNecessary() {
        guard isActive else { return }
        Logger.correctionUI.debug("[Sub \(startIndex)] will layout \(view.bounds)")

        if lastBounds != view.bounds {
            updateVisibleViews()

            lastBounds = view.bounds
        }
    }

    func updateVisibleViews() {
        let compactWidth = calculateCompactWidth()

        Logger.correctionUI.debug("[Sub \(startIndex)] did layout \(view.bounds)")
        let count = calculateCorrectionCount(bounds: view.bounds, widthFactor: compactWidth)
        Logger.correctionUI.debug("[Sub \(startIndex)] counts: \(count) \(correctionViewControllers.count)")
        if count != correctionViewControllers.count {
            updateStackViewContent(count: count)

            // inform delegate
            delegate?.subproblemPageContent(self, didChangeCorrectionCount: correctionViewControllers.count)
        }

        updateWidth(percent: compactWidth)
    }

    func updateStackViewContent(count: Int) {
        while correctionViewControllers.count > count {
            let last = correctionViewControllers.removeLast()
            removeChild(last)
            Logger.correctionUI.debug("[Sub \(startIndex)] removed vc (should \(count), is \(correctionViewControllers.count))")
        }

        while correctionViewControllers.count < count {
            let index = startIndex + correctionViewControllers.count
            guard let viewController = delegate?.subproblemPageContent(self, correctionControllerAtIndex: index) else {
                // important: break out of while loop if no further vcs can be created
                Logger.correctionUI.error("[Sub \(startIndex)] could not get vc with index \(index)")

                if correctionViewControllers.isEmpty && !isLast {
                    // Couldnt create ANY controllers -> probably reached end
                    isLast = true
                    let endVC = EndViewController()
                    addChild(endVC) { view in
                        view.translatesAutoresizingMaskIntoConstraints = false
                        self.view.addSubview(view)
                        view.pinToSuperview()
                    }
                }

                break
            }

            correctionViewControllers.append(viewController)

            addChild(viewController) { (view) in
                let insertAt = max(stackView.arrangedSubviews.count - 1, 0)
                stackView.insertArrangedSubview(view, at: insertAt)
                Logger.correctionUI.debug("[Sub \(startIndex)] Inserted vc in stack view \(pointerString(viewController)) for index \(index)")
            }

            Logger.correctionUI.debug("[Sub \(startIndex)] Added vc (should \(count), is \(correctionViewControllers.count))")
        }
    }

    func save() {
        guard let savingManager = self.parent(conformingTo: SavingManager.self) else { fatalError() }
        for controller in correctionViewControllers {
            if let correction = controller.contentViewController?.page.containingCorrection {
                savingManager.save(correction: correction)
            }
        }
    }

    func undo() {
        contextUndoManager.undo()
    }

    func redo() {
        contextUndoManager.redo()
    }
}

extension SubproblemPageContentViewController: ContextUndoManagerProvider { }
