// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Logging

protocol SubproblemDataSource: AnyObject {
    var rect: RectOnPage { get }
    var currentCount: Int { get }
    var overallCount: Int? { get }

    var correctionContext: CorrectionContext { get }

    var title: String { get }

    func correctionController(atIndex index: Int) -> SubproblemCanvasWrapperViewController?
}

class SubproblemInteractiveDataSource {
    let rect: RectOnPage
    let page: Int
    let correctionController: InteractiveCorrectionPromiseControllerType
    let correctionContext = CorrectionContext()

    var cache: CorrectionPromiseViewControllerCache<SubproblemCanvasWrapperViewController>

    private var correctionPromises: [CorrectionPromiseType] { correctionController.correctionPromises }

    init(rect: RectOnPage, page: Int, correctionController: InteractiveCorrectionPromiseControllerType) {
        self.rect = rect
        self.page = page
        self.correctionController = correctionController
        self.cache = CorrectionPromiseViewControllerCache()

        cache.creation = { [weak self] promise in
            guard let self = self else { return nil }

            let resource = PromisePageLoadingResource(promise: promise, pageNumber: self.page, rect: self.rect)
            let viewController = SubproblemCanvasWrapperViewController(
                resource
            ) { [weak self] (page, promise) -> CorrectionCanvasViewController in
                    guard let self = self else { fatalError() }
                    let correctionVC = CorrectionCanvasViewController(
                        page: page,
                        normalizedPageRect: self.rect.normalizedRect,
                        correctionContext: self.correctionContext,
                        lockReference: promise?.takeLockOwnership()
                    )
                    correctionVC.addDefaultElements()
                    correctionVC.addERIDElement()
                    return correctionVC
                }

            let aspectRatio = self.rect.aspectRatio
            if let view = viewController.view {
                view.widthAnchor.constraint(equalTo: view.heightAnchor, multiplier: aspectRatio).isActive = true
            }

            return viewController
        }
    }
}

extension SubproblemInteractiveDataSource: SubproblemDataSource {
    var currentCount: Int {
        correctionController.correctionPromises.count
    }

    var title: String {
        L10n.SubproblemController.PagePass.title("\(page)", "\(correctionController.pass)")
    }

    var overallCount: Int? {
        return nil
    }

    func correctionController(atIndex index: Int) -> SubproblemCanvasWrapperViewController? {
        // out of bounds -> create new promise
        if index >= correctionPromises.count {
            Logger.correctionUI.debug("[SubIntDataSource] no correction controller for index \(index) in promises, trying to getNextCorrection()")
            if let correction = correctionController.getNextCorrection() {
                Logger.correctionUI.debug("[SubIntDataSource] got getNextCorrection()")
                let viewController = cache.viewController(for: correction)
                Logger.correctionUI.debug("[SubIntDataSource] resulting view controller \(pointerString(viewController)) for promise \(correction.identifier)")
                return viewController
            } else {
                Logger.correctionUI.debug("[SubIntDataSource] no next corrections found, no more available :(")
                // no more corrections available
                return nil
            }
        }

        // in bounds -> use existing promise
        Logger.correctionUI.debug("[SubIntDataSource] correction promise already exists in array for index \(index)")
        let correction = correctionPromises[index]
        let viewController = cache.viewController(for: correction)
        Logger.correctionUI.debug("[SubIntDataSource] resulting view controller \(pointerString(viewController)) for promise \(correction.identifier)")
        return viewController
    }
}
