// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Combine
import Logging

class SubproblemPageViewController: UIPageViewController {
    var correctionDataSource: SubproblemDataSource

    var correctionContext: CorrectionContext { correctionDataSource.correctionContext }

//	var toolbarManager: ToolbarManager?
	var pencilInteraction: UIPencilInteraction?

	var toastViewController: ToastViewController?

	var undoManagerObserver: AnyCancellable?

    var activeSavingHandlers: [SubproblemSavingHandler] = []

    let toastBubble: ToastBubbleView = ToastBubbleView()

    let startIndex: Int

    init(dataSource: SubproblemDataSource, startIndex: Int = 0) {
        self.correctionDataSource = dataSource
        self.startIndex = startIndex

		super.init(transitionStyle: .scroll, navigationOrientation: .vertical, options: nil)
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

    override func viewDidLoad() {
        super.viewDidLoad()

		// don't extend below navigation bar
		self.edgesForExtendedLayout = []

		view.backgroundColor = Asset.correctionBackground.color

        self.title = correctionDataSource.title

        let first = createViewController(startIndex: self.startIndex)
		setViewControllers([first], direction: .forward, animated: false, completion: nil)

		self.disablePencilForScrolling()
		self.dataSource = self
		self.delegate = self

		setupPencilInteraction()
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        parent(conformingTo: CorrectionTitleTarget.self)?.updateTitle()
    }

	func createViewController(startIndex: Int) -> SubproblemPageContentViewController {
        let viewController = SubproblemPageContentViewController(
            rect: correctionDataSource.rect,
            startIndex: startIndex
        )
		viewController.delegate = self
		return viewController
	}
}

extension SubproblemPageViewController: UIPageViewControllerDataSource {
	func pageViewController(_ pageViewController: UIPageViewController,
							viewControllerBefore viewController: UIViewController) -> UIViewController? {
		guard let current = viewController as? SubproblemPageContentViewController else { return nil }
		// make sure it's not the first VC
		guard current.startIndex > 0 else { return nil }

        if !current.isLast {
            let newStart = max(0, current.startIndex - current.correctionCount)
            return createViewController(startIndex: newStart)
        } else {
            // Last
            let newStart = max(0, current.startIndex - current.possibleCorrectionCount)
            return createViewController(startIndex: newStart)
        }
	}

	func pageViewController(_ pageViewController: UIPageViewController,
							viewControllerAfter viewController: UIViewController) -> UIViewController? {
		guard let current = viewController as? SubproblemPageContentViewController else { return nil }

        let newStart = current.startIndex + current.correctionCount
        if let overallCount = correctionDataSource.overallCount {
            if newStart < overallCount {
                return createViewController(startIndex: newStart)
            } else {
                return nil
            }
        } else if current.isLast {
            return nil
        } else {
            return createViewController(startIndex: newStart)
        }
	}
}

extension SubproblemPageViewController: UIPageViewControllerDelegate {
	func pageViewController(_ pageViewController: UIPageViewController,
							willTransitionTo pendingViewControllers: [UIViewController]) {
		guard let current = self.viewControllers?.first as? SubproblemPageContentViewController else { return }
		guard let target = pendingViewControllers.first as? SubproblemPageContentViewController else { return }

		if target.startIndex > current.startIndex {
			target.startIndex = current.startIndex + current.correctionCount
		} else {
            if current.isLast {
                target.startIndex = max(0, current.startIndex - current.possibleCorrectionCount)
            } else {
                target.startIndex = max(0, current.startIndex - current.correctionCount)
            }
		}

        Logger.correctionUI.debug("Subproblem PageVC will transition, current: \(current), target: \(target)")
	}

	func pageViewController(_ pageViewController: UIPageViewController,
							didFinishAnimating finished: Bool,
							previousViewControllers: [UIViewController],
							transitionCompleted completed: Bool) {
		// don't care
        parent(conformingTo: CorrectionTitleTarget.self)?.updateTitle()
	}
}

extension SubproblemPageViewController: SubproblemPageContentViewControllerDelegate {
	func subproblemPageContent(_ viewController: SubproblemPageContentViewController,
							   didChangeCorrectionCount correctionCount: Int) {

	}

	func subproblemPageContent(_ viewController: SubproblemPageContentViewController,
							   correctionControllerAtIndex index: Int) -> SubproblemCanvasWrapperViewController? {
        correctionDataSource.correctionController(atIndex: index)
	}
}

extension SubproblemPageViewController: DefaultToastPresentingViewController {
    func show(_ toast: Toast) {
        switch toast.style {
        case .error:
            let viewController = ensureToastViewController()
            viewController.add(toast)
        case .info:

            if toast.accessory != .activityIndicator {
                let viewController = ensureToastViewController()
                viewController.add(toast)
            } else {
                // Activity
                toastBubble.add(toast)
                guard let nav = parent(conformingTo: UINavigationController.self) else {
                    return
                }
                toastBubble.present(in: nav.view)
            }
        default:
            return
        }
    }
}

extension SubproblemPageViewController: DefaultSavingManager { }
