// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit

class SubproblemStaticDataSource {
    let rect: RectOnPage
    let page: Int
    let correctionController: CorrectionControllerType
    let correctionContext = CorrectionContext()

    var cache: CorrectionViewControllerCache<SubproblemCanvasWrapperViewController>

    private var corrections: [CorrectionType] { correctionController.corrections }

    init(rect: RectOnPage, page: Int, correctionController: CorrectionControllerType) {
        self.rect = rect
        self.page = page
        self.correctionController = correctionController
        self.cache = CorrectionViewControllerCache()

        cache.creation = { [weak self] correction in
            guard let self = self else { return nil }

            let resource = SubproblemPageLoadingResource(correction: correction, pageNumber: self.page, rect: self.rect)
            let viewController = SubproblemCanvasWrapperViewController(
                resource
            ) { [weak self] (page, _) -> CorrectionCanvasViewController in
                guard let self = self else { fatalError() }
                let correctionVC = CorrectionCanvasViewController(
                    page: page,
                    normalizedPageRect: self.rect.normalizedRect,
                    correctionContext: self.correctionContext,
                    lockReference: resource.lockReference
                )
                correctionVC.addDefaultElements()
                correctionVC.addERIDElement()
                return correctionVC
            }

            let aspectRatio = self.rect.aspectRatio
            if let view = viewController.view {
                view.widthAnchor.constraint(equalTo: view.heightAnchor, multiplier: aspectRatio).isActive = true
            }

            return viewController
        }
    }
}

extension SubproblemStaticDataSource: SubproblemDataSource {
    var currentCount: Int {
        corrections.count
    }

    var title: String {
        L10n.SubproblemController.PagePass.title("\(page)", "\(correctionController.filter.pass ?? 1)")
    }

    var overallCount: Int? {
        correctionController.estimatedCount
    }

    func correctionController(atIndex index: Int) -> SubproblemCanvasWrapperViewController? {
        guard index < corrections.count else { return nil }
        let correction = corrections[index]
        return cache.viewController(for: correction)
    }
}
