// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit

enum ToolbarItem: Int {
	case undo
	case redo
	case highlighter
	case pen
	case eraser
	case save
	case edit
	case expand
	case lasso
	case preview
	case pageCount
	case dismiss
    case flexibleSpace
    case separator
    case empty
    case subproblem
    case toolbarSettings

	var image: UIImage? {
		switch self {
		case .undo: return UIImage(systemName: "arrow.uturn.left")
		case .redo: return UIImage(systemName: "arrow.uturn.right")
		case .pen: return Asset.penIcon.image
		case .highlighter: return Asset.highlighterIcon.image
		case .eraser: return Asset.eraserIcon.image
		case .save, .edit, .preview, .pageCount, .dismiss: return nil
		case .expand: return UIImage(systemName: "arrow.left.and.right.square")
		case .lasso: return UIImage(systemName: "lasso")
        case .flexibleSpace: return nil
        case .separator: return nil
        case .empty: return Asset.emptySetIcon.image
        case .subproblem: return UIImage(systemName: "crop")
        case .toolbarSettings: return UIImage(systemName: "ellipsis.circle")
		}
	}

	func createSelectableBarButtonItem(_ onSelect: @escaping ((Bool) -> Void)) -> SelectableBarButtonItem {
		let button = SelectableBarButtonItem(with: self.image ?? UIImage(),
																	 onValueChange: onSelect)
		button.tag = self.rawValue
		return button
	}

	func createBarButtonItemAtTarget(_ target: Any?, withAction action: Selector?) -> UIBarButtonItem {
		let button: UIBarButtonItem
		switch self {
		case .edit:
            button = UIBarButtonItem(
                title: L10n.CorrectionController.editSolution,
                style: .plain,
                target: target,
                action: action
            )
		case .save:
			button = UIBarButtonItem(barButtonSystemItem: .save, target: target, action: action)
		case .dismiss:
			let title = L10n.CorrectionController.dismiss
			button =  UIBarButtonItem(title: title, style: .done, target: target, action: action)
		case .preview:
			let title = L10n.Solution.Submission.tbItemPreview
			button = UIBarButtonItem(title: title, style: .plain, target: target, action: action)
        case .flexibleSpace:
            button = UIBarButtonItem(systemItem: .flexibleSpace, primaryAction: nil, menu: nil)
            button.tag = 10
        case .separator:
            let sepBackground = UIView().autolayout()
            sepBackground.width(20).height(40)
            sepBackground.backgroundColor = .clear
            let sep = UIView().autolayout()
            sepBackground.addSubview(sep)
            sep.width(1).height(30).centerX(to: \.centerXAnchor).centerY(to: \.centerYAnchor)
            sep.backgroundColor = .systemGray4
            button = UIBarButtonItem(customView: sepBackground)
		default:
			button = UIBarButtonItem(image: self.image, style: .plain, target: target, action: action)
		}
		button.tag = self.rawValue
		return button
	}

    func createBarButtonItemWithMenu(_ menu: UIMenu) -> UIBarButtonItem {
        return UIBarButtonItem(image: self.image, menu: menu)
    }
}

struct ToolbarButtonConfiguration {
	let items: [ToolbarItem]

	func newConfigurationReplacingItem(_ itemToReplace: ToolbarItem,
									   with newItem: ToolbarItem) -> ToolbarButtonConfiguration? {
		var newItems = items
		guard let idx = items.firstIndex(of: itemToReplace) else { return nil }
		newItems.remove(at: idx)
		newItems.insert(newItem, at: idx)
		return ToolbarButtonConfiguration(items: newItems)
	}

	func appendItem(_ item: ToolbarItem) -> ToolbarButtonConfiguration {
		let newItems = items + [item]
		return ToolbarButtonConfiguration(items: newItems)
	}

	func insertingItem(_ item: ToolbarItem, atIndex index: Int) -> ToolbarButtonConfiguration? {
		guard items.count > index && index >= 0 else { return nil }
		var newItems = items
		newItems.insert(item, at: index)
		return ToolbarButtonConfiguration(items: newItems)
	}

	func popItem() -> ToolbarButtonConfiguration {
		var newItems = items
		newItems.removeLast()
		return ToolbarButtonConfiguration(items: newItems)
	}
}

extension ToolbarButtonConfiguration {
	static var defaultRight: ToolbarButtonConfiguration {
		let tbItems: [ToolbarItem] = [.undo, .redo, .save]
		return ToolbarButtonConfiguration(items: tbItems)
	}

    static var solutionViewerRight: ToolbarButtonConfiguration {
        let tbItems: [ToolbarItem] = [.edit]
        return ToolbarButtonConfiguration(items: tbItems)
    }

	static var diffLeft: ToolbarButtonConfiguration {
		let tbItems: [ToolbarItem] = [.expand, .lasso, .preview].reversed()
		return ToolbarButtonConfiguration(items: tbItems)
	}

    static var diffRight: ToolbarButtonConfiguration {
        let tbItems: [ToolbarItem] = [.pen, .highlighter, .eraser, .undo, .redo, .save]
        return ToolbarButtonConfiguration(items: tbItems)
    }

    static var drawing: ToolbarButtonConfiguration {
        let tbItems: [ToolbarItem] = [
            .subproblem,
            .empty,
            .separator,
            .pen,
            .highlighter,
            .eraser,
            .separator,
            .toolbarSettings,
            .flexibleSpace
        ].reversed()
        return ToolbarButtonConfiguration(items: tbItems)
    }

    static var drawingWithoutSubproblemSelection: ToolbarButtonConfiguration {
        let tbItems: [ToolbarItem] = [
            .empty,
            .separator,
            .pen,
            .highlighter,
            .eraser,
            .separator,
            .toolbarSettings,
            .flexibleSpace
        ].reversed()
        return ToolbarButtonConfiguration(items: tbItems)
    }

    static var drawingForSubproblem: ToolbarButtonConfiguration {
        let tbItems: [ToolbarItem] = [
            .pen,
            .highlighter,
            .eraser,
            .separator,
            .toolbarSettings,
            .flexibleSpace
        ].reversed()
        return ToolbarButtonConfiguration(items: tbItems)
    }
}
