// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit

extension ToolbarManager {
	func createDismissItem() -> UIBarButtonItem {
		let title = L10n.CorrectionController.dismiss
		return UIBarButtonItem(title: title, style: .done, target: self, action: #selector(dismissItemTapped))
	}

	func createEmptyItem() -> SelectableBarButtonItem {
		let image = Asset.emptySetIcon.image
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.emptyItemDidSelect(selected)
		}
        emptyObserver = context.$currentPageIsEmpty.receiveOnMain().sink { [weak self] isEmpty in
            self?.emptyBarItem?.isSelected = isEmpty
		}
		return item
	}

    func createSolutionItem(_ subproblemMode: Bool = false) -> SelectableBarButtonItem {
		let image = UIImage(systemName: "doc.text")!
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.solutionItemDidSelect(selected, subproblemMode: subproblemMode)
		}
        item.isSelected = context.showSolution
		return item
	}

    func createPersonalizedSolutionItem() -> SelectableBarButtonItem {
        let image = UIImage(named: "personalized.solution")!
        let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
            self?.personalizedSolutionItemDidSelect(selected)
        }
        item.isSelected = context.showsPersonalizedSolution
        return item
    }

	func createOverlayItem() -> SelectableBarButtonItem {
		let image = UIImage(systemName: "eye")!
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.overlayItemDidSelect(selected)
		}
        item.isSelected = context.showOverlay
		return item
	}

	func createSubproblemItem() -> SelectableBarButtonItem {
		let image = UIImage(systemName: "crop")!
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.subproblemItemDidSelect(selected)
		}
		return item
	}

	// swiftlint:disable:next cyclomatic_complexity
	func createBarButtonItems(_ config: ToolbarButtonConfiguration,
							  at context: CorrectionContext) -> [UIBarButtonItem] {
		var barButtonItems: [UIBarButtonItem] = []
		for item in config.items {
			let barItem: UIBarButtonItem
			switch item {
			case .preview:
				barItem = item.createBarButtonItemAtTarget(self, withAction: #selector(showPreview))
				self.previewBarItem = barItem
			case .undo:
				barItem = item.createBarButtonItemAtTarget(self, withAction: #selector(undo))
				self.undoBarItem = barItem
			case .redo:
				barItem = item.createBarButtonItemAtTarget(self, withAction: #selector(redo))
				self.redoBarItem = barItem
			case .eraser:
				let selectableBarItem = item.createSelectableBarButtonItem({ [weak self] _ in
					self?.switchTool(to: .eraser)
                    self?.eraserBarItem?.isSelected = true
				})
                selectableBarItem.isSelected = context.drawingTool == .eraser
				barItem = selectableBarItem
				self.eraserBarItem = selectableBarItem
			case .highlighter:
				let selectableBarItem = item.createSelectableBarButtonItem({ [weak self] _ in
					self?.switchTool(to: .highlighter)
                    self?.highlighterBarItem?.isSelected = true
				})
                selectableBarItem.isSelected = context.drawingTool == .highlighter
				barItem = selectableBarItem
				self.highlighterBarItem = selectableBarItem
			case .pen:
				let selectableBarItem = item.createSelectableBarButtonItem({ [weak self] _ in
					self?.switchTool(to: .pen)
                    self?.penBarItem?.isSelected = true
				})
                selectableBarItem.isSelected = context.drawingTool == .pen
				barItem = selectableBarItem
				self.penBarItem = selectableBarItem
			case .save:
				barItem = item.createBarButtonItemAtTarget(self, withAction: #selector(save))
				self.saveBarItem = barItem
			case .edit:
				barItem = item.createBarButtonItemAtTarget(self, withAction: #selector(edit))
			case .expand:
				barItem = item.createBarButtonItemAtTarget(self, withAction: #selector(shouldExpand))
			case .dismiss:
				barItem = item.createBarButtonItemAtTarget(self, withAction: #selector(dismiss))
			case .pageCount:
				barItem = UpdateableBarButtonItem()
			case .lasso:
				let selectableBarItem = item.createSelectableBarButtonItem({ [weak self] isSelected in
					self?.enableLassoTool(isSelected)
				})
				barItem = selectableBarItem
				self.lassoBarItem = selectableBarItem
            case .flexibleSpace:
                barItem = item.createBarButtonItemAtTarget(nil, withAction: nil)
            case .separator:
                barItem = item.createBarButtonItemAtTarget(nil, withAction: nil)
            case .empty:
                let selectableBarItem = item.createSelectableBarButtonItem({ [weak self] isSelected in
                    self?.emptyItemDidSelect(isSelected)
                })
                emptyObserver = context.$currentPageIsEmpty.receiveOnMain().sink { [weak self] isEmpty in
                    self?.emptyBarItem?.isSelected = isEmpty
                }
                barItem = selectableBarItem
                self.emptyBarItem = selectableBarItem
            case .subproblem:
                let selectableBarItem = item.createSelectableBarButtonItem({ [weak self] isSelected in
                    self?.subproblemItemDidSelect(isSelected)
                })
                barItem = selectableBarItem
                self.subproblemSelectionBarItem = selectableBarItem
            case .toolbarSettings:
                barItem = item.createBarButtonItemWithMenu(
                    createToolbarSettingsMenu(
                        selectedPosition: App.shared.settings.drawingToolbarPosition
                    )
                )
                self.toolbarSettingsItem = barItem
			}
			barButtonItems.append(barItem)
		}
		return barButtonItems.reversed()
	}

}
