// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import Combine
import TUMexamKit
import Logging

enum DrawingToolbarPosition: Int {
    case left = 0
    case right
    case top
}

class ToolbarManager: NSObject {
	let navigationItem: UINavigationItem
    let drawingToolbar: UIToolbar?
	let context: CorrectionContext

    var dismissBarItem: UIBarButtonItem?
	var emptyBarItem: SelectableBarButtonItem?
    var solutionBarItem: SelectableBarButtonItem?
    var personalizedSolutionBarItem: SelectableBarButtonItem?
    var rawBarItem: SelectableBarButtonItem?
    var overlayBarItem: SelectableBarButtonItem?

	var highlighterBarItem: SelectableBarButtonItem?
	var penBarItem: SelectableBarButtonItem?
	var eraserBarItem: SelectableBarButtonItem?
	var lassoBarItem: SelectableBarButtonItem?

	var undoBarItem: UIBarButtonItem?
	var redoBarItem: UIBarButtonItem?
	var previewBarItem: UIBarButtonItem?
	var saveBarItem: UIBarButtonItem?
    var toolbarSettingsItem: UIBarButtonItem?

	var currentUndoManager: ContextUndoManager? {
		didSet {
			if oldValue !== currentUndoManager {
                Logger.undoRedo.debug("Switched to new UndoManager: \(pointerString(currentUndoManager))")
				undoManagerChanged()
			}
		}
	}
	var undoObserver: AnyCancellable?
	var redoObserver: AnyCancellable?

	var emptyObserver: AnyCancellable?
	var toolObserver: AnyCancellable?

	var subproblemSelectionBarItem: SelectableBarButtonItem?

	weak var delegate: ToolbarManagerDelegate?

	private var rightButtonConfig: ToolbarButtonConfiguration = .defaultRight

    var drawingToolbarPosition: DrawingToolbarPosition = .left

	// MARK: - Toolbar Setup

    init(navigationItem: UINavigationItem, context: CorrectionContext, toolbar: UIToolbar? = nil) {
		self.navigationItem = navigationItem
		self.context = context
        self.drawingToolbar = toolbar
	}

	func setupLeftItems(includeDismiss: Bool, includeSubproblem: Bool) {
		var items = [UIBarButtonItem]()
		if includeDismiss {
            let dismiss = createDismissItem()
			items.append(dismiss)
            dismissBarItem = dismiss
		}
//		let empty = createEmptyItem()
//		items.append(empty)
//		emptyBarItem = empty
        let solution = createSolutionItem()
		items.append(solution)
        solutionBarItem = solution
        let personalizedSolution = createPersonalizedSolutionItem()
        items.append(personalizedSolution)
        personalizedSolutionBarItem = personalizedSolution
        let raw = createRawItem()
		items.append(raw)
        rawBarItem = raw
        let overlay = createOverlayItem()
		items.append(overlay)
        overlayBarItem = overlay

//		if includeSubproblem {
//			let subproblemSelection = createSubproblemItem()
//			subproblemSelectionBarItem = subproblemSelection
//			items.append(subproblemSelection)
//		}

		navigationItem.leftBarButtonItems = items
	}

    func setupLeftItemsForSuproblemCorrection() {
        var items = [UIBarButtonItem]()

        let solution = createSolutionItem(true)
        items.append(solution)
        solutionBarItem = solution
        let personalizedSolution = createPersonalizedSolutionItem()
        items.append(personalizedSolution)
        personalizedSolutionBarItem = personalizedSolution
        let raw = createRawItem()
        items.append(raw)
        rawBarItem = raw
        let overlay = createOverlayItem()
        items.append(overlay)
        overlayBarItem = overlay

        navigationItem.leftBarButtonItems = items
    }

	@discardableResult
	func setupLeftItemsForDiffView() -> [UIBarButtonItem] {
		if App.shared.coordinator?.window.windowScene?.interfaceOrientation.isLandscape == true {
			navigationItem.leftBarButtonItems = createBarButtonItems(.diffLeft, at: context)
		} else {
			if let conf = ToolbarButtonConfiguration.diffLeft.newConfigurationReplacingItem(.expand, with: .pageCount) {
				navigationItem.leftBarButtonItems = createBarButtonItems(conf.appendItem(.dismiss), at: context)
			}
		}
		return navigationItem.leftBarButtonItems ?? []
	}

	func setupRightItems() {
		navigationItem.rightBarButtonItems = createBarButtonItems(.defaultRight, at: context)
		toolObserver = context.$drawingTool.receiveOnMain().sink { [weak self] tool in
			self?.toolChanged(tool)
		}
	}

    func setupRightItemsForDiffing() {
        navigationItem.rightBarButtonItems = createBarButtonItems(.diffRight, at: context)
        toolObserver = context.$drawingTool.receiveOnMain().sink { [weak self] tool in
            self?.toolChanged(tool)
        }
    }

    func setupDrawingItems(editorToolbarMode: EditorToolbarMode, includeSubproblem: Bool) {
        let app = UIToolbarAppearance()
        app.configureWithTransparentBackground()
        app.backgroundColor = .secondarySystemBackground
        drawingToolbar?.tintColor = Asset.barButtonColor.color
        drawingToolbar?.standardAppearance = app
        if #available(iOS 15.0, *) {
            drawingToolbar?.scrollEdgeAppearance = app
        } else {
            // Fallback on earlier versions
        }
        var config: ToolbarButtonConfiguration = .drawing
        switch editorToolbarMode {
        case .standard, .solution:
            config = includeSubproblem ? .drawing : .drawingWithoutSubproblemSelection
        case .single, .singleModal:
            config = .drawingWithoutSubproblemSelection
        case .subproblem:
            config = .drawingForSubproblem
        case .diffing:
            config = .drawing
        }
        drawingToolbar?.items = createBarButtonItems(config, at: context)
    }

	func createRawItem() -> SelectableBarButtonItem {
		var systemName = "photo"
		if #available(iOS 14.0, *) {
			systemName = "scanner"
		}
		let image = UIImage(systemName: systemName)!
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.rawItemDidSelect(selected)
		}
        item.isSelected = context.showsRaw
		return item
	}

	// MARK: - Button Actions

	func switchTool(to tool: Tool) {
        BreadcrumbManager.addInteraction("[Toolbar] Switched drawing tool to \(tool)")
		guard tool != context.drawingTool else {
            BreadcrumbManager.addNavigation("[Toolbar] Showing Drawing Options VC")
			showDrawingOptions(for: tool)
			return
		}

		context.drawingTool = tool
	}

	func rawItemDidSelect(_ selected: Bool) {
        BreadcrumbManager.addInteraction("[Toolbar] Raw Item Button selected = \(selected)")
		context.showsRaw = selected
	}

	func toolChanged(_ tool: Tool) {
        penBarItem?.isSelected = (tool == .pen)
        highlighterBarItem?.isSelected = (tool == .highlighter)
        eraserBarItem?.isSelected = (tool == .eraser)
	}

	func showDrawingOptions(for tool: Tool) {
		let viewController = DrawingOptionsViewController()

        guard !viewController.isBeingPresented else {
            // swiftlint:disable:next line_length
            Logger.correctionUI.error("\(pointerString(self)) DrawingOptionsVC is already being presented by \(pointerString(viewController.presentingViewController))")
            return
        }

		viewController.modalPresentationStyle = .popover
		let sourceItem: UIBarButtonItem?
		let options: GlobalDrawingOptions.ToolOptions
		switch tool {
		case .pen:
			sourceItem = penBarItem
			options = App.shared.globalDrawingOptions.pen
		case .highlighter:
			sourceItem = highlighterBarItem
			options = App.shared.globalDrawingOptions.highlighter
		case .eraser:
			sourceItem = eraserBarItem
			options = App.shared.globalDrawingOptions.eraser
		}

		guard let popover = viewController.popoverPresentationController else { return }

		viewController.toolOptions = options
		popover.barButtonItem = sourceItem

		delegate?.presentDrawingOptions(viewController)
	}

	func enableSaveButton(_ isEnabled: Bool) {
        BreadcrumbManager.addLifecycle("[Toolbar] Save Button enabled = \(isEnabled)")
		saveBarItem?.isEnabled = isEnabled
	}

	@objc func showPreview() {
        BreadcrumbManager.addInteraction("[Toolbar] show diff preview")
		delegate?.showPreview()
	}

	@objc func undo() {
        BreadcrumbManager.addInteraction("[Toolbar] undo tapped")
		currentUndoManager?.undo()
	}

	@objc func redo() {
        BreadcrumbManager.addInteraction("[Toolbar] redo tapped")
		currentUndoManager?.redo()
	}

	@objc func save() {
        BreadcrumbManager.addInteraction("[Toolbar] save tapped")
		delegate?.save()
	}

	@objc func edit() {
        BreadcrumbManager.addInteraction("[Toolbar] edit solution tapped")
		delegate?.editSampleSolution()
	}

	@objc
	func dismiss() {
        BreadcrumbManager.addInteraction("[Toolbar] dismiss tapped")
		delegate?.toolbarWantsToDismiss()
	}

	func emptyItemDidSelect(_ selected: Bool) {
        BreadcrumbManager.addInteraction("[Toolbar] Empty Item Button selected = \(selected)")
		delegate?.setEmpty(selected)
	}

    func solutionItemDidSelect(_ selected: Bool, subproblemMode: Bool) {
        BreadcrumbManager.addInteraction("[Toolbar] Solution Item Button selected = \(selected)")
		context.showSolution = selected
        if selected {
            context.showsPersonalizedSolution = false
            personalizedSolutionBarItem?.isSelected = false
        }

        let newConfig: ToolbarButtonConfiguration?
        if selected && !subproblemMode {
            newConfig = .solutionViewerRight
        } else {
            newConfig = .defaultRight
        }

		applyNewToolbarConfiguration(newConfig)
	}

    func personalizedSolutionItemDidSelect(_ selected: Bool) {
        BreadcrumbManager.addInteraction("[Toolbar] Personalized Solution Item Button selected = \(selected)")
        context.showsPersonalizedSolution = selected
        if selected {
            context.showSolution = false
            solutionBarItem?.isSelected = false
            solutionItemDidSelect(false, subproblemMode: false)
        }
    }

	@objc
	func shouldExpand() {
		delegate?.expandParentView()
	}

	func enableLassoTool(_ isEnabled: Bool) {
        BreadcrumbManager.addInteraction("[Toolbar] Lasso tool isEnabled = \(isEnabled)")
		context.enableLassoTool = isEnabled
	}

	private func applyNewToolbarConfiguration(_ cfg: ToolbarButtonConfiguration?) {
		guard let cfg = cfg else { return }
		navigationItem.rightBarButtonItems = createBarButtonItems(cfg, at: context)
		rightButtonConfig = cfg
		undoManagerChanged()
	}

	func overlayItemDidSelect(_ selected: Bool) {
        BreadcrumbManager.addInteraction("[Toolbar] overlay item selected = \(selected)")
		context.showOverlay = selected
	}

	func subproblemItemDidSelect(_ selected: Bool) {
        BreadcrumbManager.addInteraction("[Toolbar] subproblem item selected = \(selected)")
		context.setMode(selected ? .subproblemSelection : .normal)
	}

	func updateSubproblemCorrectionMode(_ selected: Bool) {
        subproblemSelectionBarItem?.isSelected = selected
	}

	@objc func dismissItemTapped() {
        BreadcrumbManager.addInteraction("[Toolbar] dismiss item tapped")
		delegate?.toolbarWantsToDismiss()
	}

	private func undoManagerChanged() {
		guard let undoManager = self.currentUndoManager else {
			undoBarItem?.isEnabled = false
			redoBarItem?.isEnabled = false
			undoObserver = nil
			redoObserver = nil
			return
		}

		undoObserver = undoManager.$canUndo.receiveOnMain().sink { [weak self] canUndo in
			self?.undoBarItem?.isEnabled = canUndo
		}

		redoObserver = undoManager.$canRedo.receiveOnMain().sink { [weak self] canRedo in
			self?.redoBarItem?.isEnabled = canRedo
		}
	}

    func updateDrawingToolbarPosition(position: DrawingToolbarPosition) {
        self.drawingToolbarPosition = position
        delegate?.changeDrawingToolbarPosition(position)

        self.toolbarSettingsItem?.menu = createToolbarSettingsMenu(selectedPosition: position)
    }

    func createToolbarSettingsMenu(selectedPosition: DrawingToolbarPosition) -> UIMenu {
        let menu = UIMenu(
            title: L10n.CorrectionController.toolbarSettings,
            options: .displayInline,
            children: [
                UIMenu(
                    title: "Position",
                    options: .displayInline,
                    children: [
                        UIAction(
                            title: L10n.left,
                            image: UIImage(systemName: "rectangle.leftthird.inset.filled"),
                            state: selectedPosition == .left ? .on : .off,
                            handler: { [weak self] _ in
                                self?.updateDrawingToolbarPosition(position: .left)
                            }
                        ),
                        UIAction(
                            title: L10n.right,
                            image: UIImage(systemName: "rectangle.rightthird.inset.filled"),
                            state: selectedPosition == .right ? .on : .off,
                            handler: { [weak self] _ in
                                self?.updateDrawingToolbarPosition(position: .right)
                            }
                        ),
                        UIAction(
                            title: L10n.top,
                            image: UIImage(systemName: "rectangle.topthird.inset.filled"),
                            state: selectedPosition == .top ? .on : .off,
                            handler: { [weak self] _ in
                                self?.updateDrawingToolbarPosition(position: .top)
                            }
                        )
                    ]
                )
            ]
        )
        return menu
    }
}
