// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import Combine
import TUMexamKit
import os.log

protocol ToolbarManagerDelegate: class {
	func setEmpty(_ empty: Bool)

	func save()

	func toolbarWantsToDismiss()

	func presentDrawingOptions(_ viewController: UIViewController)
}

class ToolbarManager: NSObject {
	let navigationItem: UINavigationItem
	let context: CorrectionContext

    var dismissBarItem: UIBarButtonItem?
	var emptyBarItem: SelectableBarButtonItem?
    var solutionBarItem: SelectableBarButtonItem?
    var rawBarItem: SelectableBarButtonItem?
    var overlayBarItem: SelectableBarButtonItem?

    var penBarItem: SelectableBarButtonItem?
    var eraserBarItem: SelectableBarButtonItem?
    var highlighterBarItem: SelectableBarButtonItem?
    var saveBarItem: UIBarButtonItem?

	var undoBarItem: UIBarButtonItem?
	var redoBarItem: UIBarButtonItem?

	var currentUndoManager: ContextUndoManager? {
		didSet {
			if oldValue !== currentUndoManager {
                os_log(.debug, log: .undoRedo, "Current: %@", pointerString(currentUndoManager))
				undoManagerChanged()
			}
		}
	}
	var undoObserver: AnyCancellable?
	var redoObserver: AnyCancellable?

	var emptyObserver: AnyCancellable?
	var toolObserver: AnyCancellable?

	var subproblemSelectionBarItem: SelectableBarButtonItem?

	lazy var drawingOptionsViewController = DrawingOptionsViewController()

	weak var delegate: ToolbarManagerDelegate?

	init(navigationItem: UINavigationItem, context: CorrectionContext) {
		self.navigationItem = navigationItem
		self.context = context
	}

    func setupLeftItems(includeDismiss: Bool, includeSubproblem: Bool) {
		var items = [UIBarButtonItem]()
		if includeDismiss {
            let dismiss = createDismissItem()
			items.append(dismiss)
            dismissBarItem = dismiss
		}
		let empty = createEmptyItem()
		items.append(empty)
		emptyBarItem = empty
        let solution = createSolutionItem()
		items.append(solution)
        solutionBarItem = solution
        let raw = createRawItem()
        items.append(raw)
        rawBarItem = raw
        let overlay = createOverlayItem()
		items.append(overlay)
        overlayBarItem = overlay
        if includeSubproblem {
			let subproblemSelection = createSubproblemItem()
			subproblemSelectionBarItem = subproblemSelection
			items.append(subproblemSelection)
        }

		navigationItem.leftBarButtonItems = items
	}

	func setupRightItems() {
		let undo = createUndoItem()
		self.undoBarItem = undo

		let redo = createRedoItem()
		self.redoBarItem = redo

        let highlighter = createHighlighterItem()
		let pen = createPenItem()
		let eraser = createEraserItem()
		let save = createSaveItem()

        highlighter.selected = context.drawingTool == .highlighter
		pen.selected = context.drawingTool == .pen
		eraser.selected = context.drawingTool == .eraser

        self.highlighterBarItem = highlighter
		self.penBarItem = pen
		self.eraserBarItem = eraser
        self.saveBarItem = save

		toolObserver = context.$drawingTool.receiveOnMain().sink { [weak self] tool in
			self?.toolChanged(tool)
		}

		navigationItem.rightBarButtonItems = [pen, highlighter, eraser, undo, redo, save].reversed()

		undoManagerChanged()
	}

	func createDismissItem() -> UIBarButtonItem {
		let title = L10n.CorrectionController.dismiss
		return UIBarButtonItem(title: title, style: .done, target: self, action: #selector(dismissItemTapped))
	}

	func createEmptyItem() -> SelectableBarButtonItem {
		let image = Asset.emptySetIcon.image
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.emptyItemDidSelect(selected)
		}
		emptyObserver = context.$currentPageIsEmpty.receiveOnMain().sink { [weak self] isEmpty in
			self?.emptyBarItem?.selected = isEmpty
		}
		return item
	}

	func createSolutionItem() -> SelectableBarButtonItem {
		let image = UIImage(systemName: "doc.text")!
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.solutionItemDidSelect(selected)
		}
		item.selected = context.showSolution
		return item
	}

    func createRawItem() -> SelectableBarButtonItem {
        var systemName = "photo"
        if #available(iOS 14.0, *) {
            systemName = "scanner"
        }
        let image = UIImage(systemName: systemName)!
        let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
            self?.rawItemDidSelect(selected)
        }
        item.selected = context.showsRaw
        return item
    }

	func createOverlayItem() -> SelectableBarButtonItem {
		let image = UIImage(systemName: "eye")!
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.overlayItemDidSelect(selected)
		}
		item.selected = context.showOverlay
		return item
	}

	func createSubproblemItem() -> SelectableBarButtonItem {
		let image = UIImage(systemName: "crop")!
		let item = SelectableBarButtonItem(with: image) { [weak self] (selected) in
			self?.subproblemItemDidSelect(selected)
		}
		return item
	}

	func createUndoItem() -> UIBarButtonItem {
		let image = UIImage(systemName: "arrow.uturn.left")!
		return UIBarButtonItem(image: image, style: .plain, target: self, action: #selector(undo))
	}

	func createRedoItem() -> UIBarButtonItem {
		let image = UIImage(systemName: "arrow.uturn.right")!
		return UIBarButtonItem(image: image, style: .plain, target: self, action: #selector(redo))
	}

	func createPenItem() -> SelectableBarButtonItem {
		let image = Asset.penIcon.image
		let item = SelectableBarButtonItem(with: image) { [weak self] _ in
			self?.switchTool(to: .pen)
			self?.penBarItem?.selected = true
		}
		return item
	}

    func createHighlighterItem() -> SelectableBarButtonItem {
        let image = Asset.highlighterIcon.image
        let item = SelectableBarButtonItem(with: image) { [weak self] _ in
            self?.switchTool(to: .highlighter)
            self?.highlighterBarItem?.selected = true
        }
        return item
    }

	func createEraserItem() -> SelectableBarButtonItem {
		let image = Asset.eraserIcon.image
		let item = SelectableBarButtonItem(with: image) { [weak self] _ in
			self?.switchTool(to: .eraser)
            self?.eraserBarItem?.selected = true
		}
		return item
	}

	func createSaveItem() -> UIBarButtonItem {
		return UIBarButtonItem(barButtonSystemItem: .save, target: self, action: #selector(save))
	}

	func switchTool(to tool: Tool) {
		guard tool != context.drawingTool else {
			showDrawingOptions(for: tool)
			return
		}

		context.drawingTool = tool
	}

	func toolChanged(_ tool: Tool) {
        penBarItem?.selected = (tool == .pen)
        highlighterBarItem?.selected = (tool == .highlighter)
        eraserBarItem?.selected = (tool == .eraser)
	}

	func showDrawingOptions(for tool: Tool) {
		let viewController = drawingOptionsViewController
        guard !viewController.isBeingPresented else { return }
		viewController.modalPresentationStyle = .popover
		let sourceItem: UIBarButtonItem?
		let options: GlobalDrawingOptions.ToolOptions
        switch tool {
        case .pen:
			sourceItem = penBarItem
			options = App.shared.globalDrawingOptions.pen
        case .highlighter:
            sourceItem = highlighterBarItem
            options = App.shared.globalDrawingOptions.highlighter
        case .eraser:
			sourceItem = eraserBarItem
			options = App.shared.globalDrawingOptions.eraser
		}

		guard let popover = viewController.popoverPresentationController else { return }

		viewController.toolOptions = options
		popover.barButtonItem = sourceItem

		delegate?.presentDrawingOptions(viewController)
	}

	@objc func undo() {
		currentUndoManager?.undo()
	}

	@objc func redo() {
		currentUndoManager?.redo()
	}

	@objc func save() {
		delegate?.save()
	}

	func emptyItemDidSelect(_ selected: Bool) {
		delegate?.setEmpty(selected)
	}

	func solutionItemDidSelect(_ selected: Bool) {
		context.showSolution = selected
	}

    func rawItemDidSelect(_ selected: Bool) {
        context.showsRaw = selected
    }

	func overlayItemDidSelect(_ selected: Bool) {
		context.showOverlay = selected
	}

	func subproblemItemDidSelect(_ selected: Bool) {
		context.setMode(selected ? .subproblemSelection : .normal)
	}

	func updateSubproblemCorrectionMode(_ selected: Bool) {
		subproblemSelectionBarItem?.selected = selected
	}

	@objc func dismissItemTapped() {
		delegate?.toolbarWantsToDismiss()
	}

	private func undoManagerChanged() {
		guard let undoManager = self.currentUndoManager else {
			undoBarItem?.isEnabled = false
			redoBarItem?.isEnabled = false
			undoObserver = nil
			redoObserver = nil
			return
		}

		undoObserver = undoManager.$canUndo.receiveOnMain().sink { [weak self] canUndo in
			self?.undoBarItem?.isEnabled = canUndo
		}

		redoObserver = undoManager.$canRedo.receiveOnMain().sink { [weak self] canRedo in
			self?.redoBarItem?.isEnabled = canRedo
		}
	}
}
