// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import Foundation
import TUMexamKit
import UIKit

extension DiffingPageViewController: ToolbarManagerDelegate {
	func setEmpty(_ empty: Bool) {}
	func save() {
		handleSaveProcess()
	}

	func toolbarWantsToDismiss() {
		dismiss(animated: true, completion: nil)
	}

	func presentDrawingOptions(_ viewController: UIViewController) { }
    func changeDrawingToolbarPosition(_ newPosition: DrawingToolbarPosition) { }
	func editSampleSolution() {}

	func setupToolbarManager() {
		let manager = ToolbarManager(navigationItem: navigationItem,
									 context: context)
		manager.delegate = self
		self.toolbarManager = manager
		// observe changes to undo manager
		undoManagerObserver = context
			.$currentUndoManager
			.receiveOnMain()
			.assignNoRetain(to: \.currentUndoManager, on: manager)

		manager.setupRightItemsForDiffing()
		setupLeftToolBarItems()
	}

	func setupLeftToolBarItems() {
		let barButtons = toolbarManager?.setupLeftItemsForDiffView()
		pageIndexButton = barButtons?.compactMap({ $0 as? UpdateableBarButtonItem }).first
		let currentPage = currentIndex + 1
		let lastPage = arrayDataSource?.viewControllers.count ?? 0
		pageIndexButton?.buttonTitle = L10n.Solution.Submission.currentPage(currentPage, lastPage)
	}

	func expandParentView() {
		shouldExpand.send()
	}

	func handleSaveProcess() {
		guard let viewControllers = arrayDataSource?.viewControllers else { return }
		let splits = viewControllers
			.compactMap { $0 as? DiffingSplitViewController }
		let rejection: [Int: SolutionRejectionModel] = diffDataSource.rejectionModels
			.reduce(into: [:]) { res, next in
			res[next.page] = next
		}

		let selectedPages: [(SolutionRejectionModel, SolutionPageType)] = splits.compactMap { page in
			let solPage = page.getChosenSolutionPage()
			guard let rejectionModel = rejection[solPage.pageNumber] else {
				return nil
			}
			return (rejectionModel, solPage)
		}

        (self.presentingViewController as? UINavigationController)?.popToRootViewController(animated: true)

        self.submissionObserver = App.shared.currentExam?.solution.submissionState.receiveOnMain()
            .sink(receiveValue: { [weak self] state in
                if state == .loaded {
                    self?.dismiss(animated: true, completion: nil)
                }
                if case .error(let err) = state, case NetworkError.rejected(let pages) = err {
                    // Our conflict resolution got rejected again
                    let readableString = pages.map { "\($0.page)" }.joined(separator: ", ")
                    let msg: String
                    if pages.count == 1 {
                        msg = L10n.Solution.Submission.Rejected.messageSingle(readableString)
                    } else {
                        msg = L10n.Solution.Submission.Rejected.message(readableString)
                    }
                    let title = L10n.Solution.Submission.Rejected.title

                    let alert = UIAlertController(title: title,
                                                  message: msg,
                                                  preferredStyle: .alert)
                    alert.addAction(UIAlertAction(title: L10n.cancel, style: .cancel) { _ in
                        App.shared.currentExam?.solution.resetSubmission()
                        self?.dismiss(animated: true, completion: nil)
                    })

                    let diffTitle = L10n.Solution.Submission.Rejected.inspect
                    alert.addAction(UIAlertAction(title: diffTitle, style: .default) { _ in
                        self?.dismiss(animated: true, completion: nil)
                        App.shared.coordinator?.startDiffOfPages(pages)
                    })
                    alert.show(animated: true)
                }
                self?.submissionObserver = nil
        })

        App.shared.currentExam?.solution.saveMergedPages(selectedPages)
	}

	func showPreview() {
		guard let currentVC = currentViewController,
			  let lhs = currentVC.lhsVC.page as? SolutionPageType,
			  let rhs = currentVC.rhsVC.page as? SolutionPageType else {
			return
		}
		let prev = PreviewViewController(lhs: lhs, rhs: rhs)

		present(prev, animated: true, completion: nil)
	}
}
