// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import Foundation
import Logging
import TUMexamKit
import UIKit

final class DiffingPageViewController: UIPageViewController, IndexRepresentingViewController {

	@Published
	var currentIndex: Int = 0

	lazy var shouldExpand = PassthroughSubject<Void, Never>()
	lazy var onShouldExpand = shouldExpand.eraseToAnyPublisher()

	// MARK: - Toolbar
	var modeObserver: AnyCancellable?
	var toolbarManager: ToolbarManager?
	var undoManagerObserver: AnyCancellable?

	var context: CorrectionContext = .init()

	var pageIndexButton: UpdateableBarButtonItem?
    var pencilInteraction: UIPencilInteraction?

	// MARK: - DataSource

	var arrayDataSource: ArrayDataSource?
	var representingIndexInCollection: Int?

	var diffDataSource: DiffingDataSource
	private var dsSubscriber: AnyCancellable?
    var submissionObserver: AnyCancellable?

	private var segmentedControl: UISegmentedControl?

	var currentViewController: DiffingSplitViewController? {
		return viewControllers?.first as? DiffingSplitViewController
	}

	init(dataSource: DiffingDataSource) {
		self.diffDataSource = dataSource
		super.init(
			transitionStyle: .scroll,
			navigationOrientation: .horizontal,
			options: [.interPageSpacing: 32.0 as NSNumber]
		)
		createDataSource()
	}

	override func viewDidLoad() {
		super.viewDidLoad()
		dsSubscriber = diffDataSource.$localPages
			.receive(on: DispatchQueue.main)
			.sink { [weak self] _ in
				self?.createDataSource()
		}
		disablePencilForScrolling()
		setupToolbarManager()
		setupNavBar()
        setupPencilInteraction()
	}

	private func setupNavBar() {
		guard segmentedControl == nil else { return }
		let items = DiffSelection.allCases.map(\.description)
		let segmentedControl = UISegmentedControl(items: items)
		segmentedControl.selectedSegmentIndex = 1
		segmentedControl.addTarget(self, action: #selector(changeDiffSelection), for: .valueChanged)
		navigationItem.titleView = segmentedControl
		self.segmentedControl = segmentedControl
	}

	@objc
	private func changeDiffSelection() {
		guard let segmentedControl = segmentedControl,
			  let vc = arrayDataSource?.viewController(at: currentIndex) as? DiffingSplitViewController else { return }

		let choice = DiffSelection.allCases[segmentedControl.selectedSegmentIndex]
		vc.diffSelection = choice

		diffDataSource.diffingStateSelection[currentIndex] = choice
		diffDataSource.visitedPages.insert(currentIndex)

        Logger.solutionUI.debug("Changed Selection to \(choice.description)")

		evaluteDisplayOfPreview(choice)
	}

	private func evaluteDisplayOfPreview(_ choice: DiffSelection) {
		if choice == .merge {
			toolbarManager?.previewBarItem?.isEnabled = true
			toolbarManager?.previewBarItem?.tintColor = nil
		} else {
			toolbarManager?.previewBarItem?.isEnabled = false
			toolbarManager?.previewBarItem?.tintColor = .clear
		}
	}

	override func viewWillTransition(to size: CGSize, with coordinator: UIViewControllerTransitionCoordinator) {
		super.viewWillTransition(to: size, with: coordinator)
		coordinator.animate(alongsideTransition: { [weak self] _ in
			self?.setupLeftToolBarItems()
		}, completion: nil)
	}

	deinit {
		dsSubscriber?.cancel()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	private func createDataSource() {
		let localPages = diffDataSource.localPages
		let copies = diffDataSource.remotePages

		let allPages: [DiffingSplitViewController] = Array(zip(localPages, copies)).compactMap { localSol, remoteSol in
			let ctx = CanvasContextBridge()
			guard let rej = diffDataSource.recjectionModelForPage(localSol) else {
				return nil
			}
			let vc = DiffingSplitViewController(
				lhs: .init(page: localSol, context: context, sharedContext: ctx),
				rhs: .init(page: remoteSol, context: context, sharedContext: ctx),
				sharedContext: ctx,
				rejection: rej
			)
			return vc
		}

		let dataSource = ArrayDataSource(allPages)
		arrayDataSource = dataSource
		self.dataSource = dataSource
		self.delegate = self
		showPage(at: 0)
	}

	private func showPage(at index: Int) {
		if let current = arrayDataSource?.viewController(at: index) as? DiffingSplitViewController {

			diffDataSource.diffingStateSelection[index] = current.diffSelection
			diffDataSource.visitedPages.insert(index)
			current.hasBeenVisited = true
			evaluateSaveButtonState()

			context.currentUndoManager = current.contextUndoManager
			self.setViewControllers([current], direction: .forward, animated: false, completion: nil)
		}
	}

	func evaluateSaveButtonState() {
		let isSaveEnabled = arrayDataSource?.viewControllers
			.compactMap({ $0 as? DiffingSplitViewController})
			.allSatisfy(\.hasBeenVisited) ?? false
		toolbarManager?.enableSaveButton(isSaveEnabled)
	}
}

extension DiffingPageViewController: SplitViewNavigationDelegate {
	func didTapItemAt(_ index: Int) {
		showPage(at: index)
	}
}

extension DiffingPageViewController: UIPageViewControllerDelegate {
	func pageViewController(_ pageViewController: UIPageViewController,
							didFinishAnimating finished: Bool,
							previousViewControllers: [UIViewController],
							transitionCompleted completed: Bool) {
		guard let vc = pageViewController.viewControllers?.first as? DiffingSplitViewController,
			  let index = arrayDataSource?.indexOfViewController(vc) else {
			return
		}

		if completed {
			vc.hasBeenVisited = true
			segmentedControl?.selectedSegmentIndex = vc.diffSelection.index
			diffDataSource.diffingStateSelection[index] = vc.diffSelection
			diffDataSource.visitedPages.insert(index)
			evaluteDisplayOfPreview(vc.diffSelection)
			evaluateSaveButtonState()
		}
		currentIndex = index
		pageIndexButton?.buttonTitle = L10n.Solution.Submission.currentPage(currentIndex + 1,
                                                                            arrayDataSource?.viewControllers.count ?? 0)
	}
}
