// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import os.log
import TUMexamKit
import UIKit
import UIKitCore

final class DiffingSplitViewController: UIViewController {

	// MARK: - ContextUndoManager
	let contextUndoManager = ContextUndoManager()

	// MARK: - DiffState
	var diffSelection: DiffSelection = .merge
	var hasBeenVisited: Bool = false
	// MARK: - Container

	@UsesAutoLayout
	private var lhsContainer = UIView(frame: .zero)
	@UsesAutoLayout
	private var rhsContainer = UIView(frame: .zero)

	@UsesAutoLayout
	private var localAuthorLabel = UILabel(frame: .zero)
	@UsesAutoLayout
	private var remoteAuthorLabel = UILabel(frame: .zero)

	/// Represents the local solution changes
	let lhsVC: CorrectionViewController
	/// Represents the remote solution changes
	let rhsVC: CorrectionViewController

	private let sharedCanvasContext: CanvasContextBridge

	private let rejectionModel: SolutionRejectionModel

	init(lhs: CorrectionViewController,
		 rhs: CorrectionViewController,
		 sharedContext: CanvasContextBridge,
		 rejection: SolutionRejectionModel) {
		self.lhsVC = lhs
		self.rhsVC = rhs
		self.sharedCanvasContext = sharedContext
		self.rejectionModel = rejection
		super.init(nibName: nil, bundle: nil)
		setupConstraints()
	}

	override func viewDidLoad() {
		super.viewDidLoad()
		edgesForExtendedLayout = []

		configure(with: rejectionModel)
	}

	func getChosenSolutionPage() -> SolutionPageType {
		switch diffSelection {
		case .left:
			guard let page = lhsVC.page as? SolutionPageType else {
				preconditionFailure("Error casting PageType to a SolutionPageType")
			}
			return page
		case .right:
			guard let page = rhsVC.page as? SolutionPageType else {
				preconditionFailure("Error casting PageType to a SolutionPageType")
			}
			return page
		case .merge:
			guard let lhs = lhsVC.page as? SolutionPageType, let rhs = rhsVC.page as? SolutionPageType else {
				preconditionFailure("Error casting PageType to a SolutionPageType")
			}
			lhs.imageToMerge = rhs.layerImage.value?.cgImage
			return lhs
		}
	}

	private func setupConstraints() {
		add(lhsVC)
		add(rhsVC)

		lhsContainer.add(to: view)
			.top(to: \.topAnchor, constant: 16)
			.left(to: \.leftAnchor, constant: 8)

		rhsContainer.add(to: view)
			.top(to: \.topAnchor, constant: 16)
			.left(to: \.rightAnchor, of: lhsContainer, constant: 16)
			.right(to: \.rightAnchor, constant: 8)

		rhsContainer.width(to: \.widthAnchor, of: lhsContainer)

		lhsVC.view.add(to: lhsContainer)
			.pinToEdges()
		rhsVC.view.add(to: rhsContainer)
			.pinToEdges()

		localAuthorLabel
			.add(to: view)
			.top(to: \.bottomAnchor, of: lhsContainer, constant: 8)
			.bottom(to: \.bottomAnchor, constant: 16)
			.left(to: \.leftAnchor, of: lhsContainer)
			.right(to: \.rightAnchor, of: lhsContainer)

		remoteAuthorLabel
			.add(to: view)
			.top(to: \.bottomAnchor, of: rhsContainer, constant: 8)
			.bottom(to: \.bottomAnchor, constant: 16)
			.left(to: \.leftAnchor, of: rhsContainer)
			.right(to: \.rightAnchor, of: rhsContainer)	}

	func configure(with model: SolutionRejectionModel) {

		let dateformatter = DateFormatter()
		dateformatter.dateFormat = "d.MM.yyyy, HH:mm"

		localAuthorLabel.textAlignment = .center
		localAuthorLabel.text = L10n.Diffing.Splitview.yourChanges

		remoteAuthorLabel.textAlignment = .center
		let dateString = dateformatter.string(from: model.date)

		if let corrector = model.corrector {
			remoteAuthorLabel.text = "\(corrector) (\(dateString))"
		} else {
			remoteAuthorLabel.text = L10n.Diffing.Splitview.Author.unknown(dateString)
		}
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}
}

extension DiffingSplitViewController: ContextUndoManagerProvider {}
