// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import Foundation
import TUMexamKit
import UIKit
import UIKitCore

final class DiffNavigationViewController: UITableViewController {

	private var dataSource: DiffingDataSource
	private var dsSubscriber: AnyCancellable?
	private var stateSubscriber: AnyCancellable?
	private var visitedSubscriber: AnyCancellable?

	private var pages: [SolutionPageType] = []
	private var visitedPages: Set<Int> = .init()

	var currentIndex: Int = 0 {
		didSet {
			selectCurrentRow()
		}
	}

	var currentCell: SolutionPageDiffCell? {
		tableView.cellForRow(at: IndexPath(row: currentIndex, section: 0)) as? SolutionPageDiffCell
	}

	weak var navigationDelegate: SplitViewNavigationDelegate?

	init(dataSource: DiffingDataSource) {
		self.dataSource = dataSource
		super.init(style: .insetGrouped)
	}

	override func viewDidLoad() {
		super.viewDidLoad()
		tableView.dataSource = self
		tableView.delegate = self
		setCustomBackButton()

		pages = dataSource.localPages

		dsSubscriber = dataSource.$localPages
			.receiveOnMain()
			.sink { [weak self] pages in
				self?.pages = pages
				self?.visitedPages = .init()
				self?.tableView.reloadData()
				self?.selectCurrentRow()
		}

		stateSubscriber = dataSource.$diffingStateSelection
			.receiveOnMain()
			.sink { [weak self] _ in
				guard let self = self else { return }
				let visited = self.visitedPages.contains(self.currentIndex)
				self.currentCell?.configureWith(
					value: self.pages[self.currentIndex],
					state: self.dataSource.diffingStateSelection[self.currentIndex],
					visited: visited)
			}

		visitedSubscriber = dataSource.$visitedPages
			.receiveOnMain()
			.sink { [weak self] visits in
				guard let self = self,
					  let newVisit = visits.subtracting(self.visitedPages).first else { return }
				self.currentCell?.configureWith(
					value: self.pages[newVisit],
					state: self.dataSource.diffingStateSelection[newVisit],
					visited: true)
				self.visitedPages.formUnion(visits)
			}

		tableView.register(cellType: SolutionPageDiffCell.self)

	}

	deinit {
		dsSubscriber?.cancel()
		stateSubscriber?.cancel()
		visitedSubscriber?.cancel()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	private func setCustomBackButton() {
		navigationItem.setHidesBackButton(true, animated: false)

		let button = UIBarButtonItem(title: L10n.CorrectionController.dismiss,
									 style: .done,
									 target: self, action: #selector(onDismiss))
		navigationItem.leftBarButtonItem = button
	}

	@objc
	private func onDismiss() {
		dismiss(animated: true, completion: nil)
	}

	func selectCurrentRow() {
		for row in 0..<pages.count {
			let idx = IndexPath(row: row, section: 0)
			tableView.deselectRow(at: idx, animated: false)
		}

		let indexPath = IndexPath(row: currentIndex, section: 0)
		tableView.selectRow(at: indexPath, animated: false, scrollPosition: .none)
	}

	override func viewDidAppear(_ animated: Bool) {
		super.viewDidAppear(animated)
		let indexPath = IndexPath(row: currentIndex, section: 0)
		tableView.selectRow(at: indexPath, animated: false, scrollPosition: .none)
	}

	// MARK: - UITableView DataSource

	override func numberOfSections(in tableView: UITableView) -> Int { 1 }

	override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
		return pages.count
	}

	override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
		let cell: SolutionPageDiffCell = tableView.dequeueReusableCell(at: indexPath)
		let visited = visitedPages.contains(indexPath.row)
		cell.configureWith(value: pages[indexPath.row],
						   state: dataSource.diffingStateSelection[indexPath.row],
						   visited: visited)
		return cell
	}

	// MARK: - UITableView Delegate

	override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
		navigationDelegate?.didTapItemAt(indexPath.row)
	}
}
