// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Combine
import MBProgressHUD

protocol EditFlagViewControllerDelegate: AnyObject {
    func editFlagControllerDidSave()
}

@available(iOS 14.0, *)
class EditFlagViewController: UIViewController {

    // MARK: Section and Items

    enum Mode {
        case createFlag
        case editFlagDef
        case createFlagDef
    }

    enum Section: Int {
        case main = 0
        case actions
    }

    enum CollectionItem: Hashable {
        case editCell
        case deleteButton
    }

    // MARK: Properties
    let mode: Mode

    let exam: ExamType
    var flagDef: FlagDefinitionType?

    var erid: String?
    var problem: ProblemType?
    var correctionPass: Int?

    weak var delegate: EditFlagViewControllerDelegate?

    private var saveFlagPublisher: AnyCancellable?

    let collectionView = UICollectionView(
        frame: .zero,
        collectionViewLayout: createCollectionViewLayout()
    ).autolayout()
    private var dataSource: UICollectionViewDiffableDataSource<Section, CollectionItem>!

    var hud: MBProgressHUD?

    // MARK: Initializers

    init(exam: ExamType, mode: Mode = .createFlag) {
        self.exam = exam
        self.mode = mode
        super.init(nibName: nil, bundle: nil)
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    // MARK: UIViewController Lifecycle

    override func viewDidLoad() {
        super.viewDidLoad()

        // Navbar
        switch mode {
        case .createFlag:
            self.title = L10n.CorrectionFlags.AddFlag.title
        case .createFlagDef:
            self.title = L10n.Flags.CreateFlagDef.title
        case .editFlagDef:
            self.title = L10n.Flags.EditFlagDef.title
        }

        let saveButton = UIBarButtonItem(barButtonSystemItem: .save, target: self, action: #selector(saveButtonTapped))
        self.navigationItem.rightBarButtonItem = saveButton

        if mode != .createFlag {
            // We are editing an existing flag def, add dismiss button
            let cancelButton = UIBarButtonItem(systemItem: .cancel, primaryAction: UIAction { [unowned self] _ in
                self.dismiss(animated: true, completion: nil)
            })
            self.navigationItem.leftBarButtonItem = cancelButton
        }

        // Setup Collection View
        self.view.addSubview(collectionView)
        collectionView.pinToSuperview()

        // swiftlint:disable:next line_length
        let cellRegistration = UICollectionView.CellRegistration<EditFlagCollectionViewCell, CollectionItem> { cell, _, _ in
            guard let flag = self.flagDef else {
                return
            }
            cell.textField.text = flag.text
            cell.colorWell.selectedColor = UIColor(hex: flag.color ?? "#FFFFFF")
        }

        // swiftlint:disable:next line_length
        let deleteButtonRegistration = UICollectionView.CellRegistration<UICollectionViewListCell, CollectionItem> { cell, _, _ in
            var config = cell.defaultContentConfiguration()
            config.text = L10n.delete // TODO
            config.textProperties.color = self.flagDef?.count == 0 ? .systemRed : .secondaryLabel
            config.textProperties.alignment = .center
            cell.contentConfiguration = config
        }

        // swiftlint:disable:next line_length
        let dataSource = UICollectionViewDiffableDataSource<Section, CollectionItem>(collectionView: collectionView) { collectionView, indexPath, item in
            switch item {
            case .editCell:
                return collectionView.dequeueConfiguredReusableCell(
                    using: cellRegistration,
                    for: indexPath,
                    item: item
                )
            case .deleteButton:
                return collectionView.dequeueConfiguredReusableCell(
                    using: deleteButtonRegistration,
                    for: indexPath,
                    item: item
                )
            }
        }

        self.dataSource = dataSource
        collectionView.dataSource = dataSource
        collectionView.delegate = self

        applyDataSourceSnapshot()
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)

        guard self.collectionView.visibleCells.count >= 1,
              let cell = self.collectionView.visibleCells[0] as? EditFlagCollectionViewCell else {
            return
        }

        cell.textField.becomeFirstResponder()
    }

    // MARK: Functions

    @objc private func saveButtonTapped() {
        guard self.collectionView.visibleCells.count >= 1,
              let cell = self.collectionView.visibleCells[0] as? EditFlagCollectionViewCell else {
            return
        }

        resignCellFirstResponder()

        guard let text = cell.textField.text, let color = cell.colorWell.selectedColor else {
            let alert = UIAlertController(title: L10n.CorrectionFlags.AddFlag.MissingAttrErr.title,
                                          message: L10n.CorrectionFlags.AddFlag.MissingAttrErr.message,
                                          preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: nil))
            alert.show(animated: true)
            return
        }

        switch mode {
        case .createFlag:
            guard let erid = self.erid else {
                return
            }
            // Create new flag
            hud = MBProgressHUD.showAdded(to: view, animated: true)
            let publisher = self.exam.createFlag(text: text,
                                                 color: color.hexString,
                                                 erid: erid,
                                                 problem: self.problem?.identifier.problem,
                                                 subproblem: self.problem?.identifier.subproblem,
                                                 pass: self.correctionPass)
            self.saveFlagPublisher = publisher.receive(on: RunLoop.main).sink { [weak self] completion in
                switch completion {
                case .finished:
                    self?.delegate?.editFlagControllerDidSave()
                case .failure(let error):
                    let alert = UIAlertController(title: L10n.CorrectionFlags.AddFlag.SaveError.title,
                                                  message: error.localizedDescription,
                                                  preferredStyle: .alert)
                    alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: nil))
                    alert.show(animated: true)
                }
                self?.saveFlagPublisher = nil
                self?.hud?.hide(animated: true)
            } receiveValue: { _ in
            }
        case .editFlagDef:
            guard let flagDef = self.flagDef else {
                return
            }
            // Flag already exists, update
            flagDef.text = text
            flagDef.color = color.hexString
            flagDef.submissionCompletion = { [weak self] result in
                switch result {
                case .failure(let error):
                    let alert = UIAlertController(
                        title: "Error",
                        message: "Could not save Flag Definition: \(error.localizedDescription)",
                        preferredStyle: .alert
                    )
                    alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: nil))
                    alert.show(animated: true)
                case .success:
                    self?.delegate?.editFlagControllerDidSave()
                    self?.dismiss(animated: true, completion: nil)
                }
                self?.hud?.hide(animated: true)
            }
            hud = MBProgressHUD.showAdded(to: view, animated: true)
            flagDef.submitChanges()
        case .createFlagDef:
            hud = MBProgressHUD.showAdded(to: view, animated: true)
            let publisher = self.exam.createFlagDef(text: text, color: color.hexString)
            self.saveFlagPublisher = publisher.receive(on: RunLoop.main).sink { [weak self] completion in
                switch completion {
                case .finished:
                    self?.delegate?.editFlagControllerDidSave()
                    self?.dismiss(animated: true, completion: nil)
                case .failure(let error):
                    let alert = UIAlertController(title: L10n.CorrectionFlags.AddFlag.SaveError.title,
                                                  message: error.localizedDescription,
                                                  preferredStyle: .alert)
                    alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: nil))
                    alert.show(animated: true)
                }
                self?.hud?.hide(animated: true)
            } receiveValue: { _ in
            }
        }
    }

    private func applyDataSourceSnapshot() {
        var snapshot = NSDiffableDataSourceSnapshot<Section, CollectionItem>()
        snapshot.appendSections([.main])
        snapshot.appendItems([.editCell])
        snapshot.appendSections([.actions])
        if mode == .editFlagDef {
            snapshot.appendItems([.deleteButton])
        }
        dataSource.apply(snapshot)
    }

    private func resignCellFirstResponder() {
        guard collectionView.visibleCells.count > 0,
              let flagCell = collectionView.visibleCells[0] as? EditFlagCollectionViewCell else {
            return
        }

        flagCell.textField.resignFirstResponder()
    }

    // MARK: Static functions

    static func createCollectionViewLayout() -> UICollectionViewLayout {
        let layout = UICollectionViewCompositionalLayout { _, layoutEnv in
            let config = UICollectionLayoutListConfiguration(appearance: .insetGrouped)
            let list = NSCollectionLayoutSection.list(using: config, layoutEnvironment: layoutEnv)
            return list
        }
        layout.configuration.interSectionSpacing = 0
        return layout
    }
}

@available(iOS 14.0, *)
extension EditFlagViewController: UICollectionViewDelegate {
    func collectionView(_ collectionView: UICollectionView, shouldSelectItemAt indexPath: IndexPath) -> Bool {
        // Delete Button
        return indexPath.section != Section.main.rawValue && flagDef?.count == 0
    }

    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        guard indexPath.section == Section.actions.rawValue else {
            return
        }

        resignCellFirstResponder()

        // Delete
        hud = MBProgressHUD.showAdded(to: view, animated: true)
        flagDef?.deletionCompletion = { [weak self] result in
            switch result {
            case .failure(let error):
                let alert = UIAlertController(
                    title: "Error",
                    message: "Could not delete flag definition: \(error.localizedDescription)",
                    preferredStyle: .alert
                )
                alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: nil))
                alert.show(animated: true)
            case .success:
                self?.delegate?.editFlagControllerDidSave()
                self?.dismiss(animated: true, completion: nil)
            }
            self?.hud?.hide(animated: true)
        }
        flagDef?.delete()
        collectionView.deselectItem(at: indexPath, animated: true)
    }
}
