// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Combine

class FlagSelectionCollectionViewController: UIViewController {
    private enum CollectionItem: Hashable {
        case flag(FlagType)
        case flagDefinition(FlagDefinitionType)

        static func == (lhs: FlagSelectionCollectionViewController.CollectionItem,
                        rhs: FlagSelectionCollectionViewController.CollectionItem) -> Bool {
            if case let .flag(lhs) = lhs, case let .flag(rhs) = rhs {
                return lhs.flagId == rhs.flagId &&
                    lhs.flagTypeId == rhs.flagTypeId &&
                    lhs.color == rhs.color &&
                    lhs.text == rhs.text &&
                    lhs.erid == rhs.erid &&
                    lhs.problem == rhs.problem &&
                    lhs.subproblem == rhs.subproblem &&
                    lhs.correctionPass == rhs.correctionPass
            }
            if case let .flagDefinition(lhs) = lhs, case let .flagDefinition(rhs) = rhs {
                return lhs.flagTypeId == rhs.flagTypeId &&
                    lhs.text == rhs.text &&
                    lhs.color == rhs.color &&
                    lhs.count == rhs.count &&
                    lhs.user == rhs.user
            }
            return false
        }

        func hash(into hasher: inout Hasher) {
            switch self {
            case .flag(let flag):
                hasher.combine("flag")
                hasher.combine(flag.flagId)
            case .flagDefinition(let flagDef):
                hasher.combine("flagDefinition")
                hasher.combine(flagDef.flagTypeId)
            }
        }
    }

    enum Section {
        case activeFlags
        case definitions
    }

    let erid: String
    let exam: ExamType
    let problem: ProblemType?
    let correctionPass: Int?

    private var flags = [FlagType]()
    private var flagsObserver: AnyCancellable?
    private var flagList: FlagListType?

    private var flagDefinitions = [FlagDefinitionType]()
    private var shownFlagDefinitions = [FlagDefinitionType]()
    private var flagDefObserver: AnyCancellable?
    private var flagDefList: FlagDefinitionListType?

    private var dataSource: UICollectionViewDiffableDataSource<Section, CollectionItem>! = nil
    private var collectionView: UICollectionView! = nil

    private var currentActionPublisher: AnyCancellable?

    init(problem: ProblemType?, erid: String, correctionPass: Int? = nil) {
        guard let exam = App.shared.currentExam else {
            fatalError("trying to create a FlagTableViewController without a selected exam")
        }
        self.erid = erid
        self.exam = exam
        self.problem = problem
        self.correctionPass = correctionPass
        super.init(nibName: nil, bundle: nil)

        self.title = L10n.CorrectionFlags.title(
            erid,
            "\(problem?.identifier.problem ?? -1)\(problem?.identifier.subproblem ?? "")"
        )
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    // MARK: - UIVIewController Lifecycle

    override func viewDidLoad() {
        // Setup Navigation Item
        let createItem = UIBarButtonItem(barButtonSystemItem: .add, target: self, action: #selector(createButtonTapped))
        self.navigationItem.rightBarButtonItem = createItem

        // Setup Collection View
        let collectionView = UICollectionView(frame: .zero, collectionViewLayout: makeLayout()).autolayout()
        collectionView.register(FlagCollectionViewCell.self,
                                forCellWithReuseIdentifier: FlagCollectionViewCell.reuseIdentifier)
        collectionView.register(TitleSupplementaryView.self,
                                forSupplementaryViewOfKind: TitleSupplementaryView.kind,
                                withReuseIdentifier: TitleSupplementaryView.reuseIdentifier)
        collectionView.backgroundColor = .systemGroupedBackground
        collectionView.delegate = self
        self.view.addSubview(collectionView)
        // Add Constraints
        NSLayoutConstraint.activate([
            collectionView.topAnchor.constraint(equalTo: self.view.safeAreaLayoutGuide.topAnchor),
            collectionView.bottomAnchor.constraint(equalTo: self.view.safeAreaLayoutGuide.bottomAnchor),
            collectionView.leadingAnchor.constraint(equalTo: self.view.safeAreaLayoutGuide.leadingAnchor),
            collectionView.trailingAnchor.constraint(equalTo: self.view.safeAreaLayoutGuide.trailingAnchor)
        ])
        self.collectionView = collectionView

        configureDataSource()
        updateArrays()

        super.viewDidLoad()
    }

    // MARK: Helper funcs

    func updateArrays() {
        let filter = FlagFilter(
            erid: erid,
            problem: problem?.identifier.problem,
            subproblem: problem?.identifier.subproblem,
            correctionPass: correctionPass
        )
        if self.flagList == nil {
            self.flagList = exam.flags(filter: filter)
        }
        self.flagList?.filter = filter

        if self.flagsObserver == nil {
            self.flagsObserver = flagList?.flagList.receiveOnMain().sink { [weak self] flags in
                self?.flags = flags
                self?.updateDataSource()
            }
        }
        self.flagList?.forceReload()

        if self.flagDefList == nil {
            self.flagDefList = exam.flagDefinitions()
            self.flagDefObserver = flagDefList?.flagDefinitions.receiveOnMain().sink { [weak self] flagDefs in
                self?.flagDefinitions = flagDefs
                self?.updateDataSource()
            }
        }

        flagDefList?.forceReload()
    }

    func makeLayout() -> UICollectionViewLayout {
        let itemSize = NSCollectionLayoutSize(widthDimension: .estimated(100),
                                             heightDimension: .fractionalHeight(1.0))
        let item = NSCollectionLayoutItem(layoutSize: itemSize)

        let groupSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                              heightDimension: .absolute(40))
        let group = NSCollectionLayoutGroup.horizontal(layoutSize: groupSize,
                                                         subitems: [item])
        group.interItemSpacing = NSCollectionLayoutSpacing.fixed(16)

        let section = NSCollectionLayoutSection(group: group)
        section.interGroupSpacing = 16
        section.contentInsets = NSDirectionalEdgeInsets(top: 0, leading: 16, bottom: 0, trailing: 16)

        let headerFooterSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                     heightDimension: .estimated(44))
        let sectionHeader = NSCollectionLayoutBoundarySupplementaryItem(
            layoutSize: headerFooterSize,
            elementKind: TitleSupplementaryView.kind,
            alignment: .top)
        section.boundarySupplementaryItems = [sectionHeader]

        let config = UICollectionViewCompositionalLayoutConfiguration()
        config.interSectionSpacing = 16
        let layout = UICollectionViewCompositionalLayout(section: section, configuration: config)
        return layout
    }

    func configureDataSource() {
        // swiftlint:disable:next line_length
        dataSource = UICollectionViewDiffableDataSource<Section, CollectionItem>(collectionView: collectionView) { (collectionView, indexPath, item) -> UICollectionViewCell? in
            // This API is just verbose as hell, I'd rather have long lines than wrap this around 10 lines
            // swiftlint:disable:next line_length
            guard let cell = collectionView.dequeueReusableCell(withReuseIdentifier: FlagCollectionViewCell.reuseIdentifier, for: indexPath) as? FlagCollectionViewCell else {
                fatalError("Could not dequeue cell for reuse: \(FlagCollectionViewCell.reuseIdentifier)")
            }

            if case let .flag(flag) = item {
                let flagColor = UIColor(hex: flag.color ?? "#FFFFFF")
                let textColor: UIColor = flagColor.brightness < 0.5 ? .white : .black

                cell.label.text = flag.text
                cell.label.textColor = textColor
                cell.imageView.tintColor = textColor
                cell.color = flagColor
                cell.showsDetailDisclosureIndicator = false

                return cell
            }

            if case let .flagDefinition(flagDef) = item {
                let flagColor = UIColor(hex: flagDef.color ?? "#FFFFFF")
                let textColor: UIColor = flagColor.brightness < 0.5 ? .white : .black

                cell.label.text = flagDef.text
                cell.label.textColor = textColor

                cell.imageView.tintColor = textColor
                cell.color = flagColor
                cell.showsDetailDisclosureIndicator = false

                return cell
            }

            return nil
        }
        dataSource.supplementaryViewProvider = { (collectionView, kind, indexPath) -> UICollectionReusableView? in
            if kind == TitleSupplementaryView.kind {
                guard let supplementaryView = collectionView.dequeueReusableSupplementaryView(
                    ofKind: kind,
                    withReuseIdentifier: TitleSupplementaryView.reuseIdentifier,
                    for: indexPath) as? TitleSupplementaryView else { fatalError("Cannot create new supplementary") }

                supplementaryView.label.text =
                    indexPath.section == 0 ?
                    L10n.CorrectionFlags.active.uppercased() :
                    L10n.CorrectionFlags.available.uppercased()

                return supplementaryView
            }
            return nil
        }
    }

    func updateDataSource() {
        // Remove active flags from defs
        self.shownFlagDefinitions = flagDefinitions.filter({ !self.flags.contains($0.flagTypeId, for: \.flagTypeId) })

        var snapshot = NSDiffableDataSourceSnapshot<Section, CollectionItem>()
        snapshot.appendSections([.activeFlags])
        snapshot.appendItems(self.flags.map({ CollectionItem.flag($0) }))
        snapshot.appendSections([.definitions])
        snapshot.appendItems(self.shownFlagDefinitions.map({ CollectionItem.flagDefinition($0) }))
        dataSource.apply(snapshot)
    }

    @objc private func createButtonTapped() {
        if #available(iOS 14.0, *) {
            let viewController = EditFlagViewController(exam: App.shared.currentExam!)
            viewController.erid = erid
            viewController.problem = problem
            viewController.correctionPass = correctionPass
            viewController.delegate = self
            self.navigationController?.pushViewController(viewController, animated: true)
        } else {
            let alert = UIAlertController(title: L10n.CorrectionFlags.AddFlagError.title,
                                          message: L10n.CorrectionFlags.AddFlagError.message,
                                          preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: nil))
            alert.show(animated: true)
        }
    }
}

extension FlagSelectionCollectionViewController: UICollectionViewDelegate {

    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        if indexPath.section == 0 {
            let flag = self.flags[indexPath.row]
            flag.deletionCompletion = { [weak self] result in
                switch result {
                case .failure(let error):
                    DispatchQueue.main.async {
                        let alert = UIAlertController(
                            title: "Error",
                            message: "Could not delete flag: \(error.localizedDescription)",
                            preferredStyle: .alert
                        )
                        alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: nil))
                        alert.show(animated: true)
                    }
                case .success:
                    self?.updateArrays()
                }
            }
            flag.delete()
        }
        if indexPath.section == 1 {
            // Create flag from definition
            let flagDef = self.shownFlagDefinitions[indexPath.row]
            let flagPub = flagDef.addTo(exam: self.exam.identifier,
                                        erid: self.erid,
                                        problem: self.problem?.identifier.problem,
                                        subproblem: self.problem?.identifier.subproblem,
                                        pass: correctionPass)
            currentActionPublisher = flagPub.receive(on: RunLoop.main).sink(receiveCompletion: { [weak self] _ in
                self?.updateArrays()
            }, receiveValue: { _ in
            })
        }
    }
}

extension FlagSelectionCollectionViewController: EditFlagViewControllerDelegate {
    func editFlagControllerDidSave() {
        self.navigationController?.popViewController(animated: true)
        self.updateArrays()
    }
}
