// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit
import Combine
import TUMexamKit

@available(iOS 14.0, *)
class FlagsCollectionViewController: UIViewController {

    // MARK: CollectionItem

    private enum Section: Int, CaseIterable {
        case loading = 0
        case definitions
        case flags
    }

    private enum CollectionItem: Hashable {
        case loading
        case flagDefinition(FlagDefinitionType)
        case flag(FlagType)
        case deleteButton

        static func == (lhs: FlagsCollectionViewController.CollectionItem,
                        rhs: FlagsCollectionViewController.CollectionItem) -> Bool {
            if case .loading = lhs,
               case .loading = rhs {
                return true
            }
            if case let .flagDefinition(ldef) = lhs,
               case let .flagDefinition(rdef) = rhs {
                return ldef.flagTypeId == rdef.flagTypeId &&
                    ldef.text == rdef.text &&
                    ldef.color == rdef.color &&
                    ldef.count == rdef.count &&
                    ldef.user == rdef.user
            }
            if case let .flag(lhs) = lhs,
               case let .flag(rhs) = rhs {
                return lhs.flagId == rhs.flagId &&
                    lhs.flagTypeId == rhs.flagTypeId &&
                    lhs.color == rhs.color &&
                    lhs.text == rhs.text &&
                    lhs.erid == rhs.erid &&
                    lhs.problem == rhs.problem &&
                    lhs.subproblem == rhs.subproblem &&
                    lhs.correctionPass == rhs.correctionPass
            }
            if case .deleteButton = lhs,
               case .deleteButton = rhs {
                return true
            }

            return false
        }

        func hash(into hasher: inout Hasher) {
            switch self {
            case .loading:
                hasher.combine("loading")
            case let .flagDefinition(def):
                hasher.combine(def.flagTypeId)
            case let .flag(flag):
                hasher.combine(flag.flagId)
            case .deleteButton:
                hasher.combine("delete")
            }
        }
    }

    // MARK: Properties

    let collectionView = UICollectionView(
        frame: .zero,
        collectionViewLayout: createLayout()
    ).autolayout()

    private var dataSource: UICollectionViewDiffableDataSource<Section, CollectionItem>!

    private var flagDefinitionList: FlagDefinitionListType?
    private var flagDefinitionObserver: AnyCancellable?
    private var flagDefinitions: [FlagDefinitionType] = [] {
        didSet {
            applyDatasourceSnapshot()
        }
    }

    private var flagList: FlagListType?
    private var flagObserver: AnyCancellable?
    private var flags: [FlagType] = [] {
        didSet {
            applyDatasourceSnapshot()
        }
    }

    private let columnCount = 6
    private let columnWidths: [CGFloat] = [200, 200, 60, 60, 70, 60]
    private let resizeHidePreferences = [5, 4, 3, 2, 1, 0]
    private let cellSpacing: CGFloat = 8

    private var modalTransitionController: UIViewControllerAnimatedTransitioning?

    // MARK: UIViewController Lifecycle

    override func viewDidLoad() {
        super.viewDidLoad()

        // Navigation Bar
        self.title = L10n.flags

        let addButton = UIBarButtonItem(
            systemItem: .add,
            primaryAction: UIAction { [unowned self] _ in
                let editFlagController = EditFlagViewController(exam: App.shared.currentExam!, mode: .createFlagDef)
                editFlagController.delegate = self
                let nav = UINavigationController(rootViewController: editFlagController)
                self.present(nav, animated: true, completion: nil)
            },
            menu: nil
        )
        self.navigationItem.rightBarButtonItem = addButton

        // Setup collection view
        collectionView.backgroundColor = .systemBackground
        view.addSubview(collectionView)
        collectionView.pinToSuperview()

        // Register Cells

        // swiftlint:disable:next line_length
        let loadingCellRegistration = UICollectionView.CellRegistration<ActivityIndicatorCell, CollectionItem> { cell, _, _ in
            cell.activityIndicator.startAnimating()
        }

        // swiftlint:disable:next line_length
        let definitionCellRegistration = UICollectionView.CellRegistration<FlagCollectionViewCell, CollectionItem> { cell, _, item in
            guard case let .flagDefinition(def) = item else {
                return
            }

            let flagColor = UIColor(hex: def.color ?? "#FFFFFF")
            let textColor: UIColor = flagColor.brightness < 0.5 ? .white : .black

            cell.label.text = "\(def.text ?? "") (\(def.count ?? 0))"
            cell.label.textColor = textColor
            cell.imageView.tintColor = textColor
            cell.color = flagColor
            cell.showsDetailDisclosureIndicator = true
            cell.disclosureIndicator.tintColor = textColor
        }

        // swiftlint:disable:next line_length
        let flagCellRegistration = UICollectionView.CellRegistration<DatacolCell, CollectionItem> { [weak self] cell, _, item in
            guard case let .flag(flag) = item else {
                return
            }

            cell.accessories = [.disclosureIndicator()]
            cell.dataConfig = self?.getCellDataConfig(flag)
        }

        // swiftlint:disable:next line_length
        let flagHeaderRegistration = UICollectionView.SupplementaryRegistration<DatacolHeaderView>(elementKind: DatacolHeaderView.elementKind) { [weak self] view, _, _ in
            view.dataConfig = self?.getHeaderDataConfig()
        }

        // swiftlint:disable:next line_length
        let defHeaderRegistration = UICollectionView.SupplementaryRegistration<TitleSupplementaryView>(elementKind: TitleSupplementaryView.kind) { view, _, _ in
            view.label.text = L10n.flagDefinitions
        }

        // swiftlint:disable:next line_length
        let dataSource = UICollectionViewDiffableDataSource<Section, CollectionItem>(collectionView: collectionView) { (collectionView, indexPath, item) -> UICollectionViewCell? in
            switch indexPath.section {
            case Section.loading.rawValue:
                return collectionView.dequeueConfiguredReusableCell(
                    using: loadingCellRegistration,
                    for: indexPath,
                    item: item
                )
            case Section.definitions.rawValue:
                // Definiton Cell
                return collectionView.dequeueConfiguredReusableCell(
                    using: definitionCellRegistration,
                    for: indexPath,
                    item: item
                )
            case Section.flags.rawValue:
                return collectionView.dequeueConfiguredReusableCell(
                    using: flagCellRegistration,
                    for: indexPath,
                    item: item
                )
            default:
                return nil
            }
        }
        dataSource.supplementaryViewProvider = { collectionView, _, indexPath in
            switch indexPath.section {
            case Section.definitions.rawValue:
                return collectionView.dequeueConfiguredReusableSupplementary(
                    using: defHeaderRegistration,
                    for: indexPath
                )
            case Section.flags.rawValue:
                return collectionView.dequeueConfiguredReusableSupplementary(
                    using: flagHeaderRegistration,
                    for: indexPath
                )
            default:
                return nil
            }
        }
        self.dataSource = dataSource
        collectionView.dataSource = dataSource
        collectionView.delegate = self

        // Refresh control
        let refreshControl = UIRefreshControl(frame: .zero, primaryAction: UIAction { [unowned self] _ in
            BreadcrumbManager.addInteraction("[FlagsCollectionViewController] refresh triggered")
            self.refreshFlags()
        })
        collectionView.refreshControl = refreshControl

        // Setup Data Observers

        flagDefinitionList = App.shared.currentExam?.flagDefinitions()
        flagDefinitionObserver = flagDefinitionList?.flagDefinitions.receiveOnMain()
            .assignNoRetain(to: \.flagDefinitions, on: self)
        flagDefinitionList?.load()

        flagList = App.shared.currentExam?.flags(filter: nil)
        flagObserver = flagList?.flagList.receiveOnMain().assignNoRetain(to: \.flags, on: self)
        flagList?.load()

        // Apply Initial Empty Snapshot

        applyDatasourceSnapshot(false)
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        BreadcrumbManager.addNavigation("[FlagsCollectionViewController] viewWillAppear")
        guard let selected = collectionView.indexPathsForSelectedItems else {
            return
        }
        for sel in selected {
            collectionView.deselectItem(at: sel, animated: true)
        }
    }

    // MARK: Functions

    private func applyDatasourceSnapshot(_ animated: Bool = true) {
        if collectionView.refreshControl?.isRefreshing ?? false {
            collectionView.refreshControl?.endRefreshing()
        }
        var snapshot = NSDiffableDataSourceSnapshot<Section, CollectionItem>()
        snapshot.appendSections([.loading])
        if flagDefinitionList?.state.value != .loaded && flagDefinitions.isEmpty {
            snapshot.appendItems([CollectionItem.loading])
        }
        snapshot.appendSections([.definitions])
        snapshot.appendItems(
            flagDefinitions.map(CollectionItem.flagDefinition)
        )
        snapshot.appendSections([.flags])
        snapshot.appendItems(
            flags.map(CollectionItem.flag)
        )
        snapshot.reloadSections([.definitions])
        dataSource.apply(
            snapshot,
            animatingDifferences: animated,
            completion: nil
        )

    }

    private func refreshFlags() {
        App.shared.currentExam?.flagDefinitions().forceReload()
        flagList?.forceReload()
    }

    // MARK: Datacol Cell Config

    private func getCellDataConfig(_ flag: FlagType) -> DatacolCell.Configuration {
        return DatacolCell.Configuration(
            columnCount: columnCount,
            columnWidths: columnWidths,
            columnGenerator: { index in
                switch index {
                case 0:
                    let container = UIView().autolayout()
                    let flagView = FlagView().autolayout()
                    container.addSubview(flagView)
                    flagView.addLeadingConstraint(to: \.leadingAnchor)
                    flagView.addTopConstraint(to: \.topAnchor)
                    flagView.addBottomConstraint(to: \.bottomAnchor)
                    flagView.addTrailingConstraint(to: \.trailingAnchor, relation: .equalOrLess)
                    return container
                case 1...5:
                    return UILabel().autolayout()
                default:
                    fatalError("Index out of range")
                }
            },
            columnConfigurator: { index, view in
                switch index {
                case 0:
                    let flagColor = UIColor(hex: flag.color ?? "#FFFFFF")
                    let textColor: UIColor = flagColor.brightness < 0.5 ? .white : .black
                    let flagView = view.subviews[0] as? FlagView
                    flagView?.label.textColor = textColor
                    flagView?.label.text = flag.text
                    flagView?.imageView.tintColor = textColor
                    flagView?.backgroundColor = flagColor
                case 1:
                    (view as? UILabel)?.text = flag.user
                case 2:
                    (view as? UILabel)?.text = flag.erid
                case 3:
                    (view as? UILabel)?.text = flag.problem.map(String.init)
                case 4:
                    (view as? UILabel)?.text = flag.subproblem
                case 5:
                    (view as? UILabel)?.text = flag.correctionPass.map(String.init)
                default:
                    return
                }
            },
            resizeHidePreferences: resizeHidePreferences,
            cellSpacing: cellSpacing
        )
    }

    private func getHeaderDataConfig() -> DatacolHeaderView.Configuration {
        return DatacolHeaderView.Configuration(
            columnCount: columnCount,
            columnWidths: columnWidths,
            columnGenerator: { _ in
                let label = UILabel().autolayout()
                label.font = .systemFont(ofSize: UIFont.smallSystemFontSize)
                return label
            },
            columnConfigurator: { index, view in
                switch index {
                case 0:
                    (view as? UILabel)?.text = L10n.FlagsController.Columns.flag
                case 1:
                    (view as? UILabel)?.text = L10n.FlagsController.Columns.user
                case 2:
                    (view as? UILabel)?.text = L10n.FlagsController.Columns.erid
                case 3:
                    (view as? UILabel)?.text = L10n.FlagsController.Columns.problem
                case 4:
                    (view as? UILabel)?.text = L10n.FlagsController.Columns.subproblem
                case 5:
                    (view as? UILabel)?.text = L10n.FlagsController.Columns.pass
                default:
                    return
                }
            },
            resizeHidePreferences: resizeHidePreferences,
            cellSpacing: cellSpacing
        )
    }

    // MARK: Static Functions

    static func createLayout() -> UICollectionViewLayout {
        let config = UICollectionViewCompositionalLayoutConfiguration()
        config.interSectionSpacing = 20
        let layout = UICollectionViewCompositionalLayout(sectionProvider: { section, layoutEnv in
            switch section {
            case Section.definitions.rawValue:
                // Definitions
                let itemSize = NSCollectionLayoutSize(widthDimension: .estimated(100),
                                                     heightDimension: .fractionalHeight(1.0))
                let item = NSCollectionLayoutItem(layoutSize: itemSize)

                let groupSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                      heightDimension: .absolute(40))
                let group = NSCollectionLayoutGroup.horizontal(layoutSize: groupSize,
                                                                 subitems: [item])
                group.interItemSpacing = NSCollectionLayoutSpacing.fixed(16)

                let section = NSCollectionLayoutSection(group: group)
                section.interGroupSpacing = 16
                section.contentInsets = NSDirectionalEdgeInsets(top: 0, leading: 16, bottom: 0, trailing: 16)

                let headerFooterSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                             heightDimension: .estimated(22))
                let sectionHeader = NSCollectionLayoutBoundarySupplementaryItem(
                    layoutSize: headerFooterSize,
                    elementKind: TitleSupplementaryView.kind,
                    alignment: .top)
                section.boundarySupplementaryItems = [sectionHeader]

                return section
            case Section.flags.rawValue:
                var config = UICollectionLayoutListConfiguration(appearance: .plain)
                config.headerMode = .supplementary
//                config.trailingSwipeActionsConfigurationProvider = { indexPath in
//                    let deleteAction = UIContextualAction(style: .destructive, title: "Delete") { _, _, completion in
//                        completion(true)
//                    }
//                    deleteAction.image = UIImage(systemName: "trash")
//                    return UISwipeActionsConfiguration(actions: [deleteAction])
//                }
                let list = NSCollectionLayoutSection.list(using: config, layoutEnvironment: layoutEnv)

                let headerSize = NSCollectionLayoutSize(
                    widthDimension: .fractionalWidth(1),
                    heightDimension: .estimated(20)
                )
                let header = NSCollectionLayoutBoundarySupplementaryItem(
                    layoutSize: headerSize,
                    elementKind: DatacolHeaderView.elementKind,
                    alignment: .top
                )
                header.pinToVisibleBounds = true
                header.zIndex = 1023
                list.boundarySupplementaryItems = [header]
                return list
            case Section.loading.rawValue:
                let height: NSCollectionLayoutDimension = .absolute(80)
                let itemSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                      heightDimension: height)
                let item = NSCollectionLayoutItem(layoutSize: itemSize)

                let groupSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                       heightDimension: .fractionalHeight(1.0))
                let group = NSCollectionLayoutGroup.vertical(layoutSize: groupSize, subitems: [item])

                let section = NSCollectionLayoutSection(group: group)
                return section
            default:
                fatalError("Unknown section")
            }
        }, configuration: config)

        return layout
    }
}

@available(iOS 14.0, *)
extension FlagsCollectionViewController: UICollectionViewDelegate {
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        guard let cell = collectionView.cellForItem(at: indexPath),
              let exam = App.shared.currentExam,
              let item = dataSource.itemIdentifier(for: indexPath) else {
            return
        }

        switch indexPath.section {
        case Section.definitions.rawValue:
            BreadcrumbManager.addInteraction("[FlagsCollectionViewController] User selected definition cell")
            guard case let .flagDefinition(def) = item else {
                return
            }

            let edit = EditFlagViewController(exam: exam, mode: .editFlagDef)
            edit.delegate = self
            edit.flagDef = def
            let nav = UINavigationController(rootViewController: edit)
            nav.view.cornerRadius = 10
            nav.view.clipsToBounds = true

            let transition = FlipTransition(cell)
            nav.transitioningDelegate = transition
            self.modalTransitionController = transition
            self.present(nav, animated: true, completion: nil)
        case Section.flags.rawValue:
            BreadcrumbManager.addInteraction("[FlagsCollectionViewController] User selected flag cell")
            guard case let .flag(flag) = item,
                  let erid = flag.erid else {
                return
            }
            let correctionId = CorrectionIdentifier(
                exam: exam.identifier,
                erid: erid,
                problem: flag.problem ?? 1,
                subproblem: flag.subproblem,
                pass: flag.correctionPass ?? 1
            )
            let correction = exam.getSingleCorrection(correctionId)
            App.shared.coordinator?.showCorrectionModally(correction, target: self)
        default:
            return
        }
    }
}

@available(iOS 14.0, *)
extension FlagsCollectionViewController: EditFlagViewControllerDelegate {
    func editFlagControllerDidSave() {
        self.refreshFlags()
    }
}
