// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit
import TUMexamKit
import Combine

@available(iOS 14.0, *)
class LocksViewController: UIViewController {

    enum Section: Int, CaseIterable {
        case loading = 0
        case main
    }

    enum CollectionItem: Hashable {
        case loading
        case lock(LockListEntryType)

        func hash(into hasher: inout Hasher) {
            switch self {
            case .loading:
                hasher.combine("loading")
            case .lock(let entry):
                hasher.combine(entry.erid)
            }
        }

        static func == (lhs: LocksViewController.CollectionItem, rhs: LocksViewController.CollectionItem) -> Bool {
            if case .loading = lhs, case .loading = rhs {
                return true
            }

            if case let .lock(lhs) = lhs,
               case let .lock(rhs) = rhs {
                return lhs.erid == rhs.erid
            }

            return false
        }
    }

    // MARK: Properties

    let collectionView = UICollectionView(
        frame: .zero,
        collectionViewLayout: createCollectionViewLayout()
    ).autolayout()
    let columnCount = 7
    let columnWidths: [CGFloat] = [60, 180, 50, 70, 50, 150, 150]
    let resizeHidePreferences: [Int] = [5, 3, 6, 4, 2, 1, 0]
    let cellSpacing: CGFloat = 8

    var dataSource: UICollectionViewDiffableDataSource<Section, CollectionItem>!

    // subscriptions
    var examSubscription: AnyCancellable?
    var currentExam: ExamType? {
        didSet {
            examChanged()
        }
    }

    var lockList: LockListType?
    var lockListSubscription: AnyCancellable?
    var lockListEntries: [LockListEntryType] = [] {
        didSet {
            if collectionView.refreshControl?.isRefreshing ?? false {
                collectionView.refreshControl?.endRefreshing()
            }
            applyDatasourceSnapshot()
        }
    }

    // MARK: View Controller Lifecycle

    override func viewDidLoad() {
        super.viewDidLoad()

        self.title = L10n.locks

        collectionView.backgroundColor = .systemBackground
        view.addSubview(collectionView)
        collectionView.pinToSuperview()

        collectionView.allowsSelection = false

        let refreshControl = UIRefreshControl(frame: .zero, primaryAction: UIAction { [unowned self] _ in
            BreadcrumbManager.addInteraction("[LocksViewController] refresh triggered")
            self.lockList?.forceReload()
        })
        collectionView.refreshControl = refreshControl

        // swiftlint:disable:next line_length
        let loadingRegistration = UICollectionView.CellRegistration<ActivityIndicatorCell, CollectionItem> { cell, _, _ in
            cell.activityIndicator.startAnimating()
        }

        // swiftlint:disable:next line_length
        let cellRegistration = UICollectionView.CellRegistration<DatacolCell, CollectionItem> { [weak self] cell, _, item in
            guard case let .lock(entry) = item else { return }
            cell.dataConfig = self?.getCellDataConfig(lock: entry)
        }

        // swiftlint:disable:next line_length
        let headerRegistration = UICollectionView.SupplementaryRegistration<DatacolHeaderView>(elementKind: DatacolHeaderView.elementKind) { [weak self] view, _, _ in
            view.dataConfig = self?.getHeaderDataConfig()
            view.trailingConstant = 0
        }

        // swiftlint:disable:next line_length
        let dataSource = UICollectionViewDiffableDataSource<Section, CollectionItem>(collectionView: collectionView) { collectionView, indexPath, item in
            switch item {
            case .loading:
                return collectionView.dequeueConfiguredReusableCell(
                    using: loadingRegistration,
                    for: indexPath,
                    item: item
                )
            case .lock:
                return collectionView.dequeueConfiguredReusableCell(
                    using: cellRegistration,
                    for: indexPath,
                    item: item
                )
            }
        }

        dataSource.supplementaryViewProvider = { collectionView, _, indexPath in
            return collectionView.dequeueConfiguredReusableSupplementary(using: headerRegistration, for: indexPath)
        }
        self.dataSource = dataSource
        collectionView.dataSource = dataSource

        // Setup Subscriptions
        examSubscription = App.shared.$currentExam.receiveOnMain().assignNoRetain(to: \.currentExam, on: self)
        App.shared.currentExam?.load()

        applyDatasourceSnapshot(false)
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        BreadcrumbManager.addNavigation("[LocksViewController] viewWillAppear")
    }

    // MARK: Private functions

    private func getCellDataConfig(lock: LockListEntryType) -> DatacolCell.Configuration {
        return DatacolCell.Configuration(
            columnCount: columnCount,
            columnWidths: columnWidths,
            columnGenerator: { index in
                switch index {
                case 0...6:
                    return UILabel().autolayout()
                default:
                    fatalError("Index out of range")
                }
            },
            columnConfigurator: { index, view in
                switch index {
                case 0:
                    (view as? UILabel)?.text = lock.erid
                case 1:
                    (view as? UILabel)?.text = lock.user
                case 2:
                    (view as? UILabel)?.text = lock.problem.map(String.init)
                case 3:
                    (view as? UILabel)?.text = lock.subproblem
                case 4:
                    (view as? UILabel)?.text = lock.correctionPass.map(String.init)
                case 5:
                    (view as? UILabel)?.text = lock.created.map {
                        let format = DateFormatter()
                        format.dateStyle = .short
                        format.timeStyle = .short
                        return format.string(from: $0)
                    }
                case 6:
                    (view as? UILabel)?.text = lock.lastUpdated.map {
                        let format = DateFormatter()
                        format.dateStyle = .short
                        format.timeStyle = .short
                        return format.string(from: $0)
                    }
                default:
                    return
                }
            },
            resizeHidePreferences: resizeHidePreferences,
            cellSpacing: cellSpacing
        )
    }

    private func getHeaderDataConfig() -> DatacolHeaderView.Configuration {
        return DatacolHeaderView.Configuration(
            columnCount: columnCount,
            columnWidths: columnWidths,
            columnGenerator: { _ in
                let label = UILabel().autolayout()
                label.font = .systemFont(ofSize: UIFont.smallSystemFontSize)
                return label
            },
            columnConfigurator: { index, view in
                switch index {
                case 0:
                    (view as? UILabel)?.text = L10n.LocksController.Columns.erid
                case 1:
                    (view as? UILabel)?.text = L10n.LocksController.Columns.user
                case 2:
                    (view as? UILabel)?.text = L10n.LocksController.Columns.problem
                case 3:
                    (view as? UILabel)?.text = L10n.LocksController.Columns.subproblem
                case 4:
                    (view as? UILabel)?.text = L10n.LocksController.Columns.pass
                case 5:
                    (view as? UILabel)?.text = L10n.LocksController.Columns.created
                case 6:
                    (view as? UILabel)?.text = L10n.LocksController.Columns.lastUpdated
                default:
                    return
                }
            },
            resizeHidePreferences: resizeHidePreferences,
            cellSpacing: cellSpacing
        )
    }

    private func examChanged() {
        lockList = currentExam?.lockList()
        lockListSubscription = lockList?.lockList.receiveOnMain().assignNoRetain(to: \.lockListEntries, on: self)
        lockList?.load()
    }

    private func unlockAtIndexPath(_ indexPath: IndexPath) {

    }

    private func applyDatasourceSnapshot(_ animated: Bool = true) {
        var snap = NSDiffableDataSourceSnapshot<Section, CollectionItem>()
        snap.appendSections([.loading])
        if lockList?.state.value != .loaded && lockListEntries.isEmpty {
            snap.appendItems([.loading])
        }
        snap.appendSections([.main])
        snap.appendItems(lockListEntries.map(CollectionItem.lock))
        dataSource.apply(snap, animatingDifferences: animated, completion: nil)
    }

    // MARK: Static functions

    static func createCollectionViewLayout() -> UICollectionViewLayout {
        let layout = UICollectionViewCompositionalLayout { section, layoutEnv in
            if section == Section.main.rawValue {
                var config = UICollectionLayoutListConfiguration(appearance: .plain)
                config.headerMode = .supplementary
                let list = NSCollectionLayoutSection.list(using: config, layoutEnvironment: layoutEnv)

                let headerSize = NSCollectionLayoutSize(
                    widthDimension: .fractionalWidth(1),
                    heightDimension: .estimated(20)
                )
                let header = NSCollectionLayoutBoundarySupplementaryItem(
                    layoutSize: headerSize,
                    elementKind: DatacolHeaderView.elementKind,
                    alignment: .top
                )
                header.pinToVisibleBounds = true
                header.zIndex = 1023
                list.boundarySupplementaryItems = [header]
                return list
            } else {
                let itemSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                      heightDimension: .fractionalHeight(1.0))
                let item = NSCollectionLayoutItem(layoutSize: itemSize)

                let groupSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                       heightDimension: .absolute(60))
                let group = NSCollectionLayoutGroup.vertical(layoutSize: groupSize, subitems: [item])

                let section = NSCollectionLayoutSection(group: group)
                return section
            }
        }

        return layout
    }

}
