// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import UIKit
import TUMexamKit
import Sentry

// swiftlint:disable private_over_fileprivate
@available(iOS 14.0, *)
fileprivate enum SidebarListSection {
    /// For displaying the Instance name, the selected Exam and the signed in user
    case header
    /// For displaying the " main menu" sidebar
    case main
}

@available(iOS 14.0, *)
fileprivate struct SidebarListItem: Hashable {
    let text: String
    var detail: String?
    let icon: UIImage?
    var enabled: Bool
}
// swiftlint:enable private_over_fileprivate

@available(iOS 14.0, *)
class SidebarCollectionViewController: UIViewController {

    // Subscriptions
    private var currentExamSubscription: AnyCancellable?
    private var nameSubscription: AnyCancellable?
    private var statsSubscription: AnyCancellable?
    private var metadataSubscription: AnyCancellable?
    private var accessRightsSubscription: AnyCancellable?

    private var currentExam: ExamType? {
        didSet {
            examChanged()
        }
    }
    private var currentExamName: String? {
        didSet {
            guard let name = currentExamName else { return }
            self.headerView?.examSelectionButton?.setTitle(name, for: .normal)

            guard let url = App.shared.backendURL,
                  let scheme = url.scheme else {
                return
            }

            if scheme == "https" {
                headerView?.state = .secure
            } else if scheme == "http" {
                headerView?.state = .unsecure
            }
        }
    }
    private var currentExamStats: [CorrectionProblemStatusType] = [] {
        didSet {
            applyDatasourceSnapshot(true)
        }
    }
    private var instanceName: String? {
        didSet {
            guard let name = instanceName else { return }
            headerView?.instanceNameLabel.text = name

            guard let url = App.shared.backendURL,
                  let scheme = url.scheme else {
                return
            }

            if scheme == "https" {
                headerView?.state = .secure
            } else if scheme == "http" {
                headerView?.state = .unsecure
            }
        }
    }

    private var examLeftToCorrect: Int {
        var corrected = 0
        var total = 0
        for problemStatus in currentExamStats {
            for pass in problemStatus.passes {
                corrected += pass.corrected
                total += pass.total
            }
        }
        return total - corrected
    }

    private var locksEnabled: Bool? {
        didSet {
            applyDatasourceSnapshot()
        }
    }

    // Collection View related stuff
    weak var collectionView: UICollectionView! = nil
    private var dataSource: UICollectionViewDiffableDataSource<SidebarListSection, SidebarListItem>! = nil
    private var secondaryControllers = [
        UINavigationController(rootViewController: StatusCollectionViewController()),
        UINavigationController(rootViewController: ResultsViewController()),
        UINavigationController(rootViewController: FlagsCollectionViewController()),
        UINavigationController(rootViewController: LocksViewController())
    ]

    weak var headerView: SidebarHeaderView?

    override func viewDidLoad() {
        super.viewDidLoad()

        // Title
        self.title = "TUMexam"
        navigationController?.navigationBar.prefersLargeTitles = true

        // Settings Bar Button Item
        let settingsItem = UIBarButtonItem(
            title: L10n.Settings.title,
            image: UIImage(systemName: "gearshape"),
            primaryAction: UIAction { [unowned self] _ in
                BreadcrumbManager.addNavigation("Open Settings Modal from SidebarCollectionViewController")
                App.shared.coordinator?.showSettings(using: self)
            },
            menu: nil
        )
        navigationItem.setRightBarButton(settingsItem, animated: false)

        // Setup Subscriptions
        App.shared.kit?.metadata.load()
        App.shared.kit?.accessRights.load()

        currentExamSubscription = App.shared.$currentExam.receiveOnMain().assignNoRetain(to: \.currentExam, on: self)
        metadataSubscription = App.shared.kit?.metadata.name
            .receiveOnMain()
            .assignNoRetain(to: \.instanceName, on: self)
        instanceName = App.shared.kit?.metadata.name.value
        accessRightsSubscription = App.shared.kit?.accessRights.state
            .receiveOnMain()
            .sink { [weak self] state in
                guard state == .loaded else { return }
                self?.locksEnabled = App.shared.kit?.accessRights.examManagement
            }

        // Setup View
        setupViews()
        splitViewController?.showDetailViewController(secondaryControllers[0], sender: self)
        collectionView.selectItem(at: IndexPath(row: 0, section: 0),
                                  animated: false,
                                  scrollPosition: .centeredVertically)
    }

    private func setupViews() {
        // Collection View Setup
        let collectionView = UICollectionView(
            frame: view.bounds,
            collectionViewLayout: createCollectionViewLayout()
        ).autolayout()
        collectionView.delegate = self
        self.collectionView = collectionView
        self.view.addSubview(collectionView)
        collectionView.pinToSuperview()

        // swiftlint:disable:next line_length
        let cellRegistration = UICollectionView.CellRegistration<UICollectionViewListCell, SidebarListItem> { cell, _, item in
            var content = cell.defaultContentConfiguration()
            content.text = item.text
            content.image = item.icon
            content.textProperties.colorTransformer = UIConfigurationColorTransformer { color in
                item.enabled ? color : .secondaryLabel
            }

            content.secondaryText = item.detail
            content.prefersSideBySideTextAndSecondaryText = true

            cell.tintColor = item.enabled ? Asset.tuMexamBlue.color : UIColor.secondaryLabel
            cell.contentConfiguration = content
        }

        // swiftlint:disable:next line_length
        let headerRegistration = UICollectionView.SupplementaryRegistration<SidebarHeaderView>(elementKind: SidebarHeaderView.elementKind) { [unowned self] view, _, _ in
            self.headerView = view
            if self.instanceName != nil {
                guard let url = App.shared.backendURL,
                      let scheme = url.scheme else {
                    headerView?.state = .loading
                    return
                }

                if scheme == "https" {
                    headerView?.state = .secure
                } else if scheme == "http" {
                    headerView?.state = .unsecure
                }
            } else {
                headerView?.state = .loading
            }
            view.examSelectionButton?.addAction(UIAction { [unowned self] action in
                // Show Exam selection popover
                guard let examList = App.shared.kit?.examList else { return }
                let alertController = UIAlertController(
                    title: L10n.Sidebar.chooseExam,
                    message: nil,
                    preferredStyle: .actionSheet
                )
                for exam in examList.exams.value {
                    alertController.addAction(
                        UIAlertAction(
                            title: exam.name.value,
                            style: .default,
                            handler: { _ in
                                App.shared.setCurrentExam(exam)
                            })
                    )
                }

                alertController.modalPresentationStyle = .popover
                alertController.view.tintColor = Asset.tuMexamBlue.color
                guard let popover = alertController.popoverPresentationController else { return }
                guard let sourceView = action.sender as? UIButton else { return }
                popover.sourceView = sourceView.imageView // Present the popover from the "down arrow" image view
                popover.permittedArrowDirections = [.up]
                self.present(alertController, animated: true, completion: nil)
            }, for: .touchUpInside)
            view.userLabel.text = L10n.Sidebar.Header.signedInAs(App.shared.kit?.userFullName ?? "n/a")
            view.instanceNameLabel.text = self.instanceName ?? L10n.Sidebar.loading
            view.examSelectionButton.setTitle(self.currentExamName ?? L10n.Sidebar.loading, for: .normal)
        }

        // swiftlint:disable:next line_length
        let dataSource = UICollectionViewDiffableDataSource<SidebarListSection, SidebarListItem>(collectionView: collectionView) { collectionView, indexPath, item in
            return collectionView.dequeueConfiguredReusableCell(using: cellRegistration, for: indexPath, item: item)
        }
        dataSource.supplementaryViewProvider = { collectionView, kind, indexPath in
            switch kind {
            case SidebarHeaderView.elementKind:
                return collectionView.dequeueConfiguredReusableSupplementary(using: headerRegistration, for: indexPath)
            default:
                return nil
            }
        }
        collectionView.dataSource = dataSource
        self.dataSource = dataSource

        // Setup Items
        applyDatasourceSnapshot()
    }

    private func createCollectionViewLayout() -> UICollectionViewLayout {
        let layout = UICollectionViewCompositionalLayout { _, layoutEnv in
            let config = UICollectionLayoutListConfiguration(appearance: .sidebar)
            let list = NSCollectionLayoutSection.list(using: config, layoutEnvironment: layoutEnv)

            let headerSize = NSCollectionLayoutSize(
                widthDimension: .fractionalWidth(1),
                heightDimension: .estimated(88)
            )
            let headerItem = NSCollectionLayoutBoundarySupplementaryItem(
                layoutSize: headerSize,
                elementKind: SidebarHeaderView.elementKind,
                alignment: .top
            )

            list.boundarySupplementaryItems = [headerItem]
            return list
        }

        return layout
    }

    private func applyDatasourceSnapshot(_ animated: Bool = false) {
        let selected = collectionView.indexPathsForSelectedItems
        var snapshot = NSDiffableDataSourceSnapshot<SidebarListSection, SidebarListItem>()
        snapshot.appendSections([.main])
        snapshot.appendItems([
            SidebarListItem(
                text: L10n.StatusController.title,
                detail: "\(examLeftToCorrect) left",
                icon: Asset.activityRingIcon.image.withRenderingMode(.alwaysTemplate),
                enabled: true
            ),
            SidebarListItem(
                text: L10n.ResultsController.title,
                icon: UIImage(systemName: "magnifyingglass")!,
                enabled: true
            ),
            SidebarListItem(
                text: L10n.flags,
                icon: UIImage(systemName: "flag")!,
                enabled: true
            ),
            SidebarListItem(
                text: L10n.locks,
                icon: UIImage(systemName: "lock")!,
                enabled: self.locksEnabled ?? false
            )
        ], toSection: .main)
        dataSource.apply(snapshot, animatingDifferences: animated) {
            guard let selected = selected else { return }
            for item in selected {
                self.collectionView.selectItem(at: item, animated: false, scrollPosition: .centeredVertically)
            }
        }
    }

    private func examChanged() {
        guard let exam = currentExam else { return }
        exam.load()
        exam.correctionStatus.load()
        nameSubscription = exam.name
            .receiveOnMain()
            .assignNoRetain(to: \.currentExamName, on: self)
        statsSubscription = exam.correctionStatus.problems
            .receiveOnMain()
            .assignNoRetain(to: \.currentExamStats, on: self)
    }
}

@available(iOS 14.0, *)
extension SidebarCollectionViewController: UICollectionViewDelegate {
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        // Set secondary VC of splitViewController
        splitViewController?.showDetailViewController(secondaryControllers[indexPath.row], sender: self)
    }

    func collectionView(_ collectionView: UICollectionView, shouldSelectItemAt indexPath: IndexPath) -> Bool {
        let item = dataSource.itemIdentifier(for: indexPath)
        return item?.enabled ?? false
    }
}
