// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit
import Combine
import TUMexamKit
import Logging

@available(iOS 14.0, *)
class ResultsViewController: UIViewController {

    private enum Section: Int, CaseIterable {
        case main
        case loading
    }

    private enum CollectionItem: Hashable, Equatable {
        case correctionRow(CorrectionType)
        case loadingRow

        func hash(into hasher: inout Hasher) {
            switch self {
            case let .correctionRow(correction):
                hasher.combine(correction.identifier)
            case .loadingRow:
                hasher.combine("loading")
            }
        }

        static func == (lhs: ResultsViewController.CollectionItem, rhs: ResultsViewController.CollectionItem) -> Bool {
            if case .correctionRow(let lhs) = lhs, case .correctionRow(let rhs) = rhs {
                return lhs.erid == rhs.erid &&
                    lhs.focusedProblem == rhs.focusedProblem &&
                    lhs.focusedSubproblem == rhs.focusedSubproblem &&
                    lhs.focusedPass == rhs.focusedPass &&
                    lhs.corrected == rhs.corrected &&
                    lhs.available == rhs.available &&
                    lhs.locked == rhs.locked &&
                    lhs.access == rhs.access &&
                    lhs.flagTypeIds == rhs.flagTypeIds
            }
            if case .loadingRow = lhs, case .loadingRow = rhs {
                return true
            }
            return false
        }
    }

    static let filterActiveImage = UIImage(systemName: "line.horizontal.3.decrease.circle.fill")
    static let filterInactiveImage = UIImage(systemName: "line.horizontal.3.decrease.circle")

    // MARK: Properties

    var currentExam: TUMexamKit.ExamType? {
        didSet {
            examChanged()
        }
    }
    var currentExamSubscription: AnyCancellable?
    var filterPublisher: AnyCancellable?
    var correctionController: CorrectionControllerType? {
        didSet {
            oldValue?.delegate = nil
            correctionControllerChanged()
        }
    }
    var correctionsCount = 0
    var flagDefinitionSubscription: AnyCancellable?
    var flagDefinitions = [FlagDefinitionType]() {
        didSet {
            applyDataSourceSnapshot()
            updateFlagLegend()
        }
    }

    var collectionView = UICollectionView(
        frame: .zero,
        collectionViewLayout: createCollectionViewLayout()
    ).autolayout()
    private var dataSource: UICollectionViewDiffableDataSource<Section, CollectionItem>?
    private let columnCount = 9
    private let columnWidths: [CGFloat] = [60, 60, 70, 60, 60, 60, 60, 60, 200]
    private let resizeHidePreferences = [7, 6, 8, 5, 4, 3, 2, 1, 0]
    private let cellSpacing: CGFloat = 8

    let flagLegendView = FlagLegendView().autolayout()

    var filterBarButtonItem: UIBarButtonItem?

    // MARK: UIViewController Lifecycle

    override func viewDidLoad() {
        super.viewDidLoad()

        // Bar
        self.title = L10n.ResultsController.title

        // Refresh Control
        let refreshControl = UIRefreshControl()
                refreshControl.addTarget(self, action: #selector(refreshControlTriggered), for: .valueChanged)
        collectionView.refreshControl = refreshControl
//        refreshControl.beginRefreshing()

        // Filter Bar Item
        let filter = UIBarButtonItem(
            image: Self.filterInactiveImage,
            style: .plain,
            target: self,
            action: #selector(showFilter(sender:))
        )
        filterBarButtonItem = filter
        self.navigationItem.leftBarButtonItem = filter

        // Collection View
        setupCollectionView()

        // Setup Flag legend
        view.addSubview(flagLegendView)
        flagLegendView.addLeadingConstraint(to: \.layoutMarginsGuide.leadingAnchor)
        flagLegendView.addTrailingConstraint(to: \.layoutMarginsGuide.trailingAnchor)
        flagLegendView.addBottomConstraint(to: \.layoutMarginsGuide.bottomAnchor, constant: 8)
        flagLegendView.addHeight(44)

        // Setup exam subscription
        currentExamSubscription = App.shared.$currentExam.receiveOnMain().assignNoRetain(to: \.currentExam, on: self)
        App.shared.currentExam?.load()
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        guard let selected = collectionView.indexPathsForSelectedItems else {
            return
        }
        for sel in selected {
            collectionView.deselectItem(at: sel, animated: true)
        }
    }

    // MARK: Functions

    private func setupCollectionView() {
        // Collection View Setup
        collectionView.backgroundColor = .systemBackground
        view.addSubview(collectionView)
        collectionView.pinToSuperview()

        // Cell Registration
        // swiftlint:disable:next line_length
        let cellRegistration = UICollectionView.CellRegistration<DatacolCell, CollectionItem> { [weak self] cell, _, item in
            guard case .correctionRow(let correction) = item else {
                return
            }

            // let cellConfig = cell.defaultContentConfiguration()
            cell.accessories = [.disclosureIndicator()]
            // cell.contentConfiguration = cellConfig

            cell.dataConfig = self?.getCellDataConfig(correction)
        }

        // swiftlint:disable:next line_length
        let loadingCellRegistration = UICollectionView.CellRegistration<ActivityIndicatorCell, CollectionItem> { cell, _, _ in
            cell.activityIndicator.startAnimating()
        }

        // swiftlint:disable:next line_length
        let headerConfiguration = UICollectionView.SupplementaryRegistration<DatacolHeaderView>(elementKind: DatacolHeaderView.elementKind) { [weak self] view, _, _ in
            view.dataConfig = self?.getHeaderDataConfig()
        }

        // swiftlint:disable:next line_length
        let dataSource = UICollectionViewDiffableDataSource<Section, CollectionItem>(collectionView: collectionView) { collectionView, indexPath, item in
            switch indexPath.section {
            case Section.main.rawValue:
                return collectionView.dequeueConfiguredReusableCell(
                    using: cellRegistration,
                    for: indexPath,
                    item: item
                )
            case Section.loading.rawValue:
                return collectionView.dequeueConfiguredReusableCell(
                    using: loadingCellRegistration,
                    for: indexPath,
                    item: item
                )
            default:
                fatalError("Section out of range")
            }
        }

        dataSource.supplementaryViewProvider = { collectionView, _, indexPath in
            let view = collectionView.dequeueConfiguredReusableSupplementary(using: headerConfiguration, for: indexPath)
            return view
        }

        self.dataSource = dataSource
        collectionView.dataSource = dataSource
        collectionView.delegate = self
    }

    @objc func showFilter(sender: UIBarButtonItem) {
        BreadcrumbManager.addInteraction("ResultsTableViewController - filter button tapped")
        // create view controllers if necessary
        let filterVC = ResultsFilterViewController(filter: self.correctionController?.filter ?? ResultFilter())
        filterPublisher = filterVC.rootView.filter.$filter
            .debounce(for: 0.5, scheduler: RunLoop.main)
            .sink { [weak self] filter in
                guard filter != self?.correctionController?.filter else { return }
                Logger.mainUI.info("result filter changed to \(filter)")
                self?.correctionController?.filter = filter
                self?.updateFilterBarButtonItem()
                self?.correctionController?.forceLoad()

                self?.refreshTableByApplyingCurrentSnapshot()
        }

        let nav = UINavigationController(rootViewController: filterVC)

        nav.modalPresentationStyle = .popover

        // set current filter and use sender for popover
        if let current = self.correctionController?.filter {
            filterVC.rootView.filter.filter = current
        }
        nav.popoverPresentationController?.barButtonItem = sender

        // present
        present(nav, animated: true, completion: nil)
    }

    @objc func refreshControlTriggered() {
        BreadcrumbManager.addInteraction("ResultsTableViewController - refresh control triggered")
        guard let exam = currentExam else { return }
        guard let filter = self.correctionController?.filter else { return }
        correctionController = exam.correctionController(withFilter: filter)
    }

    private func examChanged() {
        guard let exam = currentExam else { return }
        exam.load()

        correctionController = exam.correctionController(withFilter: ResultFilter())
        exam.flagDefinitions().load()
        flagDefinitionSubscription = exam.flagDefinitions().flagDefinitions
            .receiveOnMain()
            .assignNoRetain(to: \.flagDefinitions, on: self)
    }

    private func correctionControllerChanged() {
        guard let controller = correctionController else {
            applyDataSourceSnapshot()
            return
        }

        controller.delegate = self
        controller.load()
    }

    private func applyDataSourceSnapshot(_ animated: Bool = true) {
        guard let controller = correctionController else {
            return
        }
        var snap = NSDiffableDataSourceSnapshot<Section, CollectionItem>()
        snap.appendSections([.main])
        snap.appendItems(controller.corrections.map { CollectionItem.correctionRow($0) }.uniques)
        snap.appendSections([.loading])
        if controller.state.value == .loading {
            snap.appendItems([CollectionItem.loadingRow])
        }

        DispatchQueue.main.async {
            self.dataSource?.apply(snap, animatingDifferences: animated)
        }
    }

    private func refreshTableByApplyingCurrentSnapshot() {
        if collectionView.refreshControl?.isRefreshing ?? false {
            collectionView.refreshControl?.endRefreshing()
        }
        applyDataSourceSnapshot()
    }

    private func updateFilterBarButtonItem() {
        guard let filter = filterBarButtonItem else { return }
        let image: UIImage?
        let filterActive = correctionController?.filter.filterActive ?? false
        if filterActive {
            image = Self.filterActiveImage
        } else {
            image = Self.filterInactiveImage
        }
        filter.image = image
    }

    // swiftlint:disable:next cyclomatic_complexity
    private func getCellDataConfig(_ correction: CorrectionType) -> DatacolCell.Configuration {
        return DatacolCell.Configuration(
            columnCount: columnCount,
            columnWidths: columnWidths,
            columnGenerator: { index in
                switch index {
                case 0...3:
                    return UILabel().autolayout()
                case 4...7:
                    let imageView = UIImageView().autolayout()
                    imageView.contentMode = .left
                    return imageView
                case 8:
                    let stackView = UIStackView().autolayout()
                    stackView.axis = .horizontal
                    stackView.alignment = .center
                    stackView.spacing = 8
                    return stackView
                default:
                    fatalError("Index out of range")
                }
            },
            columnConfigurator: { [weak self] index, view in
                switch index {
                case 0:
                    (view as? UILabel)?.text = correction.erid
                case 1:
                    (view as? UILabel)?.text = String(correction.focusedProblem)
                case 2:
                    (view as? UILabel)?.text = correction.focusedSubproblem
                case 3:
                    (view as? UILabel)?.text = String(correction.focusedPass)
                case 4:
                    (view as? UIImageView)?.image = UIImage(systemName: correction.corrected ? "checkmark" : "")
                case 5:
                    (view as? UIImageView)?.image = UIImage(systemName: correction.available ? "checkmark" : "")
                case 6:
                    (view as? UIImageView)?.image = UIImage(systemName: correction.locked ? "checkmark" : "")
                case 7:
                    (view as? UIImageView)?.image = UIImage(systemName: correction.access ? "checkmark" : "")
                case 8:
                    let stack = view as? UIStackView
                    stack?.arrangedSubviews.forEach({
                        $0.removeFromSuperview()
                    })
                    correction.flagTypeIds.compactMap { id in
                        UIColor(hex: self?.flagDefinitions.first(where: { $0.flagTypeId == id })?.color ?? "#FFFFFF")
                    }.forEach {
                        let imageView = UIImageView(image: UIImage(systemName: "flag.fill"))
                        imageView.tintColor = $0
                        imageView.contentMode = .scaleAspectFit
                        stack?.addArrangedSubview(imageView)
                    }
                    let spacer = UIView().autolayout()
                    spacer.backgroundColor = .clear
                    spacer.setContentHuggingPriority(.fittingSizeLevel, for: .horizontal)
                    spacer.setContentCompressionResistancePriority(.fittingSizeLevel, for: .horizontal)
                    stack?.addArrangedSubview(spacer)
                default:
                    return
                }
            },
            resizeHidePreferences: resizeHidePreferences,
            cellSpacing: cellSpacing
        )
    }

    private func getHeaderDataConfig() -> DatacolHeaderView.Configuration {
        return DatacolHeaderView.Configuration(
            columnCount: columnCount,
            columnWidths: columnWidths,
            columnGenerator: { _ in
                let label = UILabel().autolayout()
                label.font = .systemFont(ofSize: UIFont.smallSystemFontSize)
                return label
            },
            columnConfigurator: { index, view in
                switch index {
                case 0:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.erid
                case 1:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.problem
                case 2:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.subproblem
                case 3:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.pass
                case 4:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.corrected
                case 5:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.available
                case 6:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.locked
                case 7:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.access
                case 8:
                    (view as? UILabel)?.text = L10n.ResultsController.Columns.flags
                default:
                    return
                }
            },
            resizeHidePreferences: resizeHidePreferences,
            cellSpacing: cellSpacing
        )
    }

    private func updateFlagLegend() {
        collectionView.contentInset = UIEdgeInsets(
            top: 0,
            left: 0,
            bottom: flagDefinitions.isEmpty ? 0 : 44 + 8 + 8,
            right: 0
        )
        flagLegendView.isHidden = flagDefinitions.isEmpty
        flagLegendView.flagDefinitions = self.flagDefinitions
    }

    // MARK: Static functions

    static private func createCollectionViewLayout() -> UICollectionViewLayout {
        let layout = UICollectionViewCompositionalLayout { section, layoutEnv in
            if section == Section.main.rawValue {
                var config = UICollectionLayoutListConfiguration(appearance: .plain)
                config.headerMode = .supplementary
                let list = NSCollectionLayoutSection.list(using: config, layoutEnvironment: layoutEnv)

                let headerSize = NSCollectionLayoutSize(
                    widthDimension: .fractionalWidth(1),
                    heightDimension: .estimated(20)
                )
                let header = NSCollectionLayoutBoundarySupplementaryItem(
                    layoutSize: headerSize,
                    elementKind: DatacolHeaderView.elementKind,
                    alignment: .top
                )
                header.pinToVisibleBounds = true
                header.zIndex = 1023
                list.boundarySupplementaryItems = [header]
                return list
            } else {
                let itemSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                      heightDimension: .fractionalHeight(1.0))
                let item = NSCollectionLayoutItem(layoutSize: itemSize)

                let groupSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                       heightDimension: .absolute(60))
                let group = NSCollectionLayoutGroup.vertical(layoutSize: groupSize, subitems: [item])

                let section = NSCollectionLayoutSection(group: group)
                return section
            }
        }

        return layout
    }
}

@available(iOS 14.0, *)
extension ResultsViewController: UICollectionViewDelegate {
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        BreadcrumbManager.addInteraction("ResultsTableViewController - selected row at index \(indexPath.row) ")
        guard let controller = correctionController else { return }
        App.shared.coordinator?.startCorrection(using: controller, atIndex: indexPath.row, target: self)
    }

    func scrollViewDidScroll(_ scrollView: UIScrollView) {
        guard let controller = correctionController,
              controller.state.value != .loading else {
            return
        }

        let height = scrollView.frame.size.height
        let contentOffset = scrollView.contentOffset.y
        let distanceFromBottom = scrollView.contentSize.height - contentOffset
        if distanceFromBottom < height, controller.corrections.count < controller.estimatedCount {
            controller.loadMore()
            applyDataSourceSnapshot()
        }
    }
}

@available(iOS 14.0, *)
extension ResultsViewController: CorrectionControllerDelegate {
    func controller(_ controller: CorrectionControllerType, didLoadIndexes indexes: IndexSet) {
        refreshTableByApplyingCurrentSnapshot()
        if correctionsCount != controller.estimatedCount {
            correctionsCount = controller.estimatedCount
        }

        scrollViewDidScroll(collectionView)
    }

    func controllerDidReset(_ controller: CorrectionControllerType) {
        refreshTableByApplyingCurrentSnapshot()
    }

    func controllerGotEmptyResults(_ controller: CorrectionControllerType) {
        refreshTableByApplyingCurrentSnapshot()
    }
}
