// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import UIKit
import TUMexamKit

@available(iOS 14.0, *)
class ExamSetupViewController: UIViewController {

    // MARK: Properties

    enum Section: Int, CaseIterable {
        case loading = 0
        case main
    }

    struct ExamSetupItem: Hashable {
        // swiftlint:disable:next nesting
        enum ItemType {
            case main
            case loading
        }

        let exam: ExamType?
        let type: ItemType

        func hash(into hasher: inout Hasher) {
            hasher.combine(exam?.identifier)
            hasher.combine(type)
        }

        static func == (lhs: ExamSetupViewController.ExamSetupItem,
                        rhs: ExamSetupViewController.ExamSetupItem) -> Bool {
            return lhs.exam?.identifier == rhs.exam?.identifier && lhs.type == rhs.type
        }
    }

    let titleLabel = UILabel().autolayout()
    let stackView = UIStackView().autolayout()
    let collectionView = UICollectionView(
        frame: .zero,
        collectionViewLayout: createCollectionViewLayout()
    ).autolayout()
    let refreshControl = UIRefreshControl()

    var examsObserver: AnyCancellable?
    var dataSource: UICollectionViewDiffableDataSource<Section, ExamSetupItem>!

    // MARK: UIViewController Lifecycle

    override func viewDidLoad() {
        super.viewDidLoad()

        // Load Exams & Setup Datasource
        App.shared.kit?.examList.load()
        examsObserver = App.shared.kit?.examList.exams.receiveOnMain().sink { [weak self] exams in
            if self?.refreshControl.isRefreshing ?? false {
                self?.refreshControl.endRefreshing()
            }
            self?.applyDataSourceSnapshot(exams.map({ ExamSetupItem(exam: $0, type: .main) }))
        }
        // swiftlint:disable:next line_length
        let cellRegistration = UICollectionView.CellRegistration<UICollectionViewListCell, ExamSetupItem> { cell, _, item in
            var content = cell.defaultContentConfiguration()
            content.text = item.exam?.name.value
            cell.contentConfiguration = content
            cell.accessories = [.disclosureIndicator()]
        }
        // swiftlint:disable:next line_length
        let loadingRegistration = UICollectionView.CellRegistration<ActivityIndicatorCell, ExamSetupItem> { cell, _, _ in
            cell.activityIndicator.startAnimating()
        }
        // swiftlint:disable:next line_length
        let dataSource = UICollectionViewDiffableDataSource<Section, ExamSetupItem>(collectionView: collectionView) { collectionView, indexPath, item in
            switch item.type {
            case .main:
                return collectionView.dequeueConfiguredReusableCell(using: cellRegistration,
                                                                    for: indexPath,
                                                                    item: item)
            case .loading:
                return collectionView.dequeueConfiguredReusableCell(using: loadingRegistration,
                                                                    for: indexPath,
                                                                    item: item)
            }
        }
        self.dataSource = dataSource

        // View
        view.backgroundColor = .systemBackground

        // Label
        titleLabel.font = .systemFont(ofSize: 36.0, weight: .semibold)
        titleLabel.text = L10n.Setup.Exam.title

        // Collection View
        collectionView.backgroundColor = .systemBackground
        collectionView.delegate = self
        collectionView.dataSource = dataSource
        collectionView.refreshControl = refreshControl
        refreshControl.addAction(UIAction { _ in
            App.shared.kit?.examList.forceRefresh()
        }, for: .valueChanged)

        // Stack View
        stackView.axis = .vertical
        stackView.alignment = .center
        stackView.spacing = 32.0
        stackView.addArrangedSubview(titleLabel)
        stackView.addArrangedSubview(collectionView)

        view.addSubview(stackView)

        // Constraints
        stackView.centerInSuperview()
        collectionView.width(to: \.widthAnchor, of: self.view, multiplier: 0.5)
        collectionView.height(to: \.heightAnchor, of: self.view, multiplier: 0.5)

        applyDataSourceSnapshot([], animated: false)
    }

    private func applyDataSourceSnapshot(_ items: [ExamSetupItem], animated: Bool = true) {
        var snapshot = NSDiffableDataSourceSnapshot<Section, ExamSetupItem>()
        snapshot.appendSections([.loading])
        if App.shared.kit?.examList.state.value != .loaded {
            snapshot.appendItems([ExamSetupItem(exam: nil, type: .loading)])
        }
        snapshot.appendSections([.main])
        snapshot.appendItems(items)
        dataSource.apply(snapshot, animatingDifferences: animated)
    }

    private static func createCollectionViewLayout() -> UICollectionViewLayout {
        let layout = UICollectionViewCompositionalLayout { sec, layoutEnv in
            var config = UICollectionLayoutListConfiguration(appearance: .plain)
            if sec == Section.loading.rawValue {
                config.showsSeparators = false
            }
            let list = NSCollectionLayoutSection.list(using: config, layoutEnvironment: layoutEnv)
            return list
        }

        return layout
    }
}

@available(iOS 14.0, *)
extension ExamSetupViewController: UICollectionViewDelegate {
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        guard let examItem = dataSource.itemIdentifier(for: indexPath),
              let exam = examItem.exam else {
            collectionView.deselectItem(at: indexPath, animated: false)
            return
        }
        App.shared.setCurrentExam(exam)
        App.shared.coordinator?.showMain()
    }
}
