// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import UIKit
import TUMexamKit

class ExamTableViewCell: UITableViewCell, CellClassReuseIdentifier {
	var exam: ExamType? { didSet { updateCell() } }

	override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
		super.init(style: .default, reuseIdentifier: reuseIdentifier)
		setup()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	func setup() {
		self.accessoryType = .disclosureIndicator
	}

	func updateCell() {
		guard let name = exam?.name.value else {
			self.textLabel?.text = ""
			return
		}
		self.textLabel?.text = name
	}
}

/// This is deprecated and for iOS 13 compability only
class LegacyExamSetupViewController: UIViewController {
	let tableView: UITableView = UITableView(frame: CGRect(x: 0, y: 0, width: 320, height: 200), style: .plain)
		.autolayout()
	let refreshControl = UIRefreshControl()

	let titleLabel: UILabel = UILabel().autolayout()
	let stackView: UIStackView = UIStackView().autolayout()

	var examsObserver: AnyCancellable?
	var exams: [ExamType] = []

    override func viewDidLoad() {
        super.viewDidLoad()

		view.backgroundColor = .systemBackground

		titleLabel.font = .systemFont(ofSize: 36.0, weight: .semibold)
		titleLabel.text = L10n.Setup.Exam.title

		stackView.axis = .vertical
		stackView.alignment = .center
		stackView.spacing = 32.0
		stackView.addArrangedSubview(titleLabel)

		stackView.addArrangedSubview(tableView)

		view.addSubview(stackView)
		stackView.centerInSuperview()

		NSLayoutConstraint.activate([
			tableView.heightAnchor.constraint(equalTo: view.heightAnchor, multiplier: 0.5),
			tableView.widthAnchor.constraint(equalTo: view.widthAnchor, multiplier: 0.5)
		])

		tableView.tableFooterView = UIView(frame: .zero)
		tableView.register(ExamTableViewCell.self)
		tableView.delegate = self
		tableView.dataSource = self
		tableView.refreshControl = refreshControl
		refreshControl.addTarget(self, action: #selector(refresh), for: .valueChanged)

		guard let kit = App.shared.kit else {
			fatalError("can't select an exam without a kit!")
		}

		examsObserver = kit.examList.exams.receiveOnMain().sink { [weak self] exams in
			self?.examListChanged(exams)
		}

		kit.examList.load()
    }

	func examListChanged(_ exams: [ExamType]) {
		self.exams = exams
		tableView.reloadData()
	}

	@objc func refresh() {
		App.shared.kit?.examList.load()
	}
}

extension LegacyExamSetupViewController: UITableViewDelegate, UITableViewDataSource {
	func numberOfSections(in tableView: UITableView) -> Int {
		return 1
	}

	func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
		return exams.count
	}

	func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
		let cell: ExamTableViewCell = tableView.dequeueReusableCell(for: indexPath)

		let exam = exams[indexPath.row]
		cell.exam = exam

		return cell
	}

	func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
		let exam = exams[indexPath.row]

		App.shared.setCurrentExam(exam)
		App.shared.coordinator?.showMain()
	}
}
