// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit

class SnippetTableViewCell: UITableViewCell, CellClassReuseIdentifier {
    let nameField: UITextField = UITextField().autolayout()
	let snippetImageView: UIImageView = UIImageView().autolayout()

	var snippet: SnippetType? {
		didSet {
			updateCell()
		}
	}

	override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
		super.init(style: .default, reuseIdentifier: reuseIdentifier)

		setup()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	func setup() {
		snippetImageView.contentMode = .scaleAspectFit
        nameField.font = UIFont.preferredFont(forTextStyle: .headline)
        nameField.placeholder = L10n.SnippetsController.Name.placeholder
        nameField.addTarget(self, action: #selector(editingChanged), for: .editingDidEnd)
        nameField.delegate = self

        if #available(iOS 14, *) {
            let scribble = UIScribbleInteraction(delegate: self)
            nameField.addInteraction(scribble)
        }

		contentView.addSubview(snippetImageView)
        contentView.addSubview(nameField)

        let marginGuides = contentView.layoutMarginsGuide

        NSLayoutConstraint.activate([
            nameField.topAnchor.constraint(equalTo: marginGuides.topAnchor, constant: 0.0),
            nameField.leftAnchor.constraint(equalTo: marginGuides.leftAnchor, constant: 0.0),
            nameField.rightAnchor.constraint(equalTo: marginGuides.rightAnchor, constant: 0.0),
            nameField.bottomAnchor.constraint(equalTo: snippetImageView.topAnchor, constant: -5.0),

            snippetImageView.bottomAnchor.constraint(equalTo: contentView.bottomAnchor),
            snippetImageView.leftAnchor.constraint(equalTo: contentView.leftAnchor),
            snippetImageView.rightAnchor.constraint(equalTo: contentView.rightAnchor)
        ])
	}

	func updateCell() {
		if let snippet = snippet {
			snippetImageView.image = snippet.image

            nameField.text = snippet.name
		}
	}

    @objc
    func editingChanged() {
        guard let snippet = snippet else { return }
        snippet.name = nameField.text ?? ""
    }
}

extension SnippetTableViewCell: UITextFieldDelegate {
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        textField.resignFirstResponder()
        return false
    }
}

@available(iOS 14.0, *)
extension SnippetTableViewCell: UIScribbleInteractionDelegate {
    func scribbleInteraction(_ interaction: UIScribbleInteraction, shouldBeginAt location: CGPoint) -> Bool {
        return false
    }
}
