// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Instructions

protocol SnippetsTableViewControllerDelegate: AnyObject {
	func snippetsViewControllerWantsToAdd(_ viewController: SnippetsTableViewController)

    func snippetsViewController(_ viewController: SnippetsTableViewController, didSelect snippet: SnippetType)
}

class SnippetsTableViewController: UITableViewController {
	weak var delegate: SnippetsTableViewControllerDelegate?

	let problem: ProblemType
	let exam: ExamType

    var focus: SnippetType?

	var matching: [SnippetType] = []
	var other: [SnippetType] = []

    var manager: SnippetManagerType {
        exam.snippets
    }

    let coachMarksController = CoachMarksController()
    var coachContent = [(String, UIView?, CGPoint?, CutoutPathMaker?)]()

    init(problem: ProblemType, focus: SnippetType? = nil) {
		self.problem = problem
        self.focus = focus
		guard let exam = App.shared.currentExam else {
			fatalError("trying to create a snippet table view without a selected exam")
		}
		self.exam = exam

		super.init(style: .grouped)
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override func viewDidLoad() {
		super.viewDidLoad()

		// Uncomment the following line to preserve selection between presentations
		// self.clearsSelectionOnViewWillAppear = false
		navigationItem.title = L10n.SnippetsController.title
		navigationItem.rightBarButtonItem
			= UIBarButtonItem(barButtonSystemItem: .add, target: self, action: #selector(addSnippet))

		// Uncomment the following line to display an Edit button in the navigation bar for this view controller.
		// self.navigationItem.rightBarButtonItem = self.editButtonItem
		tableView.register(SnippetTableViewCell.self)
		tableView.allowsSelection = true

        tableView.dragInteractionEnabled = true
        tableView.dragDelegate = self

		updateArrays()
	}

    override func viewDidAppear(_ animated: Bool) {
        BreadcrumbManager.addNavigation("[SnippetsTableViewController] viewDidAppear")
        super.viewDidAppear(animated)

        if let focus = self.focus, let indexPath = indexPathOf(focus) {
            tableView.scrollToRow(at: indexPath, at: .none, animated: false)
            if let cell = tableView.cellForRow(at: indexPath) as? SnippetTableViewCell {
                cell.nameField.becomeFirstResponder()
            }
            self.focus = nil
        }

        if !((App.shared.settings[HiddenSettings.didShowSnippetTutorial] as? Bool) ?? false) {
            BreadcrumbManager.addNavigation("[SnippetsTableViewController] showing tutorial")
            setupCoachMarks()
            coachMarksController.dataSource = self
            coachMarksController.delegate  = self
            let skipView = CoachMarkSkipDefaultView()
            skipView.setTitle("Skip Tutorial", for: .normal)
            coachMarksController.skipView = skipView
            coachMarksController.start(in: .window(over: self))
        }
    }

	func updateArrays() {
        matching = manager.snippets(forProblem: problem.identifier)
        other = manager.snippets(excludingProblem: problem.identifier)
	}

    func reload() {
        updateArrays()
        tableView.reloadData()
    }

	// MARK: - Table view data source

	override func numberOfSections(in tableView: UITableView) -> Int {
		return 2
	}

	func targetArray(forSection section: Int) -> [SnippetType] {
		switch section {
		case 0:
			return matching
		case 1:
			return other
		default:
			return []
		}
	}

    func snippetAt(_ indexPath: IndexPath) -> SnippetType {
        return targetArray(forSection: indexPath.section)[indexPath.row]
    }

    func indexPathOf(_ snippet: SnippetType) -> IndexPath? {
        if let idx = matching.firstIndex(where: { $0 === snippet }) {
            return IndexPath(row: idx, section: 0)
        }
        if let idx = other.firstIndex(where: { $0 === snippet }) {
            return IndexPath(row: idx, section: 1)
        }
        return nil
    }

	override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
		return targetArray(forSection: section).count
	}

	@objc func addSnippet() {
        BreadcrumbManager.addInteraction("[SnippetsTableViewController] user tapped add button")
		delegate?.snippetsViewControllerWantsToAdd(self)
	}

	override func tableView(_ tableView: UITableView, titleForHeaderInSection section: Int) -> String? {
		switch section {
		case 0:
			return L10n.SnippetsController.Section.Matching.title
		case 1:
			return L10n.SnippetsController.Section.Others.title
		default:
			return nil
		}
	}

	override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
		let cell: SnippetTableViewCell = tableView.dequeueReusableCell(for: indexPath)

		cell.snippet = snippetAt(indexPath)

		return cell
	}

	override func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
		return 88.0
	}

    override func tableView(_ tableView: UITableView,
                            commit editingStyle: UITableViewCell.EditingStyle,
                            forRowAt indexPath: IndexPath) {
        if editingStyle == .delete {
            // swiftlint:disable:next line_length
            BreadcrumbManager.addInteraction("[SnippetsTableViewController] user deleted snippet at index \(indexPath.row)")
            let snippet = snippetAt(indexPath)
            manager.removeSnippet(snippet)
            reload()
        }
    }

    override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        // swiftlint:disable:next line_length
        BreadcrumbManager.addInteraction("[SnippetsTableViewController] user selected snippet at index \(indexPath.row)")
        tableView.deselectRow(at: indexPath, animated: true)
        let snippet = snippetAt(indexPath)
        delegate?.snippetsViewController(self, didSelect: snippet)
    }

    // MARK: Coach Marks

    private func setupCoachMarks() {
        coachContent = [
            (L10n.Tutorial.Snippets.overview, self.view, nil, nil),
            (L10n.Tutorial.Snippets.add,
             self.navigationItem.rightBarButtonItem?.value(forKey: "view") as? UIView,
             nil,
             nil)
        ]
    }

}

extension SnippetsTableViewController: UITableViewDragDelegate {
    func tableView(_ tableView: UITableView,
                   itemsForBeginning session: UIDragSession,
                   at indexPath: IndexPath) -> [UIDragItem] {
        let snippet = snippetAt(indexPath)
        guard let image = snippet.image else { return [] }
        let provider = NSItemProvider(object: image)
        let item = UIDragItem(itemProvider: provider)
        item.localObject = snippet
        return [item]
    }

    func tableView(_ tableView: UITableView, dragSessionWillBegin session: UIDragSession) {
        self.navigationController?.dismiss(animated: true, completion: nil)
    }
}

extension SnippetsTableViewController: CoachMarksControllerDataSource {
    func numberOfCoachMarks(for coachMarksController: CoachMarksController) -> Int {
        return coachContent.count
    }

    func coachMarksController(_ coachMarksController: CoachMarksController, coachMarkAt index: Int) -> CoachMark {
        let content = coachContent[index]
        var mark = coachMarksController.helper.makeCoachMark(for: content.1,
                                                             pointOfInterest: content.2,
                                                             cutoutPathMaker: content.3)
        if index == 0 {
            mark.isDisplayedOverCutoutPath = true
        }

        return mark
    }

    func coachMarksController(_ coachMarksController: CoachMarksController,
                              coachMarkViewsAt index: Int,
                              madeFrom coachMark: CoachMark) -> (bodyView: (UIView & CoachMarkBodyView),
                                                                 arrowView: (UIView & CoachMarkArrowView)?) {
        let coachViews = coachMarksController.helper.makeDefaultCoachViews(withArrow: true,
                                                                           arrowOrientation: coachMark.arrowOrientation)
        coachViews.bodyView.hintLabel.text = coachContent[index].0
        coachViews.bodyView.nextLabel.text = L10n.gotIt

        return (coachViews.bodyView, coachViews.arrowView)
    }

    func coachMarksController(_ coachMarksController: CoachMarksController,
                              constraintsForSkipView skipView: UIView,
                              inParent parentView: UIView) -> [NSLayoutConstraint]? {
        return [
            skipView.bottomAnchor.constraint(equalTo: parentView.safeAreaLayoutGuide.bottomAnchor),
            skipView.centerXAnchor.constraint(equalTo: parentView.centerXAnchor)
        ]
    }
}

extension SnippetsTableViewController: CoachMarksControllerDelegate {
    func coachMarksController(_ coachMarksController: CoachMarksController, didEndShowingBySkipping skipped: Bool) {
        App.shared.settings[HiddenSettings.didShowSnippetTutorial] = true
    }
}
