// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import SwiftUI
import Combine
import TUMexamKit
import Instructions
import Sentry

class StatusCollectionViewController: UICollectionViewController {

    enum StatusCollectionSection: Int, CaseIterable {
        case loading = 0
        case issue
        case main
    }

    enum StatusCollectionItem: Hashable {
        case loading
        case issue
        case problem(CorrectionProblemStatusType, accessRights: [Int])

        func hash(into hasher: inout Hasher) {
            switch self {
            case .loading:
                hasher.combine("loading")
            case .issue:
                hasher.combine("issue")
            case .problem(let problem, _):
                hasher.combine(problem.problem)
            }
        }

        static func == (lhs: StatusCollectionViewController.StatusCollectionItem,
                        rhs: StatusCollectionViewController.StatusCollectionItem) -> Bool {
            if case .loading = lhs, case .loading = rhs {
                return true
            }
            if case .issue = lhs, case .issue = rhs {
                return true
            }
            if case .problem(let lhs, _) = lhs, case .problem(let rhs, _) = rhs {
                return lhs.problem == rhs.problem
            }
            return false
        }
    }

    var diffableDataSource: UICollectionViewDiffableDataSource<StatusCollectionSection, StatusCollectionItem>!

    var coldStart = true
    var shouldRefreshWhenAppearing = false

	weak var currentExam: TUMexamKit.ExamType? {
		didSet {
			examChanged()
		}
	}

	var currentExamSubscription: AnyCancellable?
	var statsSubscription: AnyCancellable?
    var scanIssuesSubscription: AnyCancellable?
    var scansStartedSubscription: AnyCancellable?
    var accessRightsSubscription: AnyCancellable?

	var refreshControl = UIRefreshControl()

	var problems: [CorrectionProblemStatusType] = [] {
		didSet {
			if problems.count > 0, self.refreshControl.isRefreshing {
				self.refreshControl.endRefreshing()
			}
            if problems.count > 0 && coldStart {
                coldStart = false
                // Unlock zombies
                currentExam?.freeAllLocksInBackground()
            }
            applyDatasourceSnapshot(true)
		}
	}

    var allScanIssuesFixed: Bool? {
        didSet {
            if let val = allScanIssuesFixed {
                checkTutorial()
                applyDatasourceSnapshot(true)
                if !val {
                    self.refreshControl.endRefreshing()
                }
            }
        }
    }

    var scansStarted: Bool? {
        didSet {
            if let val = scansStarted {
                checkTutorial()
                applyDatasourceSnapshot(true)
                if !val {
                    self.refreshControl.endRefreshing()
                }
            }
        }
    }

    var isAdmin: Bool = false
    var accessRights: [Int: [Int]] = [:]

    /// Tutorial
    let coachMarksController = CoachMarksController()
    var coachContent = [(String, UIView?, CGPoint?, CutoutPathMaker?)]()

	init() {
		let layout = Self.createLayout()

		super.init(collectionViewLayout: layout)

		self.title = L10n.StatusController.title
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	class func createLayout() -> UICollectionViewLayout {
		let layout = UICollectionViewCompositionalLayout { sec, layoutEnvironment in
            if sec == StatusCollectionSection.main.rawValue {
                var itemsPerRowCount = 2
                if layoutEnvironment.container.effectiveContentSize.width / 3 > 450 {
                    itemsPerRowCount = 3
                } else if layoutEnvironment.container.effectiveContentSize.width / 2 < 450 {
                    itemsPerRowCount = 1
                }

                let itemSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                      heightDimension: .fractionalHeight(1.0))
                let item = NSCollectionLayoutItem(layoutSize: itemSize)

                let groupSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                       heightDimension: .absolute(310))
                let group = NSCollectionLayoutGroup.horizontal(layoutSize: groupSize,
                                                               subitem: item,
                                                               count: itemsPerRowCount)
                let section = NSCollectionLayoutSection(group: group)
                return section
            } else {
                let height: NSCollectionLayoutDimension =
                    sec == StatusCollectionSection.loading.rawValue ? .absolute(80) : .estimated(120)
                let itemSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                      heightDimension: height)
                let item = NSCollectionLayoutItem(layoutSize: itemSize)

                let groupSize = NSCollectionLayoutSize(widthDimension: .fractionalWidth(1.0),
                                                       heightDimension: .fractionalHeight(1.0))
                let group = NSCollectionLayoutGroup.vertical(layoutSize: groupSize, subitems: [item])

                let section = NSCollectionLayoutSection(group: group)
                return section
            }
		}
		return layout
	}

    override func viewDidLoad() {
        super.viewDidLoad()

        // Register cell classes
		self.collectionView.register(StatusCollectionViewCell.self)
        self.collectionView.register(ActivityIndicatorCell.self)
        self.collectionView.register(StatusIssuesCell.self)

        // Setup Data Source
        // swiftlint:disable:next line_length
        let dataSource = UICollectionViewDiffableDataSource<StatusCollectionSection, StatusCollectionItem>(collectionView: collectionView) { [weak self] (collectionView, indexPath, item) -> UICollectionViewCell? in
            switch item {
            case .problem(let problem, let accessRights):
                let cell: StatusCollectionViewCell = collectionView.dequeueReusableCell(for: indexPath)
                cell.passButtons.forEach {
                    $0.isEnabled = self?.allScanIssuesFixed ?? false
                }
                cell.problem = problem
                cell.accessRights = accessRights
                cell.isAdmin = self?.isAdmin ?? false
                cell.delegate = self
                return cell
            case .loading:
                let cell: ActivityIndicatorCell = collectionView.dequeueReusableCell(for: indexPath)
                cell.activityIndicator.startAnimating()
                return cell
            case .issue:
                let cell: StatusIssuesCell = collectionView.dequeueReusableCell(for: indexPath)
                if let scansStarted = self?.scansStarted, !scansStarted {
                    cell.titleLabel.text = L10n.StatusController.ScanNotStarted.title
                    cell.label.text = L10n.StatusController.ScanNotStarted.message
                } else {
                    cell.titleLabel.text = L10n.StatusController.Issues.title
                    cell.label.text = L10n.StatusController.Issues.message
                }
                return cell
            }
        }
        self.diffableDataSource = dataSource
        collectionView.dataSource = dataSource
        collectionView.delegate = self
        applyDatasourceSnapshot(false)

        // Do any additional setup after loading the view.
		collectionView.backgroundColor = UIColor.systemBackground

		refreshControl.addTarget(self, action: #selector(triggerRefresh(sender:)), for: .valueChanged)
        collectionView.refreshControl = refreshControl

        // subscribe to current exam changes
        currentExamSubscription = App.shared.$currentExam.receiveOnMain().assignNoRetain(to: \.currentExam, on: self)
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        BreadcrumbManager.addNavigation("[StatusCollectionViewController] viewWillAppear")

        if shouldRefreshWhenAppearing {
            shouldRefreshWhenAppearing = false
            currentExam?.load()
            currentExam?.correctionStatus.forceReload()
        }
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        checkTutorial()
        self.collectionView.flashScrollIndicators()
    }

    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        self.coachMarksController.stop(immediately: true)
    }

    func checkTutorial() {
        guard let scansStarted = scansStarted, scansStarted,
              let allScanIssuesFixed = self.allScanIssuesFixed, allScanIssuesFixed else {
            return
        }

        if !((App.shared.settings[HiddenSettings.didShowStatusTutorial] as? Bool) ?? false) {
            let alert = UIAlertController(title: L10n.Tutorial.StartAlert.title,
                                          message: L10n.Tutorial.StartAlert.message,
                                          preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: L10n.Tutorial.StartAlert.ok, style: .default, handler: { _ in
                BreadcrumbManager.addInteraction("[StatusCollectionViewController] Tutorial accepted")
                self.setupCoachContent()
                self.coachMarksController.dataSource = self
                self.coachMarksController.delegate = self
                let skipView = CoachMarkSkipDefaultView()
                skipView.setTitle(L10n.skipTutorial, for: .normal)
                self.coachMarksController.skipView = skipView
                self.coachMarksController.start(in: .window(over: self))
            }))
            alert.addAction(UIAlertAction(title: L10n.Tutorial.StartAlert.skip, style: .destructive, handler: { _ in
                BreadcrumbManager.addInteraction("[StatusCollectionViewController] Tutorial skipped")
                App.shared.settings[HiddenSettings.didShowStatusTutorial] = true
                App.shared.settings[HiddenSettings.didShowPagingCorrectionTutorial] = true
                App.shared.settings[HiddenSettings.didShowSnippetTutorial] = true
            }))
            alert.show(animated: true)
            BreadcrumbManager.addNavigation("[StatusCollectionViewController] Show Tutorial Dialog")
        }
    }

	func examChanged() {
		guard let exam = currentExam else { return }
        scansStarted = nil
        allScanIssuesFixed = nil
		exam.load()
		exam.correctionStatus.load()
        exam.issues.load()
        App.shared.kit?.accessRights.load()
		// add subscriptions for name and stats
		statsSubscription = exam.correctionStatus.problems
            .receiveOnMain().assignNoRetain(to: \.problems, on: self)
        scanIssuesSubscription = exam.issues.allScanIssuesFixed
            .receiveOnMain().assignNoRetain(to: \.allScanIssuesFixed, on: self)
        scansStartedSubscription = exam.issues.scansStarted
            .receiveOnMain().assignNoRetain(to: \.scansStarted, on: self)
        accessRightsSubscription = App.shared.kit?.accessRights.state
            .receiveOnMain()
            .sink { [weak self] state in
                guard state == .loaded else { return }
                guard let accessRights = App.shared.kit?.accessRights.getAccessRights(exam.identifier) else {
                    return
                }

                for rule in accessRights {
                    if let pass = rule.correctionPass {
                        self?.accessRights[rule.problem] = (self?.accessRights[rule.problem] ?? []) + [pass]
                    } else {
                        self?.accessRights[rule.problem] = [1, 2]
                    }
                }

                self?.isAdmin = App.shared.kit?.accessRights.admin ?? false
                self?.applyDatasourceSnapshot()
            }
	}

    @objc func triggerRefresh(sender: UIRefreshControl) {
        BreadcrumbManager.addInteraction("[StatusCollectionViewController] refresh triggered")
		guard let exam = currentExam else { return }
        refreshControl.beginRefreshing()
		exam.load()
		exam.correctionStatus.forceReload()
        exam.issues.forceReload()
        accessRights = [:]
        App.shared.kit?.accessRights.forceReload()
    }

    // MARK: Private Functions

    private func applyDatasourceSnapshot(_ animated: Bool = false) {
        var snapshot = NSDiffableDataSourceSnapshot<StatusCollectionSection, StatusCollectionItem>()
        snapshot.appendSections([.loading])
        if currentExam?.correctionStatus.state.value != .loaded &&
            (allScanIssuesFixed ?? true) &&
            (scansStarted ?? true) {
            snapshot.appendItems([.loading])
        }
        snapshot.appendSections([.issue])
        if let scansStarted = scansStarted, !scansStarted {
            snapshot.appendItems([.issue])
        }
        if let issuesFixed = allScanIssuesFixed, !issuesFixed {
            snapshot.appendItems([.issue])
        }
        if let issuesFixed = allScanIssuesFixed, issuesFixed, let scansStarted = scansStarted, scansStarted {
            snapshot.appendSections([.main])
            snapshot.appendItems(problems.map {
                let accessRights = self.accessRights[$0.problem] ?? []
                return .problem($0, accessRights: accessRights)
            })
        }

        diffableDataSource.apply(snapshot, animatingDifferences: animated, completion: nil)
    }

    private func setupCoachContent() {
        var cellView: UIView?
        if collectionView.visibleCells.count > 0,
           let path = collectionView.indexPathsForVisibleItems.sorted().first {
            cellView = collectionView.cellForItem(at: path)
        } else {
            cellView = collectionView
        }
        // Split View with Sidebar

        // swiftlint:disable:next line_length
        guard let sidebarVC = self.splitViewController?.viewController(for: .primary) as? SidebarCollectionViewController,
              let statusCell = sidebarVC.collectionView.cellForItem(at: IndexPath(row: 0, section: 0)),
              let resultsCell = sidebarVC.collectionView.cellForItem(at: IndexPath(row: 1, section: 0)) else {
            return
        }

        coachContent = [
            (L10n.Tutorial.Status.Tabbar.status,
             statusCell,
             nil,
             nil),
            (L10n.Tutorial.Status.navbar, self.navigationController?.navigationBar, nil, nil),
            (L10n.Tutorial.Status.cell, cellView, nil, nil),
            (L10n.Tutorial.Status.Tabbar.results,
             resultsCell,
             nil,
             nil)
        ]
    }
}

extension StatusCollectionViewController: StatusCollectionViewCellDelegate {
	func statusCell(_ cell: StatusCollectionViewCell, didTapButtonWithPass pass: Int) {
		guard let problem = cell.problem, let exam = currentExam else { return }
        // swiftlint:disable:next line_length
        BreadcrumbManager.addInteraction("StatusCollectionViewCell tapped for exam \(exam) problem \(problem) pass \(pass)")
        let controller = exam.interactiveCorrectionPromiseController(forProblem: problem.problem, pass: pass)

		App.shared.coordinator?.startCorrection(using: controller, target: self)
        shouldRefreshWhenAppearing = true
	}

    func statusCellTappedInfoButton(_ cell: StatusCollectionViewCell) {
        guard let problem = cell.problem, let exam = currentExam else { return }
        BreadcrumbManager.addInteraction("StatusCollectionViewCell tapped info button exam \(exam) problem \(problem)")
        App.shared.coordinator?.showSubproblemStatus(problem, target: self, source: cell)
        shouldRefreshWhenAppearing = true
    }
}

extension StatusCollectionViewController: CoachMarksControllerDataSource {
    func numberOfCoachMarks(for coachMarksController: CoachMarksController) -> Int {
        return coachContent.count
    }

    func coachMarksController(_ coachMarksController: CoachMarksController, coachMarkAt index: Int) -> CoachMark {
        let content = coachContent[index]
        return coachMarksController.helper.makeCoachMark(for: content.1,
                                                         pointOfInterest: content.2,
                                                         cutoutPathMaker: content.3)
    }

    func coachMarksController(_ coachMarksController: CoachMarksController,
                              coachMarkViewsAt index: Int,
                              madeFrom coachMark: CoachMark) -> (bodyView: (UIView & CoachMarkBodyView),
                                                                 arrowView: (UIView & CoachMarkArrowView)?) {
        let coachViews = coachMarksController.helper.makeDefaultCoachViews(withArrow: true,
                                                                           arrowOrientation: coachMark.arrowOrientation)
        coachViews.bodyView.hintLabel.text = coachContent[index].0
        coachViews.bodyView.nextLabel.text = L10n.gotIt

        return (coachViews.bodyView, coachViews.arrowView)
    }

    func coachMarksController(_ coachMarksController: CoachMarksController,
                              constraintsForSkipView skipView: UIView,
                              inParent parentView: UIView) -> [NSLayoutConstraint]? {
        return [
            skipView.bottomAnchor.constraint(equalTo: parentView.safeAreaLayoutGuide.bottomAnchor),
            skipView.centerXAnchor.constraint(equalTo: parentView.centerXAnchor)
        ]
    }
}

extension StatusCollectionViewController: CoachMarksControllerDelegate {
    func coachMarksController(_ coachMarksController: CoachMarksController, didEndShowingBySkipping skipped: Bool) {
        BreadcrumbManager.addNavigation("[StatusCollectionViewController] tutorial ended (skipped = \(skipped))")
        App.shared.settings[HiddenSettings.didShowStatusTutorial] = true
    }
}
