// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

class ToastBubbleView: UIView {

    var isPresented = false
    var toasts: [Toast] = []

    private let titleLabel = UILabel().autolayout()
    private let subtitleLabel = UILabel().autolayout()
    private let imageView = UIImageView().autolayout()
    private let activityIndicator = UIActivityIndicatorView().autolayout()
    private let stackView = UIStackView().autolayout()
    private let containerView = ShadowView().autolayout()

    private var showAnimator: UIViewPropertyAnimator?
    private var hideAnimator: UIViewPropertyAnimator?

    override init(frame: CGRect) {
        super.init(frame: frame)
        setup()
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    private func setup() {
        containerView.backgroundColor = UIColor.systemBackground
        containerView.clipsToBounds = false
        containerView.layer.cornerRadius = 25
        addSubview(containerView)
        containerView.pinToSuperview(.init(16))

        // Image View
        containerView.addSubview(imageView)
        imageView.addTopConstraint(to: \.topAnchor)
        imageView.addLeadingConstraint(to: \.leadingAnchor, constant: 8)
        imageView.addBottomConstraint(to: \.bottomAnchor)
        imageView.addAspectRatio(heightToWidth: 1.0)
        imageView.addWidth(32)

        // Activity Indicator
        containerView.addSubview(activityIndicator)
        activityIndicator.pinToEdges(of: imageView)

        // Stack View
        stackView.axis = .vertical
        stackView.alignment = .leading
        stackView.distribution = .equalSpacing

        containerView.addSubview(stackView)
        stackView.addTopConstraint(to: \.topAnchor, constant: 8)
        stackView.addLeadingConstraint(to: \.trailingAnchor, of: imageView, constant: 8)
        stackView.addTrailingConstraint(to: \.trailingAnchor, constant: 16)
        stackView.addBottomConstraint(to: \.bottomAnchor, constant: 8)

        // Title Label
        titleLabel.textColor = .label
        stackView.addArrangedSubview(titleLabel)

        // Subtitle Label
        subtitleLabel.font = .systemFont(ofSize: 12.0, weight: .regular)
        subtitleLabel.textColor = .secondaryLabel
        stackView.addArrangedSubview(subtitleLabel)
    }

    private func updateView() {
        if toasts.first?.accessory == .activityIndicator {
            activityIndicator.isHidden = false
            activityIndicator.startAnimating()
            imageView.isHidden = true
        } else if case let .image(image) = toasts.first?.accessory {
            activityIndicator.stopAnimating()
            activityIndicator.isHidden = true
            imageView.isHidden = false
            imageView.image = image
        } else {
            activityIndicator.stopAnimating()
            activityIndicator.isHidden = true
            imageView.isHidden = true
        }

        guard !toasts.isEmpty else { return }

        self.titleLabel.text = toasts.first?.title
        if toasts.count == 1 {
            self.subtitleLabel.text = toasts.first?.subtitle
        } else {
            guard let subtitle = toasts.first?.subtitle else {
                self.subtitleLabel.text = "\(toasts.count)"
                return
            }
            self.subtitleLabel.text = L10n.Toast.Saving.Multiple.subtitle(subtitle, toasts.count - 1)
        }
    }

    // MARK: Presentation

    func add(_ toast: Toast) {
        toasts.append(toast)
        toast.delegate = self
        updateView()
    }

    func remove(_ toast: Toast) {
        guard let index = toasts.firstIndex(where: { $0.uuid == toast.uuid }) else { return }
        toasts.remove(at: index)

        guard !toasts.isEmpty else {
            // Hide view
            hide()
            return
        }
        updateView()
    }

    func present(in view: UIView) {
        guard !isPresented else {
            return
        }

        isPresented = true
        self.translatesAutoresizingMaskIntoConstraints = false
        view.addSubview(self)

        self.centerX(to: \.centerXAnchor)
        self.top(to: \.topAnchor, constant: 16)
        self.width(200)
        self.height(80)

        self.layer.transform = CATransform3DTranslate(CATransform3DIdentity, 0, -100, 0)

        let animator = UIViewPropertyAnimator(duration: 0.5, dampingRatio: 0.4) {
            self.layer.transform = CATransform3DIdentity
        }
        animator.startAnimation()
        self.showAnimator = animator
    }

    func hide() {
        guard isPresented else {
            return
        }

        isPresented = false
        self.showAnimator?.stopAnimation(true)

        let animator = UIViewPropertyAnimator(duration: 0.5, dampingRatio: 0.4) {
            self.layer.transform = CATransform3DTranslate(CATransform3DIdentity, 0, -100, 0)
        }

        animator.addCompletion { pos in
            guard pos == .end else { return }
            self.removeFromSuperview()
        }
        animator.startAnimation(afterDelay: 0.5)
        self.hideAnimator = animator
    }

//    override func layoutSubviews() {
//        super.layoutSubviews()
//        containerView.layer.cornerRadius = (bounds.height - 10) / 2
//    }
}

extension ToastBubbleView: ToastDelegate {
    func toastWantsToBeDismissed(_ toast: Toast) {
        remove(toast)
        toast.delegate = nil
    }
}
