// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

protocol ToastDelegate: AnyObject {
	func toastWantsToBeDismissed(_ toast: Toast)
}

class Toast {
    let uuid = UUID()

	enum Style: Equatable {
		case info
		case success
		case error
	}
	enum Accessory: Equatable {
		case none
		case image(UIImage)
		case activityIndicator
	}
	struct Action {
		let title: String
		let handler: (Toast) -> Void
	}

	var style: Style
	var accessory: Accessory

	var title: String
	var subtitle: String?

	var action: Action?

	var dismissable: Bool

	var isBeingPresented: Bool { return delegate != nil }

	weak var delegate: ToastDelegate?

	// swiftlint:disable vertical_parameter_alignment
	init(_ style: Style,
		 title: String,
		 subtitle: String? = nil,
		 accessory: Accessory = .none,
		 action: Action? = nil,
		 dismissable: Bool = true) {
		self.style = style
		self.accessory = accessory
		self.title = title
		self.subtitle = subtitle
		self.action = action
		self.dismissable = dismissable
	}

	func dismiss() {
		self.delegate?.toastWantsToBeDismissed(self)
	}

	func dismiss(after interval: TimeInterval, completion: (() -> Void)? = nil) {
		assert(interval > 0.0)
		DispatchQueue.main.asyncAfter(deadline: .now() + interval) { [weak self] in
			self?.dismiss()
			completion?()
		}
	}
}

extension Toast.Accessory {
	static func symbol(_ name: String) -> Toast.Accessory {
		let symbolConfiguration = UIImage.SymbolConfiguration.init(scale: .large)
		let image = UIImage(systemName: name, withConfiguration: symbolConfiguration)!
		return .image(image)
	}
}

extension Toast: Hashable {
    static func == (lhs: Toast, rhs: Toast) -> Bool {
        return lhs.uuid == rhs.uuid
    }

    func hash(into hasher: inout Hasher) {
        hasher.combine(self.uuid)
    }
}
