// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

extension Toast.Style {
	var backgroundColor: UIColor {
		switch self {
		case .info:
			return .systemGray
		case .success:
			return .systemGreen
		case .error:
			return .systemPink
		}
	}
}

class ToastCell: UITableViewCell, CellClassReuseIdentifier {
	var toast: Toast? { didSet { updateCell() } }
	lazy var containerView: UIView = UIView().autolayout()
	lazy var backdrop = UIVisualEffectView(effect: UIBlurEffect(style: .prominent)).autolayout()

	lazy var colorView: UIView = UIView().autolayout()
	lazy var titleLabel: UILabel = UILabel().autolayout()
	lazy var subtitleLabel: UILabel = UILabel().autolayout()
	lazy var actionButton: UIButton = UIButton(type: .system).autolayout()
	lazy var actionSeparator: UIView = UIView().autolayout()

	var accessoryActivityIndicator: UIActivityIndicatorView?
	var accessoryImageView: UIImageView?

	lazy var accessoryContainerView: UIView = UIView().autolayout()
	lazy var labelContainerView: UIView = UIView().autolayout()
	lazy var actionContainerView: UIView = UIView().autolayout()

	var panGestureRecognizer: UIPanGestureRecognizer?
	var panStartOrigin: CGFloat = 0.0
	var panAnimator: UIViewPropertyAnimator?

	lazy var accessorySwitcher: ConditionalViewSwitcher
		= ConditionalViewSwitcher(self.accessoryContainerView, in: self.containerView)

	lazy var subtitleSwitcher: ConditionalViewSwitcher
		= ConditionalViewSwitcher(self.subtitleLabel, in: self.labelContainerView)

	lazy var actionSwitcher: ConditionalViewSwitcher
		= ConditionalViewSwitcher(self.actionContainerView, in: self.containerView)

	override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
		super.init(style: .default, reuseIdentifier: reuseIdentifier)
		setup()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	func setup() {
		backgroundColor = .clear
		contentView.backgroundColor = .clear
		selectionStyle = .none

		containerView.backgroundColor = .clear
		containerView.layer.cornerRadius = 4.0
		containerView.layer.masksToBounds = true
		contentView.addSubview(containerView)

		containerView.addSubview(backdrop)
		containerView.addSubview(colorView)

		containerView.addSubview(labelContainerView)
		labelContainerView.addSubview(titleLabel)

		titleLabel.font = .preferredFont(forTextStyle: .headline)
		subtitleLabel.font = .preferredFont(forTextStyle: .subheadline)
		subtitleLabel.numberOfLines = 2

		actionContainerView.addSubview(actionButton)
		actionSeparator.backgroundColor = .separator
		actionContainerView.addSubview(actionSeparator)

		actionButton.contentEdgeInsets = UIEdgeInsets(top: 0, left: 16, bottom: 0, right: 16)
		actionButton.titleLabel?.font = .preferredFont(forTextStyle: .headline)
		actionButton.tintColor = titleLabel.textColor
		actionButton.addTarget(self, action: #selector(actionTapped(_:)), for: .touchUpInside)

		let pan = UIPanGestureRecognizer(target: self, action: #selector(didPan(_:)))
		addGestureRecognizer(pan)
		self.panGestureRecognizer = pan

		setupFixedConstraints()
		setupAccessorySwitcher()
		setupSubtitleSwitcher()
		setupActionSwitcher()
	}

	override func prepareForReuse() {
		super.prepareForReuse()
		// Somehow, this is required... see https://stackoverflow.com/a/36793776/1065801
		self.accessoryActivityIndicator?.startAnimating()
	}

	func setupFixedConstraints() {
		containerView.pinToSuperview(UIEdgeInsets(8.0))
		actionButton.pinToSuperview()
		backdrop.pinToSuperview()

		let spacing: CGFloat = 10.0
		NSLayoutConstraint.activate([
			labelContainerView.topAnchor.constraint(greaterThanOrEqualTo: containerView.topAnchor, constant: spacing),
			labelContainerView.bottomAnchor.constraint(lessThanOrEqualTo: containerView.bottomAnchor, constant: -spacing),
			labelContainerView.centerYAnchor.constraint(equalTo: containerView.centerYAnchor),

			titleLabel.topAnchor.constraint(equalTo: labelContainerView.topAnchor),
			titleLabel.leadingAnchor.constraint(equalTo: labelContainerView.leadingAnchor),
			titleLabel.trailingAnchor.constraint(equalTo: labelContainerView.trailingAnchor),

			actionSeparator.topAnchor.constraint(equalTo: actionContainerView.topAnchor),
			actionSeparator.bottomAnchor.constraint(equalTo: actionContainerView.bottomAnchor),
			actionSeparator.leadingAnchor.constraint(equalTo: actionContainerView.leadingAnchor),
			actionSeparator.widthAnchor.constraint(equalToConstant: 1.0),

			colorView.leadingAnchor.constraint(equalTo: containerView.leadingAnchor),
			colorView.bottomAnchor.constraint(equalTo: containerView.bottomAnchor),
			colorView.topAnchor.constraint(equalTo: containerView.topAnchor),
			colorView.widthAnchor.constraint(equalToConstant: 5.0)
		])
	}

	@objc func actionTapped(_ sender: UIButton) {
		guard let toast = self.toast, let action = toast.action else { return }
		action.handler(toast)
	}

	@objc func didPan(_ recognizer: UIPanGestureRecognizer) {
		let xOffset = recognizer.translation(in: self).x
		var containerFrame = self.containerView.frame

		switch recognizer.state {
		case .began:
			panStartOrigin = containerFrame.origin.x
			fallthrough
		case .changed:
			if xOffset > 0.0 {
				containerFrame.origin.x = panStartOrigin + pow(xOffset, 0.8)
			} else {
				containerFrame.origin.x = panStartOrigin + xOffset
			}
			self.containerView.frame = containerFrame
		case .ended:
			let velocity = recognizer.velocity(in: self).x
			// decide whether to remove, or to spring back
			if let toast = self.toast,
				toast.dismissable,
				shouldDismiss(offset: xOffset, velocity: velocity) {
				// dismiss it
				let startX = containerFrame.origin.x
				containerFrame.origin.x = -(containerFrame.size.width)
				let endX = containerFrame.origin.x
				let distance = abs(startX - endX)
				let duration = distance / abs(velocity)
				let animator = UIViewPropertyAnimator(duration: TimeInterval(duration), curve: .easeOut) {
					self.containerView.frame = containerFrame
				}
				animator.addCompletion { [weak self] _ in
					self?.toast?.dismiss()
				}
				animator.startAnimation()
			} else {
				// put it back
				containerFrame.origin.x = panStartOrigin
				let animator = UIViewPropertyAnimator(duration: 0.25, dampingRatio: 0.5) {
					self.containerView.frame = containerFrame
				}
				animator.startAnimation()
			}
		default:
			break
		}
	}

	func shouldDismiss(offset: CGFloat, velocity: CGFloat) -> Bool {
		// don't dismiss if dragged to the right
		if offset > 0.0 {
			return false
		}
		// dragged with >power< to the left
		if velocity < -500.0 {
			return true
		}

		// dragged more than half of it to left
		if panStartOrigin + offset + (containerView.frame.size.width * 0.5) < 0.0 {
			return true
		}
		return false
	}

	func setupAccessorySwitcher() {
		accessorySwitcher.visibleConstraints = { [unowned self] in
			let container = self.containerView
			let labels = self.labelContainerView
			let accessory = self.accessoryContainerView

			let spacing: CGFloat = 8.0
			let width: CGFloat = 32.0

			return [
				accessory.widthAnchor.constraint(equalToConstant: width),
				accessory.heightAnchor.constraint(equalToConstant: width),
				accessory.leadingAnchor.constraint(equalTo: container.leadingAnchor, constant: spacing + 5.0),
				accessory.topAnchor.constraint(greaterThanOrEqualTo: container.topAnchor, constant: spacing),
				accessory.bottomAnchor.constraint(lessThanOrEqualTo: container.bottomAnchor, constant: -spacing),
				accessory.centerYAnchor.constraint(equalTo: container.centerYAnchor),
				accessory.trailingAnchor.constraint(equalTo: labels.leadingAnchor, constant: -spacing)
			]
		}

		accessorySwitcher.hiddenConstraints = { [unowned self] in
			let container = self.containerView
			let labels = self.labelContainerView
			let spacing: CGFloat = 8.0 + 5.0

			return [
				labels.leadingAnchor.constraint(equalTo: container.leadingAnchor, constant: spacing)
			]
		}
	}

	func setupSubtitleSwitcher() {
		subtitleSwitcher.visibleConstraints = { [unowned self] in
			let subtitle = self.subtitleLabel
			let title = self.titleLabel
			let container = self.labelContainerView
			let spacing: CGFloat = 4.0

			return [
				subtitle.leadingAnchor.constraint(equalTo: container.leadingAnchor),
				subtitle.trailingAnchor.constraint(equalTo: container.trailingAnchor),
				subtitle.bottomAnchor.constraint(equalTo: container.bottomAnchor),

				subtitle.topAnchor.constraint(equalTo: title.bottomAnchor, constant: spacing)
			]
		}

		subtitleSwitcher.hiddenConstraints = { [unowned self] in
			let title = self.titleLabel
			let container = self.labelContainerView

			return [
				title.bottomAnchor.constraint(equalTo: container.bottomAnchor)
			]
		}
	}

	func setupActionSwitcher() {
		actionSwitcher.visibleConstraints = { [unowned self] in
			let container = self.containerView
			let labels = self.labelContainerView
			let action = self.actionContainerView

			let spacing: CGFloat = 4.0

			return [
				action.trailingAnchor.constraint(equalTo: container.trailingAnchor),
				action.topAnchor.constraint(equalTo: container.topAnchor),
				action.bottomAnchor.constraint(equalTo: container.bottomAnchor),
				action.leadingAnchor.constraint(greaterThanOrEqualTo: labels.trailingAnchor, constant: spacing)
			]
		}

		actionSwitcher.hiddenConstraints = { [unowned self] in
			let container = self.containerView
			let labels = self.labelContainerView

			let spacing: CGFloat = 8.0

			return [
				labels.trailingAnchor.constraint(equalTo: container.trailingAnchor, constant: -spacing)
			]
		}
	}

	func updateCell() {
		guard let toast = toast else { return }

		// color
		colorView.backgroundColor = toast.style.backgroundColor
		// backdrop?.isHidden = containerView.backgroundColor != UIColor.clear

		// title
		titleLabel.text = toast.title

		// subtitle
		subtitleSwitcher.visible = (toast.subtitle != nil)
		if let subtitle = toast.subtitle {
			subtitleLabel.text = subtitle
		}

		// accessory
		updateAccessory(toast)

		actionSwitcher.visible = (toast.action != nil)
		if let action = toast.action {
			actionButton.setTitle(action.title, for: .normal)
		}
	}

	func updateAccessory(_ toast: Toast) {
		accessorySwitcher.visible = (toast.accessory != .none)
		switch toast.accessory {
		case .activityIndicator:
            accessoryImageView?.removeFromSuperview()
            accessoryImageView = nil

			if self.accessoryActivityIndicator == nil {
				let indicator = UIActivityIndicatorView(style: .medium).autolayout()
				indicator.hidesWhenStopped = false
                indicator.color = titleLabel.textColor
				accessoryContainerView.addSubview(indicator)
				indicator.centerInSuperview()
				self.accessoryActivityIndicator = indicator
				indicator.startAnimating()
			}
		case let .image(image):
            accessoryActivityIndicator?.removeFromSuperview()
            accessoryActivityIndicator = nil

			if self.accessoryImageView == nil {
				let imageView = UIImageView().autolayout()
				accessoryContainerView.addSubview(imageView)
				imageView.centerInSuperview()
				imageView.tintColor = titleLabel.textColor
				self.accessoryImageView = imageView
			}
			self.accessoryImageView?.image = image
		default:
			print("[ACCESSORY REMOVED]")
			self.accessoryActivityIndicator?.removeFromSuperview()
			self.accessoryActivityIndicator = nil
			self.accessoryImageView?.removeFromSuperview()
			self.accessoryImageView = nil
		}
	}
}
