// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import Logging

protocol ToastPresentingViewController: UIViewController {
	func show(_ toast: Toast)
}

extension UIViewController {
	func toastPresentingParent() -> ToastPresentingViewController? {
		guard let presenting = parent(conformingTo: ToastPresentingViewController.self) else {
            Logger.correctionUI.error("[Toast] No toast presenting vc in hierarchy above \(self.debugDescription)")
			return nil
		}
		return presenting
	}
}

protocol DefaultToastPresentingViewController: ToastPresentingViewController {
    var toastViewController: ToastViewController? { get set }

    func ensureToastViewController() -> ToastViewController
}

extension DefaultToastPresentingViewController {
    func ensureToastViewController() -> ToastViewController {
        if let viewController = self.toastViewController {
            return viewController
        }

        let viewController = ToastViewController()
        self.toastViewController = viewController

        addChild(viewController) { childView in
            childView.translatesAutoresizingMaskIntoConstraints = false

            self.view.addSubview(childView)

            NSLayoutConstraint.activate([
                childView.leftAnchor.constraint(equalTo: self.view.leftAnchor, constant: 0),
                childView.topAnchor.constraint(equalTo: self.view.topAnchor, constant: 0),
                childView.bottomAnchor.constraint(equalTo: self.view.bottomAnchor, constant: 0),
                childView.widthAnchor.constraint(equalToConstant: 400.0)
            ])
        }

        return viewController
    }

    func show(_ toast: Toast) {
        let viewController = ensureToastViewController()
        viewController.add(toast)
    }
}
