// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import Logging

final class ToastTableView: UITableView {
	// adjust the hit test to return nil if event was not on content (but for example, the background)
	override func hitTest(_ point: CGPoint, with event: UIEvent?) -> UIView? {
		if point.y > contentSize.height {
			return nil
		}
		return super.hitTest(point, with: event)
	}
}

class ToastViewController: UITableViewController {

    enum Section {
        case main
    }

	var toasts: [Toast] = []

    var dataSource: UITableViewDiffableDataSource<Section, Toast>!

	convenience init() {
		self.init(style: .plain)
	}

	override func loadView() {
		let view = ToastTableView(frame: CGRect(x: 0, y: 0, width: 320, height: 400), style: .plain)
		view.delegate = self

        let dataSource = UITableViewDiffableDataSource<Section, Toast>(
            tableView: view
        ) { tableView, indexPath, itemIdentifier in
            let cell: ToastCell = tableView.dequeueReusableCell(for: indexPath)
            cell.toast = itemIdentifier
            return cell
        }
        self.dataSource = dataSource
		view.dataSource = dataSource
		self.view = view
	}

	override func viewDidLoad() {
		super.viewDidLoad()

		tableView.register(ToastCell.self)
		tableView.alwaysBounceVertical = false

		view.backgroundColor = .clear
		tableView.separatorStyle = .none

		tableView.clipsToBounds = false
	}

	override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        BreadcrumbManager.addInteraction("[ToastVC] tapped toast at index \(indexPath.row), toast count \(toasts.count)")
        guard toasts.count > indexPath.row else {
            Logger.correctionUI.error("[ToastVC] The user managed to select a nonexistent index in the toast table")
            return
        }
		let toast = toasts[indexPath.row]
		if toast.dismissable {
			toast.dismiss()
		}
	}

	func add(_ toast: Toast) {
        BreadcrumbManager.addNavigation("[ToastVC] showing title = `\(toast.title)`, acessory = `\(toast.accessory)`, dismissable = \(toast.dismissable)")
        toast.delegate = self
        toasts.append(toast)
        updateTableView()
	}

	func remove(_ toast: Toast) {
        BreadcrumbManager.addNavigation("[ToastVC] removing title =`\(toast.title)`, acessory = `\(toast.accessory)`, dismissable = \(toast.dismissable)")
        toasts.removeAll(where: { $0.uuid == toast.uuid })
        updateTableView()
	}

    func updateTableView() {
        DispatchQueue.main.async {
            var snap = NSDiffableDataSourceSnapshot<Section, Toast>()
            snap.appendSections([.main])
            snap.appendItems(self.toasts, toSection: .main)
            self.dataSource.apply(snap, animatingDifferences: true, completion: nil)
        }
    }
}

extension ToastViewController: ToastDelegate {
	func toastWantsToBeDismissed(_ toast: Toast) {
		remove(toast)
		toast.delegate = nil
	}
}
