// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import CoreGraphics

// swiftlint:disable identifier_name

extension CGSize {
	var aspectRatio: CGFloat {
		guard !height.isZero else { return 0.0 } // don't divide by zero
		return width / height
	}

    var area: CGFloat {
        return width * height
    }

    var smallerSide: CGFloat {
        return min(width, height)
    }
}

extension CGPoint {
	func multiply(with size: CGSize) -> CGPoint {
		return CGPoint(x: self.x * size.width, y: self.y * size.height)
	}

	func distance(_ to: CGPoint) -> CGFloat {
		let deltaX = abs(self.x - to.x)
		let deltaY = abs(self.y - to.y)
		return sqrt((deltaX * deltaX) + (deltaY * deltaY))
	}
}

extension CGRect {
    init(center: CGPoint, size: CGSize) {
        self.init(x: center.x - size.width / 2,
                  y: center.y - size.height / 2,
                  width: size.width,
                  height: size.height)
    }

    init(minX: CGFloat, maxX: CGFloat, minY: CGFloat, maxY: CGFloat) {
        self.init(x: minX, y: minY, width: (maxX - minX), height: (maxY - minY))
    }

    var area: CGFloat {
        return self.size.area
    }

    var center: CGPoint {
        return CGPoint(x: origin.x + size.width / 2,
                       y: origin.y + size.height / 2)
    }
}

extension CGSize {
    var rotate90: CGSize {
        return CGSize(width: self.height, height: self.width)
    }
}

extension CGPath {

    func normalizedCopy(using normalizedRect: CGRect, size: CGSize) -> CGPath? {
        let fullSize = normalizedRect.originalSize(knowing: size)
        var transform = CGAffineTransform
            .identity
            .translatedBy(x: normalizedRect.origin.x, y: normalizedRect.origin.y)
            .scaledBy(x: 1.0 / fullSize.width, y: 1.0 / fullSize.height)
        let result = copy(using: &transform)
        return result
    }

	func normalizedCopy(using size: CGSize) -> CGPath? {
		var transform = CGAffineTransform.identity.scaledBy(x: 1.0 / size.width, y: 1.0 / size.height)
		let result = copy(using: &transform)
		return result
	}

	func mappedCopy(using size: CGSize) -> CGPath? {
		var transform = CGAffineTransform.identity.scaledBy(x: size.width, y: size.height)
		let result = copy(using: &transform)
		return result
	}
}
