// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

protocol IndexRepresentingViewController: UIViewController {
	var representingIndexInCollection: Int? { get set }
}

class IndexedPageViewDataSource
	<ObjectType, ViewType: IndexRepresentingViewController>
	: NSObject, UIPageViewControllerDataSource {
	// typed callback blocks for object access
	typealias AccessBlock = (Int) -> ObjectType?
	typealias CreationBlock = (ObjectType) -> ViewType?

	let access: AccessBlock
	let creation: CreationBlock
	weak var target: UIPageViewController?

	init(for target: UIPageViewController, access: @escaping AccessBlock, creation: @escaping CreationBlock) {
		self.target = target
		self.access = access
		self.creation = creation
	}

	func viewController(at index: Int) -> ViewType? {
		guard let object = access(index) else {
			return nil
		}
		let created = creation(object)
		created?.representingIndexInCollection = index
		return created
	}

	func show(index: Int, animated: Bool) {
		// check if index differs from current one
		if let current = target?.viewControllers?.first as? IndexRepresentingViewController {
			if index == current.representingIndexInCollection {
				return
			}
		}

		guard let target = self.target else { return }
		guard let viewController = viewController(at: index) else { return }

		var direction = UIPageViewController.NavigationDirection.forward
		if let current = target.viewControllers?.first as? IndexRepresentingViewController,
			let currentIndex = current.representingIndexInCollection {
			direction = (index > currentIndex ? .forward : .reverse)
		}

		target.setViewControllers([viewController], direction: direction, animated: animated, completion: nil)
	}

	func pageViewController(
		_ pageViewController: UIPageViewController, viewControllerBefore viewController: UIViewController
	) -> UIViewController? {
		guard let current = viewController as? IndexRepresentingViewController else { return nil }
		guard let currentIndex = current.representingIndexInCollection else { return nil }

		return self.viewController(at: currentIndex - 1)
	}

	func pageViewController(
		_ pageViewController: UIPageViewController, viewControllerAfter viewController: UIViewController
	) -> UIViewController? {
		guard let current = viewController as? IndexRepresentingViewController else { return nil }
		guard let currentIndex = current.representingIndexInCollection else { return nil }

		return self.viewController(at: currentIndex + 1)
	}
}

class ArrayDataSource: NSObject, UIPageViewControllerDataSource {
	var viewControllers: [UIViewController]

	init(_ viewControllers: [UIViewController]) {
		self.viewControllers = viewControllers
	}

	func indexOfViewController(_ vc: UIViewController) -> Int? {
		return viewControllers.firstIndex(of: vc)
	}

	func viewController(at index: Int) -> UIViewController? {
		guard index >= 0 && index < viewControllers.count else {
			return nil
		}
		return viewControllers[index]
	}

	func pageViewController(
		_ pageViewController: UIPageViewController, viewControllerBefore viewController: UIViewController
	) -> UIViewController? {
		guard let currentIndex = viewControllers.firstIndex(of: viewController) else { return nil }

		return self.viewController(at: currentIndex - 1)
	}

	func pageViewController(
		_ pageViewController: UIPageViewController, viewControllerAfter viewController: UIViewController
	) -> UIViewController? {
		guard let currentIndex = viewControllers.firstIndex(of: viewController) else { return nil }

		return self.viewController(at: currentIndex + 1)
	}
}
