// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import TUMexamKit

class IndexedPrefetcher<ObjectType> {
	typealias AccessBlock = (Int) -> ObjectType?
	typealias PrefetchBlock = (ObjectType) -> Void

	let before: Int
	let after: Int
	var currentIndex: Int { didSet { updateWindow() } }
	var count: Int { didSet { updateClampingWindow() } }
	private var clampingWindow: Range<Int>
	private var currentWindow: Range<Int>
	private let access: AccessBlock
	private let prefetch: PrefetchBlock

	init(
		currentIndex: Int,
		count: Int,
		before: Int = 1,
		after: Int = 3,
		access: @escaping AccessBlock,
		prefetch: @escaping PrefetchBlock
	) {
		self.currentIndex = currentIndex
		self.count = count
		self.before = before
		self.after = after
		self.clampingWindow = 0..<0
		self.currentWindow = currentIndex..<currentIndex
		self.access = access
		self.prefetch = prefetch

		self.updateClampingWindow()
	}

	func updateWindow() {
		let newWindow = ((currentIndex - before)..<(currentIndex + after + 1)).clamped(to: self.clampingWindow)
		// do nothing if no changes occured
		guard newWindow != self.currentWindow else { return }

		// iterate over indices and load
		for idx in newWindow where !self.currentWindow.contains(idx) {
			if let object = access(idx) {
				prefetch(object)
			}
		}

		self.currentWindow = newWindow
	}

	func updateClampingWindow() {
		self.clampingWindow = 0..<count
		updateWindow()
	}
}
