// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import TUMexamKit

protocol StateObserving: AnyObject {
	var stateObserver: AnyCancellable? { get set }

	func observeState(_ state: Observable<ResourceObjectState>)
	func stateChanged(_ state: ResourceObjectState)
}

extension StateObserving {
	func observeState(_ state: Observable<ResourceObjectState>) {
		self.stateObserver = state.receiveOnMain().sink(receiveValue: { [weak self] (state) in
			self?.stateChanged(state)
		})
	}

	func observeState(_ object: Stateful) {
		self.observeState(object.state)
	}
}

extension ProblemIdentifier {
	var shortName: String {
		var str = "\(problem)"
		if let sub = subproblem {
			str += "\(sub)"
		}
		return str
	}
}

extension SubmissionError: LocalizedError {
	public var errorDescription: String? {
		switch self {
		case .invalidLockToken:
			return L10n.Error.Submission.invalidLockToken
		}
	}
}

extension NetworkError: LocalizedError {
	public var errorDescription: String? {
		switch self {
		case .invalidValue:
			return L10n.Error.Network.invalidValue
		case .forbidden:
			return L10n.Error.Network.forbidden
		case .notFound:
			return L10n.Error.Network.notFound
		case .unauthorized:
			return L10n.Error.Network.unauthorized
		case let .rejected(layers):
			return L10n.Error.Network.rejected(layers)
        case .lockTaken:
            return "The ERID is already locked by someone else"
		}
	}
}

extension NetworkDataError: LocalizedError {
    public var errorDescription: String? {
        switch self {
        case .noData:
            return L10n.Error.Network.noData
        case .invalidResponse:
            return L10n.Error.Network.invalidResponse
        }
    }
}
