// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

extension Optional {
	func nilOr<T: Equatable>(_ keyPath: KeyPath<Wrapped, T>, equal value: T) -> Bool {
		guard let existing = self else {
			return false
		}
		return existing[keyPath: keyPath] == value
	}

	func nilOr<T: Equatable>(_ keyPath: KeyPath<Wrapped, T>, notEqual value: T) -> Bool {
		guard let existing = self else {
			return false
		}
		return existing[keyPath: keyPath] != value
	}

	func nilOr(_ keyPath: KeyPath<Wrapped, Bool>) -> Bool {
		return nilOr(keyPath, equal: true)
	}
}

extension DispatchTimeInterval {
	var humanReadableString: String {
		switch self {
		case let .seconds(seconds):
			return "\(seconds)s"
		case let .milliseconds(millis):
			return "\(millis)ms"
		case let .microseconds(micros):
			return "\(micros)us"
		case let .nanoseconds(nanos):
			return "\(Double(nanos) / Double(NSEC_PER_SEC))s"
		case .never:
			return "never"
		default:
			return "-"
		}
	}
}

func measure<T>(_ label: String, problemBlock: @autoclosure () -> T) -> T {
    let start = DispatchTime.now()
    let result = problemBlock()
    let end = DispatchTime.now()

	let duration = start.distance(to: end)

    Logger.measure.debug("\(label) took \(duration.humanReadableString)")
	return result
}

func measure(_ label: String, problemBlock: @autoclosure () -> Void) {
    let start = DispatchTime.now()
    problemBlock()
    let end = DispatchTime.now()

	let duration = start.distance(to: end)

    Logger.measure.debug("\(label) took \(duration.humanReadableString)")
}

// Optional String Binding for SwiftUI's TextField

extension Optional where Wrapped == String {
    private var _bound: String? {
        get {
            return self
        }
        set {
            self = newValue
        }
    }
    public var bound: String {
        get {
            return _bound ?? ""
        }
        set {
            _bound = newValue.isEmpty ? nil : newValue
        }
    }
}

extension Optional where Wrapped == Int {
    private var _bound: Int? {
        get {
            return self
        }
        set {
            self = newValue
        }
    }
    public var boundStringValue: String {
        get {
            guard let bound = _bound else {
                return ""
            }
            return  String(bound)
        }
        set {
            _bound = newValue.isEmpty ? nil : Int(newValue)
        }
    }
}
