// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

extension UIImage {
    var scaledSize: CGSize {
        let screenScale = UIScreen.main.scale
        let transform = CGAffineTransform(scaleX: 1/screenScale, y: 1/screenScale)
        return self.size.applying(transform)
    }
}

// MARK: Cutout
extension UIImage {
	func cutout(normalizedPath: CGPath) -> UIImage? {
		// map path back to image size and try getting the underlying cg image
		guard let path = normalizedPath.mappedCopy(using: size),
			let cgImage = self.cgImage else { return nil }

		// get the new bounding box
		let boundingBoxRaw = path.boundingBoxOfPath
		// ... and round it
		let boundingBox = boundingBoxRaw.integral

		// crop the source image, and create a UIImage (otherwise it'll be flipped)
		guard let cropped = cgImage.cropping(to: boundingBox) else { return nil }
		let croppedImage = UIImage(cgImage: cropped)

		// make sure the format has an alpha channel
		let format = self.imageRendererFormat
		format.opaque = false
		let renderer = UIGraphicsImageRenderer(size: boundingBox.size, format: format)

		let image = renderer.image { (rendererContext) in
			let context = rendererContext.cgContext

			// draw the path
			context.saveGState()
			context.translateBy(x: -boundingBox.origin.x, y: -boundingBox.origin.y)
			context.addPath(path)
			context.setFillColor(UIColor.black.cgColor)
			context.fillPath()
			context.restoreGState()

			// draw the image
			croppedImage.draw(at: .zero, blendMode: .sourceIn, alpha: 1.0)
		}

		return image
	}
}

// MARK: Crop

extension UIImage {
    private func findSmallestRectBasedOnAlpha() -> (CGRect, Bool)? {
        let width = Int(self.size.width)
        let height = Int(self.size.height)

        // create bitmap context
        let colorSpace = CGColorSpaceCreateDeviceRGB()
        let context = CGContext(data: nil,
                                width: width,
                                height: height,
                                bitsPerComponent: 8,
                                bytesPerRow: 4 * width, // ARGB
                                space: colorSpace,
                                bitmapInfo: CGImageAlphaInfo.premultipliedFirst.rawValue) // easier access if first
        guard let ctx = context else { return nil }

        // draw this image into the new context
        guard let cgImage = self.cgImage else { return nil }
        let rect = CGRect(x: 0, y: 0, width: width, height: height)
        ctx.draw(cgImage, in: rect)

        // get image data
        guard let dataPointer = ctx.data else { return nil }
        let data = UnsafeMutableRawBufferPointer(start: dataPointer, count: (width*height*4))

        var minX: Int = width
        var minY: Int = height
        var maxX: Int = 0
        var maxY: Int = 0
        var hasAlpha: Bool = false

        // iterate over it
        // swiftlint:disable identifier_name
        for x in 0..<width {
            for y in 0..<height {
                let offset = ((y * width) + x) * 4
                let alpha = data[offset]

                if alpha > 0 {
                    hasAlpha = true
                    if x < minX { minX = x }
                    if x > maxX { maxX = x }
                    if y < minY { minY = y }
                    if y > maxY { maxY = y }
                }
            }
        }

        return (CGRect(x: minX, y: minY, width: maxX-minX+1, height: maxY-minY+1), hasAlpha)
    }

    /// crops the receiver to the smallest possible rectangle, according to the alpha channel.
    ///
    /// e.g.: it removes unnecessary empty space (based on alpha value).
    /// - Parameter extend: how many pixels the rectangle is extended
    /// - Returns: (image, hasAlpha) the cropped image, and whether the new image contains any data at all
    func cropUsingAlphaChannel(extend: Int = 0) -> (UIImage, Bool)? {
        guard let (rect, hasAlpha) = findSmallestRectBasedOnAlpha() else { return nil }
        guard let cgImage = self.cgImage else { return nil }

        var adjustedRect = rect
        if extend > 0 {
            let off = CGFloat(extend)
            let minX = max(0, rect.minX - off)
            let minY = max(0, rect.minY - off)
            let maxX = min(CGFloat(cgImage.width), rect.maxX + off)
            let maxY = min(CGFloat(cgImage.height), rect.maxY + off)
            adjustedRect = CGRect(minX: minX, maxX: maxX, minY: minY, maxY: maxY)
        }

        guard let result = cgImage.cropping(to: adjustedRect) else { return nil }

        return (UIImage(cgImage: result), hasAlpha)
    }
}
