// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

// MARK: Hierachy

extension UIView {
	func ensureSuperview(_ superview: UIView?) {
		guard self.superview != superview else { return }
		if self.superview != nil {
			self.removeFromSuperview()
		}

		if let superview = superview {
			superview.addSubview(self)
		}
	}
}

// MARK: Layout

extension UIView {
	/// sets `translatesAutoresizingMaskIntoConstraints = false` and returns the view.
	/// basically just a shorthand for initializing any view for use with autolayout
	/// - Returns: the view itself
	func autolayout() -> Self {
		translatesAutoresizingMaskIntoConstraints = false
		return self
	}

	func pinToSuperview(_ edgeInsets: UIEdgeInsets = .zero) {
		guard let superview = self.superview else {
			fatalError("pinToSuperview(): can't add constraints without superview")
		}
		NSLayoutConstraint.activate([
			leftAnchor.constraint(equalTo: superview.leftAnchor, constant: edgeInsets.left),
			rightAnchor.constraint(equalTo: superview.rightAnchor, constant: -edgeInsets.right),
			topAnchor.constraint(equalTo: superview.topAnchor, constant: edgeInsets.top),
			bottomAnchor.constraint(equalTo: superview.bottomAnchor, constant: -edgeInsets.bottom)
		])
	}

	func centerInSuperview() {
		guard let superview = self.superview else {
			fatalError("pinToSuperview(): can't add constraints without superview")
		}
		NSLayoutConstraint.activate([
			centerYAnchor.constraint(equalTo: superview.centerYAnchor),
			centerXAnchor.constraint(equalTo: superview.centerXAnchor)
		])
	}
}

extension UIView {
    func subviews(conformingTo type: AnyClass) -> [UIView] {
        var retval = [UIView]()
        for someView in self.subviews {
            if someView.isKind(of: type) {
                retval.append(someView)
            }
            let children = someView.subviews(conformingTo: type)
            retval.append(contentsOf: children)
        }
        return retval
    }
}
