// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import Logging

private protocol AnyAction {
	var isValid: Bool { get }

	func execute()
}

protocol ContextUndoManagerProvider: AnyObject {
	var contextUndoManager: ContextUndoManager { get }
}

protocol ContextUndoManagerCoordinator: AnyObject {
    func shouldChangeUndoManager(to canvasView: CorrectionCanvasViewController) -> Bool
    /// Called by parent to notify that we are now "first responder" as visible view controller
    func becomeUndoMager()
}

class ContextUndoManager {
	// MARK: Sub Types
	typealias HandlerType<TargetType: AnyObject> = (TargetType) -> Void

	enum Mode {
		case normal
		case undo
		case redo
	}

	struct Action<TargetType: AnyObject> {
		weak var target: TargetType?
		let handler: HandlerType<TargetType>
	}

	// MARK: Properties

	var mode: Mode = .normal

	private var undoActions: [AnyAction] = []
	private var redoActions: [AnyAction] = []

	// MARK: Observable Properties

	@Published var canUndo: Bool = false
	@Published var canRedo: Bool = false

	func registerUndo<TargetType: AnyObject>(withTarget target: TargetType, handler: @escaping (TargetType) -> Void) {
		let action = Action(target: target, handler: handler)
		switch mode {
		case .normal:
			// completely new action -> discard all redo actions
			redoActions = []
			// and also do the same as if we were in a redo operation
			undoActions.append(action)
		case .redo:
			// currently in redo mode -> append to undo actions
			undoActions.append(action)
		case .undo:
			// currently in undo mode -> append to redo
			redoActions.append(action)
		}

        Logger.undoRedo.debug("Registered \(target) action")
		evaluatePossibleOperations()
	}

	// make sure events are only fired onces
	private func evaluatePossibleOperations() {
		let canUndo = !undoActions.isEmpty
		let canRedo = !redoActions.isEmpty

		if self.canUndo != canUndo {
			self.canUndo = canUndo
		}
		if self.canRedo != canRedo {
			self.canRedo = canRedo
		}
	}

	func undo() {
		guard let action = undoActions.popUntil(true, forKeyPath: \.isValid) else { return }
		let previousMode = mode
		mode = .undo

		action.execute()

		mode = previousMode

		evaluatePossibleOperations()
	}

	func redo() {
		guard let action = redoActions.popUntil(true, forKeyPath: \.isValid) else { return }
		let previousMode = mode
		mode = .redo

		action.execute()

		mode = previousMode

		evaluatePossibleOperations()
	}

	func popAllUndoActions() {
		for action in undoActions where action.isValid {
			action.execute()
		}
		reset()
	}

    func reset() {
        undoActions = []
        redoActions = []

        evaluatePossibleOperations()
    }
}

extension ContextUndoManager.Action: AnyAction {
	var isValid: Bool { return target != nil }

	func execute() {
		if let target = target {
			handler(target)
		}
	}
}
