// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

class ConstraintSwitcher<LayoutIdentifier: Hashable> {
	var currentLayout: LayoutIdentifier? {
		didSet {
			if oldValue != currentLayout {
				setLayout(currentLayout)
			}
		}
	}
	var currentConstraints: [NSLayoutConstraint] = []
	typealias ConstraintBlock = (LayoutIdentifier) -> [NSLayoutConstraint]

	var layouts: [LayoutIdentifier: ConstraintBlock] = [:]

	subscript(_ identifier: LayoutIdentifier) -> ConstraintBlock? {
		get {
			layouts[identifier]
		}
		set(block) {
			layouts[identifier] = block
		}
	}

	private func setLayout(_ layout: LayoutIdentifier?) {
		removeConstraints()

		guard let layout = layout, let constraintBlock = layouts[layout] else {
			return
		}

		let newConstraints = constraintBlock(layout)
		NSLayoutConstraint.activate(newConstraints)
		currentConstraints = newConstraints
	}

	private func removeConstraints() {
		for constraint in currentConstraints {
			constraint.isActive = false
		}
		currentConstraints = []
	}
}

class ConditionalViewSwitcher {
	typealias ConstraintBlock = () -> [NSLayoutConstraint]

	let constraintSwitcher: ConstraintSwitcher<Bool> = ConstraintSwitcher()

	let optionalView: UIView
	let containerView: UIView

	var visibleConstraints: ConstraintBlock?
	var hiddenConstraints: ConstraintBlock?

	var visible: Bool {
		get {
			constraintSwitcher.currentLayout ?? false
		}
		set {
			constraintSwitcher.currentLayout = newValue
		}
	}

	init(_ view: UIView, in container: UIView) {
		self.optionalView = view
		self.containerView = container

		constraintSwitcher[true] = { [unowned self] _ in
			let view = self.optionalView
			let container = self.containerView

			view.ensureSuperview(container)

			return self.visibleConstraints?() ?? []
		}

		constraintSwitcher[false] = { [unowned self] _ in
			let view = self.optionalView

			view.ensureSuperview(nil)

			return self.hiddenConstraints?() ?? []
		}
	}
}
