// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

// Modified to use this approach: https://www.swiftbysundell.com/articles/caching-in-swift/
// NSCache needs NSObjects as keys

import UIKit
import TUMexamKit
import Logging

// swiftlint:disable:next line_length
class ViewControllerCache<ObjectType, KeyType: Hashable, ViewControllerType: UIViewController>: NSObject, NSCacheDelegate {
	typealias CreationBlock = (ObjectType) -> ViewControllerType?

	var creation: CreationBlock = { _ in nil }

    final class WrappedKeyType: NSObject {
        let key: KeyType

        init(_ key: KeyType) {
            self.key = key
        }

        override var hash: Int {
            return key.hashValue
        }

        override func isEqual(_ object: Any?) -> Bool {
            guard let value = object as? WrappedKeyType else {
                return false
            }
            return value.key == self.key
        }
    }

    private var cache: LRUCache<WrappedKeyType, ViewControllerType>
	private let keyPath: KeyPath<ObjectType, KeyType>

    init(keyPath: KeyPath<ObjectType, KeyType>, maxItems: Int? = nil) {
		self.keyPath = keyPath
        self.cache = LRUCache<WrappedKeyType, ViewControllerType>(costLimit: 0, countLimit: maxItems ?? 0)
        super.init()
	}

    func viewController(for object: ObjectType, createIfNecessary: Bool = true) -> ViewControllerType? {
		let key: KeyType = object[keyPath: keyPath]
        let wrapped = WrappedKeyType(key)

        if let existing = cache.value(forKey: wrapped) {
			return existing
		}

        guard createIfNecessary else { return nil }

		if let new = creation(object) {
            cache.set(new, forKey: wrapped)
			return new
		}

		return nil
	}

    public func removeAllObjects() {
        self.cache.removeAll()
    }

    public func currentlyHeldControllers() -> [ViewControllerType] {
        return cache.allValues()
    }

    func cache(_ cache: NSCache<AnyObject, AnyObject>, willEvictObject obj: Any) {
        guard let controller = obj as? ViewControllerType else {
            return
        }
        Logger.correctionUI.debug("[VCCache] Evicting VC from Cache: \(controller.debugDescription)")
    }
}

extension CorrectionType {
	var typeErased: AnyObject {
		return self
	}
}

extension CorrectionPromiseType {
    var typeErased: AnyObject {
        return self
    }
}

extension PageType {
	var typeErased: AnyObject {
		return self
	}
}

class CorrectionViewControllerCache
	<ViewControllerType: UIViewController>: ViewControllerCache<CorrectionType, CorrectionIdentifier, ViewControllerType> {
	init() {
        super.init(keyPath: \.identifier, maxItems: 4)
	}
}

class CorrectionPromiseViewControllerCache
    <ViewControllerType: UIViewController>: ViewControllerCache<CorrectionPromiseType, Int, ViewControllerType> {
    init() {
        super.init(keyPath: \.identifier, maxItems: 4)
    }
}

class PageViewControllerCache
	<ViewControllerType: UIViewController>: ViewControllerCache<PageType, CorrectionPageIdentifier, ViewControllerType> {
	init() {
		super.init(keyPath: \.pageIdentifier, maxItems: 4)
	}
}
