// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

/**
 Holds a list of previously connected instances.

 The list is stored separately from the settings, which are cleared whenever the apps bundle version changes in order to prevent settings conflicts.
 */
class ConnectionHistoryStore {
    /// Default path to store the connection history
    private static let defaultPath = NSSearchPathForDirectoriesInDomains(.documentDirectory, .userDomainMask, true)[0].appending("/history") // swiftlint:disable:this line_length
    /// Singleton
    static let shared = ConnectionHistoryStore(path: ConnectionHistoryStore.defaultPath)

    /// Path to store the Connection History
    private let path: String
    /// Static file manager instance
    private let fileManager = FileManager()

    /// The actual connection history data
    var history: [URL] = []

    /// Initilizes the connection history instance with the given storage path
    init(path: String) {
        self.path = path
        Logger.settings.info("Using connection history at \(path)")

        if !fileManager.fileExists(atPath: path) {
            saveDataToDisk()
        } else {
            autoreleasepool {
                let decoder = JSONDecoder()
                guard let data = fileManager.contents(atPath: self.path),
                      let arr = try? decoder.decode([URL].self, from: data) else {
                    return
                }
                self.history = arr
            }
        }
    }

    /// Adds the given url to the connection history
    /// - Parameter url: the url to add to the connection history
    func add(_ url: URL) {
        if let index = history.firstIndex(of: url) {
            history.remove(at: index)
        }
        history.insert(url, at: 0)
        Logger.settings.info("Added \(url.absoluteString) to Connection History")
        while history.count > 3 {
            history.removeLast()
        }
        saveDataToDisk()
    }

    /// Saves the connection history to disk
    /// Happens automatically when adding a URL using ``add(_ url:)``
    private func saveDataToDisk() {
        // JSONEncoder leaks memory when not using an autoreleasepool around it...
        autoreleasepool {
            do {
                let encoder = JSONEncoder()
                let data = try encoder.encode(self.history)
                try data.write(to: URL(fileURLWithPath: self.path), options: .atomic)
            } catch {
                Logger.settings.error("Failed to encode/save connection history to disk")
            }
        }
    }
}
