// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

// https://damir.me/the-mystery-of-the-disappearing-nsuserdefaults-keys/
// Wenn dus nicht nicht selbst machst gehts halt kaputt

import Foundation
import AnyCodable
import Logging

// MARK: - UserDefaultsKeys

/// Contains all the keys for the UserDefaults
struct HiddenSettings {
    static let backendBaseURL = "backendBaseURL"
    static let backendAuthType = "backendAuthType"
    static let selectedExamId = "selectedExamId"
    static let selectedExamName = "selectedExamName"
    static let deviceIdentifier = "deviceIdentifier"
    static let installedVersion = "installedVersion"
    static let didShowStatusTutorial = "didShowStatusTutorial"
    static let didShowPagingCorrectionTutorial = "didShowPagingCorrectionTutorial"
    static let didShowSnippetTutorial = "didShowSnippetTutorial"

    static func all() -> [String] {
        return [
            backendBaseURL,
            backendAuthType,
            selectedExamId,
            selectedExamName,
            deviceIdentifier,
            installedVersion,
            didShowStatusTutorial,
            didShowPagingCorrectionTutorial,
            didShowSnippetTutorial
        ]
    }
}

// MARK: - SettingsKeys

struct SettingsKeys {
    static let askBeforeDismissingCorrectionSession = "ask_before_dismissing_correction_session"
    static let allowModificationsToAllProblems = "allow_modifications_to_all_problems"
    static let sentryCrashLogsOptIn = "sentry_crash_logs_opt_in"
    static let applePencilDoubleTapGestureEnabled = "pencil_double_tap_gesture"
    static let hideOverlays = "hide_overlays"
    static let enhancedDarkMode = "enhanced_darkmode"
    static let sampleSolutionOpacity = "sample_solution_opacity"
    static let retainPageSelection = "retain_page_selection"
    static let drawingToolbarPosition = "drawing_toolbar_position"
}

// MARK: - SettingsStore

class SettingsStore {

    // swiftlint:disable:next line_length
    private static let defaultPath = NSSearchPathForDirectoriesInDomains(.documentDirectory, .userDomainMask, true)[0].appending("/tumexam_settings")
    static let shared = SettingsStore(path: SettingsStore.defaultPath)

    // private static let saveQueue = DispatchQueue(label: Bundle.main.bundleIdentifier! + ".SettingsStore", qos: .utility)

    private let path: String
    private let fileManager = FileManager()
    private var dataDict = [String: AnyCodable]()
    private var defaults = [String: Any]()

    private var saveTimer: Timer?
    var accessSemaphore: DispatchSemaphore = {
        let sema = DispatchSemaphore(value: 0)
        sema.signal()
        return sema
    }()

    var isEmpty: Bool {
        defer {
            accessSemaphore.signal()
        }
        accessSemaphore.wait()
        return dataDict.isEmpty
    }

    private init(path: String) {
        self.path = path
        Logger.settings.info("Settings Path is \(path)")
        if !fileManager.fileExists(atPath: self.path) {
            self.saveDataToDisk()
        } else {
            autoreleasepool {
                accessSemaphore.wait()
                let jsonDecoder = JSONDecoder()
                guard let data = fileManager.contents(atPath: self.path),
                    let dict = try? jsonDecoder.decode([String: AnyCodable].self, from: data) else {
                    return
                }
                self.dataDict = dict
                accessSemaphore.signal()
            }
        }
    }

    subscript(key: String) -> Any? {
        get {
            defer {
                accessSemaphore.signal()
            }
            Logger.settings.debug("GET \(key): waiting for access sema")
            accessSemaphore.wait()
            Logger.settings.debug("GET \(key): got access sema")
            return dataDict[key]?.value ?? defaults[key]
        }
        set {
            Logger.settings.debug("SET \(key): waiting for access sema")
            accessSemaphore.wait()
            dataDict[key] = AnyCodable(newValue)
            Logger.settings.debug("SET \(key): got access, value set to \(newValue.debugDescription)")
            accessSemaphore.signal()
            self.scheduleSave()
        }
    }

    func scheduleSave() {
        saveTimer?.invalidate()
        saveTimer = Timer.scheduledTimer(withTimeInterval: 1.0, repeats: false, block: { [weak self] _ in
            DispatchQueue.global(qos: .utility).async {
                self?.saveDataToDisk()
            }
        })
        Logger.settings.debug("Scheduled save")
    }

    private func saveDataToDisk() {
        autoreleasepool {
            Logger.settings.debug("SAVING: Waiting for access sema")
            self.accessSemaphore.wait()
            do {
                let encoder = JSONEncoder()
                let data = try encoder.encode(self.dataDict)
                try data.write(to: URL(fileURLWithPath: self.path), options: .atomic)
            } catch {
                Logger.settings.error("Failed to encode/store SettingsStore dataDict!")
            }
            Logger.settings.debug("SAVING: Saved data to diska")
            self.accessSemaphore.signal()
        }
    }

    // swiftlint:disable:next cyclomatic_complexity
    func registerDefaults(_ defaults: [String: Any]) {
        for (key, value) in defaults {
            // swiftlint:disable force_cast
            switch value {
            case is Int:
                self.defaults[key] = value as! Int
            case is Double:
                self.defaults[key] = value as! Double
            case is Bool:
                self.defaults[key] = value as! Bool
            case is String:
                self.defaults[key] = value as! String
            case is URL:
                self.defaults[key] = value as! URL
            case is Date:
                self.defaults[key] = value as! Date
            case is [String: Bool]:
                self.defaults[key] = value as! [String: Bool]
            case is [String: Any?]:
                self.defaults[key] = value as! [String: Any?]
            case is [Any?]:
                self.defaults[key] = value as! [Any?]
            default:
                continue
            }
            // swiftlint:enable force_cast
        }
    }

    func reset() {
        self.dataDict = [:]
        try? fileManager.removeItem(atPath: self.path)
        // swiftlint:disable:next line_length
        guard let cfBundleVersionStr = Bundle.main.object(forInfoDictionaryKey: "CFBundleShortVersionString") as? String else {
            return
        }
        self[HiddenSettings.installedVersion] = cfBundleVersionStr
        self[HiddenSettings.deviceIdentifier] = UUID().uuidString
        scheduleSave()
    }
}

extension SettingsStore {
	var deviceIdentifier: UUID {
		if let uuidString = self[HiddenSettings.deviceIdentifier] as? String,
			let uuid = UUID(uuidString: uuidString) {
			return uuid
		} else {
			let uuid = UUID()
			self[HiddenSettings.deviceIdentifier] = uuid.uuidString
			return uuid
		}
	}

    // Selected exam
    var selectedExamId: Int? {
        get {
            return self[HiddenSettings.selectedExamId] as? Int
        }
        set {
            self[HiddenSettings.selectedExamId] = newValue
        }
    }

    var backendURL: URL? {
        get {
            guard let str = self[HiddenSettings.backendBaseURL] as? String else {
                return nil
            }
            return URL(string: str)
        }
        set {
            self[HiddenSettings.backendBaseURL] = newValue?.absoluteString
        }
    }

	var applePencilDoubleTapGestureEnabled: Bool {
		get {
			self[SettingsKeys.applePencilDoubleTapGestureEnabled] as? Bool ?? true
		}
		set {
			self[SettingsKeys.applePencilDoubleTapGestureEnabled] = newValue
		}
	}

	var askBeforeDismissingCorrectionSession: Bool {
		get {
			self[SettingsKeys.askBeforeDismissingCorrectionSession] as? Bool ?? true
		}
		set {
			self[SettingsKeys.askBeforeDismissingCorrectionSession] = newValue
		}
	}

    var drawingToolbarPosition: DrawingToolbarPosition {
        get {
            guard let raw = self[SettingsKeys.drawingToolbarPosition] as? Int else {
                return .left
            }
            return DrawingToolbarPosition(rawValue: raw) ?? .left
        }
        set {
            self[SettingsKeys.drawingToolbarPosition] = newValue.rawValue
        }
    }
}
