// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

struct SetupInfo: CustomDebugStringConvertible {
    var semester: SetupSemester
    var faculty: SetupFaculty?
    var customFaculty: String?
    var shortname: String?

    var url: URL? {
        var resultFacultyStr: String?
        if let faculty = faculty {
            resultFacultyStr = faculty.rawValue
        }

        if let customFaculty = customFaculty {
            resultFacultyStr = customFaculty
        }

        guard let faculty = resultFacultyStr,
              let name = shortname?.addingPercentEncoding(withAllowedCharacters: .urlHostAllowed) else {
            return nil
        }
        let urlString = "https://\(semester.year)\(semester.term.rawValue)-\(faculty)-\(name).hq-ipad.tumexam.de"
        return URL(string: urlString)!
    }

    var debugDescription: String {
        return url?.absoluteString ?? "<SetupInfo - invalid: semester = \(semester), faculty = \(faculty.debugDescription), shortname = \(shortname ?? "nil")>"
    }
}

enum SetupTerm: String, CustomStringConvertible {
    case summer = "ss"
    case winter = "ws"

    var description: String {
        return self == .summer ? L10n.Setup.Semester.summer : L10n.Setup.Semester.winter
    }
}

struct SetupSemester: CustomStringConvertible, Hashable {
    let year: Int
    let term: SetupTerm

    var description: String {
        return "\(term.description) \(year)"
    }

    func previousSemester() -> SetupSemester {
        if term == .summer {
            return SetupSemester(year: year - 1, term: .winter)
        } else {
            return SetupSemester(year: year, term: .summer)
        }
    }

    func nextSemester() -> SetupSemester {
        if term == .summer {
            return SetupSemester(year: year, term: .winter)
        } else {
            return SetupSemester(year: year + 1, term: .summer)
        }
    }
}

enum SetupFaculty: String, CaseIterable, CustomStringConvertible, Hashable {
    // swiftlint:disable identifier_name
    #if DEBUG
    case dev
    #endif
    case lrg
    case ar
    case ch
    case bgu
    case ei
    case `in`
    case ma
    case mw
    case ed
    case ph
    case sg
    case edu
    case wzw
    case gov
    case wi
    case med
    case manual
    // swiftlint:enable identifier_name

    var description: String {
        switch self {
        #if DEBUG
        case .dev:
            return L10n.Setup.Faculty.dev
        #endif
        case .lrg:
            return L10n.Setup.Faculty.lrg
        case .ar:
            return L10n.Setup.Faculty.ar
        case .ch:
            return L10n.Setup.Faculty.ch
        case .bgu:
            return L10n.Setup.Faculty.bgu
        case .ei:
            return L10n.Setup.Faculty.ei
        case .in:
            return L10n.Setup.Faculty.in
        case .ma:
            return L10n.Setup.Faculty.ma
        case .mw:
            return L10n.Setup.Faculty.mw
        case .ed:
            return L10n.Setup.Faculty.ed
        case .ph:
            return L10n.Setup.Faculty.ph
        case .sg:
            return L10n.Setup.Faculty.sg
        case .edu:
            return L10n.Setup.Faculty.edu
        case .wzw:
            return L10n.Setup.Faculty.wzw
        case .gov:
            return L10n.Setup.Faculty.gov
        case .wi:
            return L10n.Setup.Faculty.wi
        case .med:
            return L10n.Setup.Faculty.med
        case .manual:
            return L10n.Setup.Faculty.manual
        }
    }
}
