// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import UIKit
import TUMexamKit

class CorrectionBoxButton: UIButton {

    static let foregroundColor = UIColor { traitCollection -> UIColor in
        let enhancedDarkMode = App.shared.settings[SettingsKeys.enhancedDarkMode] as? Bool ?? false
        switch (traitCollection.userInterfaceStyle, enhancedDarkMode) {
        case (.dark, true):
            return UIColor(white: 0.8, alpha: 1.0)
        default:
            return .black
        }
    }

    static let backgroundColor = UIColor { traitCollection -> UIColor in
        let enhancedDarkMode = App.shared.settings[SettingsKeys.enhancedDarkMode] as? Bool ?? false
        switch (traitCollection.userInterfaceStyle, enhancedDarkMode) {
        case (.dark, true):
            return UIColor(white: 0.2, alpha: 1.0)
        default:
            return .white
        }
    }

    // MARK: Properties

	let box: BoxType
    var creditLabel: UILabel?

    var crossColor: UIColor = Asset.correctionPass1.color
    var overlayMode: Bool = false {
        didSet {
            self.creditLabel?.isHidden = !self.overlayMode
            self.setNeedsDisplay()
        }
    }

	var normalizedFrame: CGRect = .zero

	var selectedObserver: AnyCancellable?

    private let disabledAlpha: CGFloat = 0.5

    override var isHighlighted: Bool {
        didSet {
            self.setNeedsDisplay()
        }
    }

    override var isEnabled: Bool {
        didSet {
            self.setNeedsDisplay()
        }
    }

    init(frame: CGRect, box: BoxType, overrideModeActive: Bool = false) {
        self.box = box
        super.init(frame: frame)

        if let correctionBox = self.box as? CorrectionBoxType {
            // Add Label
            let creditLabel = UILabel()
            creditLabel.translatesAutoresizingMaskIntoConstraints = false
            let credit = correctionBox.value
            creditLabel.text = floor(credit) == credit ? String((Int(credit))) : ""
            creditLabel.textColor = Self.foregroundColor
            creditLabel.font = UIFont.systemFont(ofSize: 9, weight: .bold)
            creditLabel.adjustsFontSizeToFitWidth = true
            creditLabel.minimumScaleFactor = 0.5
            creditLabel.textAlignment = .center
            self.addSubview(creditLabel)
            NSLayoutConstraint.activate([
                creditLabel.leadingAnchor.constraint(equalTo: self.leadingAnchor, constant: 1),
                creditLabel.trailingAnchor.constraint(equalTo: self.trailingAnchor, constant: -1),
                creditLabel.topAnchor.constraint(equalTo: self.topAnchor),
                creditLabel.bottomAnchor.constraint(equalTo: self.bottomAnchor)
            ])
            creditLabel.isHidden = !self.overlayMode
            self.creditLabel = creditLabel

            // swiftlint:disable:next
            let allowModToAll = overrideModeActive && App.shared.settings[SettingsKeys.allowModificationsToAllProblems] as? Bool ?? false

            self.isEnabled = correctionBox.editable || allowModToAll
        }

        self.addTarget(self, action: #selector(touchUpInside), for: .touchUpInside)

        selectedObserver = box.selected.receiveOnMain().assignNoRetain(to: \.isSelected, on: self)
    }

    required init?(coder aDecoder: NSCoder) {
        fatalError("Init with coder not implemented on CorrectionBoxButton")
    }

    // MARK: Drawing

    override func draw(_ rect: CGRect) {
        if overlayMode {
            // Fill background in overlay mode
            let rect = UIBezierPath(rect: bounds)
            Self.backgroundColor.setFill()
            rect.fill()
        }
        if isSelected || isHighlighted {
            if isEnabled {
                crossColor.setStroke()
            } else {
                crossColor.withAlphaComponent(disabledAlpha).setStroke()
            }

            let rect = UIBezierPath(rect: bounds.insetBy(dx: 1, dy: 1))
            rect.lineWidth = 1.5
            rect.stroke()

            let inset = 0.2 * bounds.width
            let cross = UIBezierPath(cross: bounds.insetBy(dx: inset, dy: inset))
            cross.lineWidth = 0.15 * bounds.width
            cross.stroke()
        } else if overlayMode {
            // Draw Black box in overlay mode
            let rect = UIBezierPath(rect: bounds.insetBy(dx: 1, dy: 1))
            rect.lineWidth = 1.5
            Self.foregroundColor.setStroke()
            rect.stroke()
        }
    }

    // MARK: Touch responder

    @objc func touchUpInside() {
        let allowModToAll = App.shared.settings[SettingsKeys.allowModificationsToAllProblems] as? Bool ?? false
        box.setSelected(!self.isSelected, allowModificationToAll: allowModToAll)
    }

}

extension UIBezierPath {
    convenience init(cross rect: CGRect) {
        self.init()
        self.addCross(rect)
    }

    func addCross(_ rect: CGRect) {
        self.move(to: rect.origin)
        self.addLine(to: CGPoint(x: rect.maxX, y: rect.maxY))
        self.move(to: CGPoint(x: rect.minX, y: rect.maxY))
        self.addLine(to: CGPoint(x: rect.maxX, y: rect.minY))
    }
}
