// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit

@available(iOS 14.0, *)
class DatacolCell: UICollectionViewListCell {
    struct Configuration {
        let columnCount: Int
        let columnWidths: [CGFloat]
        let columnGenerator: ((Int) -> UIView)
        let columnConfigurator: ((Int, UIView) -> Void)?
        let resizeHidePreferences: [Int]
        let cellSpacing: CGFloat
    }

    // MARK: Properties

    let contentStack = UIStackView().autolayout()

    var dataConfig: Configuration? {
        didSet {
            if oldValue == nil {
                updateLayout()
            } else {
                assert(
                    dataConfig?.columnCount == contentStack.arrangedSubviews.count,
                    "Can't change the number of columns once configured once"
                )
            }
            updateData()
        }
    }

    private var currentlyVisibleIndices = Set<Int>()

    // MARK: Initializers

    override init(frame: CGRect) {
        super.init(frame: frame)
        setup()
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    // MARK: Private functions

    private func setup() {
        contentStack.axis = .horizontal
        contentStack.distribution = .equalSpacing
        contentStack.alignment = .center
        contentStack.backgroundColor = .clear
        contentStack.spacing = 8

        contentView.addSubview(contentStack)
        contentStack.addLeadingConstraint(to: \.layoutMarginsGuide.leadingAnchor)
        contentStack.addTopConstraint(to: \.layoutMarginsGuide.topAnchor)
        contentStack.addTrailingConstraint(to: \.layoutMarginsGuide.trailingAnchor)
        contentStack.addBottomConstraint(to: \.layoutMarginsGuide.bottomAnchor)
    }

    private func updateLayout() {
        guard let config = dataConfig else { return }
        assert(config.columnCount == config.columnWidths.count)

        currentlyVisibleIndices = Set<Int>(0..<config.columnCount)
        contentStack.spacing = config.cellSpacing

        for index in 0..<config.columnCount {
            let width = config.columnWidths[index]
            let columnView = config.columnGenerator(index)

            contentStack.addArrangedSubview(columnView)
            columnView.addWidth(width, priority: .init(999))
        }
    }

    private func updateData() {
        guard let config = dataConfig else { return }

        for index in 0..<config.columnCount {
            config.columnConfigurator?(index, contentStack.arrangedSubviews[index])
        }
        setNeedsLayout()
    }

    override func layoutSubviews() {
        super.layoutSubviews()
        guard let config = dataConfig else {
            return
        }
        assert(config.columnWidths.count == config.resizeHidePreferences.count)

        let availableWidth = contentView.bounds.inset(by: contentView.layoutMargins).width

        var summedVisibleWidth = currentlyVisibleIndices.map { config.columnWidths[$0] }.reduce(0, +)
        summedVisibleWidth += CGFloat(currentlyVisibleIndices.count - 1) * config.cellSpacing

        var summedTotalWidth = config.columnWidths.reduce(0, +)
        summedTotalWidth += CGFloat(config.columnCount - 1) * config.cellSpacing

        if summedVisibleWidth > availableWidth {
            // Need to hide something
            var amountHidden = config.columnCount - currentlyVisibleIndices.count
            while summedVisibleWidth > availableWidth {
                guard currentlyVisibleIndices.count > 1 else { break }
                let indexToHide = amountHidden
                // Hide view according to preference
                let prefView = contentStack.arrangedSubviews[config.resizeHidePreferences[indexToHide]]
                prefView.isHidden = true
                currentlyVisibleIndices.remove(config.resizeHidePreferences[indexToHide])
                amountHidden += 1

                summedVisibleWidth = currentlyVisibleIndices
                    .map { config.columnWidths[$0] }
                    .reduce(CGFloat(currentlyVisibleIndices.count - 1) * config.cellSpacing, +)
            }
        } else if summedVisibleWidth < availableWidth {
            // Possibly can show more
            var amountVisible = currentlyVisibleIndices.count
            while summedVisibleWidth < availableWidth {
                guard currentlyVisibleIndices.count < config.columnCount else { break }
                let indexToShow = config.columnCount - amountVisible

                if config.columnWidths[config.resizeHidePreferences[indexToShow]] + summedVisibleWidth + config.cellSpacing > availableWidth {
                    break
                }

                let prefView = contentStack.arrangedSubviews[config.resizeHidePreferences[indexToShow]]
                prefView.isHidden = false
                currentlyVisibleIndices.insert(config.resizeHidePreferences[indexToShow])
                amountVisible += 1

                summedVisibleWidth = currentlyVisibleIndices
                    .map { config.columnWidths[$0] }
                    .reduce(CGFloat(currentlyVisibleIndices.count - 1) * config.cellSpacing, +)
            }
        }
    }

    override func prepareForReuse() {
        super.prepareForReuse()
        setNeedsLayout()
    }
}
