// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit

protocol StatusCollectionViewCellDelegate: AnyObject {
	func statusCell(_ cell: StatusCollectionViewCell, didTapButtonWithPass pass: Int)
    func statusCellTappedInfoButton(_ cell: StatusCollectionViewCell)
}

class StatusCollectionViewCell: UICollectionViewCell, CellClassReuseIdentifier {
	let corneredView: ShadowView = ShadowView()
	let horizontalStackView = UIStackView().autolayout()
	let statusLabelStackView = UIStackView().autolayout()
	let problemLabel: UILabel = UILabel().autolayout()
	let ringProgressView = NestedRingProgressView().autolayout()
    let infoButton = UIButton().autolayout()

	weak var delegate: StatusCollectionViewCellDelegate?

    // Problem + Passes
	var problem: CorrectionProblemStatusType? {
		didSet {
            if oldValue?.problem != problem?.problem {
				problemDidChange()
			}
		}
	}
    var accessRights: [Int]? {
        didSet {
            problemDidChange()
        }
    }
    var isAdmin: Bool = false

	var passButtons: [UIButton] = []

	override init(frame: CGRect) {
		super.init(frame: frame)
		setup()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	func setup() {
		// setup horizontal stack
		horizontalStackView.axis = .horizontal
		horizontalStackView.alignment = .center
		horizontalStackView.distribution = .fill
		horizontalStackView.spacing = 32

		// setup problem label
		problemLabel.font = UIFont.boldSystemFont(ofSize: 27)
        problemLabel.textAlignment = .center
		problemLabel.text = "Problem 0"

		// set constraints for ring progress view
        ringProgressView.addWidth(to: \.heightAnchor, of: ringProgressView)
        ringProgressView.addWidth(200, relation: .equalOrLess)
        ringProgressView.addWidth(120, relation: .equalOrGreater)

		// set up vertical status label stack
		statusLabelStackView.axis = .vertical
		statusLabelStackView.alignment = .fill
		statusLabelStackView.spacing = 8

		statusLabelStackView.addArrangedSubview(problemLabel)

		// add subviews to horizontal stack
		horizontalStackView.addArrangedSubview(ringProgressView)
		horizontalStackView.addArrangedSubview(statusLabelStackView)

		// add to content view, and add constraints
		contentView.addSubview(corneredView)
		contentView.addSubview(horizontalStackView)
		corneredView.pinToSuperview(UIEdgeInsets(20))
		horizontalStackView.pinToSuperview(UIEdgeInsets(top: 20, left: 20+24+12, bottom: 20, right: 20+20))

		// some final appearance stuff for the cornered view
		corneredView.clipsToBounds = false
		corneredView.backgroundColor = UIColor.secondarySystemGroupedBackground
		corneredView.layer.cornerRadius = 20.0

        // Info Button
        infoButton.setImage(UIImage(systemName: "ellipsis.circle"), for: .normal)
        infoButton.tintColor = Asset.tuMexamBlue.color
        contentView.addSubview(infoButton)

        infoButton.topAnchor.constraint(equalTo: corneredView.topAnchor, constant: 20).activate()
        infoButton.leadingAnchor.constraint(equalTo: corneredView.leadingAnchor, constant: 20).activate()
        infoButton.addWidth(44)
        infoButton.addAspectRatio(heightToWidth: 1)

        infoButton.addTarget(self, action: #selector(infoButtonPressed(_:)), for: .touchUpInside)
	}

	private func createPassButton(pass: Int, completed: Bool) -> UIButton {
		let button = ZoomButton(frame: CGRect(origin: .zero, size: CGSize(width: 0, height: 44)))
		button.layer.masksToBounds = true
		button.layer.cornerRadius = 17
		button.titleLabel?.textColor = .white
		button.tintColor = .white
		button.addTarget(self, action: #selector(buttonPressed(_:)), for: .touchUpInside)
		button.tag = pass
		button.contentEdgeInsets = UIEdgeInsets(top: 8, left: 16, bottom: 8, right: 16)
		if completed {
			button.isEnabled = false
			button.backgroundColor = .systemGray
		} else {
			button.backgroundColor = Asset.tuMexamBlue.color
		}
		return button
	}

    override func layoutSubviews() {
        super.layoutSubviews()

        for button in passButtons {
            let imageWidth = button.imageView?.frame.width ?? 0.0
            let contentFrame = button.frame.inset(by: button.contentEdgeInsets)
            let buttonWidth = contentFrame.size.width
            button.imageEdgeInsets = UIEdgeInsets(top: 0, left: buttonWidth-imageWidth+4.0,
                                                  bottom: 0, right: -(buttonWidth-imageWidth))
            button.titleEdgeInsets = UIEdgeInsets(top: 0, left: -imageWidth, bottom: 0, right: imageWidth)
        }
    }

	@objc
	func buttonPressed(_ sender: UIButton) {
		delegate?.statusCell(self, didTapButtonWithPass: sender.tag)
	}

    @objc
    func infoButtonPressed(_ sender: UIButton) {
        delegate?.statusCellTappedInfoButton(self)
    }

	func problemDidChange() {
		guard let problem = problem else { return }
        problemLabel.text = L10n.StatusController.Cell.problem("\(problem.problem)")

		for button in passButtons {
			button.removeFromSuperview()
		}
		passButtons = []

        let passes = problem.passes
		var progress: [Double] = []

        for (index, pass) in passes.enumerated() {
			// calculate progress
            progress.append(pass.progressRespectingSubproblems)

			// create button
            let button = createPassButton(pass: index + 1, completed: (pass.corrected >= pass.total))
            let chevron = NSTextAttachment(image: UIImage(systemName: "chevron.right")!)
            let title = NSMutableAttributedString(
                string: "\(L10n.StatusController.Cell.pass) \(index + 1): \(pass.corrected)/\(pass.total) "
            )
            title.append(NSAttributedString(attachment: chevron))
            button.setAttributedTitle(title, for: .normal)

            if let accessRights = accessRights, accessRights.contains(pass.pass) {
                button.isEnabled = true
                button.backgroundColor = Asset.tuMexamBlue.color
            } else if isAdmin {
                button.isEnabled = true
                button.backgroundColor = Asset.tuMexamBlue.color
            } else {
                button.isEnabled = false
                button.backgroundColor = .systemGray
            }

			statusLabelStackView.addArrangedSubview(button)
			passButtons.append(button)
		}

		ringProgressView.setUsing(progress: progress)
	}
}
