// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit

@available(iOS 14.0, *)
class SidebarHeaderView: UICollectionReusableView {

    enum State {
        case loading
        case secure
        case unsecure
    }

    static let elementKind = "SidebarHeaderViewElementKind"

    var state: State = .loading {
        didSet {
            switch state {
            case .loading:
                activityIndicator.startAnimating()
                lockView.isHidden = true
            case .secure:
                activityIndicator.stopAnimating()
                lockView.isHidden = false
                lockView.image = UIImage(systemName: "lock")
                lockView.tintColor = .systemGreen
            case .unsecure:
                activityIndicator.stopAnimating()
                lockView.isHidden = false
                lockView.image = UIImage(systemName: "lock.slash")
                lockView.tintColor = .systemRed
            }
        }
    }

    var instanceNameLabel: UILabel!
    var userLabel: UILabel!
    var examSelectionButton: UIButton!
    var activityIndicator: UIActivityIndicatorView!
    var lockView: UIImageView!

    override init(frame: CGRect) {
        super.init(frame: frame)
        setupView()
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    func setupView() {
        let instanceNameLabel = UILabel().autolayout()
        instanceNameLabel.text = "Instance"
        instanceNameLabel.font = .systemFont(ofSize: 21, weight: .semibold)
        instanceNameLabel.adjustsFontSizeToFitWidth = true
        instanceNameLabel.numberOfLines = 2
        instanceNameLabel.minimumScaleFactor = 0.6

        self.addSubview(instanceNameLabel)
        self.instanceNameLabel = instanceNameLabel

        let userLabel = UILabel().autolayout()
        userLabel.font = .preferredFont(forTextStyle: .footnote)
        userLabel.textColor = .secondaryLabel

        self.addSubview(userLabel)
        self.userLabel = userLabel

        let examSelectionButton = UIButton(type: .system).autolayout()
        examSelectionButton.setTitle("Exam", for: .normal)
        examSelectionButton.titleLabel?.font = .systemFont(ofSize: 17)
        examSelectionButton.setImage(UIImage(systemName: "chevron.down")!, for: .normal)
        examSelectionButton.contentHorizontalAlignment = .trailing
        examSelectionButton.semanticContentAttribute = .forceRightToLeft
        examSelectionButton.tintColor = Asset.barButtonColor.color
        examSelectionButton.setPreferredSymbolConfiguration(
            .init(pointSize: examSelectionButton.titleLabel!.font.pointSize),
            forImageIn: .normal
        )
        examSelectionButton.imageEdgeInsets = UIEdgeInsets(top: 1, left: 1, bottom: 0, right: 0)

        self.addSubview(examSelectionButton)
        self.examSelectionButton = examSelectionButton

        // Activity Indicator
        let activityIndicator = UIActivityIndicatorView(style: .medium).autolayout()
        if state == .loading {
            activityIndicator.startAnimating()
        } else {
            activityIndicator.stopAnimating()
        }

        self.addSubview(activityIndicator)
        self.activityIndicator = activityIndicator

        // Lock View
        let lockView = UIImageView().autolayout()
        if state == .secure {
            lockView.image = UIImage(systemName: "lock")
            lockView.tintColor = .systemGreen
        } else if state == .unsecure {
            lockView.image = UIImage(systemName: "lock.slash")
            lockView.tintColor = .systemRed
        } else if state == .loading {
            lockView.isHidden = true
        }

        self.addSubview(lockView)
        self.lockView = lockView

        // Setup constraints
        activityIndicator.leading(to: \.leadingAnchor, constant: 5)
        activityIndicator.top(to: \.topAnchor, constant: -10)
        activityIndicator.trailing(to: \.leadingAnchor, of: userLabel, constant: 5)
        activityIndicator.bottom(to: \.topAnchor, of: instanceNameLabel, constant: 10)

        lockView.pinToEdges(of: activityIndicator)

        userLabel.top(to: \.topAnchor, of: activityIndicator)
        userLabel.trailing(to: \.trailingAnchor)
        userLabel.bottom(to: \.bottomAnchor, of: activityIndicator)

        instanceNameLabel.leading(to: \.leadingAnchor, constant: 5)
        instanceNameLabel.trailing(to: \.trailingAnchor, constant: 5)
        instanceNameLabel.bottom(to: \.topAnchor, of: examSelectionButton)

        examSelectionButton.leading(to: \.leadingAnchor, constant: 5)
        examSelectionButton.trailing(to: \.trailingAnchor, relation: .equalOrLess, constant: 5)
        examSelectionButton.bottom(to: \.bottomAnchor, constant: 15)
    }

}
