// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit

class FlipTransition: NSObject {

    weak var flipCell: UIView!

    init(_ source: UIView) {
        super.init()
        self.flipCell = source
    }

}

extension FlipTransition: UIViewControllerTransitioningDelegate {
    func animationController(forPresented presented: UIViewController,
                             presenting: UIViewController,
                             source: UIViewController) -> UIViewControllerAnimatedTransitioning? {
        return self
    }

    func animationController(forDismissed dismissed: UIViewController) -> UIViewControllerAnimatedTransitioning? {
        return self
    }
}

extension FlipTransition: UIViewControllerAnimatedTransitioning {
    func transitionDuration(using transitionContext: UIViewControllerContextTransitioning?) -> TimeInterval {
        return 0.3
    }

    func animateTransition(using transitionContext: UIViewControllerContextTransitioning) {
        let container = transitionContext.containerView
        let fromVC = transitionContext.viewController(forKey: .from)!
        let fromView = fromVC.view!
        let toVC = transitionContext.viewController(forKey: .to)!
        let toView = toVC.view!
        let isOpening = flipCell.isDescendant(of: fromView)
        self.flipCell.isHidden = false
        let snapshot = flipCell.snapshotView(afterScreenUpdates: true)!
        self.flipCell.isHidden = true

        snapshot.frame = container.convert(flipCell.frame, from: flipCell.superview)
        container.addSubview(snapshot)
        container.bringSubviewToFront(container)

        // Calculate Start mid and destpoints for animation

        var sourcePoint: CGPoint!
        var destPoint: CGPoint!
        var midSize: CGSize!

        if isOpening {
            sourcePoint = snapshot.center
            destPoint = transitionContext.finalFrame(for: toVC).center
            midSize = CGSize(width: (snapshot.frame.size.width + toView.frame.size.width) / 2,
                             height: (snapshot.frame.size.height + toView.frame.size.height) / 2)
        } else {
            sourcePoint = transitionContext.initialFrame(for: fromVC).center
            destPoint = snapshot.center
            midSize = CGSize(width: (snapshot.frame.size.width + fromView.frame.size.width) / 2,
                             height: (snapshot.frame.size.height + fromView.frame.size.height) / 2)
        }

        let midPoint = CGPoint(x: (sourcePoint.x + destPoint.x) / 2,
                               y: (sourcePoint.y + destPoint.y) / 2)

        // Setup Initial Transforms

        if isOpening {
            container.addSubview(toView)
            container.bringSubviewToFront(toView)
            toView.frame = transitionContext.finalFrame(for: toVC)

            let viewTranslation = CATransform3DMakeTranslation(sourcePoint.x - midPoint.x,
                                                               sourcePoint.y - midPoint.y,
                                                               0)
            let viewScale = CATransform3DScale(viewTranslation,
                                               midSize.width / toView.frame.size.width,
                                               midSize.height / toView.frame.size.height,
                                               1)
            let viewRotation = CATransform3DRotate(viewScale, -.pi / 2, 0, 1, 0)
            toView.layer.transform = viewRotation
        } else {
            let cellTranslation = CATransform3DMakeTranslation(midPoint.x - destPoint.x,
                                                               midPoint.y - destPoint.y,
                                                               0)
            let cellRotation = CATransform3DRotate(cellTranslation, .pi / 2, 0, 1, 0)
            let cellScale = CATransform3DScale(cellRotation,
                                               midSize.width / snapshot.frame.size.width,
                                               midSize.height / snapshot.frame.size.height,
                                               1.0)
            snapshot.layer.transform = cellScale
        }

        // Perform animation

        let duration = transitionDuration(using: transitionContext)
        let options = UIView.KeyframeAnimationOptions(rawValue: UIView.AnimationOptions.curveEaseInOut.rawValue)
        UIView.animateKeyframes(withDuration: duration,
                                delay: 0,
                                options: options) {
            UIView.addKeyframe(withRelativeStartTime: 0, relativeDuration: 0.5) {
                if isOpening {
                    self.flipCell.isHidden = true
                    let cellTranslation = CATransform3DMakeTranslation(midPoint.x - sourcePoint.x,
                                                                       midPoint.y - sourcePoint.y, 0)
                    let cellRotation = CATransform3DRotate(cellTranslation, .pi / 2, 0, 1, 0)
                    let cellScale = CATransform3DScale(cellRotation,
                                                       midSize.width / snapshot.frame.size.width,
                                                       midSize.height / snapshot.frame.size.height,
                                                       1.0)
                    snapshot.layer.transform = cellScale
                } else {
                    let viewTranslation = CATransform3DMakeTranslation(destPoint.x - midPoint.x,
                                                                       destPoint.y - midPoint.y,
                                                                       0)
                    let viewScale = CATransform3DScale(viewTranslation,
                                                       midSize.width / toView.frame.size.width,
                                                       midSize.height / toView.frame.size.height,
                                                       1)
                    let viewRotation = CATransform3DRotate(viewScale, -.pi / 2, 0, 1, 0)
                    fromView.layer.transform = viewRotation
                }
            }
            UIView.addKeyframe(withRelativeStartTime: 0.5, relativeDuration: 0.5) {
                if isOpening {
                    snapshot.layer.opacity = 0
                    toView.layer.transform = CATransform3DIdentity
                } else {
                    snapshot.layer.transform = CATransform3DIdentity
                }
            }
        } completion: { (completion) in
            if completion {
                if isOpening {

                } else {
                    snapshot.removeFromSuperview()
                    self.flipCell.isHidden = false
                }
            }
            transitionContext.completeTransition(completion)
        }

    }
}
