// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import MBProgressHUD

/**
 UIView which holds displays the scan and the correction layers.
 */
class CanvasContainerView: UIView {

    // MARK: Properties

    private let backgroundImageView = UIImageView()
    private let staticOverlayImageView = UIImageView() // For older correction layers
    private let solutionImageView = UIImageView()
    private var scanNotFoundLabel: UILabel?

    var showsSolution = false {
        didSet {
            solutionImageView.isHidden = !self.showsSolution
            self.isUserInteractionEnabled = !self.showsSolution
        }
    }

    var drawingView: CanvasView? {
        willSet {
            if let old = drawingView {
                old.removeFromSuperview()
            }
        }

        didSet {
            if let newDrawingView = drawingView {
                newDrawingView.layer.opacity = 0.7
                newDrawingView.frame = backgroundImageView.frame
                self.addSubview(newDrawingView)
                self.bringSubviewToFront(newDrawingView)
            }
        }
    }

    var backgroundImage: UIImage? {
        didSet {
            backgroundImageView.image = backgroundImage
        }
    }

    var solutionImage: UIImage? {
        didSet {
            solutionImageView.image = solutionImage
        }
    }

    var oldCorrectionLayers = [UIImage]() {
        didSet {
            guard !oldCorrectionLayers.isEmpty else {
                return
            }

            let format = UIGraphicsImageRendererFormat()
            format.opaque = false
            if #available(iOS 12.0, *) {
                format.preferredRange = .standard
            } else {
                format.prefersExtendedRange = false
            }
            format.scale = self.traitCollection.displayScale
            let renderer = UIGraphicsImageRenderer(size: self.bounds.size)

            let combined = renderer.image { context in
                UIGraphicsPushContext(context.cgContext)
                for layer in oldCorrectionLayers {
                    layer.draw(in: CGRect(origin: CGPoint.zero, size: self.bounds.size))
                }
                UIGraphicsPopContext()
            }
            staticOverlayImageView.image = combined
        }
    }

    let canvasInset: CGFloat = 0.0

    var canvasSize: CGSize {
        didSet {
            self.frame = CGRect(origin: .zero,
                                size: CGSize(width: 2 * canvasInset + canvasSize.width,
                                             height: 2 * canvasInset + canvasSize.height))
            backgroundImageView.frame = CGRect(origin: CGPoint(x: canvasInset, y: canvasInset), size: canvasSize)
            backgroundImageView.layer.shadowPath = UIBezierPath(rect: backgroundImageView.bounds).cgPath
            drawingView?.frame = backgroundImageView.frame
        }
    }

    var sizeWidthAddedInsets: CGSize {
        return CGSize(width: canvasSize.width + 2 * canvasInset, height: canvasSize.height + 2 * canvasInset)
    }

    var actualBounds: CGRect {
        return CGRect(origin: CGPoint.zero,
                      size: CGSize(width: bounds.size.width * transform.a,
                                   height: bounds.size.height * transform.d))
    }

    // GroupId -> Button Group
    var correctionBoxButtonGroups = [String: CorrectionBoxButtonGroup]()
    var correctionButtonRects = [CGRect]()
    var correctionMistakeButtons = [UIButton]()

    // MARK: Initializers

    required init(canvasSize: CGSize) {
        // Setup view
        self.canvasSize = canvasSize
        super.init(frame: CGRect.zero)
        self.backgroundColor = Asset.correctionBackground.color

        // Setup background image view
        backgroundImageView.contentMode = .scaleToFill
        backgroundImageView.backgroundColor = .white
        backgroundImageView.layer.shadowOffset = CGSize(width: 0.0, height: 3.0)
        backgroundImageView.layer.shadowRadius = 4.0
        backgroundImageView.layer.shadowColor = UIColor.darkGray.cgColor
        backgroundImageView.layer.shadowOpacity = 1.0
        backgroundImageView.layer.shadowPath = UIBezierPath(rect: backgroundImageView.bounds).cgPath
        self.addSubview(backgroundImageView)

        // Setup Overlay image view for displaying older correction layers
        staticOverlayImageView.contentMode = .scaleToFill
        staticOverlayImageView.alpha = 0.7
        staticOverlayImageView.backgroundColor = .clear
        staticOverlayImageView.autoresizingMask = [.flexibleHeight, .flexibleWidth]
        backgroundImageView.addSubview(staticOverlayImageView)

        // Setup Solution image view for displaying the sample solution
        solutionImageView.contentMode = .scaleToFill
        solutionImageView.backgroundColor = .white
        solutionImageView.isHidden = !self.showsSolution
        solutionImageView.autoresizingMask = [.flexibleWidth, .flexibleHeight]
        self.addSubview(solutionImageView)
    }

    required init?(coder aDecoder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func point(inside point: CGPoint, with event: UIEvent?) -> Bool {
        return super.point(inside: point, with: event)
    }

    // MARK: Functions

    func addCorrectionBoxGroups(_ pageAsset: CorrectionPageAsset, sheetAsset: CorrectionSheetAsset, overlayMode: Bool) {
        // We need these to enable/disable the buttons
        let problem = sheetAsset.problem
        let pass = sheetAsset.pass

        var pageSize = CGSize(width: pageAsset.pageWidth, height: pageAsset.pageHeight)
        if pageAsset.isLandscape {
            // Rotate 90 deg because landscape
            pageSize = pageSize.rotate90
        }
        let boxWidth = pageAsset.boxWidth
        // Create a group for each group asset
        pageAsset.boxGroupAssets.forEach { boxGroupAsset in
            let buttons = boxGroupAsset.boxAssets.map({ boxAsset -> CorrectionBoxButton in
                let originalButtonRect = CGRect(x: boxAsset.xPos, y: boxAsset.yPos, width: boxWidth, height: boxWidth)
                let transformedButtonRect = rectForCorrectionBoxInCanvas(originalButtonRect, originalPageSize: pageSize)

                // Add rect to correctionButtonRects to ignore pencil gestures
                correctionButtonRects.append(transformedButtonRect.insetBy(dx: -10, dy: -10))

                let button = CorrectionBoxButton(frame: transformedButtonRect, asset: boxAsset)
                button.overlayMode = overlayMode
                // Enable when problem matches selected problem and either the pass matches selected pass or the problem is multi. choice
                button.isSelected = boxAsset.isSelected
                // swiftlint:disable line_length
                if let allowsModificationsToAllProblems = AppStateManager.shared.settings?[SettingsKeys.allowModificationsToAllProblems] as? Bool {
                    button.isEnabled = (allowsModificationsToAllProblems || boxGroupAsset.problem == problem) && (boxGroupAsset.allowsMultipleChoice ? true : boxGroupAsset.pass == pass)
                } else {
                    button.isEnabled = boxGroupAsset.problem == problem && (boxGroupAsset.allowsMultipleChoice ? true : boxGroupAsset.pass == pass)
                }
                //swiftlint:enable line_length
                // Change button cross color if pass != 1 (red is default)
                if boxAsset.pass == 2 {
                    button.crossColor = Asset.correctionPass2.color
                } else if boxAsset.pass == 3 {
                    button.crossColor = Asset.correctionPass3.color
                }
                // Add CorrectionBoxButton to view
                self.addSubview(button)
                return button
            })
            /* Adding flags to View as buttons
            if let correctionMistakeBox: CorrectionMistakeBoxAsset = boxGroupAsset.correctionMistakeAsset {
                let originalButtonRect = CGRect(x: correctionMistakeBox.xPos,
                                                y: correctionMistakeBox.yPos,
                                                width: boxWidth,
                                                height: boxWidth)
                let transformedButtonRect = rectForCorrectionBoxInCanvas(originalButtonRect, originalPageSize: pageSize)

                // Add rect to correctionButtonRects to ignore pencil gestures
                correctionButtonRects.append(transformedButtonRect.insetBy(dx: -10, dy: -10))

                let correctionMistakeButton: UIButton = CorrectionMistakeButton(frame: transformedButtonRect, asset: correctionMistakeBox)
                correctionMistakeButton.isSelected = correctionMistakeBox.isSelected
                self.addSubview(correctionMistakeButton)
                correctionMistakeButtons.append(correctionMistakeButton)
                buttons.append(correctionMistakeButton)
            }
            */
            let group = CorrectionBoxButtonGroup(buttons: buttons,
                                                 callback: nil,
                                                 allowsMultipleChoice: boxGroupAsset.allowsMultipleChoice)
            // We need to store the group somewhere, otherwise it will dealloc
            self.correctionBoxButtonGroups[boxGroupAsset.groupId] = group
        }
        self.bringSubviewToFront(self.solutionImageView)
    }

    func updateOverlayMode(_ overlayMode: Bool) {
        self.correctionBoxButtonGroups.values.forEach {
            $0.updateOverlayMode(overlayMode)
        }
    }

    // MARK: Private Helper Functions

    private func rectForCorrectionBoxInCanvas(_ originalRect: CGRect, originalPageSize: CGSize) -> CGRect {
        let transformedBoxWidth = (originalRect.width / originalPageSize.width) * canvasSize.width
        let transformedX = (originalRect.origin.x / originalPageSize.width) * canvasSize.width
        let transformedY = canvasSize.height - (originalRect.origin.y / originalPageSize.height) * canvasSize.height
        let center = CGPoint(x: transformedX, y: transformedY)
        let size = CGSize(width: transformedBoxWidth, height: transformedBoxWidth)
        return CGRect(center: center, size: size)
    }
}
