// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Logging
import MBProgressHUD

class CanvasView: BaseView {
	var strokeGestureRecognizer: PencilStrokeGestureRecognizer?

	var drawing: Drawing
	var drawingLayer: DrawingLayer

	var tool: Tool = .pen
	var color: UIColor?

	var contextUndoManager: ContextUndoManager?

    var activityIndicator: UIActivityIndicatorView?

	init(drawing: Drawing) {
		self.drawing = drawing
		self.drawingLayer = DrawingLayer(drawing: drawing)
		super.init()
		self.drawing.delegates.add(self)
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override func setup() {
		super.setup()

        self.drawingLayer.drawingLayerDelegate = self

		let stroke = PencilStrokeGestureRecognizer(target: self, action: #selector(didUpdateStroke(_:)))
		stroke.delegate = self
		stroke.swiftAllowedTouchTypes = [.pencil]
		addGestureRecognizer(stroke)
		strokeGestureRecognizer = stroke

		layer.addSublayer(drawingLayer)
		drawingLayer.frame = layer.bounds
		drawingLayer.contentsScale = UIScreen.main.scale

//		for element in drawing.elements {
//			drawingLayer.append(element: element)
//		}

        drawingLayer.loadExistingElements()
	}

	@objc func didUpdateStroke(_ recognizer: PencilStrokeGestureRecognizer) {
		guard let stroke = recognizer.stroke else {
            Logger.drawing.error("CanvasView: Stroke Recognizer didn't provide a stroke in didUpdateStroke(_:)")
			return
		}
		switch recognizer.state {
		case .began:
			drawingLayer.activeStrokeBegan(stroke)
		case .changed:
			drawingLayer.activeStrokeChanged(stroke)
		case .ended:
			drawingLayer.activeStrokeEnded(stroke)

			let finalStroke = stroke.immutableCopy()
			drawing.append(stroke: finalStroke)

		case .failed:
			drawingLayer.activeStrokeEnded(stroke)
		default:
			break
		}
	}

	override func layoutSubviews() {
		super.layoutSubviews()
		drawingLayer.frame = layer.bounds
        activityIndicator?.center = layer.bounds.center
	}

	func undo() {
		if !drawing.elements.isEmpty {
			drawing.removeLast()
		}
	}
}

extension CanvasView: PencilStrokeGestureRecognizerDelegate {
	func inkForGestureRecognizer(_ recognizer: PencilStrokeGestureRecognizer) -> Ink {
		guard let color = color?.cgColor else {
            Logger.drawing.critical("CanvasView - inkForGestureRecognizer: color property was nil")
			fatalError("can't draw without color")
		}
		let globalDrawingOptions = App.shared.globalDrawingOptions

        let toolOptions: GlobalDrawingOptions.ToolOptions = {
            switch tool {
			case .pen:
                return globalDrawingOptions.pen
            case .eraser:
                return globalDrawingOptions.eraser
            case .highlighter:
                return globalDrawingOptions.highlighter
            }
        }()
		// calculate adjusted line width
		let normalizedLineWidth = drawing.normalizedLineWidth(toolOptions.lineWidth)

		let pressure = (tool == .pen ? globalDrawingOptions.pressureSensitiveLineWidth : false)

		return Ink(
			tool: tool,
			color: color,
			normalizedLineWidth: normalizedLineWidth,
			pressureSensitiveLineWidth: pressure
		)
	}
}

extension CanvasView: DrawingDelegate {
    func drawing(_ drawing: Drawing, didRemoveLastElement element: Drawing.Element) {
		contextUndoManager?.registerUndo(withTarget: self, handler: { (canvas) in
			canvas.drawing.append(element: element)
		})
	}

    func drawing(_ drawing: Drawing, didAppendElement element: Drawing.Element) {
		contextUndoManager?.registerUndo(withTarget: self, handler: { (canvas) in
			canvas.drawing.removeLast()
		})
	}
}

extension CanvasView: DrawingLayerDelegate {
    func drawingLayerStartedLoadingExistingLayers(_ layer: DrawingLayer) {
        // guard let superview = self.superview else { return }
        guard activityIndicator == nil else { return }
        let activity = UIActivityIndicatorView(style: .large)
        self.activityIndicator = activity
        self.addSubview(activity)
        activity.hidesWhenStopped = true
        // activity.pinToEdges()
        activity.startAnimating()
        self.backgroundColor = .systemGray.withAlphaComponent(0.4)
    }

    func drawingLayerFinishedLoadingExistingLayers(_ layer: DrawingLayer) {
        guard let activity = self.activityIndicator else {
            return
        }
        activity.stopAnimating()
        self.backgroundColor = nil
        self.activityIndicator = nil
    }
}
