// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

extension CALayer {
	@discardableResult
	func disableAnimations() -> Self {
		let newActions = [
			"onOrderIn": NSNull(),
			"onOrderOut": NSNull(),
			"sublayers": NSNull(),
			"contents": NSNull(),
			"bounds": NSNull(),
			"mask": NSNull(),
			"position": NSNull(),
			"contentsScale": NSNull()
		]
		self.actions = newActions
		return self
	}

	func ensureFrame(_ frame: CGRect) {
		if self.frame != frame {
			self.frame = frame
		}
	}

	func syncSublayerContentsScale() {
		for layer in sublayers ?? [] {
			layer.contentsScale = contentsScale
		}
	}
}

class BaseLayer: CALayer {
	override init() {
		super.init()

		setup()
	}

    override init(layer: Any) {
        super.init(layer: layer)
    }

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override var contentsScale: CGFloat {
		didSet {
			guard let sublayers = sublayers else { return }
			for layer in sublayers {
				if let layer = layer as? BaseLayer {
					layer.contentsScale = contentsScale
				}
			}
		}
	}

	override func layoutSublayers() {
		super.layoutSublayers()

		guard let sublayers = sublayers else { return }
		for layer in sublayers {
			if let layer = layer as? BaseLayer {
				layer.ensureFrame(bounds)
			}
		}
	}

	func setup() {
		disableAnimations()
	}

	override func addSublayer(_ layer: CALayer) {
		layer.frame = bounds
		layer.contentsScale = contentsScale

		super.addSublayer(layer)
	}
}
