// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import UIKit
import TUMexamKit
import Logging

extension DrawingLayer {
    func collapseIfNecessary(useBackground: Bool = true) {
        let currentCount = drawing.elements.count

        if currentCount - renderedElementCount > Self.maxLayerCount {
            self.collapse()
        }

        debounceTimer?.invalidate()
        debounceTimer = Timer.scheduledTimer(withTimeInterval: 0.5, repeats: false, block: { [weak self] _ in
            self?.collapse(useBackground: useBackground)
        })
    }

    func collapse(useBackground: Bool = true) {
        if useBackground {
            collapseInBackground()
        } else {
            collapseOnMain()
        }
    }

    func collapseInBackground() {
        guard isCurrentlyRendering.wait(timeout: .now()) == .success else {
            return
        }

        Self.renderQueue.async { [weak self] in
            guard let drawing = self?.drawing else {
                self?.isCurrentlyRendering.signal()
                return
            }

            let elements = drawing.elements
            guard let size = self?.bounds.size,
                  let scale = self?.contentsScale,
                  let refSize = self?.referenceSize(),
                  let image = drawing.getImageForLayerCollapse(
                    size: size,
                    referenceSize: refSize,
                    contentsScale: scale
                  ) else {
                Logger.drawing.error("Layer collapse failed: no image")
                self?.isCurrentlyRendering.signal()
                return
            }

            DispatchQueue.main.async {
                self?.renderingFinished(image: image, containedElements: elements)
            }
        }
    }

    func collapseOnMain() {
        guard isCurrentlyRendering.wait(timeout: .now()) == .success else {
            return
        }

        let elements = drawing.elements
        guard let image = drawing.getImageForLayerCollapse(
            size: bounds.size,
            referenceSize: self.referenceSize(),
            contentsScale: self.contentsScale
        ) else {
            Logger.drawing.error("Layer collapse (on main) failed: no image ")
            isCurrentlyRendering.signal()
            return
        }

        renderingFinished(image: image, containedElements: elements)
    }

    func renderingFinished(image: CGImage, containedElements: [Drawing.Element]) {
        // ALWAYS set to false
        defer {
            isCurrentlyRendering.signal()
            currentlyRenderingExisting = false
        }

        if self.currentlyRenderingExisting {
            drawingLayerDelegate?.drawingLayerFinishedLoadingExistingLayers(self)
        }

        // check whether elements were removed
        guard drawing.elements.count >= containedElements.count else {
            Logger.drawing.error("Layer collapse failed: elements were removed while rendering")
            return
        }

        // check whether rendered elements didn't change
        guard containedElements == Array(drawing.elements.prefix(containedElements.count)) else {
            Logger.drawing.error("Layer collapse failed: rendered elements changed while rendering")
            return
        }

        // create new layer and switch layers
        let backgroundLayer = BaseLayer()
        backgroundLayer.contents = image
        currentGroupLayer.removeFromSuperlayer()
        currentGroupLayer = ElementGroupLayer(background: backgroundLayer, drawingSize: drawing.normalizedRect.size)
        contentLayer.addSublayer(currentGroupLayer)

        // add remaining elements to layer
        if drawing.elements.count > containedElements.count {
            for element in drawing.elements.suffix(from: containedElements.count) {
                append(element: element)
            }
        }

        renderedElementCount = containedElements.count
        Logger.drawing.info("Layer collapse succeeded: Collapsed \(containedElements.count) layers")
    }

    func referenceSize() -> CGSize {
        return bounds.size.normalizedIn(size: drawing.normalizedRect.size)
    }
}
