// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit
import Logging

protocol DrawingLayerDelegate: AnyObject {
    func drawingLayerStartedLoadingExistingLayers(_ layer: DrawingLayer)
    func drawingLayerFinishedLoadingExistingLayers(_ layer: DrawingLayer)
}

class DrawingLayer: BaseLayer {
	let drawing: Drawing
	var currentGroupLayer: ElementGroupLayer!

	private var activeStrokeLayer: ActiveStrokeLayer?

	var contentLayer: BaseLayer = BaseLayer()

	var eraserIndicator: CALayer = CALayer()

    static let maxLayerCount: Int = 20
    static let renderQueue: DispatchQueue = DispatchQueue(
        label: Bundle.main.bundleIdentifier! + ".layer-collapse",
        qos: .default
    )

    var renderedElementCount: Int = 0
    var isCurrentlyRendering: DispatchSemaphore = {
        let sema = DispatchSemaphore(value: 0)
        sema.signal()
        return sema
    }()
    var debounceTimer: Timer?
    var renderExistingOnNextLayout = false
    var currentlyRenderingExisting = false

    weak var drawingLayerDelegate: DrawingLayerDelegate?

	override var contentsScale: CGFloat {
		didSet {
			syncSublayerContentsScale()
		}
	}

    init(drawing: Drawing, drawingLayerDelegate: DrawingLayerDelegate? = nil) {
		self.drawing = drawing

		let backgroundLayer = BaseLayer()
        if let originalImage = drawing.originalImage {
            backgroundLayer.contents = originalImage
        }

        self.drawingLayerDelegate = drawingLayerDelegate

		super.init()
        currentGroupLayer = ElementGroupLayer(background: backgroundLayer, drawingSize: drawing.normalizedRect.size)
		disableAnimations()
		addSublayer(contentLayer)
		contentLayer.addSublayer(currentGroupLayer)

		drawing.delegates.add(self)

		eraserIndicator.anchorPoint = CGPoint(x: 0.5, y: 0.5)
		eraserIndicator.disableAnimations()
        eraserIndicator.cornerRadius = eraserIndicator.bounds.size.width/2
        eraserIndicator.backgroundColor = UIColor.gray.withAlphaComponent(0.5).cgColor
	}

    override init(layer: Any) {
        if let other = layer as? DrawingLayer {
            self.drawing = other.drawing
            self.currentGroupLayer = other.currentGroupLayer
            self.activeStrokeLayer = other.activeStrokeLayer
            self.contentLayer = other.contentLayer
            self.eraserIndicator = other.eraserIndicator
            self.renderedElementCount = other.renderedElementCount
            self.isCurrentlyRendering = other.isCurrentlyRendering
        } else {
            fatalError("Cannot copy instance variables from different layer type")
        }
        super.init(layer: layer)
    }

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	func activeStrokeBegan(_ stroke: ActiveStroke) {
        switch stroke.ink.tool {
        case .pen:
            let activePen = ActivePenStrokeLayer(stroke: stroke, drawingSize: drawing.normalizedRect.size)
			self.activeStrokeLayer = activePen
			activePen.frame = bounds
			activePen.contentsScale = UIScreen.main.scale
			contentLayer.addSublayer(activePen)
			activePen.setNeedsDisplay()
        case .highlighter:
            let activeHighlighter = ActiveHighlighterStrokeLayer(
                stroke: stroke,
                drawingSize: drawing.normalizedRect.size
            )
            self.activeStrokeLayer = activeHighlighter
            activeHighlighter.frame = bounds
            activeHighlighter.contentsScale = UIScreen.main.scale
            contentLayer.addSublayer(activeHighlighter)
            activeHighlighter.setNeedsDisplay()
        case .eraser:
            let activeEraser = ActiveEraserStrokeLayer(
                stroke: stroke,
                drawingSize: drawing.normalizedRect.size
            )
			self.activeStrokeLayer = activeEraser
			activeEraser.frame = bounds
			activeEraser.contentsScale = UIScreen.main.scale
			contentLayer.mask = activeEraser
			activeEraser.setNeedsDisplay()

			addSublayer(eraserIndicator)
			updateEraserSize(stroke: stroke)
			updateEraser(stroke: stroke)
		}
	}

	func activeStrokeChanged(_ stroke: ActiveStroke) {
		activeStrokeLayer?.strokeChanged()
		if stroke.ink.tool == .eraser {
			updateEraser(stroke: stroke)
		}
	}

	func activeStrokeEnded(_ stroke: ActiveStroke) {
		if let active = activeStrokeLayer {
			contentLayer.mask = nil
			active.removeFromSuperlayer()
		}
		activeStrokeLayer = nil
		if eraserIndicator.superlayer != nil {
			eraserIndicator.removeFromSuperlayer()
		}
	}

    func append(element: Drawing.Element) {
        switch element {
        case let .stroke(stroke):
            append(stroke: stroke)
        case let .snippet(snippet):
            append(snippet: snippet)
		case let .selection(selection):
			append(selection: selection)
            Logger.drawing.debug(
                "Added Element \(pointerString(drawing)) \(drawing.elements.count) \(self.bounds.debugDescription)"
            )
		}
        collapseIfNecessary()
	}

    func append(stroke: Stroke) {
        if currentGroupLayer.appendIfPossible(stroke) {
            Logger.drawing.debug("Stroke added to current group")
        } else {
            Logger.drawing.debug("Stroke: Can't add to current group, creating new one")

            currentGroupLayer.removeFromSuperlayer()
            let newGroup = ElementGroupLayer(background: currentGroupLayer, drawingSize: drawing.normalizedRect.size)
            guard newGroup.appendIfPossible(stroke) else {
                fatalError()
            }

            contentLayer.addSublayer(newGroup)
            currentGroupLayer = newGroup
        }
    }

    func append(snippet: SnippetElement) {
        if currentGroupLayer.appendIfPossible(snippet) {
            Logger.drawing.debug("Snippet added to current group")
        } else {
            Logger.drawing.debug("Snippet: Cant add to current group, creating new one")

            currentGroupLayer.removeFromSuperlayer()
            let newGroup = ElementGroupLayer(background: currentGroupLayer, drawingSize: drawing.normalizedRect.size)
            guard newGroup.appendIfPossible(snippet) else {
                fatalError()
            }

            contentLayer.addSublayer(newGroup)
            currentGroupLayer = newGroup
        }
    }

	func append(selection: SelectionElement) {
		if currentGroupLayer.appendIfPossible(selection) {
            Logger.drawing.debug("Selection added to current group")
		} else {
            Logger.drawing.debug("Selection: Cant add to current group, creating new one")

			currentGroupLayer.removeFromSuperlayer()
            let newGroup = ElementGroupLayer(background: currentGroupLayer, drawingSize: drawing.normalizedRect.size)
			guard newGroup.appendIfPossible(selection) else {
				fatalError()
			}

			contentLayer.addSublayer(newGroup)
			currentGroupLayer = newGroup
		}
	}

    func removeLast() -> Drawing.Element? {
		if let element = currentGroupLayer.removeLast() {
			if !currentGroupLayer.canRemoveMore {
				if let background = currentGroupLayer.backgroundLayer as? ElementGroupLayer {
					currentGroupLayer.removeFromSuperlayer()
					contentLayer.addSublayer(background)
					currentGroupLayer = background
				} else {
					// for now, do nothing
				}
			}
			setNeedsDisplay()
			return element
		} else {
            if drawing.elements.count < Self.maxLayerCount {
                let backgroundLayer = BaseLayer()
                backgroundLayer.contents = drawing.originalImage
                currentGroupLayer.removeFromSuperlayer()
                currentGroupLayer = ElementGroupLayer(
                    background: backgroundLayer,
                    drawingSize: drawing.normalizedRect.size
                )
                contentLayer.addSublayer(currentGroupLayer)

                for element in drawing.elements {
                    append(element: element)
                }
            } else {
                collapse(useBackground: false)
            }
            return nil
		}
	}

	func updateEraserSize(stroke: ActiveStroke) {
		let mappedRadius = stroke.ink.normalizedLineWidth * max(bounds.size.width, bounds.size.height)
		let rect = CGRect(origin: .zero, size: CGSize(width: mappedRadius, height: mappedRadius))
		eraserIndicator.frame = rect
        eraserIndicator.cornerRadius = rect.size.width/2
	}

	func updateEraser(stroke: ActiveStroke) {
		guard let last = stroke.samples.last else { return }
		let location = last.location.scaled(to: bounds.size)
		eraserIndicator.position = location
	}

    func loadExistingElements() {
        // Insert potential existing elements as rendered background
        guard drawing.elements.count > 0 else { return }
        drawingLayerDelegate?.drawingLayerStartedLoadingExistingLayers(self)
        self.currentlyRenderingExisting = true
        guard self.frame != .zero else {
            self.renderExistingOnNextLayout = true
            return
        }
        collapseInBackground()
    }
    override func layoutSublayers() {
        super.layoutSublayers()

        guard self.frame != .zero else { return }
        if renderExistingOnNextLayout {
            collapseInBackground()
            renderExistingOnNextLayout = false
        }
    }
}

extension DrawingLayer: DrawingDelegate {
    func drawing(_ drawing: Drawing, didAppendElement element: Drawing.Element) {
		append(element: element)
	}

    func drawing(_ drawing: Drawing, didRemoveLastElement element: Drawing.Element) {
		_ = removeLast()
	}
}
