// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit

class ElementGroupLayer: BaseLayer {
	let backgroundLayer: CALayer
    var strokeOpacity: Float?
	var strokes: [Stroke] = []
	var strokeLayers: [StrokeLayer] = []
	var eraserMask: CombinedEraserMaskLayer?

    init(background: CALayer) {
		self.backgroundLayer = background
		super.init()

		addSublayer(backgroundLayer)
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

    func appendIfPossible(_ snippet: SnippetElement) -> Bool {
        return false
    }

	func appendIfPossible(_ stroke: Stroke) -> Bool {
        switch stroke.ink.tool {
        case .pen:
			// Adding a Pen Stroke

            // set stroke opacity if nil
            if strokeOpacity == nil {
                strokeOpacity = Tool.pen.strokeDisplayOpacity
            }

            // Check if stroke opacity is equal to required opacity
            let strokeOpacity = self.strokeOpacity!
            if strokeOpacity != Tool.pen.strokeDisplayOpacity {
                return false
            }

			// can't add another line if an eraser mask is active
			guard eraserMask == nil else { return false }
			// no eraser mask -> good to go
            let strokeLayer = StrokeLayer(stroke: stroke, opacity: stroke.ink.tool.strokeDisplayOpacity)
			addSublayer(strokeLayer)
			strokeLayers.append(strokeLayer)
			strokeLayer.setNeedsDisplay()
			strokes.append(stroke)
			return true
        case .highlighter:
            // Adding a Highliter Stroke

            // set stroke opacity if nil
            if strokeOpacity == nil {
                strokeOpacity = Tool.highlighter.strokeDisplayOpacity
            }

            // Check if stroke opacity is equal to required opacity
            let strokeOpacity = self.strokeOpacity!
            if strokeOpacity != Tool.highlighter.strokeDisplayOpacity {
                return false
            }

            // can't add another line if an eraser mask is active
            guard eraserMask == nil else { return false }
            // no eraser mask -> good to go
            let strokeLayer = StrokeLayer(stroke: stroke, opacity: stroke.ink.tool.strokeDisplayOpacity)
            addSublayer(strokeLayer)
            strokeLayers.append(strokeLayer)
            strokeLayer.setNeedsDisplay()
            strokes.append(stroke)
            return true
        case .eraser:
			// Adding an Eraser Stroke
			if let eraserMask = eraserMask {
				eraserMask.append(stroke)
			} else {
				let eraserMask = CombinedEraserMaskLayer(stroke: stroke)
				eraserMask.contentsScale = contentsScale
				eraserMask.frame = bounds
				mask = eraserMask
				self.eraserMask = eraserMask
				eraserMask.setNeedsDisplay()
			}
			return true
		}
	}

	/// whether this group contains strokes that can be removed
	var canRemoveMore: Bool {
		if let eraserMask = eraserMask {
			return !eraserMask.strokes.isEmpty
		}
		return !strokes.isEmpty
	}

	func removeLast() -> Stroke? {
		if let eraserMask = eraserMask {
			let stroke = eraserMask.removeLast()
			if eraserMask.canRemoveMore == false {
				// no strokes left in eraser mask -> remove
				self.eraserMask = nil
				self.mask = nil
			}
			return stroke
		}
		let strokeLayer = strokeLayers.removeLast()
		strokeLayer.removeFromSuperlayer()
		return strokes.removeLast()
	}

}
