// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit

class CombinedEraserMaskLayer: BaseLayer {
    let drawingSize: CGSize
	var strokes: [Stroke]

    init(strokes: [Stroke], drawingSize: CGSize) {
		assert(strokes.allSatisfy({ $0.ink.tool == .eraser }))
		self.strokes = strokes
        self.drawingSize = drawingSize
		super.init()
	}

    convenience init(stroke: Stroke, drawingSize: CGSize) {
        self.init(strokes: [stroke], drawingSize: drawingSize)
	}

    override init(layer: Any) {
        if let other = layer as? CombinedEraserMaskLayer {
            self.strokes = other.strokes
            self.drawingSize = other.drawingSize
        } else {
            self.strokes = []
            self.drawingSize = .zero
        }
        super.init(layer: layer)
    }

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override func draw(in ctx: CGContext) {
		ctx.setFillColor(UIColor.black.cgColor)
		ctx.fill(bounds)

		ctx.setBlendMode(.copy)

		for stroke in strokes {
			stroke.draw(in: ctx, withSize: bounds.size, referenceSize: referenceSize())
		}
	}

	func append(_ stroke: Stroke) {
		assert(stroke.ink.tool == .eraser)

		strokes.append(stroke)
		setNeedsDisplay()
	}

	var canRemoveMore: Bool {
		return !strokes.isEmpty
	}

	func removeLast() -> Stroke {
		let last = strokes.removeLast()
		setNeedsDisplay()
		return last
	}

    func referenceSize() -> CGSize {
        return bounds.size.normalizedIn(size: drawingSize)
    }
}
