// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit
import TUMexamKit

class ActiveStroke {
	let ink: Ink
	/// indicates the number of samples, which won't change their value (e.g. can be used for fixed drawing)
	var finishedCount: Int = 0
	let viewSize: CGSize

	var samples: [StrokeSample] = []

	var predictedSamples: [StrokeSample] = []

	private var estimatedSampleMapping: [NSNumber: Int] = [:]

	init(ink: Ink, viewSize: CGSize) {
		self.ink = ink
		self.viewSize = viewSize
	}

	func add(sample: StrokeSample) {
		samples.append(sample)
	}

	func immutableCopy() -> Stroke {
		assert(predictedSamples.isEmpty)
		return Stroke(samples: samples, ink: ink)
	}

	func immutableCopy(forPrefix prefix: Int) -> Stroke {
		assert(prefix <= samples.count)
		return Stroke(samples: Array(samples[0..<prefix]), ink: ink)
	}
}

protocol PencilStrokeGestureRecognizerDelegate: UIGestureRecognizerDelegate {
	func inkForGestureRecognizer(_ recognizer: PencilStrokeGestureRecognizer) -> Ink
}

class PencilStrokeGestureRecognizer: UIGestureRecognizer {
	var trackedTouch: UITouch?
	var stroke: ActiveStroke?

	override func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent) {
		guard let delegate = delegate as? PencilStrokeGestureRecognizerDelegate else {
			fatalError("delegate is required to provide ink")
		}

		let ink = delegate.inkForGestureRecognizer(self)

		if touches.count != 1 {
			self.state = .failed
		}

		guard trackedTouch == nil else {
			return
		}

		// get first touch. should always work (see touches.count != 1)
		guard let touch = touches.first,
			let view = self.view else { fatalError() }

		// create new stroke!

		let stroke = ActiveStroke(ink: ink, viewSize: view.bounds.size)
		self.stroke = stroke

		if let coalesced = event.coalescedTouches(for: touch) {
			if !stroke.addSamples(for: coalesced, in: view) {
				finishAndEnd()
			}
		}
	}

	override func touchesMoved(_ touches: Set<UITouch>, with event: UIEvent?) {
		guard let touch = touches.first,
			let view = self.view,
			let stroke = self.stroke,
			let coalesced = event?.coalescedTouches(for: touch)
			else { return }

		if !stroke.addSamples(for: coalesced, in: view) {
			finishAndEnd()
			return
		}

		if let predicted = event?.predictedTouches(for: touch) {
			stroke.addPredictedSamples(for: predicted, in: view)
		}
		state = .changed
	}

	override func touchesEstimatedPropertiesUpdated(_ touches: Set<UITouch>) {
		guard let touch = touches.first,
			let stroke = self.stroke
			else { return }

		stroke.updateSamples(for: touch)
	}

	override func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?) {
		guard let touch = touches.first,
			let view = self.view,
			let stroke = self.stroke,
			let coalesced = event?.coalescedTouches(for: touch)
			else { return }

		// don't care whether there were points out of bounds
		_ = stroke.addSamples(for: coalesced, in: view)
		stroke.predictedSamples = []

		state = .ended
	}

	override func touchesCancelled(_ touches: Set<UITouch>, with event: UIEvent?) {
		stroke?.predictedSamples = []
		state = .cancelled
	}

	override func reset() {
		stroke = nil
		self.trackedTouch = nil
	}

	func finishAndEnd() {
		stroke?.predictedSamples = []
		state = .ended
	}
}

extension ActiveStroke {
	func addSamples(for touches: [UITouch], in view: UIView) -> Bool {
		// Add all of the touches to the active stroke.
		for touch in touches {
			let viewLocation = touch.preciseLocation(in: view)
			let location = viewLocation.normalized(using: view.bounds.size)

			if !view.bounds.contains(viewLocation) {
				return false
			}

			if touch == touches.last {
				let sample = StrokeSample(
					location: location,
					coalescedSample: false,
					force: touch.force,
                    maximumPossibleForce: touch.maximumPossibleForce
				)

				self.add(sample: sample)
			} else {
				// If the touch is not the last one in the array,
				//  it was a coalesced touch.
				let sample = StrokeSample(
					location: location,
					coalescedSample: true,
					force: touch.force,
                    maximumPossibleForce: touch.maximumPossibleForce
				)

				self.add(sample: sample)
			}

			if touch.estimatedPropertiesExpectingUpdates.contains(.force),
				let updateIndex = touch.estimationUpdateIndex {
				estimatedSampleMapping[updateIndex] = samples.count - 1
			}
		}

		if let maxSampleIndex = estimatedSampleMapping.values.max() {
			self.finishedCount = maxSampleIndex
		}

		return true
	}

	func addPredictedSamples(for touches: [UITouch], in view: UIView) {
		predictedSamples = touches.map {
			StrokeSample(location: $0.preciseLocation(in: view).normalized(using: view.bounds.size),
						 coalescedSample: false,
						 force: $0.force,
                         maximumPossibleForce: $0.maximumPossibleForce
			)
		}
	}

	func updateSamples(for touch: UITouch) {
		guard let updateIndex = touch.estimationUpdateIndex,
			let index = estimatedSampleMapping[updateIndex] else { return }

		samples[index].force = touch.force
	}

	func toPath() -> CGPath {
		let path = CGMutablePath()
		// a single point is not a stroke
		guard samples.count >= 2 else { return path }

		path.move(to: samples[0].location)
		for sample in samples[1..<samples.count] {
			path.addLine(to: sample.location)
		}
		for sample in predictedSamples {
			path.addLine(to: sample.location)
		}

		return path
	}
}
