// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import SwiftUI
import TUMexamKit

extension Int: Identifiable {
    public var id: String {
        String(self)
    }
}

struct ResultsFilterView: View {

    @ObservedObject var filter: ResultFilterBox

    var body: some View {
        Form {
            // Reset Section
            Section {
                // Reset Button
                Button(action: {
                    self.filter.filter = ResultFilter()
                }, label: {
                    Text(L10n.ResultsFilter.reset)
                        .modifier(CenteredFormLabel())
                })
            }
            // Filter Section
            Section(header: Text(L10n.ResultsFilter.filters)) {
                // ERID
                HStack {
                    Text(L10n.ResultsFilter.Filters.erid)
                    Spacer()
                    TextField(L10n.ResultsFilter.Filters.Erid.eg, text: $filter.filter.erid.bound)
                        .multilineTextAlignment(.trailing)
                }
                // Problem
                HStack {
                    Text(L10n.ResultsFilter.Filters.problem)
                    Spacer()
                    TextField(L10n.ResultsFilter.Filters.Problem.eg, text: $filter.filter.problem.boundStringValue)
                        .multilineTextAlignment(.trailing)
                        // .keyboardType(.numbersAndPunctuation)
                }
                // Subroblem
                HStack {
                    Text(L10n.ResultsFilter.Filters.subproblem)
                    Spacer()
                    TextField(L10n.ResultsFilter.Filters.Subproblem.eg, text: $filter.filter.subproblem.bound)
                        .multilineTextAlignment(.trailing)
                        .autocapitalization(.none)
                }
                // Pass
                HStack {
                    Text(L10n.ResultsFilter.Filters.pass)
                    Spacer()
                    TextField(L10n.ResultsFilter.Filters.Pass.eg, text: $filter.filter.pass.boundStringValue)
                        .multilineTextAlignment(.trailing)
                        .keyboardType(.numberPad)
                }
                // Flag Type Ids
                MultiSelector(
                    label: Text("Flags"),
                    options: filter.availableFlagTypeIds,
                    optionToString: { val in
                        return filter.availableFlagDefinitions.first { $0.flagTypeId == val }?.text ?? ""
                    },
                    optionToColor: { val in
                        return Color(
                            UIColor(
                                hex: filter.availableFlagDefinitions.first { $0.flagTypeId == val }?.color ?? "#FFFFFF"
                            )
                        )
                    },
                    selected: $filter.filter.flagTypeIds
                )
                // Corrected
                HStack {
                    Text(L10n.ResultsFilter.Filters.corrected)
                    Spacer()
                    Picker("", selection: $filter.filter.corrected) {
                        Text(L10n.ResultsFilter.all).tag(FilterBoolValue.all)
                        Text(L10n.ResultsFilter.yes).tag(FilterBoolValue.yes)
                        Text(L10n.ResultsFilter.no).tag(FilterBoolValue.no)
                    }.pickerStyle(SegmentedPickerStyle())
                }
                // Available
                HStack {
                    Text(L10n.ResultsFilter.Filters.available)
                    Spacer()
                    Picker("", selection: $filter.filter.available) {
                        Text(L10n.ResultsFilter.all).tag(FilterBoolValue.all)
                        Text(L10n.ResultsFilter.yes).tag(FilterBoolValue.yes)
                        Text(L10n.ResultsFilter.no).tag(FilterBoolValue.no)
                    }.pickerStyle(SegmentedPickerStyle())
                }
                // Locked
                HStack {
                    Text(L10n.ResultsFilter.Filters.locked)
                    Spacer()
                    Picker("", selection: $filter.filter.locked) {
                        Text(L10n.ResultsFilter.all).tag(FilterBoolValue.all)
                        Text(L10n.ResultsFilter.yes).tag(FilterBoolValue.yes)
                        Text(L10n.ResultsFilter.no).tag(FilterBoolValue.no)
                    }.pickerStyle(SegmentedPickerStyle())
                }
                // Access
                HStack {
                    Text(L10n.ResultsFilter.Filters.access)
                    Spacer()
                    Picker("", selection: $filter.filter.access) {
                        Text(L10n.ResultsFilter.all).tag(FilterBoolValue.all)
                        Text(L10n.ResultsFilter.yes).tag(FilterBoolValue.yes)
                        Text(L10n.ResultsFilter.no).tag(FilterBoolValue.no)
                    }.pickerStyle(SegmentedPickerStyle())
                }
            }
            // Sort Section
            Section(header: Text(L10n.ResultsFilter.sort)) {
                // Sort By
                HStack {
                    Text(L10n.ResultsFilter.Sort.column)
                    Spacer()
                    Picker("", selection: $filter.filter.sortColum) {
                        Text(L10n.ResultsFilter.Sort.Column.erid).tag(ResultSortByColumn.erid)
                        Text(L10n.ResultsFilter.Sort.Column.problem).tag(ResultSortByColumn.problem)
                        Text(L10n.ResultsFilter.Sort.Column.pass).tag(ResultSortByColumn.pass)
                        Text(L10n.ResultsFilter.Sort.Column.date).tag(ResultSortByColumn.date)
                    }.pickerStyle(SegmentedPickerStyle())
                }
                // Sort Direction
                HStack {
                    Text(L10n.ResultsFilter.Sort.direction)
                    Spacer()
                    Picker("", selection: $filter.filter.sortDirection) {
                        Text(L10n.ResultsFilter.Sort.Direction.asc).tag(FilterSortDirection.ascending)
                        Text(L10n.ResultsFilter.Sort.Direction.desc).tag(FilterSortDirection.descending)
                    }.pickerStyle(SegmentedPickerStyle())
                }
            }
        }.onAppear(perform: {
            if #available(iOS 15.0, *) { } else {
                UITableView.appearance().contentInset.top = -35
            }
        }).onDisappear(perform: {
            if #available(iOS 15.0, *) { } else {
                UITableView.appearance().contentInset.top = 0
            }
        })
    }
}

struct ResultsFilterView_Previews: PreviewProvider {
    static var previews: some View {
        // Approx. Popover size on iPad Pro 12,9
        let layout = PreviewLayout.fixed(width: 380, height: 700)
        return Group {
            ResultsFilterView(filter: ResultFilterBox())
                .previewLayout(layout)
                .previewDisplayName("Light Mode")
            ResultsFilterView(filter: ResultFilterBox())
                .environment(\.colorScheme, .dark)
                .previewLayout(layout)
                .previewDisplayName("Dark Mode")
        }
    }
}
