// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import SwiftUI
import UIKit
import Combine

enum FeedbackSubmissionError: Error, LocalizedError {
    case networkError(response: URLResponse)
    case statusCode(Int)

    var errorDescription: String? {
        switch self {
        case .networkError(response: let resp):
            return "Erroneous URLResponse: \(resp)"
        case .statusCode(let code):
            return "Unnaceptable status code: \(code)"
        }
    }
}

var cancellables = Set<AnyCancellable>()

struct SentryFeedbackView: View {

    // swiftlint:disable:next line_length
    private static let feedbackURL = URL(string: "https://sentry.tumexam.de/api/0/projects/tumexam/ipad-app/user-feedback/")!
    private static let clientDSN = "DSN https://fd78b2f7b267467497caeb8cc37b2c56@sentry.tumexam.de/2"

    var dismissHandler: (() -> Void)?

    @State var eventId: String = ""
    @State var fullName: String = ""
    @State var email: String = ""
    @State var comments: String = ""

    var body: some View {
        NavigationView {
            ZStack {
                Form {
                    Section(header: sectionHeader,
                            footer: Text(L10n.SentryFeedback.footer)) {
                                TextField(L10n.SentryFeedback.fullName,
                                          text: $fullName)
                                TextField(L10n.SentryFeedback.email,
                                          text: $email)
                                TextView(L10n.SentryFeedback.comments,
                                         text: $comments)
                                    .frame(height: 100.0)
                    }
                }
                .navigationBarTitle(LocalizedStringKey(L10n.SentryFeedback.navbar), displayMode: .inline)
                .navigationBarItems(leading: Button(L10n.skip) {
                    guard let handler = self.dismissHandler else { return }
                    handler()
                }.disabled(!cancellables.isEmpty), trailing: Button(action: {
                    self.submitFeedback()
                }, label: {
                    Text(L10n.submit).bold()
                }).disabled(comments.isEmpty || !cancellables.isEmpty))
                if #available(iOS 14.0, *) {
                    if !cancellables.isEmpty {
                        ProgressView(L10n.SentryFeedback.sending)
                            .padding()
                            .background(
                                RoundedRectangle(cornerRadius: 15)
                                    .foregroundColor(Color(.tertiarySystemBackground))
                                    .shadow(color: Color(.tertiarySystemFill), radius: 30)
                            )
                    }
                }
            }
        }
    }

    private var sectionHeader: some View {
        VStack(alignment: .leading, spacing: 8.0) {
            Text(L10n.SentryFeedback.title)
                .fontWeight(.heavy)
                .font(.system(size: 21))
                .foregroundColor(.primary)
                .lineLimit(nil)
            Text(L10n.SentryFeedback.body)
                .fontWeight(.regular)
        }
        .padding(.vertical)
    }

    private func submitFeedback() {
        BreadcrumbManager.addInteraction("SentryFeedback - Submitting sentry feedback")

        let name = self.fullName.isEmpty ? "TUMexam User" : self.fullName
        let email = self.email.isEmpty ? "noreply@tumexam.de" : self.email
        let jsonDict = [
            "event_id": eventId,
            "name": name,
            "email": email,
            "comments": comments
        ]

        let jsonEnc = JSONEncoder()
        let jsonData = try? jsonEnc.encode(jsonDict)

        var req = URLRequest(url: SentryFeedbackView.feedbackURL)
        req.addValue("application/json", forHTTPHeaderField: "Content-Type")
        req.addValue("application/json", forHTTPHeaderField: "Accept")
        req.addValue(SentryFeedbackView.clientDSN, forHTTPHeaderField: "Authorization")
        req.httpMethod = "POST"
        req.httpBody = jsonData

        URLSession.shared.dataTaskPublisher(for: req)
            .tryMap { _, resp -> Void in
                guard let httpResponse = resp as? HTTPURLResponse else {
                    throw FeedbackSubmissionError.networkError(response: resp)
                }

                guard 200..<400 ~= httpResponse.statusCode else {
                    throw FeedbackSubmissionError.statusCode(httpResponse.statusCode)
                }
            }
            .receive(on: RunLoop.main)
            .sink(receiveCompletion: { completion in
                switch completion {
                case .failure(let err):
                    let errorDescription = err.localizedDescription
                    let alert = UIAlertController(title: L10n.SentryFeedback.FailureAlert.title,
                                                  message: L10n.SentryFeedback.FailureAlert.message(errorDescription),
                                                  preferredStyle: .alert)
                    alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: { _ in
                        guard let handler = self.dismissHandler else { return }
                        handler()
                    }))
                    alert.show(animated: true)
                case .finished:
                    let alert = UIAlertController(title: L10n.SentryFeedback.SuccessAlert.title,
                                                  message: L10n.SentryFeedback.SuccessAlert.message,
                                                  preferredStyle: .alert)
                    alert.addAction(UIAlertAction(title: L10n.ok, style: .default, handler: { _ in
                        guard let handler = self.dismissHandler else { return }
                        handler()
                    }))
                    alert.show(animated: true)
                }
                cancellables.removeAll()
            }, receiveValue: {  })
            .store(in: &cancellables)
    }
}

struct SentryFeedbackView_Previews: PreviewProvider {
    static var previews: some View {
        // Usual iPad OS Form Sheet size, found no better way to set this
        let layout = PreviewLayout.fixed(width: 540, height: 680)
        return Group {
            SentryFeedbackView(eventId: "")
                .previewLayout(layout)
                .previewDisplayName("Light Mode")
            SentryFeedbackView(eventId: "")
                .environment(\.colorScheme, .dark)
                .previewLayout(layout)
                .previewDisplayName("Dark Mode")
        }
    }
}
