// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import SwiftUI
import TUMexamKit

struct ExamWrapper: Equatable, CustomStringConvertible, Hashable {
    static func == (lhs: ExamWrapper, rhs: ExamWrapper) -> Bool {
        return lhs.identifier == rhs.identifier && lhs.name == rhs.name
    }

    let exam: ExamType

    var identifier: ExamIdentifier { exam.identifier }
    var name: String { exam.name.value ?? ""}
    var description: String { name }

    func hash(into hasher: inout Hasher) {
        hasher.combine(exam.identifier)
        hasher.combine(exam.name.value)
    }
}

final class SettingsProxy: ObservableObject {
    // swiftlint:disable force_cast
    var exam = App.shared.currentExam.map { ExamWrapper(exam: $0) } {
        didSet {
            guard let exam = exam?.exam else {
                return
            }
            App.shared.setCurrentExam(exam)
        }
    }

    var hideOverlays = App.shared.settings[SettingsKeys.hideOverlays] as! Bool {
        didSet {
            App.shared.settings[SettingsKeys.hideOverlays] = hideOverlays
        }
    }
    var allowModificationToAll = App.shared.settings[SettingsKeys.allowModificationsToAllProblems] as! Bool {
        didSet {
            App.shared.settings[SettingsKeys.allowModificationsToAllProblems] = allowModificationToAll
        }
    }
    var retainPageSelection = App.shared.settings[SettingsKeys.retainPageSelection] as! Bool {
        didSet {
            App.shared.settings[SettingsKeys.retainPageSelection] = retainPageSelection
        }
    }
    var applePencilGesture = App.shared.settings[SettingsKeys.applePencilDoubleTapGestureEnabled] as! Bool {
        didSet {
            App.shared.settings[SettingsKeys.applePencilDoubleTapGestureEnabled] = applePencilGesture
        }
    }
    var askBeforeDismiss = App.shared.settings[SettingsKeys.askBeforeDismissingCorrectionSession] as! Bool {
        didSet {
            App.shared.settings[SettingsKeys.askBeforeDismissingCorrectionSession] = askBeforeDismiss
        }
    }

    var enhancedDarkMode = App.shared.settings[SettingsKeys.enhancedDarkMode] as! Bool {
        didSet {
            App.shared.settings[SettingsKeys.enhancedDarkMode] = enhancedDarkMode
        }
    }

    // swiftlint:disable:next line_length
    var sampleSolutionOpacity: Double = (App.shared.settings[SettingsKeys.sampleSolutionOpacity] as? Double) != nil ? App.shared.settings[SettingsKeys.sampleSolutionOpacity] as! Double : Double(App.shared.settings[SettingsKeys.sampleSolutionOpacity] as! Int) {
        didSet {
            // For some reason, this helped resolve a bug with implicit conversion to int. Dont understand y tho
            App.shared.settings[SettingsKeys.sampleSolutionOpacity] = sampleSolutionOpacity
        }
    }

    var collectCrashInfo = App.shared.settings[SettingsKeys.sentryCrashLogsOptIn] as! Bool {
        didSet {
            App.shared.settings[SettingsKeys.sentryCrashLogsOptIn] = collectCrashInfo
        }
    }

    // swiftlint:enable force_cast
}

struct SettingsView: View {

    private let versionString: String  = {
        guard let shortVersion = Bundle.main.object(forInfoDictionaryKey: "CFBundleShortVersionString") as? String,
              let version = Bundle.main.object(forInfoDictionaryKey: kCFBundleVersionKey as String) as? String else {
                return ""
        }
        return "\(shortVersion) (\(version))"
    }()

    // swiftlint:disable:next force_cast
    private var _oldCrashSetting = App.shared.settings[SettingsKeys.sentryCrashLogsOptIn] as! Bool

    @ObservedObject private var proxy = SettingsProxy()

    @State private var cacheSizeString = SettingsView.cacheSizeString()

    @State private var showsDisconnectAlert = false
    @State private var showsZombieLocksAlert = false
    @State private var showsTutorialAlert = false

    var body: some View {
        Form {
            // Header
            Section {
                // Header
                self.headerView
                // Check for Updates
                Button(action: {
                    App.shared.checkForUpdates(userInitiated: true)
                }, label: {
                    Text(L10n.Settings.checkForUpdates)
                        .foregroundColor(Color(Asset.tuMexamBlue.color))
                        .modifier(CenteredFormLabel())
                })
            }

            // Backend
            App.shared.kit.map { kit in
                Section(header: Text("TUMexam Backend"),
                        footer: Text(L10n.Settings.backendInfo(
                                        App.shared.backendURL?.host ?? "unknown",
                                        "\(kit.userFullName ?? "unknown") (\(kit.userId ?? "unknown"))",
                                        App.shared.deviceIdentifier.uuidString)
                        )) {
                    // Exam Picker
                    if #available(iOS 14.0, *) {} else {
                        Picker("Exam", selection: self.$proxy.exam) {
                            ForEach(kit.examList.exams.value.map({ ExamWrapper(exam: $0) }), id: \.self) {
                                Text($0.name).tag($0 as ExamWrapper?)
                            }
                        }
                    }
                    // Disconnect
                    Button(action: {
                        BreadcrumbManager.addNavigation("Settings - reset alert")
                        self.showsDisconnectAlert = true
                    }, label: {
                        Text(L10n.Settings.disconnect)
                            .modifier(CenteredFormLabel())
                            .foregroundColor(.red)
                    })
                    // Disconnect Alert
                    .alert(isPresented: self.$showsDisconnectAlert) {
                        Alert(title: Text(L10n.Settings.DisconnectAlert.title),
                              message: Text(
                                L10n.Settings.DisconnectAlert.message(App.shared.backendURL?.host ??
                                "unknown"
                                )),
                              primaryButton: .cancel(),
                              secondaryButton: .destructive(Text(L10n.Settings.disconnect), action: {
                                BreadcrumbManager.addInteraction("Settings - reset confirmed")
                                App.shared.reset()
                        }))
                    }
                    // Unlock Zombies
                    Button(action: {
                        self.showsZombieLocksAlert = true
                    }, label: {
                        Text(L10n.Settings.unlockZombieLocks)
                            .modifier(CenteredFormLabel())
                            .foregroundColor(.red)
                    })
                    // Zombie Locks Alert
                    .alert(isPresented: self.$showsZombieLocksAlert) {
                        Alert(title: Text(L10n.Settings.UnlockZombieLocksAlert.title),
                              message: Text(L10n.Settings.UnlockZombieLocksAlert.message),
                              primaryButton: .cancel(),
                              secondaryButton: .destructive(Text(L10n.Settings.UnlockZombieLocksAlert.unlock),
                                                            action: {
                                BreadcrumbManager.addInteraction("Settings - free zombie locks tapped")
                                App.shared.currentExam?.freeAllLocksInBackground()
                        }))
                    }
                }
            }
            // General Settings
            Section(header: Text(L10n.Settings.Features.title)) {
                Toggle(isOn: self.$proxy.hideOverlays) {
                    Text(L10n.Settings.Features.hideOverlays)
                }
                Toggle(isOn: self.$proxy.allowModificationToAll) {
                    Text(L10n.Settings.Features.modificationsToAny)
                }
                Toggle(isOn: self.$proxy.retainPageSelection) {
                    Text(L10n.Settings.Features.retainPageSelection)
                }
                Toggle(isOn: self.$proxy.applePencilGesture) {
                    Text(L10n.Settings.Features.pencilTapGesture)
                }
                Toggle(isOn: self.$proxy.askBeforeDismiss) {
                    Text(L10n.Settings.Features.askBeforeDismiss)
                }
            }
            Section(header: Text(L10n.Settings.Experimental.sampleSolutionOverlay)) {
                HStack {
                    Image(systemName: "circle.lefthalf.fill")
                    Slider(value: self.$proxy.sampleSolutionOpacity, in: 0.0...1.0)
                    Image(systemName: "circle.fill")
                }
            }
            // Experimental Settings
            Section(header: Text(L10n.Settings.Experimental.title),
                    footer: Text(L10n.Settings.Experimental.footer)) {
                Toggle(isOn: self.$proxy.enhancedDarkMode) {
                    Text(L10n.Settings.Experimental.edm)
                }
            }
            // Cache
            Section(header: Text("Cache"), footer: Text(cacheSizeString)) {
                // Clear Cache
                Button(action: {
                    BreadcrumbManager.addInteraction("Settings - clear cache tapped")
                    guard let cache = App.shared.kit?.fileCache else { return }
                    cache.removeAll()
                    cache.flush()
                    self.cacheSizeString = SettingsView.cacheSizeString()
                }, label: {
                    Text(L10n.Settings.Cache.clear)
                        .foregroundColor(.init(Asset.tuMexamBlue.color))
                        .modifier(CenteredFormLabel())
                })
            }
            // Tutorial
            Section(header: Text(L10n.Settings.Tutorial.title)) {
                Button(action: {
                    App.shared.settings[HiddenSettings.didShowStatusTutorial] = false
                    App.shared.settings[HiddenSettings.didShowPagingCorrectionTutorial] = false
                    App.shared.settings[HiddenSettings.didShowSnippetTutorial] = false
                    self.showsTutorialAlert = true
                }, label: {
                    Text(L10n.Settings.Tutorial.reset)
                        .foregroundColor(.init(Asset.tuMexamBlue.color))
                        .modifier(CenteredFormLabel())
                })
                .alert(isPresented: $showsTutorialAlert, content: {
                    Alert(title: Text(L10n.Settings.Tutorial.Alert.title),
                          message: Text(L10n.Settings.Tutorial.Alert.message),
                          dismissButton: .default(Text(L10n.ok)))
                })
            }
            // Crash Settings
            Section(header: Text(L10n.Settings.crashInfoTitle), footer: Text(L10n.Settings.crashInfo)) {
                // Crash Info
                Toggle(isOn: self.$proxy.collectCrashInfo) {
                    Text(L10n.Settings.collectCrashInfo)
                }
            }
            // Licenses
            Section {
                Button {
                    UIApplication.shared.open(
                        URL(string: UIApplication.openSettingsURLString)!,
                        options: [:],
                        completionHandler: nil)
                } label: {
                    Text("Licenses")
                        .foregroundColor(.init(Asset.tuMexamBlue.color))
                        .modifier(CenteredFormLabel())
                }

            }

        }.onAppear(perform: {
            UITableView.appearance().contentInset.top = -35
        }).onDisappear(perform: {
            UITableView.appearance().contentInset.top = 0
        })

    }

    var headerView: some View {
        HStack {
            Image(uiImage: Asset.tuMexamIcon76.image)
                .cornerRadius(18)
                .padding([.vertical], 8)
            VStack(alignment: .leading, spacing: 5.0) {
                Text("TUMexam for iPad")
                    .fontWeight(.bold)
                Text(self.versionString)
                    .font(.subheadline)
                    .foregroundColor(.secondary)
                HStack {
                    Text("Made with ❤️ at")
                    Button("net.in.tum.de") {
                        UIApplication.shared.open(URL(string: "https://net.in.tum.de")!,
                        options: [:],
                        completionHandler: nil)
                    }
                    .foregroundColor(Color(Asset.tuMexamBlue.color))
                    .buttonStyle(PlainButtonStyle())
                }
            }
        }
    }

    private static func cacheSizeString() -> String {
        guard let cache = App.shared.kit?.fileCache else {
            return "Cache size unknown"
        }
        let byteFormatter = ByteCountFormatter()
        let current = byteFormatter.string(fromByteCount: Int64(cache.totalSize))
        let max = byteFormatter.string(fromByteCount: Int64(cache.sizeLimit))
        return L10n.Settings.Cache.size(current, max)
    }
}

struct SettingsView_Previews: PreviewProvider {
    static var previews: some View {
        // Approx. Popover size on iPad Pro 12,9
        let layout = PreviewLayout.fixed(width: 380, height: 900)
        return Group {
            SettingsView()
                .previewLayout(layout)
                .previewDisplayName("Light Mode")
            SettingsView()
                .environment(\.colorScheme, .dark)
                .previewLayout(layout)
                .previewDisplayName("Dark Mode")
        }
    }
}
