// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import SwiftUI
import TUMexamKit

struct LinearProgressView: View {

    let gradient: [Color]
    @State var progress: CGFloat

    func progressWidth(totalWidth: CGFloat) -> CGFloat {
        if progress == 0 {
            return 0
        } else {
            return max(totalWidth * 0.03, totalWidth * progress)
        }
    }

    var body: some View {
        ZStack {
            // Background
            GeometryReader { geometryReader in
                Capsule()
                    .fill(
                        gradient[0].opacity(0.15)
                    )
                    .frame(width: geometryReader.size.width)
                    .animation(.easeIn)

                Capsule()
                    .fill(
                        LinearGradient(gradient: Gradient(colors: gradient), startPoint: .leading, endPoint: .trailing)
                    )
                    .frame(width: progressWidth(totalWidth: geometryReader.size.width))
                    .shadow(color: Color.black.opacity(0.15), radius: 8, x: 0.0, y: 0.0)
                    .animation(.easeIn)
            }

        }

    }
}

struct SubproblemStatusView: View {

    var dismissHandler: (() -> Void)?
    @State var status: CorrectionProblemStatusType

    var body: some View {
        ScrollView(.vertical) {
            VStack(alignment: .leading) {
                ForEach(status.subproblems, id: \.subproblem) { sub in
                    Text(L10n.StatusController.Cell.subproblem(sub.subproblem))
                        .fontWeight(.bold)
                        .padding([.top, .leading, .trailing])
                    HStack(alignment: .center) {
                        VStack(alignment: .leading) {
                            ForEach(sub.passes, id: \.pass) { pass in
                                HStack {
                                    Text("\(L10n.StatusController.Cell.pass) \(pass.pass)")
                                    Divider()
                                    Text("\(pass.corrected) / \(pass.total)")
                                    Divider()
                                    Text("\(Int(pass.progress * 100)) %")
                                }.frame(height: 15)
                            }
                        }
                        // This is ugly, but it works. Research if there is a way to use max of all left colums for sizing all left cols
                        .frame(width: L10n.StatusController.Cell.pass.count > 4 ? 230 : 200, alignment: .leading)
                        VStack(alignment: .trailing) {
                            ForEach(sub.passes, id: \.pass) { pass in
                                LinearProgressView(
                                    gradient: passColor(pass.pass),
                                    progress: CGFloat(pass.progress)
                                )
                                .frame(height: 15)
                            }
                        }
                    }
                    .padding([.leading, .trailing, .bottom])
                    Divider()
                }
            }
        }
        .navigationBarTitle("Problem \(status.problem)", displayMode: .large)
        .navigationViewStyle(StackNavigationViewStyle())
        .navigationBarItems(trailing: Button(L10n.done) {
            dismissHandler?()
        })
    }

    private func passColor(_ pass: Int) -> [Color] {
        switch pass {
        case 1:
            return [
                Color(Asset.gradientRed1.color),
                Color(Asset.gradientRed2.color)
            ]
        case 2:
            return [
                Color(Asset.gradientGreen1.color),
                Color(Asset.gradientGreen2.color)
            ]
        default:
            return [
                Color(Asset.gradientPurple1.color),
                Color(Asset.gradientPurple2.color)
            ]
        }
    }
}

struct SubproblemStatusView_Previews: PreviewProvider {
    static var previews: some View {
        let layout = PreviewLayout.fixed(width: 540, height: 680)
        let view = SubproblemStatusView(
            status: MockCorrectionProblemStatus(
                problem: 1,
                passes: [
                    MockCorrectionPassStatus(
                        pass: 1,
                        corrected: 7,
                        total: 20,
                        subCorrected: 7,
                        subTotal: 20
                    ),
                    MockCorrectionPassStatus(
                        pass: 1,
                        corrected: 2,
                        total: 20,
                        subCorrected: 7,
                        subTotal: 20
                    )
                ],
                subproblems: [
                    MockCorrectionSubproblemStatus(
                        subproblem: "a",
                        passes: [
                            MockCorrectionSubproblemPassStatus(
                                pass: 1,
                                corrected: 10,
                                total: 20
                            ),
                            MockCorrectionSubproblemPassStatus(
                                pass: 2,
                                corrected: 7,
                                total: 20
                            )
                        ]
                    ),
                    MockCorrectionSubproblemStatus(
                        subproblem: "b",
                        passes: [
                            MockCorrectionSubproblemPassStatus(
                                pass: 1,
                                corrected: 7,
                                total: 20
                            ),
                            MockCorrectionSubproblemPassStatus(
                                pass: 2,
                                corrected: 2,
                                total: 20
                            )
                        ]
                    )
                ]
            )
        )
        return Group {
            view
        }.previewLayout(layout)
    }
}

// Mock types for preview

// swiftlint:disable private_over_fileprivate

fileprivate struct MockCorrectionProblemStatus: CorrectionProblemStatusType {
    var problem: Int
    var passes: [CorrectionPassStatusType]

    var subproblems: [CorrectionSubproblemStatusType]
}

fileprivate struct MockCorrectionSubproblemStatus: CorrectionSubproblemStatusType {
    var subproblem: String
    var passes: [CorrectionSubproblemPassStatusType]
}

fileprivate struct MockCorrectionPassStatus: CorrectionPassStatusType {
    var pass: Int
    var corrected: Int
    var total: Int
    var subCorrected: Int
    var subTotal: Int

    var progress: Double {
        return Double(corrected) / Double(total)
    }

    var progressRespectingSubproblems: Double {
        return Double(subCorrected) / Double(subTotal)
    }
}

fileprivate struct MockCorrectionSubproblemPassStatus: CorrectionSubproblemPassStatusType {
    var pass: Int
    var corrected: Int
    var total: Int

    var progress: Double {
        return Double(corrected) / Double(total)
    }
}

// swiftlint:enable private_over_fileprivate
