// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

class CutoutLayer: CALayer {
	var normalizedCutoutRect: CGRect = .zero {
		didSet {
			setNeedsLayout()
		}
	}
	let backgroundMask: CAShapeLayer = CAShapeLayer()

	override init() {
		super.init()

		setup()
	}

	override init(layer: Any) {
		if let other = layer as? CutoutLayer {
			normalizedCutoutRect = other.normalizedCutoutRect
		}
		super.init(layer: layer)

		setup()
	}

	func setup() {
		backgroundMask.fillColor = UIColor.black.cgColor
		self.backgroundColor = UIColor.black.withAlphaComponent(0.5).cgColor
		self.mask = backgroundMask
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	override func layoutSublayers() {
		super.layoutSublayers()

		CATransaction.begin()
		CATransaction.setDisableActions(true)

		let cutoutRect = normalizedCutoutRect.mappedTo(size: self.bounds.size)
		backgroundMask.frame = self.bounds

		let maskPath = UIBezierPath(rect: cutoutRect)
		maskPath.append(UIBezierPath(rect: self.bounds))
		backgroundMask.fillRule = .evenOdd
		backgroundMask.path = maskPath.cgPath

		CATransaction.commit()
	}
}

class DynamicRectangleHandle {
	weak var target: DynamicRectangle?

	init(target: DynamicRectangle) {
		self.target = target
	}

	func set(using point: CGPoint) {

	}
}

class DynamicRectangleReszieHandle: DynamicRectangleHandle {
	typealias TargetKeyPath = WritableKeyPath<DynamicRectangle, CGFloat>
	// swiftlint:disable identifier_name
	let xTarget: TargetKeyPath?
	let yTarget: TargetKeyPath?

	init(x: TargetKeyPath?, y: TargetKeyPath?, for target: DynamicRectangle) {
		self.xTarget = x
		self.yTarget = y
		super.init(target: target)
	}

	override func set(using point: CGPoint) {
		guard var target = target else { return }
		if let x = xTarget {
			target[keyPath: x] = point.x
		}
		if let y = yTarget {
			target[keyPath: y] = point.y
		}
	}

	var point: CGPoint {
		guard let target = target else { return .zero }
		var p: CGPoint = .zero
		if let x = xTarget {
			p.x = target[keyPath: x]
		} else {
			// mid x
			p.x = (target.x1 + target.x2) * 0.5
		}
		if let y = yTarget {
			p.y = target[keyPath: y]
		} else {
			// mid y
			p.y = (target.y1 + target.y2) * 0.5
		}
		return p
	}
}

class DynamicRectangleMoveHandle: DynamicRectangleHandle {
	let initialPoint: CGPoint
	let initialX1: CGFloat
	let initialX2: CGFloat
	let initialY1: CGFloat
	let initialY2: CGFloat

	init(initial: CGPoint, target: DynamicRectangle) {
		initialPoint = initial
		initialX1 = target.x1
		initialX2 = target.x2
		initialY1 = target.y1
		initialY2 = target.y2
		super.init(target: target)
	}

	override func set(using point: CGPoint) {
		guard let target = target else { return }

		let deltaX = point.x - initialPoint.x
		let deltaY = point.y - initialPoint.y
		target.x1 = initialX1 + deltaX
		target.x2 = initialX2 + deltaX
		target.y1 = initialY1 + deltaY
		target.y2 = initialY2 + deltaY
	}
}

extension CGRect {
	var horizontalRange: ClosedRange<CGFloat> {
		return minX...maxX
	}

	var verticalRange: ClosedRange<CGFloat> {
		return minY...maxY
	}
}

extension CGFloat {
	func clamp(to range: ClosedRange<CGFloat>) -> CGFloat {
		return Swift.min(Swift.max(range.lowerBound, self), range.upperBound)
	}

	func clampVertically(to rect: CGRect) -> CGFloat {
		return clamp(to: rect.verticalRange)
	}

	func clampVertically(to optionalRect: CGRect?) -> CGFloat {
		guard let rect = optionalRect else { return self }
		return clampVertically(to: rect)
	}

	func clampHorizontally(to rect: CGRect) -> CGFloat {
		return clamp(to: rect.horizontalRange)
	}

	func clampHorizontally(to optionalRect: CGRect?) -> CGFloat {
		guard let rect = optionalRect else { return self }
		return clampHorizontally(to: rect)
	}
}

class DynamicRectangle {
	// swiftlint:disable identifier_name
	var _x1: CGFloat = 0.0
	var _x2: CGFloat = 0.0
	var _y1: CGFloat = 0.0
	var _y2: CGFloat = 0.0

	var x1: CGFloat {
		get { _x1 }
		set { _x1 = newValue.clampHorizontally(to: clippingRect) }
	}
	var x2: CGFloat {
		get { _x2 }
		set { _x2 = newValue.clampHorizontally(to: clippingRect) }
	}
	var y1: CGFloat {
		get { _y1 }
		set { _y1 = newValue.clampVertically(to: clippingRect) }
	}
	var y2: CGFloat {
		get { _y2 }
		set { _y2 = newValue.clampVertically(to: clippingRect) }
	}

	var rect: CGRect {
		get {
			return CGRect(x: x1, y: y1, width: x2-x1, height: y2-y1).standardized
		}
		set {
			x1 = newValue.minX
			x2 = newValue.maxX
			y1 = newValue.minY
			y2 = newValue.maxY
		}
	}
	var clippingRect: CGRect?

	var handles: [DynamicRectangleReszieHandle] = []

	init() {
		handles = [
			DynamicRectangleReszieHandle(x: \.x1, y: \.y1, for: self),
			DynamicRectangleReszieHandle(x: \.x1, y: nil, for: self),
			DynamicRectangleReszieHandle(x: \.x1, y: \.y2, for: self),
			DynamicRectangleReszieHandle(x: nil, y: \.y2, for: self),
			DynamicRectangleReszieHandle(x: \.x2, y: \.y2, for: self),
			DynamicRectangleReszieHandle(x: \.x2, y: nil, for: self),
			DynamicRectangleReszieHandle(x: \.x2, y: \.y1, for: self),
			DynamicRectangleReszieHandle(x: nil, y: \.y1, for: self)
		]
	}

	func nearestHandle(for point: CGPoint) -> (CGFloat, DynamicRectangleHandle)? {
		var curDistance: CGFloat = CGFloat.greatestFiniteMagnitude
		var curHandle: DynamicRectangleReszieHandle?
		for handle in handles {
			let dist = point.distance(handle.point)
			if dist < curDistance {
				curDistance = dist
				curHandle = handle
			}
		}

		guard let handle = curHandle else { return nil }
		return (curDistance, handle)
	}

	func createMoveHandle(initialPoint point: CGPoint) -> DynamicRectangleHandle {
		return DynamicRectangleMoveHandle(initial: point, target: self)
	}

}

class RectangleSelectionView: BaseView {
	let cutoutLayer: CutoutLayer = CutoutLayer()
	let resizeHandlesLayer: ResizeHandlesLayer = ResizeHandlesLayer()
	var rectangle: DynamicRectangle? = nil {
		didSet {
			if oldValue !== rectangle {
				self.didChangeRectangle()
			}
		}
	}
	var currentHandle: DynamicRectangleHandle?

	var panGestureRecognizer: UIPanGestureRecognizer?
	var tapGestureRecognizer: UITapGestureRecognizer?
	var penOnly: Bool = false {
		didSet {
			panGestureRecognizer?.swiftAllowedTouchTypes = (penOnly ? [.pencil] : [.direct])
		}
	}

	var onTapInside: (() -> Void)?
	var onDidBegin: (() -> Void)?
	var onDidEnd: (() -> Void)?

	override func setup() {
		super.setup()

		layer.addSublayer(cutoutLayer)
		layer.addSublayer(resizeHandlesLayer)

		let pan = UIPanGestureRecognizer(target: self, action: #selector(didPan(_:)))
		self.addGestureRecognizer(pan)
		panGestureRecognizer = pan

		let tap = UITapGestureRecognizer(target: self, action: #selector(didTap(_:)))
		self.addGestureRecognizer(tap)
		tapGestureRecognizer = tap

		didChangeRectangle()
	}

	@objc func didTap(_ recognizer: UITapGestureRecognizer) {
		guard let rectangle = rectangle else { return }
		let location = recognizer.location(in: self)
		if rectangle.rect.contains(location) {
			// touch inside -> probably use this rect
			onTapInside?()
		} else {
			// touch outside -> exit rectangle mode
			self.rectangle = nil
			currentHandle = nil
			onDidEnd?()
		}
	}

	@objc func didPan(_ recognizer: UIPanGestureRecognizer) {
		let location = recognizer.location(in: self)
		switch recognizer.state {
		case .began:
			if let rectangle = rectangle {
				// already existing rectangle... check if gripping a handle...
				if let (dist, handle) = rectangle.nearestHandle(for: location), dist < 40.0 {
					// began near a handle -> use it as target
					currentHandle = handle
				} else if rectangle.rect.contains(location) {
					// began inside rect, but no handle was close enough -> move
					currentHandle = rectangle.createMoveHandle(initialPoint: location)
				}
			} else {
				// no rectangle, or outside the current one
				let r = DynamicRectangle()
				r.rect = CGRect(origin: location, size: .zero)
				rectangle = r
				currentHandle = r.handles.first
				onDidBegin?()
			}
		case .changed:
			if let handle = currentHandle {
				handle.set(using: location)
				setNeedsLayout()
			}
		case .ended:
			currentHandle = nil
		default:
			break
		}
	}

	override func layoutSubviews() {
		super.layoutSubviews()

		CATransaction.begin()
		CATransaction.setDisableActions(true)
		cutoutLayer.frame = self.bounds
		resizeHandlesLayer.frame = self.bounds

		if let rect = rectangle {
			rect.clippingRect = self.bounds
			cutoutLayer.normalizedCutoutRect = rect.rect.normalizedUsing(size: bounds.size)
			resizeHandlesLayer.normalizedTargetRect = rect.rect.normalizedUsing(size: bounds.size)
		}

		CATransaction.commit()
	}

	private func didChangeRectangle() {
		if rectangle == nil {
			tapGestureRecognizer?.isEnabled = false
			cutoutLayer.isHidden = true
			resizeHandlesLayer.isHidden = true
		} else {
			tapGestureRecognizer?.isEnabled = true
			cutoutLayer.isHidden = false
			resizeHandlesLayer.isHidden = false
		}
	}
}
