// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import UIKit

private class HandleLayer: CAShapeLayer {
	let rectangleAnchor: CGPoint

	init(_ rectangleAnchor: CGPoint = .zero) {
		self.rectangleAnchor = rectangleAnchor

		super.init()
		setup()
	}

	override init(layer: Any) {
		if let other = layer as? HandleLayer {
			rectangleAnchor = other.rectangleAnchor
		} else {
			rectangleAnchor = .zero
		}
		super.init(layer: layer)

		setup()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	func setup() {
		bounds = CGRect(x: 0, y: 0, width: 14, height: 14)
		path = UIBezierPath(ovalIn: bounds).cgPath
		fillColor = UIColor.white.cgColor
		strokeColor = UIColor.systemBlue.cgColor
		lineWidth = 1.0
	}
}

class ResizeHandlesLayer: CALayer {
	var normalizedTargetRect: CGRect = .zero {
		didSet {
			setNeedsLayout()
		}
	}

	let outlineLayer: CAShapeLayer = CAShapeLayer()
	fileprivate let handles: [HandleLayer] = [
		HandleLayer(CGPoint(x: 0.0, y: 0.0)),
		HandleLayer(CGPoint(x: 0.5, y: 0.0)),
		HandleLayer(CGPoint(x: 1.0, y: 0.0)),
		HandleLayer(CGPoint(x: 1.0, y: 0.5)),
		HandleLayer(CGPoint(x: 1.0, y: 1.0)),
		HandleLayer(CGPoint(x: 0.5, y: 1.0)),
		HandleLayer(CGPoint(x: 0.0, y: 1.0)),
		HandleLayer(CGPoint(x: 0.0, y: 0.5))
	]

	override init() {
		super.init()

		setup()
	}

	override init(layer: Any) {
		if let other = layer as? ResizeHandlesLayer {
			normalizedTargetRect = other.normalizedTargetRect
		} else {
			fatalError("couldn't copy")
		}
		super.init(layer: layer)

		setup()
	}

	required init?(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}

	func setup() {
		outlineLayer.strokeColor = UIColor.systemBlue.cgColor
		outlineLayer.lineWidth = 1.0
		outlineLayer.fillColor = nil
		self.addSublayer(outlineLayer)

		for handle in handles {
			self.addSublayer(handle)
		}
	}

	override func layoutSublayers() {
		super.layoutSublayers()

		CATransaction.begin()
		CATransaction.setDisableActions(true)

		let targetRect = normalizedTargetRect.mappedTo(size: self.bounds.size)

		outlineLayer.frame = targetRect
		outlineLayer.path = UIBezierPath(rect: outlineLayer.bounds).cgPath

		for handle in handles {
			var pos = CGPoint()
			pos.x = targetRect.origin.x + (handle.rectangleAnchor.x * targetRect.size.width)
			pos.y = targetRect.origin.y + (handle.rectangleAnchor.y * targetRect.size.height)
			handle.position = pos
		}

		CATransaction.commit()
	}
}
