// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import CoreGraphics

//swiftlint:disable identifier_name

extension CGPoint {
    /// Point in Size -> Point in Normal
    /// - Parameter size: size where the point is in
    /// - Returns: point in normal
    public func normalized(using size: CGSize) -> CGPoint {
        return CGPoint(x: x / size.width, y: y / size.height)
    }

    /// Point in Normal -> Point in Size
    /// - Parameter size: size where the point should be in
    /// - Returns: point in size
    public func scaled(to size: CGSize) -> CGPoint {
        return CGPoint(x: x * size.width, y: y * size.height)
    }

    /// Point in Normal -> Point in Sub-Normal
    /// - Parameter rect: sub-rect
    /// - Returns: point in sub-normal
    public func convertTo(normalizedSubRect rect: CGRect) -> CGPoint {
        return CGPoint(
            x: (x - rect.origin.x) / rect.size.width,
            y: (y - rect.origin.y) / rect.size.height
        )
    }

    /// Point in Sub-Normal -> Point in Normal
    /// - Parameter rect: sub-rect
    /// - Returns: point in normal
    public func convertFrom(normalizedSubRect rect: CGRect) -> CGPoint {
        return CGPoint(
            x: (x * rect.size.width) + rect.origin.x,
            y: (y * rect.size.height) + rect.origin.y
        )
    }

    public func midpointTowards(other: CGPoint) -> CGPoint {
        CGPoint(x: (self.x + other.x) / 2, y: (self.y + other.y) / 2)
    }
}

extension CGRect {
    init(center: CGPoint, size: CGSize) {
        self.init(x: center.x - size.width / 2,
                  y: center.y - size.height / 2,
                  width: size.width,
                  height: size.height)
    }
}

extension CGRect {
    /// Rect in Size -> Rect in Normal
    /// - Parameter s: size
    /// - Returns: rect in normal
    public func normalizedUsing(size s: CGSize) -> CGRect {
        return CGRect(
            x: origin.x / s.width,
            y: origin.y / s.height,
            width: size.width / s.width,
            height: size.height / s.height
        )
    }

    /// Rect in Page -> Rect in Normal (y-axis mirrored)
    /// - Parameter s: page size
    /// - Returns: rect in normal
    public func normalizedFromPageUsing(size s: CGSize) -> CGRect {
        return CGRect(
            x: origin.x / s.width,
            y: (s.height - origin.y) / s.height,
            width: size.width / s.width,
            height: size.height / s.height
        )
    }

    /// Rect in Rect -> Rect in normal
    /// - Parameter r: the current rect
    /// - Returns: the rect in a normalized rect
    public func normalizedIn(rect r: CGRect) -> CGRect {
        return CGRect(
            x: (origin.x - r.origin.x) / r.size.width,
            y: (origin.y - r.origin.y) / r.size.height,
            width: size.width / r.size.width,
            height: size.height / r.size.height
        )
    }

    /// Rect in Normal -> Rect in sub-normal
    /// - Parameter rect: the sub-rect
    /// - Returns: rect in sub-normal
    public func convertTo(normalizedSubRect rect: CGRect) -> CGRect {
        return CGRect(
            x: (origin.x - rect.origin.x) / rect.size.width,
            y: (origin.y - rect.origin.y) / rect.size.height,
            width: width / rect.size.width,
            height: height / rect.size.height
        )
    }

    /// Rect in Sub-Normal -> Rect in Normal
    /// - Parameter rect: sub-rect
    /// - Returns: rect in normal
    public func convertFrom(normalizedSubRect rect: CGRect) -> CGRect {
        return CGRect(
            x: (origin.x * rect.size.width) + rect.origin.x,
            y: (origin.y * rect.size.height) + rect.origin.y,
            width: width * rect.size.width,
            height: height * rect.size.height)
    }

    /// Rect in Normal -> Rect in Size
    /// - Parameter s: the size
    /// - Returns: rect in size
    public func mappedTo(size s: CGSize) -> CGRect {
        return CGRect(
            x: origin.x * s.width,
            y: origin.y * s.height,
            width: size.width * s.width,
            height: size.height * s.height
        )
    }

    /// Assumes receiver to be a Rect in Normal, and returns size if extrapolated to size
    /// - Parameter size: size
    /// - Returns: size of rect in size
    public func originalSize(knowing size: CGSize) -> CGSize {
        CGSize(width: size.width / self.width, height: size.height / self.height)
    }

    public func centeredRect(size: CGSize) -> CGRect {
        CGRect(x: self.minX + (self.width / 2) - (size.width / 2),
               y: self.minY + (self.height / 2) - (size.height / 2),
               width: size.width,
               height: size.height)
    }
}

#if os(iOS)
import UIKit

extension CGColor {
    class var clear: CGColor {
        UIColor.clear.cgColor
    }
}
#endif

extension CGColor {
    func withAlphaComponent(_ alpha: CGFloat) -> CGColor {
        guard let new = copy(alpha: alpha) else {
            fatalError("couldn't create a copy from color")
        }
        return new
    }
}

extension CGSize {
    var rotate90: CGSize {
        return CGSize(width: self.height, height: self.width)
    }

    public func normalizedIn(size: CGSize) -> CGSize {
        CGSize(width: self.width / size.width,
               height: self.height / size.height)
    }
}

// taken from https://stackoverflow.com/a/55906075/1065801
extension CGImage {
    // MARK: - CoreGraphics

    /// Resize image from given size.
    ///
    /// - Parameter newSize: Size of the image output.
    /// - Returns: Resized image.
    func resize(to newSize: CGSize) -> CGImage? {
        guard let colorSpace = self.colorSpace else { return nil }

        let width = Int(newSize.width)
        let height = Int(newSize.height)

        guard let context = CGContext(data: nil, width: width, height: height,
                                      bitsPerComponent: bitsPerComponent,
                                      bytesPerRow: bytesPerRow, space: colorSpace,
                                      bitmapInfo: bitmapInfo.rawValue) else { return nil }
        context.interpolationQuality = .high
        let rect = CGRect(origin: CGPoint.zero, size: newSize)
        context.draw(self, in: rect)

        return context.makeImage()
    }
}
