// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import CoreGraphics

extension CorrectionProblemModel {
    func groupBoxesByPass() -> [Int: [CorrectionBoxModel]] {
        return boxes.reduce(into: [Int: [CorrectionBoxModel]]()) { (result, box) in
            var current = result[box.correctionPass] ?? []
            current.append(box)
            result[box.correctionPass] = current
        }
    }

    func creditForPass(_ pass: Int) -> Double? {
        let passCredits = credits.first { $0.pass == pass }
        return passCredits?.credit
    }
}

extension CorrectionDataModel {
    var maxPageDimension: CGFloat {
        CGFloat(max(pageWidth, pageHeight))
    }
    var normalizedPageSize: CGSize {
        let maxDimension = maxPageDimension
        return CGSize(width: CGFloat(pageWidth) / maxDimension, height: CGFloat(pageHeight) / maxDimension)
    }
}

extension CorrectionBoxModel {
    func normalizedFrame(using correction: CorrectionDataModel) -> CGRect {
        let pageWidth: CGFloat
        let pageHeight: CGFloat
        if let page = correction.pages[page], page.landscape {
            pageWidth = CGFloat(correction.pageHeight)
            pageHeight = CGFloat(correction.pageWidth)
        } else {
            pageWidth = CGFloat(correction.pageWidth)
            pageHeight = CGFloat(correction.pageHeight)
        }

        let boxWidth: CGFloat = CGFloat(correction.boxWidth) / pageWidth
        let boxHeight: CGFloat = CGFloat(correction.boxWidth) / pageHeight

        let boxCenter = CGPoint(x: CGFloat(xPos) / pageWidth, y: (pageHeight - CGFloat(yPos)) / pageHeight)

        return CGRect(center: boxCenter, size: CGSize(width: boxWidth, height: boxHeight))
    }
}

extension CorrectionEmptyBoxModel {
    func normalizedFrame(using correction: CorrectionDataModel) -> CGRect {
        let pageWidth: CGFloat
        let pageHeight: CGFloat
        if let page = correction.pages[page], page.landscape {
            pageWidth = CGFloat(correction.pageHeight)
            pageHeight = CGFloat(correction.pageWidth)
        } else {
            pageWidth = CGFloat(correction.pageWidth)
            pageHeight = CGFloat(correction.pageHeight)
        }

        let boxWidth: CGFloat = CGFloat(correction.boxWidth) / pageWidth
        let boxHeight: CGFloat = CGFloat(correction.boxWidth) / pageHeight

        let boxCenter = CGPoint(x: CGFloat(xPos) / pageWidth, y: (pageHeight - CGFloat(yPos)) / pageHeight)

        return CGRect(center: boxCenter, size: CGSize(width: boxWidth, height: boxHeight))
    }
}

extension ProblemIdentifier {
    init(from model: CorrectionProblemModel) {
        if model.subproblem.isEmpty {
            self = .problem(model.problem)
        } else {
            self = .subproblem(model.problem, model.subproblem)
        }
    }

    var backendFormat: String {
        if let subproblem = self.subproblem {
            return "\(problem)\(subproblem)"
        } else {
            return "\(problem)"
        }
    }
}
