// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

// Taken from https://stackoverflow.com/a/28016692/1043495

extension DateFormatter {
    static let iso8601: DateFormatter = {
        let formatter = DateFormatter()
        formatter.locale = Locale(identifier: "en_US")
        formatter.calendar = Calendar(identifier: .iso8601)
        formatter.dateFormat = "yyyy-MM-dd'T'HH:mm:ss.SSSXXXXX"
        return formatter
    }()

    static let iso8601Fractional: DateFormatter = {
        let formatter = DateFormatter()
        formatter.locale = Locale(identifier: "en_US")
        formatter.calendar = Calendar(identifier: .iso8601)
        formatter.dateFormat = "yyyy-MM-dd'T'HH:mm:ss.SSSSSS"
        return formatter
    }()

    static let iso8601NoFractional: DateFormatter = {
        let formatter = DateFormatter()
        formatter.locale = Locale(identifier: "en_US")
        formatter.calendar = Calendar(identifier: .iso8601)
        formatter.dateFormat = "yyyy-MM-dd'T'HH:mm:ss"
        return formatter
    }()

    static let httpHeaderSpec: DateFormatter = {
        let formatter = DateFormatter()
        formatter.locale = Locale(identifier: "en_US")
        formatter.calendar = Calendar(identifier: .iso8601)
        formatter.dateFormat = "EEE, dd MMM yyyy HH:mm:ss zzz"
        return formatter
    }()

}

extension ISO8601DateFormatter {
    static let fractionalFormatter: ISO8601DateFormatter = {
        let fmt = ISO8601DateFormatter()
        fmt.formatOptions = [.withFractionalSeconds]
        return fmt
    }()
}

extension JSONDecoder.DateDecodingStrategy {
    static let iso8601withConditionalFractionalSeconds = custom {
        let container = try $0.singleValueContainer()
        let string = try container.decode(String.self)
        if string.contains(".") {
            // Has fractional seconds
            guard let date = DateFormatter.iso8601Fractional.date(from: string) else {
                throw DecodingError.dataCorruptedError(in: container,
                                                       debugDescription: "Invalid date: " + string)
            }
            return date
        } else {
            // Has no fractional seconds
            guard let date = DateFormatter.iso8601NoFractional.date(from: string) else {
                throw DecodingError.dataCorruptedError(in: container,
                                                       debugDescription: "Invalid date: " + string)
            }
            return date
        }

    }

    static let httpHeaderSpec = custom {
        let container = try $0.singleValueContainer()
        let string = try container.decode(String.self)
        guard let date = DateFormatter.httpHeaderSpec.date(from: string) else {
            throw DecodingError.dataCorruptedError(in: container,
                                                   debugDescription: "Invalid date: " + string)
        }
        return date
    }
}

extension JSONEncoder.DateEncodingStrategy {
    static let iso8601withFractionalSeconds = custom {
        var container = $1.singleValueContainer()
        try container.encode(DateFormatter.iso8601.string(from: $0))
    }

    static let httpHeaderSpec = custom {
        var container = $1.singleValueContainer()
        try container.encode(DateFormatter.httpHeaderSpec.string(from: $0))
    }
}
