// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

//swiftlint:disable identifier_name

// based on https://www.swiftbysundell.com/articles/the-power-of-key-paths-in-swift/
extension Sequence {
    /// returns a sorted sequence by key path
    /// - Parameter keyPath: the key path which is be used for sorting (has to be comparable)
    /// - Returns: the sorted elements
    public func sorted<T: Comparable>(by keyPath: KeyPath<Element, T>) -> [Element] {
        return sorted { a, b in
            return a[keyPath: keyPath] < b[keyPath: keyPath]
        }
    }

    /// returns the smallest element in the sequence compared by the given key path
    /// - Parameter keyPath: the key path
    /// - Returns: the smallest element (or none)
    public func min<T: Comparable>(by keyPath: KeyPath<Element, T>) -> Element? {
        return self.min { a, b in
            return a[keyPath: keyPath] < b[keyPath: keyPath]
        }
    }
}

extension Sequence {
    /// look up whether the sequence contains an element by it's keypath (which has to be a bool)
    /// - Parameter keyPath: the key path
    /// - Returns: whether an element returns true for this key path
    public func contains(_ keyPath: KeyPath<Element, Bool>) -> Bool {
        contains { $0[keyPath: keyPath] }
    }

    /// look up whether the sequence contains an element matching a value by it's keypath (which has to be a bool)
    /// - Parameters:
    ///   - matching: the value which the key path's value should be equal to
    ///   - keyPath: the key path
    /// - Returns: whether an element matches the value for this key path
    public func contains<T: Equatable>(_ matching: T, for keyPath: KeyPath<Element, T>) -> Bool {
        contains { $0[keyPath: keyPath] == matching }
    }

    /// filters the sequence by matching a key path to a value
    /// - Parameters:
    ///   - matching: the value which the key path's value should be equal to
    ///   - keyPath: the key path
    /// - Returns: the filtered sequence
    public func filter<T: Equatable>(_ matching: T, for keyPath: KeyPath<Element, T>) -> [Element] {
        filter { $0[keyPath: keyPath] == matching }
    }

    /// returns the first element matching a specific value for a key path
    /// - Parameters:
    ///   - matching: the value which the key path's value should be equal to
    ///   - keyPath: the key path
    /// - Returns: the element, if any
    public func first<T: Equatable>(_ matching: T, for keyPath: KeyPath<Element, T>) -> Element? {
        first { $0[keyPath: keyPath] == matching }
    }

    /// returns whether all elements match a specific value for a key path
    /// - Parameters:
    ///   - matching: the value which the key path's value should be equal to
    ///   - keyPath: the key path
    /// - Returns: true, if all elements satisify the condition
    public func all<T: Equatable>(_ matching: T, for keyPath: KeyPath<Element, T>) -> Bool {
        allSatisfy { $0[keyPath: keyPath] == matching }
    }
}

// from https://www.hackingwithswift.com/example-code/language/how-to-split-an-array-into-chunks
extension Array {
    /// split the array into subarrays with equal counts
    /// - Parameter size: the count of each subarray
    /// - Returns: the chunks
    public func chunked(into size: Int) -> [[Element]] {
        return stride(from: 0, to: count, by: size).map {
            Array(self[$0 ..< Swift.min($0 + size, count)])
        }
    }
}
