// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import KeychainAccess
import Nuke
import Combine
import UIKit

/**
 The kit is the main class required for communicating with a TUMexam backend.

 It coordinates between multiple internal components which are required to fetch and update corrections.
 Authentication is done using the `ShibbolethAuthentication` class.
 */
public class Kit {
    // MARK: Public Properties

    /// the url of the backend
    public let backendURL: URL

    /// the device identifier (used for locking corrections)
    public let uuid: UUID

    /// the user id (LRZ id) of the currently logged in user
    public var userId: String? { return networking.userId }
    /// the full name of the currently signed in user (for display purposes only)
    public var userFullName: String? { return networking.userFullName }

    /// The AccessRights object for the current user
    public internal(set) lazy var accessRights: AccessRightsType = object(for: AccessRightsIdentifier()) as AccessRights

    /// The Metadata object for the currently connected TUMexam instance
    public internal(set) lazy var metadata: MetadataType = object(for: MetadataIdentifier()) as Metadata

    /// the list of exams available on this backend
    public internal(set) lazy var examList: ExamListType = object(for: ExamListIdentifier()) as ExamList

    /// the controller handling authentication using shibboleth
    public internal(set) lazy var shibbolethAuthentication = ShibbolethAuthentication(kit: self)

    /// provides information about possible updates (e.g. the newest version)
    public internal(set) lazy var updateInfo: UpdateInfoType = object(for: UpdateInfoIdentifier()) as UpdateInfo

    /// provides information about required updates (e.g. the newest version)
    // swiftlint:disable:next line_length
    public internal(set) lazy var updateEnforcement: UpdateEnforcementType = object(for: UpdateEnforcementIdentifier()) as UpdateEnforcement

    /// the lock manager, which handles and observes requested locks
    public var locks: LockManagerType { lockManager }

    // MARK: Internal Properties

    /// The internal networking object handling all network requests
    lazy var networking: Networking = Networking(kit: self)
    // TODO: Implement custom DataCache and remove nuke dep again
    /// On Disk Data cache for image data (compressed)
    public lazy var fileCache: DataCache? = {
        let cache = try? DataCache(name: "de.tum.in.www1.tumexam.datacache")
        cache?.sizeLimit = 500 * 1000 * 1000
        return cache
    }()

    /// Lock Manager object for creating and freeing locks on exams
    lazy var lockManager: LockManager = LockManager(kit: self)

    /// A reference to the Keychain object for storing authorization tokens
    var keychain: Keychain?

    /// the internal queue which handles background work
    let queue: DispatchQueue

    // MARK: Private Properties

    /// some objects should only exist once in the whole kit. this dictionary keeps references to those objects (see `object(for:)`)
    private var objects: [String: Any] = [:]

    /// For objects created asynchronously (e.g. when making a web request is needed beforehand)
    private var creationObservers = [Any]()

    /// create a new kit for the specified backend
    /// - Parameters:
    ///   - backendURL: the base URL of the backend
    ///   - useKeychain: whether to store credentials and cookies in the keychain
    ///   - uuid: a uuid to use when acquiring locks for corrections (if nil, a UUID will be generated)
    public init(backendURL: URL, useKeychain: Bool = true, uuid: UUID? = nil) {
        self.backendURL = backendURL
        self.uuid = uuid ?? UUID()

        queue = DispatchQueue(label: "TUMexamKit", qos: .userInitiated)

        // assign self to subcomponents
        networking.kit = self

        if useKeychain {
            setupKeychain()
        }
    }

    /// dumps the referenced objects to the console
    public func dumpObjects() {
        print("********  Objects Dump  ********")
        let ordered = objects.keys.sorted()
        for key in ordered {
            guard let object = objects[key] else { continue }
            if let object = object as? ShortDescription {
                print("\(key): \(object.shortDescription)")
            } else {
                print("\(key): \(object)")
            }
        }
        print("********************************")
    }

    /// manually sets auth info for the backend
    /// - Parameter callbackInfo: the auth info to set
    func setAuthenticationInfo(_ callbackInfo: ShibbolethAuthentication.CallbackInfo) {
        networking.setAuthenticationInfo(callbackInfo)
        if let keychain = keychain {
            keychain.store(authInfo: callbackInfo)
        }
    }

    /// returns the exam with the specified identifier
    /// - Parameter identifier: the exam's identifier
    /// - Returns: the exam
    public func exam(withId identifier: ExamIdentifier) -> ExamType {
        return object(for: identifier) as Exam
    }

    /// Clears all caches of the kit
    public func clearCache() {
        fileCache?.removeAll()
    }

    /// Custom deinitializer to kill networking session properly
    deinit {
        // This is needed for in order for URLSession to free the reference to its delegate
        networking.invalidateSession()
    }
}

extension Kit {
    /// returns the object with the specified identifier (and type) or creates a new one if no object exists
    /// - Parameter identifier: the identifier
    /// - Returns: the object
    func object<T: KitChildObjectType>(for identifier: T.IdentifierType) -> T {
        // try fetching the object from our mapping
        let typedIdentifier = T.typedIdentifier(identifier)
        let obj = objects[typedIdentifier]

        if obj == nil {
            // create a new object
            let newObj = T(identifier: identifier, kit: self)
            // insert into mapping and return the new object
            objects[typedIdentifier] = newObj
            return newObj
        } else {
            // check the type
            guard let obj = obj as? T else {
                fatalError("object of wrong type in mapping")
            }
            return obj
        }
    }

    /// Creates a Kit object that depends on an asyncronous network call before creation, a `Creatable` Kit object
    /// - Returns: a publisher eventually returning the created object
    func createObject<T: Creatable>(
        withCreationSubmission submission: T.CreationSubmissionType
    ) -> AnyPublisher<T, Error> {
        let resource = T.creationResource(submission: submission, in: self)
        let observer = ModelResourceObserver(resource: resource)
        let publisher = observer.publisher.tryMap { [unowned self] value -> T in
            self.creationObservers.removeAll(where: {
                guard let obs = $0 as? ModelResourceObserver<T.ModelType, Resource<T.ModelType>>  else {
                    return false
                }
                return observer === obs
            })
            let identifier = T.identifier(forCreationResult: value, resource: resource)
            let typedIdentifier = T.typedIdentifier(identifier)
            let obj = self.objects[typedIdentifier]

            guard obj == nil else {
                guard let obj = obj as? T else {
                    fatalError("object of wrong type in mapping")
                }
                return obj
            }

            let newObj = T(identifier: identifier, kit: self)
            newObj.update(using: value)
            objects[typedIdentifier] = newObj
            return newObj
        }

        resource.load()
        creationObservers.append(observer)
        return publisher.eraseToAnyPublisher()
    }

    /// Replaces a kit child object with another kit child object of the same type
    func replaceObject<T: KitChildObjectType>(with identifier: T.IdentifierType, object: T) {
        let typedIdentifier = T.typedIdentifier(identifier)
        guard let old = objects[typedIdentifier] else {
            return
        }
        guard old is T else {
            fatalError("object of wrong type in mapping")
        }
        objects[typedIdentifier] = object
    }

    /// Called by the Shibboleth authentication once we have properly authorized with the backend
    func authorizationGranted() {
        for object in objects.values {
            if let obj = object as? Loadable {
                obj.load()
            }
        }
    }
}
