// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import KeychainAccess

internal enum KeychainKeys: String, CaseIterable {
    case userId
    case token
    case tokenExpirationDate
    case refreshToken // We don't have this yet
    case csrfToken
}

extension Keychain {
    static let `default`: Keychain = Keychain(service: Bundle.main.bundleIdentifier!)
    // limit possible keys to enum cases
    internal subscript(_ key: KeychainKeys) -> String? {
        get {
            return self[key.rawValue]
        }
        set {
            self[key.rawValue] = newValue
        }
    }

    func store(cookies: [HTTPCookie]) {
        KeychainCookie.store(cookies: cookies, to: self)
    }

    func store(authInfo callbackInfo: ShibbolethAuthentication.CallbackInfo) {
        self[.userId] = callbackInfo.userId
        self[.token] = callbackInfo.token
        self[.tokenExpirationDate] = callbackInfo.tokenExpirationDate
    }

    func load(usingDomain domain: String) -> [HTTPCookie] {
        return KeychainCookie.load(from: self, domain: domain)
    }

    func loadAuthInfo() -> ShibbolethAuthentication.CallbackInfo? {
        guard let userId = self[.userId],
            let token = self[.token],
            let tokenExpirationDate = self[.tokenExpirationDate] else {
                return nil
        }
        return ShibbolethAuthentication.CallbackInfo(userId: userId,
                                                     token: token,
                                                     tokenExpirationDate: tokenExpirationDate)
    }
}
