// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import KeychainAccess

struct KeychainCookie {
    typealias Key = KeychainKeys
    typealias NameAcceptFunction = (String) -> Bool
    enum Name {
        case fixed(String)
        case dynamic(Key, NameAcceptFunction)
    }

    var name: Name
    var valueKey: Key
    var fixedDomain: String?
    var path: String

    init(_ name: Name, valueKey: Key, domain: String? = nil, path: String = "/") {
        self.name = name
        self.valueKey = valueKey
        self.fixedDomain = domain
        self.path = path
    }

    init(_ name: String, valueKey: Key, domain: String? = nil, path: String = "/") {
        self.init(.fixed(name), valueKey: valueKey, domain: domain, path: path)
    }
    init(_ name: Key, accepting: @escaping NameAcceptFunction, valueKey: Key, domain: String? = nil, path: String = "/") {
        self.init(.dynamic(name, accepting), valueKey: valueKey, domain: domain, path: path)
    }

    //	static let session = KeychainCookie("session", valueKey: .session)
    //	static let proxySession = KeychainCookie("proxysession", valueKey: .proxySessionCookie)
    static let csrfToken = KeychainCookie("X-CSRF-Token", valueKey: .csrfToken)
    //	static let jsessionId = KeychainCookie("JSESSIONID", valueKey: .jsessionIdCookie, domain: "tumidp.lrz.de")
    //	static let shibIDPSession =
    //		KeychainCookie("shib_idp_session", valueKey: .shibIDPSessionCookie, domain: "tumidp.lrz.de", path: "/idp")
    //	static let shibSession = KeychainCookie(
    //		.shibsessionCookieName,
    //		accepting: { $0.starts(with: "_shibsession_") }, // specify which cookie names to accept
    //		valueKey: .shibsessionCookieValue,
    //		domain: "hq.tumexam.de"
    //	)

    static let allCookies: [KeychainCookie] = [
        //		.session,
        //		.proxySession,
        .csrfToken
        //		.jsessionId,
        //		.shibIDPSession,
        //		.shibSession
    ]
}

extension KeychainCookie {
    private func cookieName(from keychain: Keychain) -> String? {
        switch name {
        case let .fixed(string):
            return string
        case let .dynamic(key, _):
            return keychain[key]
        }
    }

    func load(from keychain: Keychain, domain: String) -> HTTPCookie? {
        guard let cookieName = cookieName(from: keychain),
            let cookieValue = keychain[valueKey] else { return nil }

        return HTTPCookie(properties: [
            .name: cookieName as NSString,
            .domain: (fixedDomain ?? domain) as NSString,
            .value: cookieValue as NSString,
            .path: path as NSString
        ])
    }

    static func load(from keychain: Keychain, domain: String) -> [HTTPCookie] {
        return allCookies.compactMap { $0.load(from: keychain, domain: domain) }
    }

    func matches(cookieName: String) -> Bool {
        switch name {
        case let .fixed(fixedName):
            return cookieName == fixedName
        case let .dynamic(_, matcher):
            return matcher(cookieName)
        }
    }

    func store(cookie: HTTPCookie, to keychain: Keychain) {
        if case let .dynamic(key, _) = name {
            keychain[key] = cookie.name
        }
        keychain[valueKey] = cookie.value
    }

    static func store(cookies: [HTTPCookie], to keychain: Keychain) {
        for cookie in cookies {
            if let keyCookie = allCookies.first(where: { $0.matches(cookieName: cookie.name) }) {
                keyCookie.store(cookie: cookie, to: keychain)
            }
        }
    }
}
