// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

/// Enum representing an authentication error
public enum ShibbolethAuthenticationFailureReason {
    /// Used when the session was canceled by the user
    case canceled
    /// Used when the user is not authorized to acces the instace or the instance doesnt exist
    case unauthorized
    /// Used when the instance is under maintenance (503 Service unavailable)
    case maintenance
    /// Used when an unknown error occurs
    case unknown
}

/// the shibboleth authentication delegate which should create and present the necessary views
public protocol ShibbolethAuthenticationDelegate: AnyObject {
    /// called when a request failed due to invalid authentication
    /// - Parameters:
    ///   - auth: the shibboleth authentication manager
    ///   - url: the url which should be presented to the user
    ///   - refreshExisting: flag indicating if an existing session is to be renewed
    func shibboleth(_ auth: ShibbolethAuthentication, requiresAuthenticationUsingURL url: URL, refreshExisting: Bool)

    /// called when the authentication was sucessful, and the presented UI can be dismissed
    /// - Parameter auth: the shibboleth authentication manager
    func shibbolethDidAuthenticate(_ auth: ShibbolethAuthentication)

    /// called when the authentication failed (e.g. if the instance doesnt exist or the user has no access)
    func shibbolethAuthenticationFailed(_ auth: ShibbolethAuthentication, reason: ShibbolethAuthenticationFailureReason)
}

/// manages the authentication with a shibboleth backend (without having any UI itself)
public class ShibbolethAuthentication {

    public struct TUMexamJWTClaims: Decodable {
        var uid: String
        var shortname: String
        var firstname: String
        var lastname: String
    }

    /// struct containing the userid/token
    public struct CallbackInfo: CustomDebugStringConvertible {
        /// the user id
        let userId: String
        /// the token
        let token: String
        /// when the token will expire
        let tokenExpirationDate: String
        // let refreshToken: String

        var claims: TUMexamJWTClaims? {
            guard let jwtPayload = token.split(separator: ".")[safeAccess: 1] else { return nil }
            let payloadString = String(jwtPayload)
            // Taken from https://github.com/Kitura/Swift-JWT/blob/master/Sources/SwiftJWT/Data%2BBase64URLEncoded.swift
            // Need to add padding to payload base64
            let paddingLength = 4 - payloadString.count % 4
            let padding = (paddingLength < 4) ? String(repeating: "=", count: paddingLength) : ""
            let paddedPayload = payloadString
                .replacingOccurrences(of: "-", with: "+")
                .replacingOccurrences(of: "_", with: "/")
                + padding
            guard let payloadData = Data(base64Encoded: paddedPayload) else { return nil }
            return autoreleasepool {
                let decoder = JSONDecoder()
                let claims = try? decoder.decode(TUMexamJWTClaims.self, from: payloadData)
                return claims
            }
        }

        public var debugDescription: String {
            let shortname = claims?.shortname ?? "nil"
            let expDouble = TimeInterval(tokenExpirationDate) ?? .zero
            let exp = Date(timeIntervalSince1970: expDouble)
            return "<\(userId), \(shortname), \(exp)>"
        }
    }

    /// the delegate which needs to be set in order to have a working authentication flow
    public weak var delegate: ShibbolethAuthenticationDelegate?
    unowned var kit: Kit

    init(kit: Kit) {
        self.kit = kit
    }

    internal func start(withURL url: URL) {
        // without delegate -> immediately cancel
        guard let delegate = delegate else {
            Logger.shibboleth.error("attempted to start authentication without a delegate")
            cancel(url: url)
            return
        }
        DispatchQueue.main.async { [unowned self] in
            let refresh = self.kit.networking.userId != nil
            delegate.shibboleth(self, requiresAuthenticationUsingURL: url, refreshExisting: refresh)
        }
    }

    /// cancel the current login process
    /// - Parameters:
    ///   - url: the url
    ///   - byUser: whether the user cancelled
    public func cancel(url: URL, byUser: Bool = false) {
        if byUser, self.kit.networking.userId != nil {
            // Only prompt again if we had a session, and the user cancelled
            DispatchQueue.main.asyncAfter(deadline: DispatchTime.now().advanced(by: .milliseconds(750))) {
                // Delay prompt by 1 second, as topmost view controller might still be from auth session
                self.start(withURL: url)
            }
        } else {
            failure(url: url, reason: .canceled)
        }
    }

    /// Call this to indicate that the shibboleth auth failed
    public func failure(url: URL, reason: ShibbolethAuthenticationFailureReason) {
        DispatchQueue.main.async {
            self.delegate?.shibbolethAuthenticationFailed(self, reason: reason)
        }
    }

    /// has to be called with the final cookies from the login session, so the kit can persist them
    /// - Parameter cookies: the cookies
    /// - Returns: currently always true (additional checks might be introduced in the future)
    public func authorizeUsing(userId: String, token: String, tokenExpiration: String) -> Bool {
        let callbackInfo = CallbackInfo(userId: userId, token: token, tokenExpirationDate: tokenExpiration)
        Logger.shibboleth.info("sucessfully authorized using new auth info \(callbackInfo.debugDescription)")
        kit.setAuthenticationInfo(callbackInfo)

        delegate?.shibbolethDidAuthenticate(self)
        kit.authorizationGranted()

        return true
    }
}
