// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

struct ExamModel: Codable, Equatable, CustomStringConvertible {
    var examId: Int
    var name: String
    var date: Date?

    var graded: Bool?
    var groupsAllowed: Bool?
    var halfCredits: Bool?
    var locked: Bool?

    // Available from WS2020 onward
    var texOptions: TeXOptionsModel?

    private enum CodingKeys: String, CodingKey {
        case name, date, graded, locked
        case examId = "exam_id"
        case groupsAllowed = "groups_allowed"
        case halfCredits = "halfcredits"
        case texOptions = "texoptions"
    }

    var description: String {
        return name
    }
}

extension ExamModel: Comparable {
    static func < (lhs: ExamModel, rhs: ExamModel) -> Bool {
        return lhs.examId < rhs.examId
    }
}

struct TeXOptionsModel: Codable, Equatable {
    var registrationSticker: Bool
    var externalProblems: Bool
    var coversheet: Bool
    var staple: Int
    var serif: Bool
    var subproblemArabic: Bool
    var noProblemCredits: Bool
    var noProblem: Bool
    var noTitle: Bool
    var noCorrectorBox: Bool
    var singleSide: Bool

    private enum CodingKeys: String, CodingKey {
        case coversheet, staple, serif
        case registrationSticker = "registrationsticker"
        case externalProblems = "externalproblems"
        case subproblemArabic = "subproblemarabic"
        case noProblemCredits = "noproblemcredits"
        case noProblem = "noproblem"
        case noTitle = "notitle"
        case noCorrectorBox = "nocorrectionbox"
        case singleSide = "singleside"
    }
}

struct ExamDetailsModel: Codable {
    var examCount: Int

    var problemsCount: Int
    var subproblemsCount: Int
    var totalCreditsCount: Int

    private enum CodingKeys: String, CodingKey {
        case examCount = "exams"
        case problemsCount = "problems"
        case subproblemsCount = "subproblems"
        case totalCreditsCount = "total_credits"
    }
}

struct ExamCorrectionsResultModel: Codable, Hashable {
    let currentPage: Int
    let maxPage: Int
    let total: Int
    let results: [ExamCorrectionModel]

    private enum CodingKeys: String, CodingKey {
        case results, total
        case currentPage = "current_page"
        case maxPage = "max_page"
    }
}

struct ExamCorrectionModel: Codable, Hashable {
    var access: Bool
    var available: Bool
    var locked: Bool
    var erid: String
    var corrected: Bool
    var correctionPass: Int
    var problem: Int
    var subproblem: String?
    var date: Date
    var flagTypeId: [Int]

    private enum CodingKeys: String, CodingKey {
        case access, available, erid, corrected, problem, subproblem, date, locked
        case correctionPass = "correction_pass"
        case flagTypeId = "flag_type_id"
    }
}

struct ExamCorrectionResultModel: Codable, Hashable {
    var erid: String

    var corrector: String?
    var comment: String?
    var credit: Double?

    var date: Date

    var correctionPass: Int
    var problem: String?
    var subproblem: String?

    private enum CodingKeys: String, CodingKey {
        case erid, corrector, comment, credit, date, problem, subproblem
        case correctionPass = "correction_pass"
    }
}

struct PaginationModel<ResultType: Codable & Hashable>: Codable, Hashable {
    let currentPage: Int
    let maxPage: Int
    let total: Int
    let results: [ResultType]

    private enum CodingKeys: String, CodingKey {
        case results, total
        case currentPage = "current_page"
        case maxPage = "max_page"
    }
}
