// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

/// This code is supposed to capsule network requests in NSOperations so that we have more control over how many network reqeusts are performed in parallel
/// This code never made it to production, because it cause crashes and memory leaks and I couldn't pin down why :(
class NetworkOperation: Operation {
    enum State: String {
        case ready = "Ready"
        case executing = "Executing"
        case finished = "Finished"
        case cancelled = "Cancelled"
        fileprivate var keyPath: String { return "is" + self.rawValue }
    }

    let resource: BaseResource & BaseNetworkResource
    let task: URLSessionTask

    private let stateQueue = DispatchQueue(label: "NetworkOperationStateQueue", attributes: .concurrent)
    var state: State {
        get {
            stateQueue.sync {
                return stateStore
            }
        }
        set {
            let oldValue = state
            willChangeValue(forKey: state.keyPath)
            willChangeValue(forKey: newValue.keyPath)
            stateQueue.sync(flags: .barrier) {
                stateStore = newValue
            }
            didChangeValue(forKey: state.keyPath)
            didChangeValue(forKey: oldValue.keyPath)
        }
    }
    private var stateStore: State = .ready

    override var isAsynchronous: Bool { true }
    override var isExecuting: Bool { state == .executing }
    override var isFinished: Bool { state == .finished }
    override var isCancelled: Bool { state == .cancelled }

    init?(_ resource: BaseResource & BaseNetworkResource, session: URLSession, networking: Networking) {
        self.resource = resource
        guard let task = resource.createTask(session: session, networking: networking) else {
            return nil
        }
        self.task = task
        super.init()

        self.queuePriority = resource.priority
        Logger.network.info("[NetworkOperation] created operation for resource \(resource)")
    }

    override func start() {
        guard !isCancelled else {
            return
        }
        task.resume()
        state = .executing
        Logger.network.info("[NetworkOperation] started operation for resource \(resource)")
    }

    override func cancel() {
        task.cancel()
        state = .cancelled
        Logger.network.info("[NetworkOperation] cancelled operation for resource \(resource)")
    }

    func taskDidCompleteWithError(error: Error?) {
        resource.task(task, didCompleteWithError: error)
        state = .finished
        Logger.network.info("[NetworkOperation] finished operation for resource \(resource)")
    }
}
