// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import Foundation
import Logging
#if DEBUG
import PulseCore
#endif

/// The class responsible for handling all network related tasks, such as performing HTTP requests or reacting to HTTP errors
class Networking: NSObject {
    /// backreference to the Kit which instantiated this object
    unowned var kit: Kit?
    /// The global URLSession object
    private var session: URLSession = URLSession(configuration: .default)

    /// Object holding the authorization token obtained via shibboleth
    private var authInfo: ShibbolethAuthentication.CallbackInfo?
    /// The base url of the connected TUMexam instance
    private var baseURL: URL
    /// The API URL of the connected TUMexam instance (usually `baseURL` + `/api/`)
    private var apiURL: URL
    /// Currently inflight URLSessionTasks
    private var currentTasks: [URLSessionTask: BaseResource] = [:]

    /// userId of the currently signed in user (LRZ id)
    var userId: String? { return authInfo?.userId }
    /// full name of the currently signed in user
    var userFullName: String? {
        guard let firstName = authInfo?.claims?.firstname,
              let lastName = authInfo?.claims?.lastname else {
            return nil
        }
        return firstName + " " + lastName
    }

    /// Initializer, supposed to be called from the Kit object only
    init(kit: Kit) {
        self.kit = kit
        baseURL = kit.backendURL
        apiURL = kit.backendURL.appendingPathComponent("api", isDirectory: true)

        super.init()

        let config = URLSessionConfiguration.default
        config.httpCookieAcceptPolicy = .always
        config.httpCookieStorage?.cookieAcceptPolicy = .always
        config.httpAdditionalHeaders = ["User-Agent": Networking.userAgent()]
        #if DEBUG
        // For Pulse network debugging screen
        let proxy = URLSessionProxyDelegate(delegate: self)
        session = URLSession(configuration: config, delegate: proxy, delegateQueue: nil)
        #else
        session = URLSession(configuration: config, delegate: self, delegateQueue: nil)
        #endif
        Logger.network.info("setup networking api client as \(Networking.userAgent())")
    }

    /// Loads the given resource on a background thread
    /// - Parameter resource: a resource to load
    func load(resource: BaseResource & BaseNetworkResource) {
        session.delegateQueue.addOperation { [weak self] in
            self?.reallyLoad(resource: resource)
        }
    }

    /// Creates a data task for the resource and loads it using the URLSession
    /// - Parameter resource: the resource object to load
    func reallyLoad(resource: BaseResource & BaseNetworkResource) {
        guard !currentTasks.values.contains(resource) else {
            fatalError("logic error: trying to load the same resource twice")
        }

        guard let task = resource.createTask(session: session, networking: self) else {
            Logger.network.error("could not create task for resource \(resource)")
            return
        }
        if let url = task.currentRequest?.url {
            Logger.network.debug("loading \(task.currentRequest?.httpMethod ?? "GET") \(url)")
        }
        currentTasks[task] = resource
        task.resume()
    }

    /// Updates the authInfo stored property
    ///
    /// Usually called from ShibbolethAuthentication
    /// - Parameter callbackInfo: the new authentication info
    func setAuthenticationInfo(_ callbackInfo: ShibbolethAuthentication.CallbackInfo) {
        self.authInfo = callbackInfo
    }

    /// Generates a URL object for the given API path
    /// - Parameter path: api path
    /// - Returns: a URL object with the given path
    func url(forPath path: String) -> URL {
        return apiURL.appendingPathComponent(path)
    }

    /// Creates a URL from the base url and the given path
    /// - Parameter path: path to append to baseURL
    /// - Returns: a URL object with the given path
    func urlAgainstBase(forPath path: String) -> URL {
        return baseURL.appendingPathComponent(path)
    }

    /// Creates a URLRequest againt the API url and adds authentication Headers
    /// - Parameter path: the api path
    /// - Returns: an authenticated URLRequest object
    func protectedRequest(forPath path: String) -> URLRequest? {
        let url = self.url(forPath: path)
        return protectedRequest(forURL: url)
    }

    /// Creates an authorized URLRequest from a URL object
    /// - Parameter url: the URL to call
    /// - Returns: a URLRequest which is has authorization headers
    func protectedRequest(forURL url: URL) -> URLRequest? {
        let req = URLRequest(url: url)
        return protectedRequest(forRequest: req)
    }

    /// Adds authorization headers to an existing URLRequest
    /// - Parameter request: a URLRequest object
    /// - Returns: a copy of the URLRequest with authorization headers
    func protectedRequest(forRequest request: URLRequest) -> URLRequest? {
        var req = request
        // We need to add the Authorization headers ourselves here
        // If we don't have an AuthInfo object, the request will most likely fail
        guard let authInfo = self.authInfo,
              let exp = Double(authInfo.tokenExpirationDate),
              Date(timeIntervalSince1970: exp) > Date() else {

            // Our AuthInfo object is invalid
            // Perform Shibboleth authentication
            if let url = req.url {
                Logger.shibboleth.info("saved authinfo \(self.authInfo != nil ? self.authInfo!.debugDescription : "") is invalid, starting shibboleth authentication")
                kit?.shibbolethAuthentication.start(withURL: url)
            }
            Logger.network.info("Cannot build protected request for req \(req.httpMethod ?? "GET") \(req.url?.absoluteString ?? "nil")")
            return nil
        }

        req.addValue("Bearer \(authInfo.token)", forHTTPHeaderField: "Authorization")

        return req
    }

    /// Looks up the resource for the given task in the `currentTasks` dict, casts it to a given targetType and executes the given block with the typecast object as input (if cast succeeded)
    /// - Parameter task: URLSessionTask object to be cast
    /// - Parameter targetType: Type to cast `task` to
    /// - Parameter block: Closure to execute with cast object
    func perform<T>(for task: URLSessionTask, as targetType: T.Type, block: ((T) -> Void)) {
        guard let operation = currentTasks[task] else {
            Logger.network.error("no operation found for task \(task)")
            return
        }

        guard let cast = operation as? T else {
            Logger.network.error("could not cast operation \(operation) to \(String(describing: T.self))")
            return
        }

        block(cast)
    }

    /// Invalidates the URLSession object in order to free the delegate reference
    ///
    /// See [the discussion in the docs](https://developer.apple.com/documentation/foundation/nsurlsession/1411530-delegate) for more info
    func invalidateSession() {
        session.invalidateAndCancel()
        Logger.network.info("invalidated urlsession of shared networking singleton")
    }
}

extension Networking: URLSessionDelegate {
    func urlSession(_ session: URLSession, task: URLSessionTask, didCompleteWithError error: Error?) {
        if let error = error {
            Logger.network.error("task failed with error \(error)")
        } else {
            Logger.network.debug("finished loading \(task.currentRequest?.httpMethod ?? "GET") \(task.currentRequest?.url?.absoluteString ?? "nil")")
        }

        perform(for: task, as: BaseNetworkResource.self) { (resource) in
            resource.task(task, didCompleteWithError: error)
        }

        defer {
            currentTasks[task] = nil
        }

        // There is no URLProtocol for HTTP Bearer Authorization (https://developer.apple.com/forums/thread/92429)
        // Thus, we wont be notified of an authentication challenge by URLSession
        // Check for 401 Unauthorized response here
        guard let response = task.response as? HTTPURLResponse else { return }

        if response.statusCode == 401, let url = response.url {
            Logger.network.info("got 401 unauthorized response")

            // Check if we should have been able to access based on our auth info
            if let authInfo = self.authInfo,
               let exp = Double(authInfo.tokenExpirationDate),
               Date(timeIntervalSince1970: exp) > Date() {
                // We probably don't have access rights for this instance
                Logger.shibboleth.error("token \(authInfo) should be valid, we are probably unauthorized to access instance at \(response.url?.host ?? "nil")")
                kit?.shibbolethAuthentication.failure(url: url, reason: .unauthorized)
                return
            }

            // Our session probably expired
            // Let's try getting a new token via shibboleth
            Logger.shibboleth.info("session has expired, stating shibboleth authentication")
            kit?.shibbolethAuthentication.start(withURL: url)
        }

        if response.statusCode == 503, let url = response.url {
            kit?.shibbolethAuthentication.failure(url: url, reason: .maintenance)
            Logger.network.info("got 503, instance is under maintenance")
        }
    }

    func urlSession(_ session: URLSession,
                    didReceive challenge: URLAuthenticationChallenge,
                    completionHandler: @escaping (URLSession.AuthChallengeDisposition, URLCredential?) -> Void) {
        completionHandler(.performDefaultHandling, nil)
    }
}

extension Networking: URLSessionTaskDelegate {
    func urlSession(_ session: URLSession,
                    task: URLSessionTask,
                    willPerformHTTPRedirection response: HTTPURLResponse,
                    newRequest request: URLRequest,
                    completionHandler: @escaping (URLRequest?) -> Void) {
        guard let original = task.originalRequest else {
            completionHandler(request)
            return
        }

        if original.value(forHTTPHeaderField: "Authorization") != nil {
            // Add Authorization header for new redirected request
            let protected = protectedRequest(forRequest: request)
            completionHandler(protected)
        } else {
            completionHandler(request)
        }
    }
}

extension Networking: URLSessionDataDelegate {
    func urlSession(_ session: URLSession, dataTask: URLSessionDataTask, didReceive data: Data) {
        perform(for: dataTask, as: BaseNetworkDataResource.self) { (resource) in
            // Notify resource
            resource.task(dataTask, didReceive: data)
        }
    }

    func urlSession(_ session: URLSession, dataTask: URLSessionDataTask,
                    didReceive response: URLResponse,
                    completionHandler: @escaping (URLSession.ResponseDisposition) -> Void) {
        let expectedContentSize = Int64(response.expectedContentLength)
        perform(for: dataTask, as: BaseNetworkDataResource.self) { resource in
            // Update resource
            resource.task(dataTask, didUpdateExpectedSizeTo: expectedContentSize)
        }
        completionHandler(.allow)
    }
}

extension Networking: URLSessionDownloadDelegate {
    func urlSession(_ session: URLSession, downloadTask: URLSessionDownloadTask, didFinishDownloadingTo location: URL) {
        perform(for: downloadTask, as: BaseDownloadResource.self) { (resource) in
            // Notify the resource that the task finished downloading
            resource.task(downloadTask, didFinishDownloadingTo: location)
        }
    }

    func urlSession(_ session: URLSession,
                    downloadTask: URLSessionDownloadTask,
                    didWriteData bytesWritten: Int64,
                    totalBytesWritten: Int64,
                    totalBytesExpectedToWrite: Int64) {
        perform(for: downloadTask, as: BaseDownloadResource.self) { resource in
            // Notify the resource of download progress update
            resource.task(downloadTask, bytesWritten: totalBytesWritten, of: totalBytesExpectedToWrite)
        }
    }
}

extension Networking {
    /// Determines the version of the install CFNetwork stack
    /// - Returns: the version fo the CFNetwork framework
    static func cfNetworkVersion() -> String? {
        // https://developer.apple.com/forums/thread/124183
        guard
            let bundle = Bundle(identifier: "com.apple.CFNetwork"),
            let versionAny = bundle.infoDictionary?[kCFBundleVersionKey as String],
            let version = versionAny as? String
            else { return nil }
        return version
    }

    /// Function for determining the darwin kernel version
    /// - Returns: the darwin kernel version string
    static func darwinVersion() -> String {
        // https://stackoverflow.com/questions/41350077/cocoa-get-darwin-version-at-runtime
        var systemInfo = utsname()
        uname(&systemInfo)
        let machineMirror = Mirror(reflecting: systemInfo.release)
        let darvinVersionString = machineMirror.children.reduce("") { identifier, element in
            guard let value = element.value as? Int8,
                value != 0 else {
                    return identifier
            }

            return identifier + String(UnicodeScalar(UInt8(value)))
        }
        return darvinVersionString
    }

    /// Generates the user agent string for HTTP requests
    /// - Returns: the User Agent string for the app
    static func userAgent() -> String {
        let bundlePlist = Bundle.main.infoDictionary
        let version = bundlePlist?["CFBundleShortVersionString"] ?? "(unknown version)"
        let buildId = (bundlePlist?[kCFBundleVersionKey as String] as? String?)
        let buildIdSanitized = buildId??.replacingOccurrences(of: "/", with: "-") ?? "(unknown build)"
        let bundleId = bundlePlist?[kCFBundleIdentifierKey as String] ?? "(unknown bundle id)"
        #if os(iOS)
        let osVersion = "iOS \(ProcessInfo.processInfo.operatingSystemVersionString)"
        #elseif os(macOS)
        let osVersion = "macOS \(ProcessInfo.processInfo.operatingSystemVersionString)"
        #endif
        let cfnetworkVersion = Networking.cfNetworkVersion() ?? "(unknown version)"
        let darwinVersion = Networking.darwinVersion()
        let agent = "TUMexam/\(version) (\(bundleId); build: \(buildIdSanitized); \(osVersion)) CFNetwork/\(cfnetworkVersion) Darwin/\(darwinVersion)"
        return agent
    }
}
