// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

protocol BaseDownloadResource {
    func task(_ task: URLSessionDownloadTask, didFinishDownloadingTo url: URL)
    func task(_ task: URLSessionDownloadTask, bytesWritten written: Int64, of total: Int64)
}

class DownloadResource<T>: NetworkResource<T> {
    override var useCache: Bool { true }
    var filename: String { fatalError("override in subclass") }
    var cachedData: Data?

    var bytesLoaded: Int64?
    var bytesTotal: Int64?
    var progress: Double? {
        return Double(bytesLoaded ?? 0) / Double(bytesTotal ?? 0)
    }

    override func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        let path = self.path
        guard let req = networking.protectedRequest(forPath: path) else {
            self.fail(with: NetworkError.unauthorized)
            return nil
        }
        let task = session.downloadTask(with: req)

        return task
    }

    override func load() {
        if self.state == .initial, useCache {
            if loadFromCache() {
                // did load from cache
                finished()
                return
            }
        }
        // disabled cache, or cache miss
        guard self.state == .initial || self.state == .error else { return }

        super.load()
    }

    @discardableResult
    func loadFromCache() -> Bool {
        guard let cache = kit?.fileCache else { return false }
        guard let data = cache.cachedData(for: self.filename) else { return false }
        return (setValue(using: data) != nil)
    }

    override func taskDidComplete(_ task: URLSessionTask) {
        if let cachedData = self.cachedData {
            setValue(using: cachedData)
            self.cachedData = nil
            finished()
        } else if useCache, let kit = kit {
            kit.queue.async { [weak self] in
                self?.loadFromCache()
                self?.finished()
            }
        }
    }
}

extension DownloadResource: BaseDownloadResource {
    func task(_ task: URLSessionDownloadTask, didFinishDownloadingTo url: URL) {
        if useCache, let kit = kit {
            // use cache -> let the cache move the file, then load it
            guard let data = try? Data(contentsOf: url) else {
                return
            }
            kit.fileCache?.storeData(data, for: self.filename)
        } else {
            // NO cache -> read file and discard
            // TODO: it would be better to open() on this thread, and read on another
            guard let data = try? Data(contentsOf: url) else { return }
            cachedData = data
        }
    }

    func task(_ task: URLSessionDownloadTask, bytesWritten written: Int64, of total: Int64) {
        self.bytesLoaded = written
        self.bytesTotal = total
    }
}

class ImageDownloadResource: DownloadResource<Data> {
    override func createValue(from data: Data) -> Data? {
        return data
    }

    func createImage() -> Image? {
        guard let data = self.value else { return nil }
        return Image(data: data)
    }
}
