// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import os.log

class ExamCorrectionsResource: PaginatedModelResource<ExamCorrectionModel>, IdentifiableResource {
    let identifier: ExamIdentifier
    var filter: ResultFilter?

    required init(_ identifier: ExamIdentifier, in kit: Kit) {
        self.identifier = identifier
        super.init(in: kit)
    }

    override func createQueryItems() -> [URLQueryItem] {
        guard let filter = filter else { return super.createQueryItems() }

        var items = super.createQueryItems()
        items.append(contentsOf: filter.queryItems)

        return items
    }

    override var path: String {
        "exam/\(identifier)/correction/"
    }
}

fileprivate extension FilterBoolValue {
    func queryItem(withName name: String) -> URLQueryItem? {
        switch self {
        case .yes:
            return URLQueryItem(name: name, value: "True")
        case .no:
            return URLQueryItem(name: name, value: "False")
        default:
            return nil
        }
    }
}

fileprivate extension ResultSortByColumn {
    var backendName: String {
        switch self {
        case .erid:
            return "erid"
        case .problem:
            return "problem"
        case .pass:
            return "pass"
        case .date:
            return "date"
        }
    }

    func backendNameWithDirection(_ direction: FilterSortDirection) -> String {
        let sortDir = direction == .ascending ? "asc" : "desc"
        return "\(backendName)(\(sortDir))"
    }
}

extension ResultFilter {
    var queryItems: [URLQueryItem] {
        var items = [URLQueryItem]()
        if let problem = problem {
            items.append(URLQueryItem(name: "filter[problem]", value: "\(problem)"))
        }
        if let subproblem = subproblem {
            items.append(URLQueryItem(name: "filter[subproblem]", value: subproblem))
        }
        if let pass = pass {
            items.append(URLQueryItem(name: "filter[correction_pass]", value: "\(pass)"))
        }
        if let erid = erid {
            let cutoff = eridCutoff.applied(to: erid)
            items.append(URLQueryItem(name: "filter[erid]", value: "\(cutoff)"))
        }
        if let corrected = corrected.queryItem(withName: "filter[corrected]") {
            items.append(corrected)
        }
        if let available = available.queryItem(withName: "filter[available]") {
            items.append(available)
        }
        if let locked = locked.queryItem(withName: "filter[locked]") {
            items.append(locked)
        }
        if let access = access.queryItem(withName: "filter[access]") {
            items.append(access)
        }
        if !flagTypeIds.isEmpty {
            items.append(
                URLQueryItem(
                    name: "filter[flag_type_id]",
                    value: flagTypeIds.map { String($0) }.joined(separator: "|")
                )
            )
        }

        items.append(URLQueryItem(name: "sort", value: sortColum.backendNameWithDirection(sortDirection)))

        return items
    }
}
