// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

class FlagDefinitionsResource: PaginatedModelResource<FlagDefinitionModel> {
    var exam: ExamIdentifier

    override var path: String {
        return "exam/\(exam)/flag/definitions"
    }

    required init(exam: ExamIdentifier, in kit: Kit) {
        self.exam = exam
        super.init(in: kit)
    }

    override func createQueryItems() -> [URLQueryItem] {
        var items = super.createQueryItems()
        items.append(URLQueryItem(name: "sort", value: "count(desc),corrector(asc)"))
        return items
    }
}

class CreateFlagDefinitionResource: ModelResource<FlagDefinitionModel> {
    var exam: ExamIdentifier
    var flagDefinitionSubmissionModel: FlagDefinitionSubmissionModel

    override var path: String {
        return "exam/\(exam)/flag/definition"
    }

    required init(exam: ExamIdentifier, flagDefinition: FlagDefinitionSubmissionModel, in kit: Kit) {
        self.exam = exam
        self.flagDefinitionSubmissionModel = flagDefinition
        super.init(in: kit)
    }

    override func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        let path = self.path

        guard var request = networking.protectedRequest(forPath: path) else {
            self.fail(with: NetworkError.unauthorized)
            return nil
        }

        request.httpMethod = "POST"

        let encoder = JSONEncoder()
        let data = try? encoder.encode(flagDefinitionSubmissionModel)
        request.httpBody = data

        request.setValue("application/json", forHTTPHeaderField: "Content-Type")

        return session.dataTask(with: request)
    }

}

class UpdateFlagDefinitionResource: ModelResource<FlagDefinitionModel> {
    var identifier: FlagDefinitionIdentifier
    var flagDefinitionSubmissionModel: FlagDefinitionSubmissionModel

    override var path: String {
        return "exam/\(identifier.examId)/flag/definition/\(identifier.flagTypeId)"
    }

    required init(identifier: FlagDefinitionIdentifier, model: FlagDefinitionSubmissionModel, in kit: Kit) {
        self.identifier = identifier
        self.flagDefinitionSubmissionModel = model
        super.init(in: kit)
    }

    override func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        let path = self.path

        guard var request = networking.protectedRequest(forPath: path) else {
            self.fail(with: NetworkError.unauthorized)
            return nil
        }

        request.httpMethod = "PUT"

        let encoder = JSONEncoder()
        let data = try? encoder.encode(flagDefinitionSubmissionModel)
        request.httpBody = data

        request.setValue("application/json", forHTTPHeaderField: "Content-Type")

        return session.dataTask(with: request)
    }
}

class DeleteFlagDefinitionResource: NetworkDataResource<Void> {
    var identifier: FlagDefinitionIdentifier

    override var path: String {
        return "exam/\(identifier.examId)/flag/definition/\(identifier.flagTypeId)"
    }

    required init(identifier: FlagDefinitionIdentifier, in kit: Kit) {
        self.identifier = identifier
        super.init(in: kit)
    }

    override func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        let path = self.path

        guard var request = networking.protectedRequest(forPath: path) else {
            self.fail(with: NetworkError.unauthorized)
            return nil
        }

        request.httpMethod = "DELETE"

        return session.dataTask(with: request)
    }
}
