// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

public struct FlagFilter {
    let erid: String?
    let problem: Int?
    let subproblem: String?
    let correctionPass: Int?

    public init(erid: String? = nil, problem: Int? = nil, subproblem: String? = nil, correctionPass: Int? = nil) {
        self.erid = erid
        self.problem = problem
        self.subproblem = subproblem
        self.correctionPass = correctionPass
    }
}

class FlagsResource: PaginatedModelResource<FlagModel> {
    var exam: ExamIdentifier
    let filter: FlagFilter?

    override var path: String {
        return "exam/\(exam)/flag/flags"
    }

    required init(exam: ExamIdentifier, filter: FlagFilter? = nil, in kit: Kit) {
        self.exam = exam
        self.filter = filter
        super.init(in: kit)
    }

    override func createQueryItems() -> [URLQueryItem] {
        var items = super.createQueryItems()

        guard let filter = self.filter else {
            return items
        }

        if let erid = filter.erid {
            items += [URLQueryItem(name: "filter[erid]", value: erid)]
        }
        if let problem = filter.problem {
            items += [URLQueryItem(name: "filter[problem]", value: String(problem))]
        }
        if let subproblem = filter.subproblem {
            items += [URLQueryItem(name: "filter[subproblem]", value: subproblem)]
        }
        if let correctionPass = filter.correctionPass {
            items += [URLQueryItem(name: "filter[correction_pass]", value: String(correctionPass))]
        }

        return items
    }
}

class CreateFlagResource: ModelResource<FlagModel> {
    var exam: ExamIdentifier
    var model: FlagSubmissionModel

    override var path: String {
        return "exam/\(exam)/flag/flag"
    }

    required init(exam: ExamIdentifier, model: FlagSubmissionModel, in kit: Kit) {
        self.exam = exam
        self.model = model
        super.init(in: kit)
    }

    override func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        let path = self.path

        guard var request = networking.protectedRequest(forPath: path) else {
            self.fail(with: NetworkError.unauthorized)
            return nil
        }

        request.httpMethod = "POST"

        let encoder = JSONEncoder()
        let data = try? encoder.encode(model)
        request.httpBody = data

        request.setValue("application/json", forHTTPHeaderField: "Content-Type")

        return session.dataTask(with: request)
    }
}

class DeleteFlagResource: NetworkResource<Void> {
    var flag: FlagIdentifier

    override var path: String {
        return "exam/\(flag.examId)/flag/flag/\(flag.flagId)"
    }

    required init(flag: FlagIdentifier, in kit: Kit) {
        self.flag = flag
        super.init(in: kit)
    }

    override func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        let path = self.path

        guard var request = networking.protectedRequest(forPath: path) else {
            self.fail(with: NetworkError.unauthorized)
            return nil
        }

        request.httpMethod = "DELETE"

        return session.dataTask(with: request)
    }
}
