// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

struct TokenModel: Codable {
    let token: String
}

class AcquireLockResource: ModelResource<TokenModel> {
    struct PostBody: Codable {
        var token: String
        var pass: Int?
        var problem: Int?
        var subproblem: String?
        var refresh: Bool?
    }

    let lockIdentifier: LockIdentifier
    var refresh: Bool = false

    init(_ lockIdentifier: LockIdentifier, in kit: Kit) {
        self.lockIdentifier = lockIdentifier
        super.init(in: kit)
    }

    override func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        let path = self.path

        // create post request
        guard var request = networking.protectedRequest(forPath: path) else {
            self.fail(with: NetworkError.unauthorized)
            return nil
        }
        if refresh {
            request.httpMethod = "PUT"
        } else {
            request.httpMethod = "POST"
        }

        // add body
        if let kit = kit {
            let body = PostBody.init(token: kit.uuid, identifier: lockIdentifier, refresh: (refresh ? true : nil))
            request.httpBody = try? JSONEncoder().encode(body)
            request.setValue("application/json", forHTTPHeaderField: "Content-Type")
            // somehow, this request requires more authorization...
        }

        return session.dataTask(with: request)
    }

    override var path: String {
        "exam/\(lockIdentifier.exam)/lock/\(lockIdentifier.erid)"
    }
}

extension AcquireLockResource.PostBody {
    enum CodingKeys: String, CodingKey {
        case token
        case pass = "correction_pass"
        case problem
        case subproblem
    }

    init(token: UUID, identifier: LockIdentifier, refresh: Bool? = nil) {
        self.init(
            token: token.uuidString,
            pass: identifier.pass,
            problem: identifier.problem,
            subproblem: identifier.subproblem,
            refresh: refresh)
    }
}
