// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

public struct LockFilter {

    public enum SortColumn {
        case erid
        case user
        case problem
        case subproblem
        case pass
        case created
        case lastUpdated

        public var fieldId: String {
            switch self {
            case .erid:
                return "erid"
            case .user:
                return "user"
            case .problem:
                return "problem"
            case .subproblem:
                return "subproblem"
            case .pass:
                return "pass"
            case .created:
                return "created"
            case .lastUpdated:
                return "last_updated"
            }
        }

        func withDirection(_ direction: FilterSortDirection) -> String {
            return "\(fieldId)(\(direction.urlRepresentation))"
        }
    }

    let erid: String?
    let user: String?
    let problem: Int?
    let subproblem: String?
    let pass: Int?

    public init(erid: String? = nil,
                user: String? = nil,
                problem: Int? = nil,
                subproblem: String? = nil,
                pass: Int? = nil) {
        self.erid = erid
        self.user = user
        self.problem = problem
        self.subproblem = subproblem
        self.pass = pass
    }

    /// column to sort by
    public var sortColum: SortColumn = .erid
    /// sort direction
    public var sortDirection: FilterSortDirection = .ascending
}

class LockListResource: PaginatedModelResource<LockListEntryModel> {

    var identifier: LockListIdentifier
    var filter: LockFilter?

    init(with identifier: LockListIdentifier, in kit: Kit) {
        self.identifier = identifier
        super.init(in: kit)
    }

    override func createQueryItems() -> [URLQueryItem] {
        var items = super.createQueryItems()

        guard let filter = self.filter else {
            return items
        }

        if let erid = filter.erid {
            items.append(.init(name: "filter[erid]", value: erid))
        }
        if let user = filter.user {
            items.append(.init(name: "filter[user]", value: user))
        }
        if let problem = filter.problem {
            items.append(.init(name: "filter[problem]", value: String(problem)))
        }
        if let subproblem = filter.subproblem {
            items.append(.init(name: "filter[subproblem]", value: subproblem))
        }
        if let pass = filter.pass {
            items.append(.init(name: "filter[pass]", value: String(pass)))
        }

        items.append(.init(name: "sort", value: filter.sortColum.withDirection(filter.sortDirection)))

        return items
    }

    override var path: String {
        return "exam/\(identifier.examId)/lock"
    }
}
