// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

class ModelResource<T: Codable>: NetworkDataResource<T> {
    override var useCache: Bool { false }

    var completion: ((T?, Error?) -> Void)?

    func decoder() -> JSONDecoder {
        let decoder = JSONDecoder()
        decoder.dateDecodingStrategy = .iso8601withConditionalFractionalSeconds
        return decoder
    }

    override func createValue(from data: Data) -> T? {
        // Apparently JSONDecoder and JSONEncoder leak memory (undocumented behaviour)
        // https://bugs.swift.org/browse/SR-5501?page=com.atlassian.jira.plugin.system.issuetabpanels%3Aall-tabpanel
        // The proposed workaround suggests adding an autoreleasepool around them
        return autoreleasepool {
            do {
                let model = try decoder().decode(T.self, from: data)
                return model
            } catch let error as NSError {
                if let string = String(data: data, encoding: .utf8) {
                    Logger.network.error("Error decoding model from resource \(error)",
                                         metadata: ["dump": Logger.MetadataValue(stringLiteral: string)])
                } else {
                    Logger.network.error("Error decoding model from resource \(error)")
                }
            }

            return nil
        }
    }

    override func finished() {
        super.finished()
        completion?(value, error)
    }
}
