// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import os.log

protocol BaseNetworkDataResource {
    func task(_ task: URLSessionDataTask, didReceive data: Data)
    func task(_ task: URLSessionDataTask, didUpdateExpectedSizeTo size: Int64)
}

/// enum representing a network response error
public enum NetworkDataError: Error {
    /// empty response / no data
    case noData
    /// data could not be parsed
    case invalidResponse
}

class NetworkDataResource<T>: NetworkResource<T> {
    var buffer: Data?
    var bytesTotal: Int64?
    var progress: Double? {
        return Double(buffer?.count ?? 0) / Double(bytesTotal ?? 0)
    }

    override func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        let path = self.path
        guard let req = networking.protectedRequest(forPath: path) else {
            self.fail(with: NetworkError.unauthorized)
            return nil
        }
        let task = session.dataTask(with: req)

        return task
    }

    override func taskDidComplete(_ task: URLSessionTask) {
        // TODO check this
		if self is SolutionSubmissionResource
			&& self.buffer == nil {
			return finished()
		}

        guard let data = self.buffer else {
            fail(with: NetworkDataError.noData)
            return
        }
        if setValue(using: data) != nil {
            finished()
        } else {
            fail(with: NetworkDataError.invalidResponse)
        }
        buffer = nil
    }
}

extension NetworkDataResource: BaseNetworkDataResource {
    func task(_ task: URLSessionDataTask, didReceive data: Data) {
        if var buffer = self.buffer {
            buffer.append(data)
            self.buffer = buffer
        } else {
            self.buffer = data
        }
    }

    func task(_ task: URLSessionDataTask, didUpdateExpectedSizeTo size: Int64) {
        self.bytesTotal = size
    }
}
