// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

protocol BaseNetworkResource {
    var priority: Operation.QueuePriority { get }

    func createTask(session: URLSession, networking: Networking) -> URLSessionTask?

    func task(_ task: URLSessionTask, didCompleteWithError error: Error?)
}

/// enum representing a network error
public enum NetworkError: Error {
    /// invalid value
    case invalidValue
    /// 404 not found
    case notFound
    /// 403 forbidden
    case forbidden
    /// 401 unauthorized
    case unauthorized // maybe delete this. this should not be handled by the UI
	/// 409 rejected pages when submitting
	case rejected([SolutionRejectionModel])
    case lockTaken
}

extension NetworkError: Equatable {}

class NetworkResource<T>: Resource<T> {
    var path: String { fatalError("implement in subclass") }
    var useCache: Bool { false }
    var priority: Operation.QueuePriority = .normal
    let maxRetries: Int = 3
    var retryCounter: Int = 0

    override func load() {
        guard self.state == .initial || self.state == .error else { return }
        guard let networking = kit?.networking else { return }
        self.state = .loading
        networking.load(resource: self)
    }

    func createValue(from data: Data) -> T? {
        return nil
    }

    @discardableResult
    func setValue(using data: Data) -> T? {
        let val = createValue(from: data)
        self.value = val
        return val
    }

    func createTask(session: URLSession, networking: Networking) -> URLSessionTask? {
        fatalError("don't use network resource directly")
    }

    func taskDidComplete(_ task: URLSessionTask) {
        finished()
    }

	func shouldFailWithError(forStatusCode statusCode: Int) -> Error? {
		switch statusCode {
        case 200...299:
            return nil
		case 401:
			return NetworkError.unauthorized
		case 404:
			return NetworkError.notFound
		case 403:
			return NetworkError.forbidden
		// TODO: 423 lock taken
		case 409:
			if let data = (self as? NetworkDataResource)?.buffer {
                return autoreleasepool {
                    let pages = try? JSONDecoder().decode([SolutionRejectionModel].self, from: data)
                    return NetworkError.rejected(pages ?? [])
                }
			}
			return NetworkError.rejected([])
        case 423:
            return NetworkError.lockTaken
		default:
            Logger.network.error("unexpected status code \(statusCode) for path \(path)")
			return nil
		}
	}

    func task(_ task: URLSessionTask, didCompleteWithError error: Error?) {
        if let error = error {
            // Retry before failing if error is recoverable
            // Check if error is NSError (URLError) and check if code is outside of HTTP status code range
            if !(100...599 ~= (error as NSError).code) {
                retry(with: error)
            } else {
                fail(with: error)
            }
        } else if let response = task.response as? HTTPURLResponse {
            if let kitError = shouldFailWithError(forStatusCode: response.statusCode) {
                fail(with: kitError)
            } else {
                taskDidComplete(task)
            }
        } else {
            taskDidComplete(task)
        }
    }

    func retry(with error: Error) {
        guard retryCounter < maxRetries else {
            Logger.network.error("Max retries reached for resource \(self)")
            fail(with: error)
            return
        }
        guard let networking = kit?.networking else { return }

        defer {
            self.state = .loading
            retryCounter += 1
        }

        // Retry request
        // swiftlint:disable:next line_length
        Logger.network.info("Retrying request for resource \(self), retryCounter = \(retryCounter), maxRetries = \(maxRetries)")
        networking.load(resource: self)
    }
}

extension NetworkResource: BaseNetworkResource {}
