// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

/// the state for a network resource
public enum ResourceState {
    /// not yet loaded
    case initial
    /// currently loading
    case loading
    /// successfully loaded
    case loaded
    /// error while loading
    case error
}

/// Delegate of a Resource, should be implemented by types consuming the resource
protocol ResourceDelegate: AnyObject {
    /// Called when loading the resource failed
    /// - Parameter resource: The resource which failed to load
    /// - Parameter error: The error indicating why the load failed
    func resource(_ resource: BaseResource, didFailWithError error: Error)

    /// Called when the resource finished loading sucessfully
    /// - Parameter resource: The resource which finished loading
    func resourceDidLoad(_ resource: BaseResource)
}

/**
 The base class for a resource.

 A resource is anything that can be fetched or created on demand.
 While most commonly used for remote resources, it could also be a local resource that requires some background computation.
 */
class BaseResource {
    /// the current state of the resource
    var state: ResourceState
    /// the error with which the resource failed to load (if it failed, nil otherwise)
    var error: Error?

    /// weak backreference to the kit
    weak var kit: Kit?
    /// weak ref to the ResourceDelegate of this resource
    weak var delegate: ResourceDelegate?

    /// Initializer, should be called by consumer of the resource
    ///
    /// Don't forget to set the resources delegate reference
    init(in kit: Kit) {
        self.state = .initial
        self.kit = kit
    }

    /// tries to load the resource
    func load() {
        // BaseResource should never be used directly, it is meant as an extension point for custom subclasses, implementing the most common features of a resource
        fatalError("override in subclass")
    }

    // MARK: - State Lifecycle

    /// sets the internal state to failed and notifies the delegate
    ///
    /// needs to be called by a subclass when the resource could not be loaded
    /// - Parameter error: the error reason why it could not be loaded
    internal func fail(with error: Error) {
        self.error = error
        self.state = .error
        delegate?.resource(self, didFailWithError: error)
    }

    /// sets the internal state to `.loaded` and notifies the delegate
    ///
    /// needs to be called by a subclass when the resource did load
    internal func finished() {
        self.state = .loaded
        delegate?.resourceDidLoad(self)
    }
}

// MARK: - Hashable Conformance

extension BaseResource: Hashable {
    static func == (lhs: BaseResource, rhs: BaseResource) -> Bool {
        return lhs === rhs
    }

    func hash(into hasher: inout Hasher) {
        return hasher.combine(ObjectIdentifier(self))
    }
}

// MARK: - Convenience Getters
extension BaseResource {
    /// Indicates whether `state == .loaded`
    var isLoaded: Bool {
        return self.state == .loaded
    }

    /// Indicated whether `state == .error`
    var hasFailed: Bool {
        return self.state == .error
    }
}

// MARK: - ShortDescription conformance

extension ResourceState: ShortDescription {
    public var shortDescription: String {
        return String(describing: self)
    }
}

// MARK: - IdentifiableResource Protocol

protocol IdentifiableResource: BaseResource {
    associatedtype IdentifierType
    var identifier: IdentifierType { get }

    init(_ identifier: IdentifierType, in kit: Kit)
}

// MARK: - Resource with associated value type

class Resource<ValueType>: BaseResource {
    var value: ValueType?
}
