// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine

/// base protocol for an editable objec
public protocol Editable: AnyObject {
    /// can currently be edited
    var editable: Bool { get }

    /// has changes
    var hasChanges: Bool { get }
}

/// enum representing the state of a submission
public enum SubmissionState: Equatable {
    /// nothing changed
    case noChanges
    /// something changed, not yet submitted
    case changed
    /// currently submitting
    case submitting
    /// error while submitting
    case error(Error)

    public static func == (lhs: SubmissionState, rhs: SubmissionState) -> Bool {
        switch (lhs, rhs) {
        case (.noChanges, .noChanges), (.changed, .changed), (.submitting, .submitting), (.error, .error):
            return true
        default:
            return false
        }
    }
}

/// error type representing an error which occured when trying to submit the changes
public enum SubmissionError: Error {
    /// no valid lock which can be used to submit the changes
    case invalidLockToken
}

/// base protocol for anything that can be submitted
public protocol Submitable: Editable {
    /// the current state of the submission (observable)
    var submissionState: Observable<ResourceObjectState> { get }

    var submissionCompletion: ((Result<Void, Error>) -> Void)? { get set }

    /// submits the changes
    func submitChanges()
}

public protocol Deletable: Editable {
    var deletionState: Observable<ResourceObjectState> { get }

    var deletionCompletion: ((Result<Void, Error>) -> Void)? { get set }

    func delete()
}

protocol Creatable: ModelKitObjectType, KitChildObjectType {
    associatedtype CreationSubmissionType

    static func creationResource(submission: CreationSubmissionType, in kit: Kit) -> Resource<ModelType>
    static func identifier(forCreationResult model: ModelType, resource: Resource<ModelType>) -> IdentifierType
}


