// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

/// Identifies a specific correction
public struct CorrectionIdentifier: Hashable {
    /// the exam
    let exam: ExamIdentifier
    /// the erid
    let erid: String
    /// the problem
    let problem: Int
    /// the subproblem
    let subproblem: String?
    /// the correctionpass
    let pass: Int

    public init(exam: ExamIdentifier,
                erid: String,
                problem: Int,
                subproblem: String?,
                pass: Int) {
        self.exam = exam
        self.erid = erid
        self.problem = problem
        self.subproblem = subproblem
        self.pass = pass
    }
}

extension CorrectionIdentifier: KitIdentifier {
    public var stringIdentifier: String {
        return "C-\(exam)-\(erid)-\(problem)-\(subproblem ?? "all")-\(pass)"
    }

    public func identifier(forPage page: Int) -> CorrectionPageIdentifier {
        return CorrectionPageIdentifier(exam: exam, erid: erid, page: page)
    }
}

/// Identifies a page for a correction
public struct CorrectionPageIdentifier: Hashable {
    /// the exam
    let exam: ExamIdentifier
    /// the erid
    let erid: String
    /// the page
    let page: Int
}

extension CorrectionPageIdentifier: KitIdentifier {
    public var stringIdentifier: String {
        return "C-\(exam)-\(erid)-\(page)"
    }

    public func identifier(forLayer layer: Int) -> CorrectionPageLayerIdentifier {
        return CorrectionPageLayerIdentifier(pageIdentifier: self, layer: layer)
    }
}

/// Identifies a layer on a correction page
public struct CorrectionPageLayerIdentifier: Hashable {
    /// the correction page
    let pageIdentifier: CorrectionPageIdentifier
    /// the layer number
    let layer: Int
}

extension CorrectionPageLayerIdentifier {
    var exam: ExamIdentifier { return pageIdentifier.exam }
    var erid: String { return pageIdentifier.erid }
    var page: Int { return pageIdentifier.page }
}

extension CorrectionPageLayerIdentifier: KitIdentifier {
    public var stringIdentifier: String {
        return pageIdentifier.stringIdentifier + "-\(layer)"
    }
}

/// Identifies a problem/pass for a correction
struct CorrectionProblemPassIdentifier: KitIdentifier {
    /// the correctios
    let correction: CorrectionIdentifier
    /// the pass
    let pass: Int?
    /// the problem
    let problem: Int?
    /// the subproblem
    let subproblem: String?

    var stringIdentifier: String {
        var base: String = "C-\(correction.exam)-\(correction.erid)-\(pass?.description ?? "ALL")"
        if let problem = problem {
            base += "-\(problem)"
            if let subproblem = subproblem {
                base += ".\(subproblem)"
            }
        } else {
            assert(subproblem == nil, "no problem, but subproblem specified")
        }
        return base
    }
}
