// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

/// wrapper to enable multiple delegates for a specific delegate protocol `P`
public class MulticastDelegate<P> {
    private var delegates: [WeakBox<AnyObject>] = []

    /// adds a delegate utilizing a weak reference. it is not required to call remove
    /// - Parameter delegate: the delegate
    public func add(_ delegate: P?) {
        guard let delegate = delegate else {
            return
        }

        let object = delegate as AnyObject
        delegates.append(WeakBox(object))
    }

    /// remove the delegate
    /// - Parameter delegate: the delegate
    public func remove(_ delegate: P) {
        let object = delegate as AnyObject
        self.delegates = delegates.filter {
            if let value = $0.value {
                return value !== object
            }
            return false
        }
    }

    func inform(_ block: (P) -> Void) {
        // if any unbox'ed value == nil -> filter delegates
        var needsFilter: Bool = false

        // iterate and cast
        for delegate in delegates {
            if let delegate = delegate.value as? P {
                // call the block
                block(delegate)
            } else {
                needsFilter = true
            }
        }

        // filter needed -> remove nil delegates
        if needsFilter {
            delegates = delegates.filter { $0.value != nil }
        }
    }
}
