// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

/// the state of an object which is bound to a specific (most commonly network) resource
public enum ResourceObjectState: Equatable {

    /// Returns a Boolean value indicating whether two values are equal.
    ///
    /// Equality is the inverse of inequality. For any values `a` and `b`,
    /// `a == b` implies that `a != b` is `false`.
    ///
    /// - Parameters:
    ///   - lhs: A value to compare.
    ///   - rhs: Another value to compare.
    public static func == (lhs: ResourceObjectState, rhs: ResourceObjectState) -> Bool {
        switch (lhs, rhs) {
        case (.idle, .idle), (.loading, .loading), (.loaded, .loaded):
            return true
        case (.error, .error):
            // currently discard the error value
            return true
        default:
            return false
        }
    }

    /// not yet loaded
    case idle

    /// currently loading
    case loading

    /// loaded and available
    case loaded

    /// error occured while loading
    case error(Error)

    /// in case the state is .error, this computed property returns the wrapped error
    public var errorValue: Error? {
        if case let .error(error) = self {
            return error
        }
        return nil
    }
}

protocol ResourceKitObjectType: ResourceDelegate, Loadable {
    associatedtype ResourceType: BaseResource
    var resource: ResourceType { get }
    var state: Observable<ResourceObjectState> { get }

    func update(using resource: ResourceType)
    func createResource() -> ResourceType
}

extension ResourceKitObjectType {
    public func load() {
        guard state.value != .loading && state.value != .loaded else { return }
        state.value = .loading
        resource.load()
    }

    func resourceDidLoad(_ resource: BaseResource) {
        guard resource === self.resource else { return }
        update(using: self.resource)
        state.value = .loaded
    }

    func resource(_ resource: BaseResource, didFailWithError error: Error) {
        Logger.network.error("failed to load \(resource): \(error)")
        state.value = .error(error)
    }

    func createDelegatedResource() -> ResourceType {
        let resource = createResource()
        resource.delegate = self
        return resource
    }
}

protocol ModelKitObjectType: AnyObject {
    associatedtype ModelType: Codable

    var model: ModelType? { get set }

    func update(using model: ModelType)
}

protocol ModelResourceKitObjectType: ModelKitObjectType, ResourceKitObjectType where ResourceType: Resource<ModelType> { }

extension ModelResourceKitObjectType {
    func update(using resource: ResourceType) {
        if let model = resource.value {
            self.model = model
            update(using: model)
        }
    }
}

extension ResourceKitObjectType where
    Self: KitChildObjectType,
    ResourceType: IdentifiableResource,
    Self.IdentifierType == ResourceType.IdentifierType {
    func createResource() -> ResourceType {
        return ResourceType(self.identifier, in: self.kit!)
    }
}
