// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine

/// Wrapper for observable values. It would be awesome to just use Combine's `@Published`
/// but Property Wrappers are not available (yet) in protocols
///
/// Important: observers are not necessarily called on the main thread (unless you use combine `receiveOnMain()`)
public class Observable<T>: Publisher {
    public typealias Output = T
    public typealias Failure = Never

    private let subject: CurrentValueSubject<Output, Failure>

    public internal(set) var value: T {
        get { subject.value }
        set { subject.value = newValue }
    }

    init(_ value: T) {
        self.subject = CurrentValueSubject(value)
    }

    public func receive<S>(subscriber: S) where S: Subscriber, Never == S.Failure, T == S.Input {
        subject.receive(subscriber: subscriber)
    }
}

extension Observable where T: ExpressibleByNilLiteral {
    convenience init() {
        self.init(nil)
    }
}

extension Observable where T: ExpressibleByArrayLiteral {
    convenience init() {
        self.init([])
    }
}
