// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

extension Correction: LockStatePropagation {
    func lockStateChanged(_ lockState: LockState) {
        Logger.correction.info("[Correction|\(identifier.stringIdentifier)] lock state changed to \(lockState)")
        editable = lockState.editable

        if case .taken = lockState, case .freed = lockState {
            dataMayHaveChanged = true
        }

        if state.value == .idle && lockState.loaded {
            // if the correction wasn't loaded so far,
            // and the lock state indicates loaded -> load the correction
            Logger.correction.info("[Correction|\(identifier.stringIdentifier)] calling load() because lockState is loaded and self.state == .idle")
            load()
        } else if state.value == .loaded, dataMayHaveChanged, case .acquired = lockState {
            if dataMayHaveChanged {
                // if the correction is loaded, we have a lock, and data may have changed -> reload
                Logger.correction.info("[Correction|\(identifier.stringIdentifier)] calling load() because lockState == .acquired and self.state == .loaded and data may have changed")
                invalidate()
                load()
                return
            } else {
                // lock was acquired and we already loaded data, release resources on next free
                awaitFreeingResources()
            }

        } else {
            Logger.correction.info("[Correction|\(identifier.stringIdentifier)] not calling load() because lockState == \(lockState) and self.state == \(state.value)")
        }

        // propagate changes
        for page in pages.value {
            if let casted = page as? Page {
                casted.lockStateChanged(lockState)
            }
        }
    }

    func invalidate() {
        self.state.value = .idle
        self.resource = createDelegatedResource()

        // propagate changes
        for page in pages.value {
            if let casted = page as? LockStatePropagation {
                casted.invalidate()
            }
        }

        for problem in problems.value {
            problem.invalidate()
        }
    }
}
