// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

extension Correction: ModelResourceKitObjectType {
    typealias ModelType = CorrectionDataModel

    func createResource() -> CorrectionDataResource {
        return CorrectionDataResource(identifier, in: kit!)
    }

    func resourceDidLoad(_ resource: BaseResource) {
        if resource === self.resource {
            update(using: self.resource)
            state.value = .loaded

            // if lock is acquired, data changes are unlikely
            if case .acquired = self.lockToken.state.value {
                dataMayHaveChanged = false
                // Dat is loaded and lock is acquired, free resource on freeing lock
                awaitFreeingResources()
            }
        } else if resource === self.submissionResource {
            submissionDidFinish()
        }
    }

    func resource(_ resource: BaseResource, didFailWithError error: Error) {
        Logger.network.error("failed to load \(resource): \(error)")
        if resource === self.resource {
            state.value = .error(error)
        } else if resource === self.submissionResource {
            submissionDidFail(error)
        }
    }

    func update(using model: ModelType) {
        var newPages: [Page] = []

        let problems: [Problem] = model.problemDict.values.map { Problem(problem: $0, in: model) }
        if !allowModificationsToAllProblems {
            for problem in problems {
                if problem.identifier.problem != focusedProblem {
                    problem.editable = false
                } else {
                    for pass in problem.passes where pass.pass != focusedPass {
                        (pass as? ProblemPass)?.editable = false
                    }
                }
            }
        }
        self.problems.value = problems

        self.emptyBoxes.value = model.emptyBoxes.map({ EmptyBox(using: $0, in: model) })

        for (pageNumber, pageModel) in model.pages {
            let pageId = identifier.identifier(forPage: pageNumber)
            let page: Page = Page(identifier: pageId.unmanaged(owner: self), kit: kit!)
            page.correction = self
            let pageProblems = problems.filter { $0.page == pageNumber }
            let emptyBox = self.emptyBoxes.value.first(where: { $0.page == pageNumber })
            page.update(correctionModel: model, pageModel: pageModel, problems: pageProblems, emptyBox: emptyBox)

            page.lockStateChanged(lockToken.state.value)

            newPages.append(page)
        }

        newPages.sort { $0.pageNumber < $1.pageNumber }
        pages.value = newPages
    }

    func updateDueToLockChange(using model: ModelType) {
        for problem in problems.value {
            let backendIdentifier = problem.identifier.backendFormat
            guard let model = model.problemDict[backendIdentifier] else {
                Logger.correction.error("correction structure changed - problem \(backendIdentifier) missing")
                continue
            }
            problem.update(using: model)
        }
    }
}
