// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

extension Correction: Submitable {
    var hasChanges: Bool {
        pages.value.contains(\.hasChanges) || problems.value.contains(\.hasChanges)
    }

	func submitChanges(_ usingFakeDate: Bool) {
		self.submitChanges()
	}

    func submitChanges() {
        guard hasChanges else { return }
        guard submissionState.value != .loading else { return }
        // check if there's a token. (no changes should've occured without a valid token)
        kit?.queue.async { [weak self] in
            self?.submitInBackground()
        }
    }

    private func submitInBackground() {
        Logger.submission.debug("attempting to submit correction \(identifier.stringIdentifier)")
        guard let lockToken = lockToken.token else {
            Logger.submission.error("[\(identifier.stringIdentifier)] can't save without a valid lock")
            self.submissionState.value = .error(SubmissionError.invalidLockToken)
            return
        }

        submissionState.value = .loading

        let problems = self.problems.value.compactMap { $0.createSubmission() }
        let images = self.pages.value.flatMap { ($0 as? Page)?.createSubmissions() ?? [] }
        let emptyBoxes = self.emptyBoxes.value.compactMap { $0.createSubmission() }

        guard !problems.isEmpty || !images.isEmpty || !emptyBoxes.isEmpty else {
            // swiftlint:disable:next line_length
            Logger.submission.info("[\(identifier.stringIdentifier)] nothing to submit, hasChanges = \(self.hasChanges), empty = [p: \(problems.isEmpty), i: \(images.isEmpty), e: \(emptyBoxes.isEmpty)]")
            return
        }

        submissionState.value = .loading

        let submission = CorrectionSubmission(lockToken: lockToken,
                                              problems: problems,
                                              images: images,
                                              emptyBoxes: emptyBoxes)
        let submissionResource = SubmissionResource(identifier, submission: submission, in: kit!)
        self.submissionResource = submissionResource

        submissionResource.delegate = self
        submissionResource.load()
        Logger.submission.info("correction \(identifier.stringIdentifier) submitting")
    }

    func submissionDidFinish() {
        // back to no changes
        Logger.submission.info("correction sumission \(identifier.stringIdentifier) succeeded")
        submissionState.value = .loaded
        DispatchQueue.main.async {
            self.submissionCompletion?(.success(()))
        }

        for page in pages.value {
            if let page = page as? Page {
                page.submissionDidFinish()
            }
        }
        for problem in problems.value {
            problem.submissionDidFinish()
        }

        for box in emptyBoxes.value {
            box.submissionDidFinish()
        }

        assert(!hasChanges)

        submissionState.value = .idle
    }

    func submissionDidFail(_ error: Error) {
        Logger.submission.error("correction sumission \(identifier.stringIdentifier) failed: \(error)")
        submissionState.value = .error(error)
        self.submissionCompletion?(.failure(error))
    }
}
