// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import Logging
import os.log
import Nuke

extension CorrectionIdentifier {
    var lockIdentifier: LockIdentifier {
        self.lockIdentifier(pass: pass, problem: problem, subproblem: nil)
    }

    func lockIdentifier(pass: Int?, problem: Int?, subproblem: String?) -> LockIdentifier {
        return LockIdentifier(exam: exam, erid: erid, pass: pass, problem: problem, subproblem: subproblem)
    }
}

class Correction: KitObject<CorrectionIdentifier> {
    lazy var resource: CorrectionDataResource = createDelegatedResource()
    var state: Observable<ResourceObjectState> = Observable(.idle)

    var submissionState: Observable<ResourceObjectState> = Observable(.idle)
    var submissionCompletion: ((Result<Void, Error>) -> Void)?
    var submissionResource: SubmissionResource?

    var editable: Bool = false

    var corrected: Bool = false
    var available: Bool = false
    var locked: Bool = false
    var access: Bool = false

    var flagTypeIds: [Int] = []

    var dataMayHaveChanged: Bool = false

    var pages: Observable<[PageType]> = Observable()

    var problems: Observable<[Problem]> = Observable()

    var emptyBoxes: Observable<[EmptyBox]> = Observable()

    var allowModificationsToAllProblems: Bool = false

    var model: CorrectionDataModel?

    // the token used for locking this correction
    let lockToken: LockToken

    var lockStateObserver: AnyCancellable?
    var freeLockStateObserver: AnyCancellable?
    var stateLogObserver: AnyCancellable? // For debugging only

    var promiseBasedResourceFreeing: Bool = false

    // keep a weak reference to the parent
    weak var controller: CorrectionController?

    required init(identifier: CorrectionIdentifier, kit: Kit) {
        lockToken = kit.object(for: identifier.lockIdentifier)
        super.init(identifier: identifier, kit: kit)

        // observe the lock state, in order to enable/disable editing and layer loading
        lockStateObserver = lockToken.state
            .sink { [weak self] (lockState) in
                self?.lockStateChanged(lockState)
        }

        stateLogObserver = state.sink(receiveValue: { [weak self] newState in
            Logger.correction.info("[Correction|\(self?.identifier.stringIdentifier ?? "nil")] state changed to \(newState)")
        })
    }
}

extension Correction: CustomStringConvertible {
    public var description: String {
        return "Correction(id: \(self.identifier), erid: \(erid), hashValue: \(self.hashValue), model: \(String(describing: model)))"
    }
}

extension Correction: CorrectionType {
    var erid: String {
        return identifier.erid
    }

    var focusedProblem: Int {
        return identifier.problem
    }

    var focusedSubproblem: String? {
        return identifier.subproblem
    }

    var focusedPass: Int {
        return identifier.pass
    }

    func addLockRequirement() -> LockReferenceType {
        return lockToken.add()
    }

    func page(forProblem problem: Int) -> PageType? {
        // return corresponding mutable page
        guard let pageNumber = pageNumber(forProblem: problem) else { return nil }
        let index = pageNumber - 1

        guard index >= 0, index < pages.value.count else { return nil }

        return pages.value[index]
    }

    func page(forProblem identifier: ProblemIdentifier) -> PageType? {
        // return corresponding mutable page
        guard let pageNumber = pageNumber(forProblem: identifier) else { return nil }
        let index = pageNumber - 1

        guard index >= 0, index < pages.value.count else { return nil }

        return pages.value[index]
    }

    func pageNumber(forProblem problem: Int) -> Int? {
        for page in self.pages.value {
            if page.problems.value.contains(where: { $0.identifier.problem == problem }) {
                return page.pageNumber
            }
        }
        return nil
    }

    func pageNumber(forProblem identifier: ProblemIdentifier) -> Int? {
        for page in self.pages.value {
            if page.problems.value.contains(where: { identifier.matches($0.identifier) }) {
                return page.pageNumber
            }
        }
        return nil
    }

    func corrector(forProblem problem: Int, pass: Int) -> ProblemPassCorrectorType {
        if let model = self.model,
           let firstSubproblem = model.problemDict.filter({ $0.key.hasPrefix(String(problem)) }).sorted(by: \.key).first,
           let creditsObj = firstSubproblem.value.credits.first(where: { $0.pass == pass }),
           let corrector = creditsObj.corrector {
            // Corrector is present in the dictionary already
            return StaticProblemPassCorrector(name: corrector)
        }

        let correctorIdentifier
            = CorrectionProblemPassIdentifier(correction: identifier, pass: pass, problem: problem, subproblem: nil)
        let problemPassCorrector: ProblemPassCorrector = kit!.object(for: correctorIdentifier)

        return problemPassCorrector
    }

    func allCorrectors(forProblem problem: Int) -> [SubproblemCorrectorType] {
        var correctors = [SubproblemCorrector]()

        guard let model = model else {
            return correctors
        }

        for (key, value) in model.problemDict.filter({ $0.key.hasPrefix(String(problem)) }) {
            var subproblemCorrectors = [SubproblemCorrectorMapping?]()
            let creditsSorted = value.credits.sorted(by: \.pass)

            for credits in creditsSorted {
                if let name = credits.corrector,
                   let date = credits.date {
                    subproblemCorrectors.append(SubproblemCorrectorMapping(name: name, date: date))
                } else {
                    subproblemCorrectors.append(nil)
                }
            }

            let correctorInfo = SubproblemCorrector(subproblem: key, correctors: subproblemCorrectors)
            correctors.append(correctorInfo)
        }

        return correctors.sorted(by: \.subproblem)
    }

    func freeResourcesAndReset() {
        guard state.value == .loaded else {
            Logger.correction.warning("[Correction|\(self.identifier)] wont free resources because state == \(state.value) != .loaded")
            return
        }
        guard submissionState.value == .idle else {
            // TODO: Handle this by delaying the free, not by cancelling it
            Logger.correction.error("Could not free correction because submission is still in progress")
            return
        }
        Logger.correction.info("[Correction|\(self.identifier.stringIdentifier)] freeing resources and resetting to unloaded state")

        self.invalidate()
        self.pages.value = []
        self.submissionState.value = .idle
    }

    func awaitFreeingResources() {
        guard !promiseBasedResourceFreeing else { return }
        guard self.state.value == .loaded && self.lockToken.state.value == .acquired else { return }

        Logger.correction.info("[Correction|\(self.identifier.stringIdentifier)] waiting for lock to be freed in order to free resources")
        self.freeLockStateObserver = self.lockToken.state.sink(receiveValue: { [weak self] state in
            if state == .freed {
                self?.freeLockStateObserver = nil
                guard !(self?.promiseBasedResourceFreeing ?? false) else { return }
                self?.freeResourcesAndReset()
            }
        })
    }

    func simulateLockLoss() {
        self.lockStateChanged(.taken)
        DispatchQueue.main.asyncAfter(deadline: .now() + .seconds(2)) {
            self.lockStateChanged(.acquired)
        }
    }
}
