// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

/// the opaque type for a single correction instance, restricted to a specific problem in a specific pass
public protocol CorrectionType: Stateful, Submitable {
    /// the identifier
    var identifier: CorrectionIdentifier { get }

    /// the erid of the correction
    var erid: String { get }

    /// the currently 'focused' problem for this correction
    var focusedProblem: Int { get }

    /// the currently focused subproblem for this correction. Might not be present for every correction
    var focusedSubproblem: String? { get }

    /// the currently 'focused' pass for this correction
    var focusedPass: Int { get }

    /// whether this correction is already corrected
    var corrected: Bool { get }
    var available: Bool { get }
    var locked: Bool { get }
    var access: Bool { get }

    var flagTypeIds: [Int] { get }

    var promiseBasedResourceFreeing: Bool { get set }

    /// the pages (observable)
    var pages: Observable<[PageType]> { get }

    /// returns a reference to a lock. the lifetime of this reference implies the
    /// duration of the lock. when all lock references are removed,
    /// the lock will be free'd in the backend
    func addLockRequirement() -> LockReferenceType

    /// returns the page for a specific problem (requires loaded)
    /// - Parameter problem: the problem for which to get the page
    func page(forProblem problem: Int) -> PageType?

    /// returns the page for a specific problem (requires loaded)
    /// - Parameter identifier: the problem identifier for which to get the page
    func page(forProblem identifier: ProblemIdentifier) -> PageType?

    /// returns the corrector information for the specified problem/pass
    /// - Parameters:
    ///   - problem: the problem
    ///   - pass: the pass
    func corrector(forProblem problem: Int, pass: Int) -> ProblemPassCorrectorType

    func allCorrectors(forProblem: Int) -> [SubproblemCorrectorType]

    func simulateLockLoss()
}

extension CorrectionType {
    /// returns the page of the correction, which contains the focused problem
    /// - Returns: the focused page
    public func pageForFocusedProblem() -> PageType? {
        return page(forProblem: focusedProblem)
    }

    public func pageForFocusedSubproblem() -> PageType? {
        guard let subproblem = focusedSubproblem else {
            return pageForFocusedProblem()
        }
        return page(forProblem: .subproblem(focusedProblem, subproblem))
    }

    /// returns the page with the specifed number
    ///
    /// **IMPORTANT** index starts at 1
    /// - Parameter number: the page number
    /// - Returns: the page
    public func page(withNumber number: Int) -> PageType? {
        return pages.value.first(where: { $0.pageNumber == number })
    }
}
