// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import os.log

extension CorrectionController.InternalState {
    var canLoadMore: Bool {
        switch self {
        case let .pagination(currentPage: current, maxPage: max):
            return current < max
        default:
            return false
        }
    }

    var nextPage: Int? {
        if case let .pagination(currentPage: current, maxPage: max) = self {
            if current >= max {
                return nil
            }
            return current + 1
        } else if case let .lookingForUnknownERIDs(currentPage: current, maxPage: max) = self {
            if current >= max {
                return nil
            }
            return current + 1
        } else {
            return nil
        }
    }

    var isRetrying: Bool {
        switch self {
        case .retry, .lookingForUnknownERIDs:
            return true
        default:
            return false
        }
    }
}

extension CorrectionController: ModelResourceKitObjectType {
    typealias ModelType = PaginationModel<ExamCorrectionModel>

    public func load() {
        guard state.value != .loading else { return }
        guard internalState == .initial else { return }
        self.state.value = .loading
        resource = createDelegatedResource()
        resource.load()
    }

    public func forceLoad() {
        guard internalState == .initial else { return }
        self.state.value = .loading
        resource = createDelegatedResource()
        resource.load()
    }

    public func loadMore() {
        guard state.value != .loading else { return }
        guard let nextPage = internalState.nextPage else {
            return
        }
        self.state.value = .loading

        resource = createDelegatedResource()
        resource.page = nextPage
        resource.load()
    }

    private func retry() {
        // can only retry when coming from pagination
        guard case .pagination = internalState else { fatalError() }
        // set state to retry
        internalState = .retry

        // reset resource and load
        resource = createDelegatedResource()
        resource.page = nil
        resource.load()
    }

    func handleEmptyResults() {
        switch internalState {
        case .initial, .retry, .lookingForUnknownERIDs:
            // initial load OR retry and no results -> done
            done()
        case .pagination:
            // pagination and no results -> retry
            retry()
        case .done:
            assertionFailure("handleEmptyResults called although already done")
        }
    }

    func done() {
        if Thread.isMainThread {
            self.delegate?.controllerGotEmptyResults(self)
        } else {
            DispatchQueue.main.async {
                self.delegate?.controllerGotEmptyResults(self)
            }
        }
    }

    func update(using model: ModelType) {
        // if empty -> retry
        guard !model.results.isEmpty else {
            handleEmptyResults()
            return
        }

        // non-empty results -> use them
        var newCorrections: [Correction] = []
        for correctionModel in model.results {
            // check if erid is already in controller
            if !listMode, correctionErids.contains(correctionModel.erid) {
                continue
            }

            let identifier = CorrectionIdentifier(
                exam: self.identifier,
                erid: correctionModel.erid,
                problem: correctionModel.problem,
                subproblem: correctionModel.subproblem,
                pass: correctionModel.correctionPass
            )

            let correction: Correction = Correction(identifier: identifier, kit: kit)
            correction.allowModificationsToAllProblems = self.allowModificationsToAllProblems
            correction.controller = self
            correction.corrected = correctionModel.corrected
            correction.available = correctionModel.available
            correction.access = correctionModel.access
            correction.locked = correctionModel.locked
            correction.flagTypeIds = correctionModel.flagTypeId

            newCorrections.append(correction)
            correctionErids.insert(correction.erid)
        }

        // if currently retrying,
        if !listMode, internalState.isRetrying {
            if newCorrections.isEmpty {
                // no new erids in this batch -> try the next one
                if model.currentPage < model.maxPage {
                    internalState = .lookingForUnknownERIDs(currentPage: model.currentPage, maxPage: model.maxPage)
                    loadMore()
                } else {
                    // reached end. done for good
                    done()
                }
                // in both cases: no results -> return
                return
            } else {
                // found some erid's
                if model.currentPage < model.maxPage {
                    // more pages available -> switch back to pagination mode
                    internalState = .pagination(currentPage: model.currentPage, maxPage: model.maxPage)
                } else {
                    internalState = .done
                }
            }
        } else {
            // not currently retrying, normal mode
            // check if last page
            if model.currentPage < model.maxPage {
                internalState = .pagination(currentPage: model.currentPage, maxPage: model.maxPage)
            } else {
                internalState = .done
            }
        }

        // calculate indexes for delegate
        let countBefore = corrections.count
        corrections.append(contentsOf: newCorrections)
        let countAfter = corrections.count

        let added = IndexSet(integersIn: countBefore..<countAfter)
        DispatchQueue.main.async {
            self.delegate?.controller(self, didLoadIndexes: added)
            if case .done = self.internalState {
                self.done()
            }
        }
    }

    func createResource() -> ExamCorrectionsResource {
        let resource = ExamCorrectionsResource(identifier, in: kit)
        resource.filter = self.filter
        return resource
    }
}
