// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import Logging

protocol CorrectionPromiseDataSource: AnyObject {
	func promiseWantsCorrection(_ promise: CorrectionPromise)
	func promise(_ promise: CorrectionPromise, failedToLockCorrection correction: Correction)
}

class CorrectionPromise {
    let uuid = UUID()

	var state: Observable<CorrectionPromiseState> = .init(.loading)

	weak var dataSource: CorrectionPromiseDataSource?

    var hasCorrection: Bool { correction != nil }

	internal private(set) var correction: Correction?
	private var correctionLock: LockReferenceType?
	private var correctionLockObserver: AnyCancellable?
    private var correctionStateObserver: AnyCancellable?
    private var correctionLockFreeObserver: AnyCancellable?

    private var prefetchProblem: ProblemIdentifier?
    private var pagesObserver: AnyCancellable?

    var correctionPass: Int? {
        return (dataSource as? InteractiveCorrectionController)?.pass
    }

    init(dataSource: CorrectionPromiseDataSource?) {
		self.dataSource = dataSource
	}

    // MARK: - Internal Logic

	/// starts by requesting a correction
	internal func start() {
		guard let dataSource = dataSource else { fatalError("no data source") }
        state.value = .loading

		// start by telling the data source that we need a correction
        Logger.dataSource.debug("[CorrectionPromise|\(pointerString(self))] wants correction")
		dataSource.promiseWantsCorrection(self)
	}

    /// called when the lock token was freed by the app (e.g. when the correction was scrolled back too far)
    ///
    /// This function should free the corrections resources from memory to free up space
    internal func expire() {
        Logger.dataSource.info("[CorrectionPromise\(pointerString(self))] \(correction?.identifier.stringIdentifier ?? "nil") is freeing up resources because lock was freed")
        // self.correction = nil
        self.correctionLock = nil
        self.correctionStateObserver = nil
        self.correctionLockObserver = nil
        self.correctionLockFreeObserver = nil
        self.pagesObserver = nil
        self.correction?.freeResourcesAndReset()
        self.state.value = .expired
    }

    public func load() {
        guard state.value == .expired else { return }
        guard let correction = self.correction else {
            Logger.correction.debug("[CorrectionPromise|\(pointerString(self))] load() called without a correction object")
            return
        }

        Logger.correction.info("[CorrectionPromise|\(pointerString(self))] reloading correction \(correction.identifier.stringIdentifier) after expiry using load()")

        state.value = .loading
        // Ensure we have lock (same logic as in `assign(correction:)`
        let lock = correction.addLockRequirement()
        self.correctionLock = lock
        self.correctionLockObserver = lock.state.sink { [weak self] (state) in
            self?.lockStateChanged(state)
        }
    }

	/// called when the lock state changes
	/// - Parameter state: the new state
	private func lockStateChanged(_ state: LockState) {
		switch state {
        case .unknown, .freed:
			// do nothing. wait for it to change
			break
		case .taken:
			// lock taken -> request new correction
			failed()
		case .acquired:
			preloadCorrection()
		}
	}

    private func preloadCorrection() {
        guard let correction = correction else { fatalError("can't preload without a correction") }

        correctionStateObserver = correction.state.sink { [weak self] (state) in
            self?.loadingStateChanged(state)
        }
    }

    private func loadingStateChanged(_ state: ResourceObjectState) {
        switch state {
        case .idle:
            correction?.load()
        case .loading:
            break
        case .loaded:
            finish()
        case .error:
            // todo
            failed()
        }
    }

	private func failed() {
		guard let correction = correction else { fatalError("failed to lock but no correction?!") }
		guard let dataSource = dataSource else { fatalError("no data source") }

        Logger.dataSource.error("[CorrectionPromise|\(pointerString(self))] failed to lock \(correction.erid)")

        // 'reset' references
        self.correction = nil
        self.correctionLock = nil
        self.correctionLockObserver = nil

		// tell the data source
		dataSource.promise(self, failedToLockCorrection: correction)

		// start over
		start()
	}

	private func finish() {
		guard let correction = correction else { fatalError("can't finish without a correction") }
        Logger.dataSource.info("[CorrectionPromise|\(pointerString(self))] ready with \(correction.identifier.stringIdentifier)")

		// don't observe the state anymore
		self.correctionLockObserver = nil

        // create new observer for free lock notification
        self.correctionLockFreeObserver = self.correctionLock?.state.sink(receiveValue: { [weak self] (state) in
            if case .freed = state {
                self?.expire()
            }
        })

		// update state
		self.state.value = .ready(correction)

        prefetchIfNecessary()
	}

    private func prefetchIfNecessary() {
        guard let problem = prefetchProblem, let correction = correction else { return }

        if !correction.pages.value.isEmpty {
            prefetchPage(for: problem)
        } else {
            pagesObserver = correction.pages.sink { [weak self] _ in
                self?.pagesChanged()
            }
        }
    }

    private func pagesChanged() {
        guard let problem = prefetchProblem else { return }
        guard let correction = correction else { return }
        guard !correction.pages.value.isEmpty else { return }

        prefetchPage(for: problem)
        pagesObserver = nil
    }

    private func prefetchPage(for problem: ProblemIdentifier) {
        guard let page = correction?.page(forProblem: problem) else { return }

        page.scan.load()
        page.solution?.load()
        page.personalizedSolution?.load()
        for layer in page.layers.value {
            if let focusedPass = page.focusedPass, layer.pass > focusedPass {
                continue
            }
            layer.image.load()
        }
    }

	// MARK: - Methods called by Controller

	/// called by the data source when a correction was made available
	/// - Parameter correction: the correction
    internal func assign(correction: Correction) {
		guard self.correction == nil else { fatalError("trying to assign another correction") }

        Logger.dataSource.debug("[CorrectionPromise|\(pointerString(self))] got correction \(correction.erid)")
		self.correction = correction
        self.correction?.promiseBasedResourceFreeing = true

		// add lock and observe
		let lock = correction.addLockRequirement()
		self.correctionLock = lock
        self.correctionLockObserver = lock.state.sink { [weak self] (state) in
			self?.lockStateChanged(state)
		}
	}

	internal func reachedEnd() {
        self.state.value = .reachedEnd(pass: self.correctionPass)
	}
}

// MARK: - Public API

extension CorrectionPromise: CorrectionPromiseType {

    var identifier: Int {
//        guard let correction = self.correction else {
//            return "".hashValue
//        }
//        return correction.identifier.hashValue
        return uuid.hashValue
    }

	func takeLockOwnership() -> LockReferenceType? {
		let lock = self.correctionLock
		self.correctionLock = nil
        Logger.lockManager.info("[CorrectionPromise|\(pointerString(self))] lock ownership of \(lock.debugDescription) transferred to callee")
		return lock
	}

    func prefetch(problem: Int) {
        prefetch(problem: .problem(problem))
    }

    func prefetch(problem: ProblemIdentifier) {
        prefetchProblem = problem
        if case .ready = state.value {
            prefetchIfNecessary()
        }
    }

    var debugDescription: String {
        if let correction = self.correction {
            return "<CorrPromise: \(pointerString(self))|\(correction.identifier.stringIdentifier)>"
        } else {
            return "<CorrPromise: \(pointerString(self))|nil>"
        }
    }
}
