// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

class InteractiveCorrectionController {
	private var exam: ExamIdentifier
	public private(set) var pass: Int
    var problem: ProblemIdentifier {
        return problems.first!
    }
	var problems: [ProblemIdentifier]
	private let loadingController: InteractiveLoadingController
	unowned var kit: Kit

	private var promises: [CorrectionPromise] = []
    private var waiting: [CorrectionPromise] = []

	private var availableCorrections: [Correction] = []

	private(set) var reachedEnd: Bool = false

    convenience init(exam: ExamIdentifier,
                     pass: Int,
                     problem: ProblemIdentifier,
                     filterLocked: Bool = true,
                     kit: Kit) {
        self.init(
            exam: exam,
            pass: pass,
            problems: [problem],
            filterLocked: filterLocked,
            kit: kit
        )
    }

    init(exam: ExamIdentifier,
         pass: Int,
         problems: [ProblemIdentifier],
         filterLocked: Bool = true,
         kit: Kit) {
        assert(problems.count > 0)
        self.exam = exam
        self.pass = pass
        self.problems = problems
        self.kit = kit

        self.loadingController = InteractiveLoadingController(
            exam: exam,
            pass: pass,
            problems: problems,
            filterLocked: filterLocked,
            kit: kit
        )
        self.loadingController.delegate = self
    }

}

extension InteractiveCorrectionController: InteractiveLoadingControllerDelegate {
    var usedERIDs: [String] {
        promises.compactMap { $0.correction?.erid } + availableCorrections.map(\.erid)
    }

    func controller(_ controller: InteractiveLoadingController, didLoadCorrections corrections: [Correction]) {
        Logger.correction.info("[IntCorrectionController] loaded \(corrections.count) corrections")
        // get waiting promises
        // var waiting = promises.filter(false, for: \.hasCorrection)

        // iterate until no descriptor or no index left
        for correction in corrections {
            // if descriptor is waiting -> assign
            if waiting.isEmpty {
                // no one waiting -> add to available
                availableCorrections.append(correction)
            } else {
                let promise = waiting.removeFirst()
                // assign the correction to the descriptor
                promise.assign(correction: correction)
            }
        }

        // if there are still waiting promises -> load more
        if !waiting.isEmpty {
            controller.loadMore()
        }
    }

    func controllerDidReachEnd(_ controller: InteractiveLoadingController) {
        Logger.correction.info("[IntCorrectionController] reached end of available corrections")
        // tell the waiting descriptors it's over
        for promise in promises where !promise.hasCorrection {
            promise.reachedEnd()
        }
        self.reachedEnd = true
    }
}

extension InteractiveCorrectionController: InteractiveCorrectionPromiseControllerType {
    var correctionPromises: [CorrectionPromiseType] {
        return promises
    }

    func getNextCorrection() -> CorrectionPromiseType? {
		// don't create new descriptors if already reached end
        guard !availableCorrections.isEmpty || !reachedEnd else { return nil }

		let promise = CorrectionPromise(dataSource: self)
        promises.append(promise)
        promise.start()
		return promise
	}
}

extension InteractiveCorrectionController: CorrectionPromiseDataSource {
	func promiseWantsCorrection(_ promise: CorrectionPromise) {
		// check if correction available
		if availableCorrections.isEmpty {
            waiting.append(promise)
            // -> wait
            loadingController.loadMore()
		} else {
            // -> assign
            let correction = availableCorrections.removeFirst()
            promise.assign(correction: correction)
		}
	}

	func promise(_ promise: CorrectionPromise, failedToLockCorrection correction: Correction) {
		// the descriptor couldn't lock the correction.
		// for now, don't do anything
        // TODO do something when locking fails
	}
}
