// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

protocol InteractiveLoadingControllerDelegate: AnyObject {
    var usedERIDs: [String] { get }

    func controller(_ controller: InteractiveLoadingController, didLoadCorrections: [Correction])
    func controllerDidReachEnd(_ controller: InteractiveLoadingController)
}

class InteractiveLoadingController {
    var underlyingCorrectionController: CorrectionController
    unowned var kit: Kit

    enum State {
        case idle
        case loading
        case retrying
    }

    var state: State = .idle

    weak var delegate: InteractiveLoadingControllerDelegate?

    convenience init(exam: ExamIdentifier, pass: Int, problem: ProblemIdentifier, filterLocked: Bool = true, kit: Kit) {
        self.init(exam: exam, pass: pass, problems: [problem], filterLocked: filterLocked, kit: kit)
    }

    init(exam: ExamIdentifier, pass: Int, problems: [ProblemIdentifier], filterLocked: Bool = true, kit: Kit) {
        self.kit = kit

        assert(problems.count > 0)
        let problem = problems.first!.problem
        guard problems.allSatisfy({ $0.problem == problem }) else {
            fatalError("ProblemIdentifiers given to InteractiveLoadingController must be from the same problem")
        }

        let subproblem = problems
            .compactMap { $0.subproblem }
            .joined(separator: "|")

        // create filter
        var filter = ResultFilter()
        filter.corrected = .no
        filter.locked = filterLocked ? .no : .all
        filter.available = .yes
        filter.access = .yes
        filter.pass = pass
        filter.problem = problem
        filter.subproblem = subproblem != "" ? subproblem : nil

        // create underlying correction
        self.underlyingCorrectionController = CorrectionController(identifier: exam, filter: filter, kit: kit)
        self.underlyingCorrectionController.delegate = self
        self.underlyingCorrectionController.listMode = false
    }

    func loadMore() {
        // don't load more while already working
        guard state == .idle else { return }
        state = .loading

        let usedERIDs = delegate?.usedERIDs

        let maxERID = usedERIDs?.max() ?? "0001"
        var filter = underlyingCorrectionController.filter
        filter.eridCutoff = .greaterEqual
        filter.erid = maxERID
        underlyingCorrectionController.filter = filter

        underlyingCorrectionController.load()
    }
}

extension InteractiveLoadingController: CorrectionControllerDelegate {
    func controller(_ controller: CorrectionControllerType, didLoadIndexes indexes: IndexSet) {
        guard let delegate = delegate else { fatalError() }

        let usedERIDs = Set(delegate.usedERIDs)

        // iterate over new corrections and gather all unused corrections/ERIDs
        var newCorrections: [Correction] = []
        for index in indexes {
            if let correction = controller.corrections[index] as? Correction {
                if !usedERIDs.contains(correction.erid) {
                    newCorrections.append(correction)
                }
            }
        }

        // anything new?
        if !newCorrections.isEmpty {
            // yes -> tell delegate
            state = .idle
            delegate.controller(self, didLoadCorrections: newCorrections)
        } else {
            // nope -> load more
            controller.loadMore()
        }
    }

    func controllerGotEmptyResults(_ controller: CorrectionControllerType) {
        switch state {
        case .loading, .idle:
            startRetry()
        case .retrying:
            retryEmpty()
        }
    }

    func startRetry() {
        state = .retrying

        var filter = underlyingCorrectionController.filter
        filter.eridCutoff = .greaterEqual
        filter.erid = "0001"
        underlyingCorrectionController.filter = filter
        underlyingCorrectionController.reset()
        underlyingCorrectionController.load()
    }

    func retryEmpty() {
        delegate?.controllerDidReachEnd(self)
    }
}
