// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

/// the delegate of a correction controller which can be used to receive state changes
public protocol CorrectionControllerDelegate: AnyObject {
    /// notifies the delegate when the controller did load indexes
    /// - Parameters:
    ///   - controller: the controller which did load indexes
    ///   - indexes: the indexes the controller did load
    func controller(_ controller: CorrectionControllerType, didLoadIndexes indexes: IndexSet)

    /// notifies the delegate when the controller got an empty result list (for example if the filter did not return any matches)
    /// - Parameter controller: the controller which got an empty results list
    func controllerGotEmptyResults(_ controller: CorrectionControllerType)
}

/**
 the opaque data type for a correction controller

 conforms to `StatefulLoadable`
 */
public protocol CorrectionControllerType: StatefulLoadable {
    /// the delegate of the controller, which receives state changes
    var delegate: CorrectionControllerDelegate? { get set }

    /// can be used to estimate how many corrections will probably be available
    var estimatedCount: Int { get }

    /// the loaded corrections (observable)
    var corrections: [CorrectionType] { get }

    /// the currently active filter
    var filter: ResultFilter { get set }

    /// whether to limit correction to the specified filter (should be set before the first `load()`)
    var allowModificationsToAllProblems: Bool { get set }

    /// loads more corrections
    func loadMore()

    func forceLoad()
}
