// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

/// state of the promise
public enum CorrectionPromiseState: Equatable, CustomStringConvertible {
    /// waiting for new correction or trying to acquire lock
    case loading
    /// locked and ready
    case ready(CorrectionType)
    /// error (not able to lock is not an error and handled internally)
    case error(Error)
    /// no more corrections available
    case reachedEnd(pass: Int?)

    case expired

    public static func == (lhs: CorrectionPromiseState, rhs: CorrectionPromiseState) -> Bool {
        if case .loading = lhs, case .loading = rhs {
            return true
        }
        if case .ready = lhs, case .ready = rhs {
            return true
        }
        if case .error = lhs, case .error = rhs {
            return true
        }
        if case .reachedEnd = lhs, case .reachedEnd = rhs {
            return true
        }
        if case .expired = lhs, case .expired = rhs {
            return true
        }
        return false
    }

    public var description: String {
        switch self {
        case .loading: return ".loading"
        case .ready: return ".ready"
        case .error: return ".error"
        case .reachedEnd: return ".reachedEnd"
        case .expired: return ".expired"
        }
    }
}

extension CorrectionPromiseState {
    /// the correction, if ready
    public var correction: CorrectionType? {
        guard case let .ready(correction) = self else { return nil }
        return correction
    }
}

/** a correction promise which tries to lock and prefetch a new correction.

 - Important: You need to take the ownership using `takeLockOwnership()` when correction was loaded
 */
public protocol CorrectionPromiseType: AnyObject, CustomDebugStringConvertible {
    var identifier: Int { get }

    /// the current state (observable)
    var state: Observable<CorrectionPromiseState> { get }

    var correctionPass: Int? { get }

    /// if ready, you need to transfer the lock ownership to a new owner.
    func takeLockOwnership() -> LockReferenceType?

    /// sets the problem to prefetch before being ready
    /// - Parameter problem: the problem / subproblem
    func prefetch(problem: ProblemIdentifier)

    /// sets the problem to prefetch before being ready
    /// - Parameter problem: the problem
    func prefetch(problem: Int)

    /// Called to load an expired promise again
    func load()
}
