// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

/// the column to sort the results by
public enum ResultSortByColumn: CustomStringConvertible {
    /// sort by erid
    case erid

    /// sort by problem
    case problem

    /// sort by pass
    case pass

    /// sort by date
    case date

    public var description: String {
        switch self {
        case .erid:
            return "ERID"
        case .problem:
            return "Problem"
        case .pass:
            return "Pass"
        case .date:
            return "Date"
        }
    }
}

/// the sort direction for the results
public enum FilterSortDirection: CustomStringConvertible {
    /// ascending (A-Z)
    case ascending

    /// desceding (Z-A)
    case descending

    public var description: String {
        switch self {
        case .ascending:
            return "Ascending"
        case .descending:
            return "Descending"
        }
    }

    public var urlRepresentation: String {
        switch self {
        case .ascending:
            return "asc"
        case .descending:
            return "desc"
        }
    }
}

/// a boolean filter for the results
public enum FilterBoolValue: CustomStringConvertible {
    /// only yes
    case yes

    /// only no
    // swiftlint:disable:next identifier_name
    case no

    /// both
    case all

    public var description: String {
        switch self {
        case .yes:
            return "True"
        case .no:
            return "False"
        default:
            return "All"
        }
    }
}

/// an additional cutoff for a filter attribute
public enum FilterAttributeCutoff: CustomStringConvertible {
    /// ==
    case none
    /// >=
    case greaterEqual
    /// <=
    case lessEqual

    public var description: String {
        switch self {
        case .none:
            return ""
        case .greaterEqual:
            return "ge:"
        case .lessEqual:
            return "le:"
        }
    }

    /// apply the cutoff to a attribute string
    /// - Parameter string: the attribute's string
    /// - Returns: the attribute with specified cutoff
    public func applied(to string: String) -> String {
        switch self {
        case .none:
            return string
        case .greaterEqual:
            return "ge:" + string
        case .lessEqual:
            return "le:" + string
        }
    }
}

/// a filter which can be used when loading corrections
public struct ResultFilter: Equatable {
    /// specif erid, or erid range (with cutoff)
    public var erid: String?
    /// erid cutoff
    public var eridCutoff: FilterAttributeCutoff = .none
    /// Filtering by specific subproblem for subproblem correction
    public var subproblem: String?
    /// specif problem, or range (with cutoff)
    public var problem: Int?
    /// problem cutoff
    public var problemCutoff: FilterAttributeCutoff = .none
    /// specif pass, or range (with cutoff)
    public var pass: Int?
    /// pass cutoff
    public var passCutoff: FilterAttributeCutoff = .none
    /// corrected (both, yes, no)
    public var corrected: FilterBoolValue = .all
    /// available  (both, yes, no)
    public var available: FilterBoolValue = .all
    /// locked  (both, yes, no)
    public var locked: FilterBoolValue = .all
    /// access (both, yes, no)
    public var access: FilterBoolValue = .all
    /// flag types (ORed together)
    public var flagTypeIds: Set<Int> = []

    /// column to sort by
    public var sortColum: ResultSortByColumn = .erid
    /// sort direction
    public var sortDirection: FilterSortDirection = .ascending

    /// swift generates a default init with internal access level.
    /// so we need to implement it manually
    public init(
        erid: String? = nil,
        eridCutoff: FilterAttributeCutoff = .none,
        subproblem: String? = nil,
        problem: Int? = nil,
        problemCutoff: FilterAttributeCutoff = .none,
        pass: Int? = nil,
        passCutoff: FilterAttributeCutoff = .none,
        corrected: FilterBoolValue = .all,
        available: FilterBoolValue = .all,
        locked: FilterBoolValue = .all,
        access: FilterBoolValue = .all,
        flagTypeIds: Set<Int> = [],
        sortColum: ResultSortByColumn = .erid,
        sortDirection: FilterSortDirection = .ascending) {
        self.erid = erid
        self.eridCutoff = eridCutoff
        self.subproblem = subproblem
        self.problem = problem
        self.problemCutoff = problemCutoff
        self.pass = pass
        self.passCutoff = passCutoff
        self.corrected = corrected
        self.available = available
        self.locked = locked
        self.access = access
        self.flagTypeIds = flagTypeIds
        self.sortColum = sortColum
        self.sortDirection = sortDirection
    }

    /// if there's any set value in this filter
    public var filterActive: Bool {
        return
            erid != nil ||
                subproblem != nil ||
                problem != nil ||
                pass != nil ||
                corrected != .all ||
                available != .all ||
                locked != .all ||
                access != .all ||
                !flagTypeIds.isEmpty ||
                sortColum != .erid ||
                sortDirection != .ascending
    }
}

/// Combine/SwiftUI wrapper for `ResultFilter`
public class ResultFilterBox: ObservableObject {
    @Published public var filter: ResultFilter
    @Published public var availableFlagDefinitions: [FlagDefinitionType] = []

    public var availableFlagTypeIds: [Int] {
        return availableFlagDefinitions.map { $0.flagTypeId }
    }

    /// create new wrapper
    /// - Parameter filter: the initial filter
    public init(filter: ResultFilter = ResultFilter()) {
        self.filter = filter
    }
}
