// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine

struct CorrectionSubproblemPassStatus: CorrectionSubproblemPassStatusType {
    var pass: Int
    var corrected: Int
    var total: Int

    var progress: Double {
        return Double(corrected) / Double(total)
    }
}

struct CorrectionPassStatus: CorrectionPassStatusType {
    var pass: Int
    var corrected: Int
    var total: Int

    var subCorrected: Int
    var subTotal: Int

    var progress: Double {
        return Double(corrected) / Double(total)
    }

    var progressRespectingSubproblems: Double {
        return Double(subCorrected) / Double(subTotal)
    }
}

struct CorrectionProblemStatus: CorrectionProblemStatusType {
    var problem: Int
    var subproblems: [CorrectionSubproblemStatusType]

    var passes: [CorrectionPassStatusType]
}

struct CorrectionSubproblemStatus: CorrectionSubproblemStatusType {
    var subproblem: String
    var passes: [CorrectionSubproblemPassStatusType]
}

class CorrectionStatus: KitObject<ExamIdentifier> {
    lazy var resource: ExamCorrectionStatusResource = createDelegatedResource()
    var model: [CorrectionStatusModel]?

    var state: Observable<ResourceObjectState> = Observable(.idle)

    // Problem -> Pass -> Problem Status
    var problems: Observable<[CorrectionProblemStatusType]> = Observable([])

    func forceReload() {
        guard kit != nil else { return }
        self.problems.value = []
        self.resource = createDelegatedResource()
        self.resource.load()
    }
}

extension CorrectionStatus: ModelResourceKitObjectType {
    func createResource() -> ExamCorrectionStatusResource {
        return ExamCorrectionStatusResource(identifier, perSubproblem: true, in: kit!)
    }

    func update(using model: [CorrectionStatusModel]) {
        guard !model.isEmpty else {
            return
        }
        var newProblems = [CorrectionProblemStatus]()
        var currentProblemNumber = model.first!.problem
        var currentProblem: CorrectionProblemStatus?

        for entry in model {
            guard let subproblemName = entry.subproblem else {
                continue
            }
            if entry.problem != currentProblemNumber && currentProblem != nil {
                // Next Problem
                newProblems.append(currentProblem!)
                currentProblemNumber = entry.problem
                currentProblem = nil
            }
            if currentProblem == nil {
                currentProblem = CorrectionProblemStatus(
                    problem: entry.problem,
                    subproblems: [
                        CorrectionSubproblemStatus(
                            subproblem: subproblemName,
                            passes: [
                                CorrectionSubproblemPassStatus(
                                    pass: entry.correctionPass,
                                    corrected: entry.corrected,
                                    total: entry.total
                                )
                            ])
                    ],
                    passes: [
                        CorrectionPassStatus(
                            pass: entry.correctionPass,
                            corrected: entry.corrected,
                            total: entry.total,
                            subCorrected: entry.corrected,
                            subTotal: entry.total
                        )
                    ]
                )
            } else {
                if let subIndex = currentProblem?.subproblems.firstIndex(where: { $0.subproblem == subproblemName }) {
                    // Subproblem already exists in currentProblem, just add pass
                    // swiftlint:disable:next force_cast
                    var sub = currentProblem!.subproblems[subIndex] as! CorrectionSubproblemStatus
                    sub.passes.append(
                        CorrectionSubproblemPassStatus(
                            pass: entry.correctionPass,
                            corrected: entry.corrected,
                            total: entry.total
                        )
                    )
                    currentProblem?.subproblems[subIndex] = sub
                } else {
                    // Create the subproblem
                    currentProblem?.subproblems.append(
                        CorrectionSubproblemStatus(
                            subproblem: subproblemName,
                            passes: [
                                CorrectionSubproblemPassStatus(
                                    pass: entry.correctionPass,
                                    corrected: entry.corrected,
                                    total: entry.total
                                )
                            ]
                        )
                    )
                }

                // Check if min changed
                if currentProblem!.passes.count <= entry.correctionPass - 1 {
                    currentProblem!.passes.append(
                        CorrectionPassStatus(
                            pass: entry.correctionPass,
                            corrected: entry.corrected,
                            total: entry.total,
                            subCorrected: entry.corrected,
                            subTotal: entry.total
                        )
                    )
                } else {
                    // swiftlint:disable:next force_cast
                    var pass = currentProblem!.passes[entry.correctionPass - 1] as! CorrectionPassStatus
                    pass.corrected = min(entry.corrected, pass.corrected)
                    pass.subCorrected += entry.corrected
                    pass.subTotal += entry.total
                    currentProblem!.passes[entry.correctionPass - 1] = pass
                }
            }
        }
        newProblems.append(currentProblem!)
        problems.value = newProblems
    }
}

extension CorrectionStatus: ShortDescription {
    public var shortDescription: String {
        return "CorrectionStatus[id=\(self.identifier)]"
    }
}

extension CorrectionStatus: CorrectionStatusType { }
