// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import CoreGraphics
#if os(iOS)
import UIKit
#endif

/// Delegate for a drawing, which is informed about element changes
public protocol DrawingDelegate: AnyObject {
    /// Called when an element is added
    /// - Parameters:
    ///   - drawing: the drawing
    ///   - element: the added element
    func drawing(_ drawing: Drawing, didAppendElement element: Drawing.Element)

    /// Called when the last element ist removed
    /// - Parameters:
    ///   - drawing: the drawing
    ///   - element: the removed, last element
    func drawing(_ drawing: Drawing, didRemoveLastElement element: Drawing.Element)

//    func drawing(_ drawing: Drawing, willExpireWithNewDrawing: Drawing)
}

/**
 Represents a drawing made out of strokes and snippets, optionally on top of an already existing image.

 While the image can be rendered using `getImage()`, it should not be used for the user interface, where CALayers are a better approach.

 It is possible to create a `Subdrawing` which targets only a small subsection of the overall drawing using `createSubdrawing(normalizedRect:)`.
 */
public class Drawing {
    /// enum for the different types of elements which can be added to a drawing
    public enum Element: Equatable {
        /// a stroke (pen, eraser, etc)
        case stroke(Stroke)
        /// a snippet
        case snippet(SnippetElement)
		/// a lasso selection
		case selection(SelectionElement)
    }

    /// the original image which should be drawn on top of
    public let originalImage: CGImage?
    /// the drawing's size
    public let size: CGSize

    public var normalizedRect: CGRect
//    public var normalizedPageRect: CGRect {
//        return CGRect(x: 0, y: 0, width: 1, height: 1)
//    }

    /// all elements added to the drawing. use `append(...)` to add elements
    public internal(set) var elements: [Element] = []

    /// the delegates to inform about changes to the drawing
    public let delegates = MulticastDelegate<DrawingDelegate>()

    init(size: CGSize, originalImage: CGImage?, elements: [Element] = []) {
        self.size = size
        self.originalImage = originalImage
        self.elements = elements
        self.normalizedRect = CGRect(x: 0, y: 0, width: 1, height: 1)
    }

    public func getImageForLayerCollapse(
        size: CGSize,
        referenceSize: CGSize? = nil,
        contentsScale: CGFloat
    ) -> CGImage? {
        return getImage(
            size: size,
            renderMode: .layerCollapse,
            referenceSize: referenceSize,
            contentsScale: contentsScale
        )
    }

    /// renders the drawing to a new image.
    /// - Note: slow. should be used carefully
    /// - Attention: not yet tested on macOS
    /// - Returns: the new image
    public func getImage() -> CGImage? {
        return getImage(size: size, renderMode: .finalImage)
    }

	// swiftlint:disable:next cyclomatic_complexity
    private func getImage(
		size: CGSize,
		renderMode: RenderMode,
        referenceSize: CGSize? = nil,
		contentsScale: CGFloat = 1.0) -> CGImage? {
        guard let ctx = createContext(size: size, contentsScale: contentsScale) else {
            print("Couldn't create context")
            return nil
        }

        let elements = self.elements
        let bounds = CGRect(origin: .zero, size: size)
            .mappedTo(size: CGSize(width: contentsScale,
                                   height: contentsScale))

        // draw original image
        if let original = originalImage {
            ctx.draw(original, in: bounds)
        }

        ctx.saveGState()
        ctx.translateBy(x: 0, y: bounds.size.height)
        ctx.scaleBy(x: 1.0, y: -1.0)

        // draw the elements
        for element in elements {
            switch element {
			case let .selection(element):
				let image = element.selection
				let rect = element.normalizedRect.mappedTo(size: bounds.size)
				// CoreGraphics flips the image. translate/scaleBy doesn't work.
				#if os(iOS)
				UIGraphicsPushContext(ctx)
				image.draw(in: rect)
				UIGraphicsPopContext()
				#else
				if let image = image.cgImage {
					ctx.draw(image, in: rect)
				}
				#endif
            case let .stroke(stroke):
                switch stroke.ink.tool {
				case .pen, .highlighter:
                    ctx.setBlendMode(.normal)
                case .eraser:
                    ctx.setBlendMode(.copy)
                }
                if let referenceSize = referenceSize {
                    let scaled = CGSize(
                        width: referenceSize.width * contentsScale,
                        height: referenceSize.height * contentsScale
                    )
                    stroke.draw(in: ctx, withSize: bounds.size, referenceSize: scaled, renderMode: renderMode)
                } else {
                    stroke.draw(in: ctx, withSize: bounds.size, renderMode: renderMode)
                }
            case let .snippet(snippet):
                guard let image = snippet.snippet.image else { continue }
                let rect = snippet.normalizedRect.mappedTo(size: bounds.size)
                // CoreGraphics flips the image. translate/scaleBy doesn't work.
                #if os(iOS)
                UIGraphicsPushContext(ctx)
                image.draw(in: rect)
                UIGraphicsPopContext()
                #else
                if let image = image.cgImage {
                    ctx.draw(image, in: rect)
                }
                #endif
            }
        }

        let image = ctx.makeImage()
        ctx.restoreGState()
        return image
    }

    func createContext(size: CGSize, contentsScale: CGFloat = 1.0) -> CGContext? {
        let width = Int(size.width * contentsScale)
        let height = Int(size.height * contentsScale)
        guard let colorSpace = CGColorSpace(name: CGColorSpace.sRGB) else {
            print("Couldn't create sRGB Color Space")
            return nil
        }
        let context = CGContext(
            data: nil,
            width: width,
            height: height,
            bitsPerComponent: 8,
            bytesPerRow: 4 * width, space: colorSpace, bitmapInfo: CGImageAlphaInfo.premultipliedFirst.rawValue)

        return context
    }

    /// append an element, either snippet or stroke
    /// - Parameter element: the element to add
    public func append(element: Element) {
        elements.append(element)
        delegates.inform { $0.drawing(self, didAppendElement: element) }
    }

    /// append a stroke
    ///
    /// same as
    ///
    ///     append(element: .stroke(stroke)
    ///
    /// - Parameter stroke: the stroke to add
    public func append(stroke: Stroke) {
        append(element: .stroke(stroke))
    }

    /// removes the last storke
    public func removeLast() {
        let last = elements.removeLast()
        delegates.inform { $0.drawing(self, didRemoveLastElement: last) }
    }

    /// normalize the line width to the drawing's size
    /// - Parameter lineWidth: the original line width
    /// - Returns: the new line width
    public func normalizedLineWidth(_ lineWidth: CGFloat) -> CGFloat {
        let maxDimension = max(size.width, size.height)
        return lineWidth / maxDimension
//        return lineWidth
    }

    /// creates a new subdrawing based on the normalizedRect, where origin=(0,0), size=(1,1) would be the full drawing
    /// - Parameter normalizedRect: the normalized sub rect
    /// - Returns: a new subdrawing, drawing internally to its parent
    public func createSubdrawing(normalizedRect: CGRect) -> Subdrawing {
        return Subdrawing(parent: self, normalizedRect: normalizedRect)
    }

}
