// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import CoreGraphics

/// a single sample of a stroke
public struct StrokeSample: Equatable {
    /// the location
    public let location: CGPoint
    /// whether it's coalesced
    public let coalescedSample: Bool

    /// the force
    public var force: CGFloat
    /// the maximum possible force.
    /// In theory, this value should never change as it is only dependent on the input device
    public let maxPossibleForce: CGFloat

    /// creates a new sample
    /// - Parameters:
    ///   - location: the location
    ///   - coalescedSample: whether its coalesced
    ///   - force: the force
    ///   - maximumPossibleForce: the max. possible force
    public init(location: CGPoint, coalescedSample: Bool, force: CGFloat, maximumPossibleForce: CGFloat) {
        self.location = location
        self.coalescedSample = coalescedSample
        self.force = force
        self.maxPossibleForce = maximumPossibleForce
    }

    func normalized(to rect: CGRect) -> StrokeSample {
        StrokeSample(
            location: location.convertTo(normalizedSubRect: rect),
            coalescedSample: coalescedSample,
            force: force,
            maximumPossibleForce: maxPossibleForce
        )
    }

    func normalized(from rect: CGRect) -> StrokeSample {
        StrokeSample(
            location: location.convertFrom(normalizedSubRect: rect),
            coalescedSample: coalescedSample,
            force: force,
            maximumPossibleForce: maxPossibleForce
        )
    }
}

/// Used to differentiate between the rendering into intermediate CALayers while still drawing and drawing a final image
public enum RenderMode {
    /// drawing to a CALayer
    case layerComposition

    /// collapsing multiple layers to an image to be drawn onscreen
    case layerCollapse

    /// drawing to a CGImage for submitting to the backend
    case finalImage
}

/// Enum representing the different drawing tools
public enum Tool: Equatable {
    /// a pen
    case pen

    /// a highlighter
    case highlighter

    /// an eraser
    case eraser

    /// Opacity at which the strokes are drawn when rendering while still drawing
    public var strokeDisplayOpacity: Float {
        switch self {
        case .pen:
            return 0.9
        case .highlighter:
            return 0.3
        case .eraser:
            return 0.0
		}
	}

	/// Opacity at which strokes are drawn when rendering to the final image
	public var strokeRenderOpacity: Float {
		switch self {
		case .pen:
			return 1.0
		case .highlighter:
			return 0.3
		case .eraser:
			return 0.0
		}
	}
}

/// A struct which assigns parameters to a tool
public struct Ink: Equatable {
    /// the tool
    public let tool: Tool

    /// the color for this tool
    public let color: CGColor

    /// the normalized line width
    public let normalizedLineWidth: CGFloat

    /// whether to use the pressure of the pencil
    public let pressureSensitiveLineWidth: Bool

    /// creates new ink using the given parameters
    /// - Parameters:
    ///   - tool: the tool
    ///   - color: the color
    ///   - normalizedLineWidth: the line width
    ///   - pressureSensitiveLineWidth: whether it uses the pencil's pressure
    public init(tool: Tool, color: CGColor, normalizedLineWidth: CGFloat, pressureSensitiveLineWidth: Bool) {
        self.tool = tool
        self.color = color
        self.normalizedLineWidth = normalizedLineWidth
        self.pressureSensitiveLineWidth = pressureSensitiveLineWidth
    }

    func normalized(to rect: CGRect) -> Ink {
        Ink(
            tool: tool,
            color: color,
            normalizedLineWidth: normalizedLineWidth / max(rect.size.width, rect.size.height),
            pressureSensitiveLineWidth: pressureSensitiveLineWidth
        )
    }

    func normalized(from rect: CGRect) -> Ink {
        Ink(
            tool: tool,
            color: color,
            normalizedLineWidth: normalizedLineWidth * max(rect.size.width, rect.size.height),
            pressureSensitiveLineWidth: pressureSensitiveLineWidth
        )
    }
}

/// a stroke which is a line drawn by a specific ink
public struct Stroke: Equatable {
    /// the individual line samples
    public let samples: [StrokeSample]
    /// the ink of the stroke
    public let ink: Ink

    /// creates a new stroke
    /// - Parameters:
    ///   - samples: the samples
    ///   - ink: the ink
    public init(samples: [StrokeSample], ink: Ink) {
        self.samples = samples
        self.ink = ink
    }

    /// empty stroke (used for convenience against optionals)
    public static let empty: Stroke = Stroke(
        samples: [],
        ink: Ink(tool: .pen, color: .clear, normalizedLineWidth: 0.0, pressureSensitiveLineWidth: false)
    )

    func normalized(to rect: CGRect) -> Stroke {
        Stroke(
            samples: samples.map({ $0.normalized(to: rect) }),
            ink: ink.normalized(to: rect)
        )
    }

    func normalized(from rect: CGRect) -> Stroke {
        Stroke(
            samples: samples.map({ $0.normalized(from: rect) }),
            ink: ink.normalized(from: rect)
        )
    }
}

extension Stroke {
    /// draws the stroke in the context, using a size (against which the de-normalization happens) and a render mode
    /// - Parameters:
    ///   - ctx: the context
    ///   - size: the target size
    ///   - renderMode: the render mode
    public func draw(in ctx: CGContext, withSize size: CGSize, renderMode: RenderMode = .layerComposition) {
        draw(in: ctx, withSize: size, referenceSize: size, renderMode: renderMode)
    }

    /// draws the stroke in the context, using a size (against which the de-normalization happens) and a render mode
    /// - Parameters:
    ///   - ctx: the context
    ///   - size: the target size
    ///   - renderMode: the render mode
    public func draw(
        in ctx: CGContext,
        withSize size: CGSize,
        referenceSize: CGSize,
        renderMode: RenderMode = .layerComposition
    ) {
        ctx.drawStrokeSamples(samples, ink: ink, size: size, referenceSize: referenceSize, renderMode: renderMode)
    }
}

extension CGContext {

    /// draw the specified stroke samples in this context
    /// - Parameters:
    ///   - samples: the samples
    ///   - ink: the ink
    ///   - size: the target size
    ///   - renderMode: the render mode
    public func drawStrokeSamples<S: Collection>(
        _ samples: S,
        ink: Ink,
        size: CGSize,
        renderMode: RenderMode = .layerComposition
    ) where S.Element == StrokeSample {
        drawStrokeSamples(samples, ink: ink, size: size, referenceSize: size, renderMode: renderMode)
    }

    /// draw the specified stroke samples in this context
    /// - Parameters:
    ///   - samples: the samples
    ///   - ink: the ink
    ///   - size: the target size
    ///   - renderMode: the render mode
    public func drawStrokeSamples<S: Collection>(
        _ samples: S,
        ink: Ink,
        size: CGSize,
        referenceSize: CGSize,
        renderMode: RenderMode = .layerComposition
    ) where S.Element == StrokeSample {
        // check if empty
        guard !samples.isEmpty else { return }

        // set color
        let color = ink.tool == .eraser ? .clear : ink.color

        switch renderMode {
        case .finalImage:
            setAlpha(CGFloat(ink.tool.strokeRenderOpacity))
        case .layerCollapse:
            setAlpha(CGFloat(ink.tool.strokeDisplayOpacity))
        case .layerComposition:
            break
        }

        setStrokeColor(color)

        let maxDimension = max(referenceSize.width, referenceSize.height)
        let lineWidth = ink.normalizedLineWidth * maxDimension

        setLineCap(.round)

        if !ink.pressureSensitiveLineWidth {
            setLineWidth(lineWidth)
        }

        if ink.tool != .eraser {
            beginTransparencyLayer(auxiliaryInfo: nil)
        }

        var lastSample: StrokeSample = samples.first!
        for sample in samples {
            let lastScaled = lastSample.location.scaled(to: size)
            let scaled = sample.location.scaled(to: size)
            move(to: lastScaled)
            addLine(to: scaled)
            // The following lien uses quad curves instead of simply connecting points, but this is too inefficient at such small scales.
            // We would need to downsample the line to reduce the amount of points, then draw a quad curve
            // addQuadCurve(to: scaled, control: lastScaled.midpointTowards(other: scaled))

            if ink.pressureSensitiveLineWidth {
                // We want the pressure sensitive line width to be normalized around the selected line width
                let lineWidthOffset: CGFloat = 0.5 // 50% offset
                let minLineWidth = lineWidth * (1.0 - lineWidthOffset)
                let maxLineWidth = lineWidth * (1.0 + lineWidthOffset)
                let widthRange = (maxLineWidth - minLineWidth)
                let lastTransposedForce = (lastSample.force / lastSample.maxPossibleForce) * widthRange + minLineWidth
                let sampleTransposedForce = (sample.force / sample.maxPossibleForce) * widthRange + minLineWidth
                setLineWidth((lastTransposedForce + sampleTransposedForce) / 2)
            }
            strokePath()
            lastSample = sample
        }

        if ink.tool != .eraser {
            endTransparencyLayer()
        }
    }
}
