// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine

/// as an exam is only identified by an integer, use int directly as type
public typealias ExamIdentifier = Int

extension ExamIdentifier: KitIdentifier {
    /// the identifier as string
    public var stringIdentifier: String {
        return "\(self)"
    }
}

class Exam: KitObject<ExamIdentifier> {
	var name: Observable<String?> = Observable()
	var date: Observable<Date?> = Observable()

	var correctionStatus: CorrectionStatusType
	var snippets: SnippetManagerType
	var solution: SolutionType
    var issues: ExamIssuesType
    var isKKK: Bool = false

	required init(identifier: ExamIdentifier, kit: Kit) {
		correctionStatus = kit.object(for: identifier) as CorrectionStatus
		snippets = kit.object(for: identifier) as SnippetManager
        issues = kit.object(for: identifier) as ExamIssues
		self.solution = kit.object(for: identifier) as Solution
		super.init(identifier: identifier, kit: kit)
	}

	func update(using model: ExamModel) {
		name.value = model.name
		if let modelDate = model.date {
			date.value = modelDate
		}
        self.isKKK = model.texOptions?.externalProblems ?? false
	}

	func load() {
		// this is a little counter intuitive...
		// but the exam name/date/etc is only available through the exam/ endpoint
		// therefore, the whole list has to be refreshed
		kit?.examList.load()
	}

	func correctionController(forProblem problem: Int, pass: Int) -> CorrectionControllerType {
		let filter = ResultFilter(problem: problem,
								  pass: pass,
								  corrected: .no,
								  available: .yes,
								  locked: .no,
								  access: .yes)
		return CorrectionController(identifier: self.identifier, filter: filter, kit: kit!)
	}

	func correctionController(withFilter filter: ResultFilter) -> CorrectionControllerType {
		return CorrectionController(identifier: self.identifier, filter: filter, kit: kit!)
	}

	func interactiveCorrectionPromiseController(forProblem problem: Int,
												pass: Int) -> InteractiveCorrectionPromiseControllerType {
		return InteractiveCorrectionController(exam: identifier, pass: pass, problem: .problem(problem), kit: kit!)
	}

    func interactiveCorrectionPromiseController(forProblem problem: Int,
                                                subproblem: String,
                                                pass: Int,
                                                filterLocked: Bool) -> InteractiveCorrectionPromiseControllerType {
        return InteractiveCorrectionController(
            exam: identifier,
            pass: pass,
            problem: .subproblem(problem, subproblem),
            filterLocked: filterLocked,
            kit: kit!
        )
    }

    func interactiveCorrectionPromiseController(forProblem problem: Int,
                                                subproblems: [String],
                                                pass: Int,
                                                filterLocked: Bool) -> InteractiveCorrectionPromiseControllerType {
        return InteractiveCorrectionController(
            exam: identifier,
            pass: pass,
            problems: subproblems.map { ProblemIdentifier.subproblem(problem, $0) },
            filterLocked: filterLocked,
            kit: kit!
        )
    }

    func getSingleCorrection(_ identifier: CorrectionIdentifier) -> CorrectionType {
        let correction: Correction = kit!.object(for: identifier)
        return correction
    }

	func freeAllLocksInBackground() {
		kit?.lockManager.freeLocksByDevice(exam: identifier)
	}

    func flags(filter: FlagFilter? = nil) -> FlagListType {
        let ident = FlagListIdentifier(examId: self.identifier)
        let list: FlagList = FlagList(identifier: ident, kit: kit!)
        list.filter = filter
        return list
    }

    func flagDefinitions() -> FlagDefinitionListType {
        let ident = FlagDefinitionListIdentifier(examId: self.identifier)
        let list: FlagDefinitionList = kit!.object(for: ident)
        return list
    }

    func lockList() -> LockListType {
        let ident = LockListIdentifier(examId: identifier)
        let list: LockList = kit!.object(for: ident)
        return list
    }

    func createFlag(text: String,
                    color: String,
                    erid: String,
                    problem: Int? = nil,
                    subproblem: String? = nil,
                    pass: Int? = nil) -> AnyPublisher<FlagType, Error> {
        let creationData = FlagCreationData(
            examId: self.identifier,
            erid: erid,
            color: color,
            text: text,
            problem: problem,
            subproblem: subproblem,
            correctionPass: pass
        )
        return Flag.create(
            data: creationData,
            in: kit!
        )
    }

    func createFlagDef(text: String,
                       color: String) -> AnyPublisher<FlagDefinitionType, Error> {
        let creationData = FlagDefinitionCreationData(
            exam: self.identifier,
            color: color,
            text: text
        )
        return FlagDefinition.create(
            data: creationData,
            in: kit!
        )
    }
}

extension Exam: ShortDescription {
	public var shortDescription: String {
		return "Exam[id=\(identifier),name=\(name.value ?? "(null)")]"
	}
}

extension Exam: ExamType { }
