// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine

/**
 The opaque type for an exam.

 It can be loaded using `load()` (see `Loadable`).

 You can create a correction controller using `correctionController(forProblem:pass:)` or `correctionController(withFilter:)`.
 For an interactive correction controller, use `interactiveCorrectionPromiseController(forProblem:pass:)`.

 It is generally a good practice to call `freeAllLocksInBackground()` after starting the app, in order to reset any orphan-locks.

 */
public protocol ExamType: AnyObject, Loadable {
    /// the identifier
    var identifier: ExamIdentifier { get }

    /// the name of the exam. available after using `load()` or when using the `ExamListType` to get the exam
    var name: Observable<String?> { get }

    /// the date of the exam. available after using `load()` or when using the `ExamListType` to get the exam
    var date: Observable<Date?> { get }

    /// the statistics for this exam
    var correctionStatus: CorrectionStatusType { get }

    /// the snippet manager which can be used to store (and reuse) commonly used feedback
    var snippets: SnippetManagerType { get }

	var solution: SolutionType { get }

    var issues: ExamIssuesType { get }

    var isKKK: Bool { get }

    /// creates a new correction controller for the specified problem/pass
    /// - Parameters:
    ///   - problem: the problem
    ///   - pass: the pass
    func correctionController(forProblem problem: Int, pass: Int) -> CorrectionControllerType

    /// creates a new interactive controller for the specified problem/pass
    /// - Parameters:
    ///   - problem: the problem
    ///   - pass: the pass
	func interactiveCorrectionPromiseController(forProblem problem: Int,
                                                pass: Int) -> InteractiveCorrectionPromiseControllerType

    /// creates a new interactive controller for the specified problem/subproblem/pass
    /// - Parameters:
    ///   - problem: the problem
    ///   - subproblem: the subproblem
    ///   - pass: the pass
    func interactiveCorrectionPromiseController(forProblem problem: Int,
                                                subproblem: String,
                                                pass: Int,
                                                filterLocked: Bool) -> InteractiveCorrectionPromiseControllerType

    func interactiveCorrectionPromiseController(forProblem problem: Int,
                                                subproblems: [String],
                                                pass: Int,
                                                filterLocked: Bool) -> InteractiveCorrectionPromiseControllerType

    /// creates a new correction controller with the specified filter
    /// - Parameter filter: the filter
    func correctionController(withFilter filter: ResultFilter) -> CorrectionControllerType

    func getSingleCorrection(_ identifier: CorrectionIdentifier) -> CorrectionType

    /// frees all locks made by this device in the background. should be called after launching the target
    func freeAllLocksInBackground()

    /// CorrectionFlags
    func flags(filter: FlagFilter?) -> FlagListType

    func flagDefinitions() -> FlagDefinitionListType

    func lockList() -> LockListType

    // swiftlint:disable:next function_parameter_count
    func createFlag(
        text: String,
        color: String,
        erid: String,
        problem: Int?,
        subproblem: String?,
        pass: Int?
    ) -> AnyPublisher<FlagType, Error>

    func createFlagDef(
        text: String,
        color: String
    ) -> AnyPublisher<FlagDefinitionType, Error>
}
