// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine

class FlagDefinition: KitObject<FlagDefinitionIdentifier> {

    // MARK: Properties
    var flagTypeId: Int
    var user: String?
    var color: String?
    var text: String?
    var count: Int?

    var editable: Bool {
        return true
    }

    var hasChanges: Bool = false

    var model: FlagDefinitionModel?

    var state: Observable<ResourceObjectState> = Observable(.idle)

    var submissionState: Observable<ResourceObjectState> = Observable(.idle)
    var submissionCompletion: ((Result<Void, Error>) -> Void)?
    var submissionResource: UpdateFlagDefinitionResource?

    var deletionState: Observable<ResourceObjectState> = Observable(.idle)
    var deletionCompletion: ((Result<Void, Error>) -> Void)?
    var deletionResource: DeleteFlagDefinitionResource?

    // MARK: - Initializers

    required init(identifier: FlagDefinitionIdentifier, kit: Kit) {
        self.flagTypeId = identifier.flagTypeId
        super.init(identifier: identifier, kit: kit)
    }

    func update(using model: FlagDefinitionModel) {
        self.model = model
        self.flagTypeId = model.flagTypeId
        self.user = model.user
        self.color = model.color
        self.text = model.text
        self.count = model.count
    }

    // MARK: - Loadable

    func load() {
        // Only way to obtain/update a flag is by loading the flag def list
        let list: FlagDefinitionList? = kit?.object(for: FlagDefinitionListIdentifier(examId: identifier.examId))
        list?.load()
    }

    // MARK: - Submitable

    func submitChanges() {
        guard let color = self.color, let text = self.text else {
            fatalError("Tried to submit an uninitialized flag definition")
        }
        let model = FlagDefinitionSubmissionModel(color: color, text: text)
        submissionResource = UpdateFlagDefinitionResource(identifier: identifier, model: model, in: kit!)
        submissionResource?.delegate = self
        submissionResource?.load()
        submissionState.value = .loading
    }

    // MARK: - Creatable

    static func creationResource(submission: FlagDefinitionCreationData, in kit: Kit) -> Resource<FlagDefinitionModel> {
        let model = FlagDefinitionSubmissionModel(color: submission.color, text: submission.text)
        return CreateFlagDefinitionResource(exam: submission.exam, flagDefinition: model, in: kit)
    }

    static func identifier(forCreationResult model: ModelType,
                           resource: Resource<ModelType>) -> FlagDefinitionIdentifier {
        guard let res = resource as? CreateFlagDefinitionResource else {
            fatalError("Expected CreateFlagDefinitionResource but got \(resource)")
        }
        let exam = res.exam
        return FlagDefinitionIdentifier(examId: exam, flagTypeId: model.flagTypeId)
    }

    // MARK: - Deletable

    func delete() {
        deletionResource = DeleteFlagDefinitionResource(identifier: identifier, in: kit!)
        deletionResource?.delegate = self
        deletionResource?.load()
        self.deletionState.value = .loading
    }

}

extension FlagDefinition: ResourceDelegate {
    func resource(_ resource: BaseResource, didFailWithError error: Error) {
        if resource == self.deletionResource {
            deletionState.value = .error(error)
            DispatchQueue.main.async {
                self.deletionCompletion?(.failure(error))
            }
        }
        if resource == self.submissionResource {
            submissionState.value = .error(error)
            DispatchQueue.main.async {
                self.submissionCompletion?(.failure(error))
            }
        }
    }

    func resourceDidLoad(_ resource: BaseResource) {
        if resource == self.deletionResource {
            deletionState.value = .loaded
            DispatchQueue.main.async {
                self.deletionCompletion?(.success(()))
            }
        }
        if resource == self.submissionResource {
            submissionState.value = .loaded
            DispatchQueue.main.async {
                self.submissionCompletion?(.success(()))
            }
        }
    }
}

extension FlagDefinition: FlagDefinitionType { }
extension FlagDefinition: Creatable {
    typealias ModelType = FlagDefinitionModel
    typealias CreationSubmissionType = FlagDefinitionCreationData

    static func create(data: FlagDefinitionCreationData, in kit: Kit) -> AnyPublisher<FlagDefinitionType, Error> {
        let pub: AnyPublisher<FlagDefinition, Error> = kit.createObject(withCreationSubmission: data)
        return pub.map({ $0 as FlagDefinitionType }).eraseToAnyPublisher()
    }
}

extension FlagDefinition {
    func addTo(exam: ExamIdentifier,
               erid: String,
               problem: Int? = nil,
               subproblem: String? = nil,
               pass: Int? = nil) -> AnyPublisher<FlagType, Error> {
        let creationData = FlagCreationData(examId: self.identifier.examId,
                                            flagTypeId: self.flagTypeId,
                                            erid: erid,
                                            problem: problem,
                                            subproblem: subproblem,
                                            correctionPass: pass)
        return Flag.create(data: creationData, in: kit!)
    }
}
