// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine

class Flag: KitObject<FlagIdentifier> {
    // MARK: Flag Type Implementation
    var flagTypeId: Int? { model?.flagTypeId }
    var flagId: Int? { model?.flagId }
    var color: String? { model?.color }
    var text: String? { model?.text }
    var erid: String? { model?.erid }

    var problem: Int? { model?.problem }
    var subproblem: String? { model?.subproblem }
    var correctionPass: Int? { model?.correctionPass }
    var user: String? { model?.user }

    var model: FlagModel?

    var state: Observable<ResourceObjectState> = Observable(.idle)

    var deletionState: Observable<ResourceObjectState> = Observable(.idle)
    var deletionCompletion: ((Result<Void, Error>) -> Void)?
    var deletionResource: DeleteFlagResource?

    var editable: Bool {
        return false
    }

    var hasChanges: Bool {
        return false
    }

    required init(identifier: FlagIdentifier, kit: Kit) {
        super.init(identifier: identifier, kit: kit)
    }

    func update(using model: FlagModel) {
        self.model = model
    }

    func load() {
        // Only way to obtain/update a flag is by loading the flag def list
        let list: FlagList? = kit?.object(for: FlagListIdentifier(examId: identifier.examId))
        list?.filter = nil
        list?.load()
    }

    // MARK: - Creatable

    static func creationResource(submission: FlagCreationData, in kit: Kit) -> Resource<FlagModel> {
        let model = FlagSubmissionModel(flagTypeId: submission.flagTypeId,
                                        color: submission.color,
                                        text: submission.text,
                                        erid: submission.erid,
                                        problem: submission.problem,
                                        subproblem: submission.subproblem,
                                        correctionPass: submission.correctionPass)
        return CreateFlagResource(exam: submission.examId, model: model, in: kit)
    }

    static func identifier(forCreationResult model: ModelType,
                           resource: Resource<ModelType>) -> FlagIdentifier {
        guard let res = resource as? CreateFlagResource else {
            fatalError("Expected CreateFlagDefinitionResource but got \(resource)")
        }
        let exam = res.exam
        return FlagIdentifier(examId: exam, flagTypeId: model.flagTypeId, flagId: model.flagId)
    }

    // MARK: - Deletable

    func delete() {
        deletionResource = DeleteFlagResource(flag: identifier, in: kit!)
        deletionResource?.delegate = self
        deletionResource?.load()
        self.deletionState.value = .loading
    }
}

extension Flag: ResourceDelegate {
    func resource(_ resource: BaseResource, didFailWithError error: Error) {
        if resource == self.deletionResource {
            deletionState.value = .error(error)
            deletionCompletion?(.failure(error))
        }
    }

    func resourceDidLoad(_ resource: BaseResource) {
        if resource == self.deletionResource {
            deletionState.value = .loaded
            deletionCompletion?(.success(()))
        }
    }
}

extension Flag: FlagType { }
extension Flag: Creatable {
    typealias ModelType = FlagModel
    typealias CreationSubmissionType = FlagCreationData

    static func create(data: FlagCreationData, in kit: Kit) -> AnyPublisher<FlagType, Error> {
        let pub: AnyPublisher<Flag, Error> = kit.createObject(withCreationSubmission: data)
        return pub.map({ $0 as FlagType }).eraseToAnyPublisher()
    }
}
