// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

class KeepaliveLockOperation: Operation {

    unowned let kit: Kit
    unowned let lockToken: LockToken

    private var keepAliveResource: AcquireLockResource?
    private var result: Result<Void, Error>?

    init(kit: Kit, lockToken: LockToken) {
        self.kit = kit
        self.lockToken = lockToken
    }

    override var isAsynchronous: Bool {
        return true
    }

    override var isExecuting: Bool {
        return result == nil
    }

    override var isFinished: Bool {
        return result != nil
    }

    override func cancel() {
        super.cancel()
        keepAliveResource = nil
    }

    override func start() {
        Logger.lockManager.info("[\(lockToken.identifier.stringIdentifier)] started keepalive lock operation")
        guard lockToken.state.value == .acquired else {
            self.cancel()
            return
        }

        let resource = AcquireLockResource(lockToken.identifier, in: kit)
        self.keepAliveResource = resource
        resource.refresh = true
        resource.delegate = self
        resource.load()
    }

}

extension KeepaliveLockOperation: ResourceDelegate {
    func resource(_ resource: BaseResource, didFailWithError error: Error) {
        willChangeValue(forKey: #keyPath(isExecuting))
        willChangeValue(forKey: #keyPath(isFinished))
        super.cancel()
        self.keepAliveResource = nil
        self.result = .failure(error)
        didChangeValue(forKey: #keyPath(isFinished))
        didChangeValue(forKey: #keyPath(isExecuting))
    }

    func resourceDidLoad(_ resource: BaseResource) {
        guard let keepAliveResource = keepAliveResource else { return }
        willChangeValue(forKey: #keyPath(isExecuting))
        willChangeValue(forKey: #keyPath(isFinished))
        if keepAliveResource.value?.token != nil {
            Logger.lockManager.info("[\(lockToken.identifier.stringIdentifier)] keep alive successful")
            result = .success(())
        } else {
            Logger.lockManager.error("[\(lockToken.identifier.stringIdentifier)] keep alive failed")
            lockToken.invalidate()
            result = .failure(NetworkError.invalidValue)
        }
        didChangeValue(forKey: #keyPath(isFinished))
        didChangeValue(forKey: #keyPath(isExecuting))
    }
}
