// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging
import BackgroundTasks

final class LockManager {
    var locks: Set<LockToken> = Set() {
        didSet {
            currentLocks.value = locks.sorted(by: \.erid)
        }
    }
    var currentLocks: Observable<[LockTokenType]> = .init()

    unowned var kit: Kit

    private var freeResources: Set<FreeLocksByDeviceResource> = Set()

    init(kit: Kit) {
        self.kit = kit
    }

    public func handleAppRefresh(_ task: BGTask) {
        Logger.lockManager.info("running keepalive background task")
        let queue = OperationQueue()

        let opcount = self.locks.count
        var counter = 0

        let ops = self.locks.map { lockToken -> KeepaliveLockOperation in
            let operation = KeepaliveLockOperation(kit: kit, lockToken: lockToken)
            operation.completionBlock = {
                counter += 1
                if counter == opcount {
                    task.setTaskCompleted(success: true)
                }
            }
            return operation
        }
        task.expirationHandler = {
            queue.cancelAllOperations()
        }

        queue.addOperations(ops, waitUntilFinished: true)

        // Schedule Lock Keepalive
        let bundleId = Bundle.main.bundleIdentifier!
        let request = BGAppRefreshTaskRequest(identifier: "\(bundleId).lock-keepalive")
        // Wait 20 min
        request.earliestBeginDate = Date(timeIntervalSinceNow: 20 * 60)
        do {
            try BGTaskScheduler.shared.submit(request)
            Logger.lockManager.info("registered BGAppRefreshTask for keepalive locks")
        } catch {
            Logger.lockManager.error("could not register BGAppRefreshTask to keepalive locks")
        }
    }

    func stopTimers() {
        self.locks.forEach { token in
            token.stopTimer()
        }
    }

    func startTimers() {
        self.locks.forEach { token in
            token.startTimer()
        }
    }

    func add(_ lock: LockToken) {
        locks.insert(lock)
        logLocks()
    }

    func remove(_ lock: LockToken) {
        locks.remove(lock)
        logLocks()
    }

    func logLocks() {
        guard !locks.isEmpty else {
            Logger.lockManager.info("no active locks")
            return
        }
        let string = locks.sorted(by: \.erid).map { $0.erid }.joined(separator: ", ")
        Logger.lockManager.info("active locks: \(string)")
    }

    func freeLocksByDevice(exam: ExamIdentifier) {
        let resource = FreeLocksByDeviceResource(exam, in: kit)
        freeResources.insert(resource)

        resource.delegate = self
        resource.load()
    }

    func notifyLocks(withExam exam: ExamIdentifier) {
        for lock in locks where lock.identifier.exam == exam {
            lock.invalidate()
        }
    }
}

extension LockManager: ResourceDelegate {
    func resourceDidLoad(_ resource: BaseResource) {
        guard let resource = resource as? FreeLocksByDeviceResource else {
            return
        }
        if resource.value == true {
            notifyLocks(withExam: resource.identifier)
        } else {
            Logger.lockManager.error("Failed to free exam \(resource.identifier)")

        }
    }

    func resource(_ resource: BaseResource, didFailWithError error: Error) {
        let exam: String
        if let resource = resource as? FreeLocksByDeviceResource {
            exam = "\(resource.identifier)"
        } else {
            exam = "(unknown)"
        }
        Logger.lockManager.error("Failed to free exam \(exam)")
    }
}

extension LockManager: LockManagerType { }
