// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

class LockReference {
    weak var token: LockToken?
    var state: Observable<LockState>

    init(_ token: LockToken) {
        self.token = token
        self.state = token.state // this can be a strong reference (no optional)
    }

    deinit {
        token?.remove(self)
    }
}

extension LockReference: LockReferenceType { }

extension LockReference: CustomDebugStringConvertible {
    var debugDescription: String {
        "<LockReference: state = \(self.state.value), token = \(self.token?.token ?? "nil")>"
    }
}

class LockToken: KitObject<LockIdentifier> {
    var acquireResource: AcquireLockResource?
    var keepAliveResource: AcquireLockResource?
    var freeResource: FreeLockResource?

    var state: Observable<LockState> = Observable(.unknown)

    internal var token: String?

    private var references: [WeakBox<LockReference>] = []
    private var keepAliveTimer: Timer?

    func add() -> LockReference {
        let ref = LockReference(self)
        references.append(WeakBox(ref))

        Logger.lockManager.info("[\(identifier.stringIdentifier)] add ref \(pointerString(ref)) - count @ \(references.count)")

        if state.value != .acquired {
            acquire()
        }
        return ref
    }

    func remove(_ reference: LockReference) {
        // make sure the array is not empty
        guard !references.isEmpty else { return }
        references = references.filter({ $0.value !== reference && $0.value != nil })

        Logger.lockManager.info("[\(identifier.stringIdentifier)] remove ref \(pointerString(reference)) - count @ \(references.count)")

        if references.isEmpty {
            // -> last reference removed
            free()
        }
    }

    func acquire() {
        guard acquireResource == nil else { return }
        guard let kit = kit else { return }

        kit.lockManager.add(self)

        let acquireResource = AcquireLockResource(identifier, in: kit)
        self.acquireResource = acquireResource

        acquireResource.delegate = self
        acquireResource.load()
        Logger.lockManager.info("[\(identifier.stringIdentifier)] acquiring lock")
    }

    func free() {
        guard freeResource == nil else { return }
        guard let kit = kit else { return }

        // Moved to `freeResourceDidLoad()`
        kit.lockManager.remove(self)

        let freeResource = FreeLockResource(identifier, in: kit)
        self.freeResource = freeResource

        freeResource.delegate = self
        freeResource.load()
        Logger.lockManager.info("[\(identifier.stringIdentifier)] freeing lock")
    }

    func startTimer() {
        guard RunLoop.current == RunLoop.main else {
            DispatchQueue.main.async {
                self.startTimer()
            }
            return
        }
        if let current = self.keepAliveTimer {
            current.invalidate()
            self.keepAliveTimer = nil
        }

        let timer = Timer.scheduledTimer(withTimeInterval: 120.0, repeats: true) { [weak self] _ in
            self?.timerFired()
        }
        self.keepAliveTimer = timer
    }

    func stopTimer() {
        guard RunLoop.current == RunLoop.main else {
            DispatchQueue.main.async {
                self.stopTimer()
            }
            return
        }
        guard let current = self.keepAliveTimer else { return }
        current.invalidate()
        self.keepAliveTimer = nil
    }

    func timerFired() {
        sendKeepAlive()
    }

    func acquireResourceDidLoad() {
        guard let acquireResource = acquireResource else { return }
        if let token = acquireResource.value?.token {
            Logger.lockManager.info("[\(identifier.stringIdentifier)] acquired")
            self.token = token
            state.value = .acquired
            startTimer()
        } else {
            Logger.lockManager.error("[\(identifier.stringIdentifier)] already taken")
            self.token = nil
            state.value = .taken
            // Reset resource so that we can retry
            self.acquireResource = nil
            stopTimer()
        }
    }

    func keepAliveResourceDidLoad() {
        guard let keepAliveResource = keepAliveResource else { return }
        if keepAliveResource.value?.token != nil {
            Logger.lockManager.info("[\(identifier.stringIdentifier)] keepalive successful")
        } else {
            Logger.lockManager.error("[\(identifier.stringIdentifier)] keepalive failed")

            invalidate()
        }
    }

    func freeResourceDidLoad() {
        guard let freeResource = freeResource else { return }
        if let success = freeResource.value {
            if success {
                Logger.lockManager.info("[\(identifier.stringIdentifier)] free sucessful")
                invalidate(freed: true)
            } else {
                Logger.lockManager.error("[\(identifier.stringIdentifier)] free failed")
            }
            // kit?.lockManager.remove(self)
        }
    }

    func sendKeepAlive() {
        guard state.value == .acquired else {
            stopTimer()
            return
        }
        guard let kit = kit else { return }

        let resource = AcquireLockResource(identifier, in: kit)
        self.keepAliveResource = resource
        resource.refresh = true
        resource.delegate = self
        resource.load()
    }

    func invalidate(freed: Bool = false) {
        self.token = nil
        self.acquireResource = nil
        self.freeResource = nil
        self.state.value = freed ? .freed : .unknown
        stopTimer()
    }
}

extension LockToken: ResourceDelegate {
    func resourceDidLoad(_ resource: BaseResource) {
        if resource === acquireResource {
            acquireResourceDidLoad()
        } else if resource === keepAliveResource {
            keepAliveResourceDidLoad()
        } else if resource === freeResource {
            freeResourceDidLoad()
        }
    }

    func resource(_ resource: BaseResource, didFailWithError error: Error) {
        if resource === acquireResource {
            acquireResourceDidLoad()
        } else if resource === keepAliveResource {
            keepAliveResourceDidLoad()
        } else if resource === freeResource {
            freeResourceDidLoad()
        }
        Logger.lockManager.error("[\(identifier.stringIdentifier)] \(resource) failed with error \(error)")
    }
}

extension LockToken: LockTokenType {
    var erid: String { identifier.erid }
}
