// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import CoreGraphics
import Combine
import Logging

class LayerImage: KitObject<CorrectionPageLayerIdentifier> {
    var image: Observable<Image?> = Observable()
    var state: Observable<ResourceObjectState> = Observable(.idle)
    var underlyingScanSize: CGSize? {
        didSet {
            self.image.value = scaledImageIfRequired(self.image.value)
        }
    }

    lazy var resource: LayerImageResource = createDelegatedResource()

    func scaledImageIfRequired(_ image: Image?) -> Image? {
        // return nil if no image is available yet
        guard let image = image else { return nil }

        // check if there's a underlying scan size set
        guard let scanImageSize = underlyingScanSize else {
            return image
        }

        // get the image size of the layer, and check if matching
        let layerImageSize = image.size
        guard layerImageSize != scanImageSize else {
            return image
        }

        // not matching -> resize
        Logger.correction.debug("[LayerImage] layer size mismatch, will scale to match size (scan = \(scanImageSize), layer = \(layerImageSize)")

        guard let cgImage = image.cgImage else {
            return image
        }

        guard let result = cgImage.resize(to: scanImageSize) else {
            Logger.correction.error("layer image resize failed")
            return image
        }

        return Image(cgImage: result)
    }

    func invalidate() {
        image.value = nil
        resource = createDelegatedResource()
        state.value = .idle
    }
}

extension LayerImage: ResourceKitObjectType {
    func createResource() -> LayerImageResource {
        return LayerImageResource(identifier, in: kit!)
    }

    func update(using resource: LayerImageResource) {
        image.value = resource.createImage()
    }
}

extension LayerImage: LoadableImageType { }
