// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import CoreGraphics
import ImageIO
#if os(iOS)
import MobileCoreServices
#endif
import Logging

class Layer: KitObject<UnmanagedIdentifier<CorrectionPageLayerIdentifier>> {
    let layerImage: LayerImage
    /// This indicates that we got a 404 for the case the we didnt get the scan size yet, but already got the layer 404
    private var blankFlag = false
    var underlyingScanSize: CGSize? {
        didSet {
            layerImage.underlyingScanSize = underlyingScanSize
            tryCreatingDrawing(blank: blankFlag)
        }
    }

    var drawing: Observable<Drawing?> = .init()

    var editable: Bool = false

    var layerStateObserver: AnyCancellable?

    private var _hasChanges: Bool = false
    var hasChanges: Bool {
        guard let drawing = self.drawing.value else {
            return false
        }
        return _hasChanges && drawing.elements.count > 0
    }

    required init(identifier: UnmanagedIdentifier<CorrectionPageLayerIdentifier>, kit: Kit) {
        layerImage = LayerImage(identifier: identifier.identifier, kit: kit)
        super.init(identifier: identifier, kit: kit)

        layerStateObserver = layerImage.state.sink { [weak self] state in
            if case let .error(error) = state,
               let err = error as? NetworkError,
               err == .notFound {
                // Layer returned 404 not found, we need to create the layer ourselves
                Logger.correction.info("[Layer|\(identifier.identifier.erid)] got 404 for page \(identifier.identifier.page) layer \(identifier.identifier.layer), creating blank drawing")
                self?.blankFlag = true
                self?.tryCreatingDrawing(blank: true)
                return
            }

            guard state == .loaded else {
                return
            }
            Logger.correction.debug("[Layer|\(identifier.identifier.erid)] got image for page \(identifier.identifier.page) layer \(identifier.identifier.layer), creating drawing")
            self?.tryCreatingDrawing(blank: false)
        }
    }

    func createSubmission() -> CorrectionSubmission.CorrectionImageSubmission? {
        // ensure drawing and changes
        guard let drawing = drawing.value, drawing.elements.count > 0 else { return nil }
        // ensure image
        guard let image = drawing.getImage() else {
            // swiftlint:disable:next line_length
            Logger.submission.error("[Layer|\(identifier.identifier.erid)] could not create image for drawing \(identifier.stringIdentifier)")
            return nil
        }

        guard let data = CFDataCreateMutable(nil, 0) else { return nil }
        guard let destination = CGImageDestinationCreateWithData(data, kUTTypePNG, 1, nil) else { return nil }
        CGImageDestinationAddImage(destination, image, nil)
        CGImageDestinationFinalize(destination)

        let page = identifier.identifier.page
        let layer = identifier.identifier.layer

        let submission = CorrectionSubmission.CorrectionImageSubmission(page: page, layer: layer, data: data as Data)
        return submission
    }

    func submissionDidFinish() {
        _hasChanges = false
    }

    func tryCreatingDrawing(blank: Bool) {
        // don't create another one
        guard drawing.value == nil else {
            Logger.correction.debug("[Layer|\(identifier.identifier.erid)] could not create drawing, drawing already exists for page \(identifier.identifier.page) layer \(identifier.identifier.layer)")
            return
        }

        // check necessary components
        guard let scanSize = underlyingScanSize else {
            Logger.correction.debug("[Layer|\(identifier.identifier.erid)] could not create drawing, scan size is missing for page \(identifier.identifier.page) layer \(identifier.identifier.layer)")
            return
        }

        if !blank && layerImage.state.value != .loaded {
            Logger.correction.debug("[Layer|\(identifier.identifier.erid)] could not create drawing, layer image missing for page \(identifier.identifier.page) layer \(identifier.identifier.layer)")
            return
        }

        let image: CGImage? = layerImage.image.value?.cgImage

        let drawing = Drawing(size: scanSize, originalImage: image)
        drawing.delegates.add(self)
        self.drawing.value = drawing
    }
}

extension Layer: LockStatePropagation {
    func lockStateChanged(_ lockState: LockState) {
        editable = lockState.editable
    }

    func invalidate() {
        layerImage.invalidate()
        drawing.value = nil
    }
}

extension Layer: DrawingDelegate {
    func drawing(_ drawing: Drawing, didAppendElement element: Drawing.Element) {
        _hasChanges = true
    }

    func drawing(_ drawing: Drawing, didRemoveLastElement element: Drawing.Element) {
        _hasChanges = true
    }
}

extension Layer: LayerType {
    var pass: Int {
        return identifier.identifier.layer
    }

    var image: LoadableImageType {
        return layerImage
    }
}
