// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import CoreGraphics
import Logging

class Page: KitObject<UnmanagedIdentifier<CorrectionPageIdentifier>> {
    let scanImage: ScanImage
    let rawImage: RawImage
    let personalizedSolutionImage: PersonalizedSolutionImage

	var solutionImage: SolutionImage? {
        (solutionPage.value as? SolutionPage)?.solutionImage
	}

	var solutionPage: Observable<SolutionPageType>

	var solutionImageObserver: AnyCancellable?

    var layers: Observable<[LayerType]> = Observable()

    var scanImageObserver: AnyCancellable?
    var rawImageObserver: AnyCancellable?
    var personalizedSolutionImageObserver: AnyCancellable?
    var emptyBoxObserver: AnyCancellable?

    var problems: Observable<[ProblemType]> = Observable()

    lazy var segmentation: PageSegmentationType? = PageSegmentation(page: self)

    var solutionObservation: AnyCancellable?

    var pageNumber: Int {
        return identifier.identifier.page
    }

	var pageSize: CGSize = .zero {
		didSet {
            (solutionPage.value as? SolutionPage)?.pageSize = pageSize
		}
	}

    var empty: Bool = false {
        didSet {
            if empty {
                self.emptyBox.value?.setSelected(false, allowModificationToAll: false)
            }
            updateEmpty()
        }
    }

    var emptyBox: Observable<EmptyBoxType?> = Observable()

    var model: CorrectionPageModel?

    var editable: Bool = false

    weak var correction: Correction? {
        didSet {
            let solId = SolutionPageIdentifier(exam: identifier.identifier.exam,
                                               page: identifier.identifier.page)
            if let correction = self.correction {
                let solutionId = correction.identifier.exam
                let solution: Solution? = kit!.object(for: solutionId)
                self.solutionObservation = solution?.pages.sink(receiveValue: { [weak self] _ in
                    guard let solutionPage: SolutionPage = self?.kit!.object(for: solId) else {
                        return
                    }
                    self?.solutionPage.value = solutionPage
                })
            }
        }
    }

	required init(identifier: UnmanagedIdentifier<CorrectionPageIdentifier>, kit: Kit) {
		scanImage = ScanImage(identifier: identifier.identifier, kit: kit)
		rawImage = RawImage(identifier: identifier.identifier, kit: kit)
        personalizedSolutionImage = PersonalizedSolutionImage(identifier: identifier.identifier, kit: kit)
		let solId = SolutionPageIdentifier(exam: identifier.identifier.exam,
										   page: identifier.identifier.page)

        let solutionPage: SolutionPage = kit.object(for: solId)
        self.solutionPage = Observable(solutionPage)

		super.init(identifier: identifier, kit: kit)

        solutionImageObserver = solutionPage.state.sink { state in
            guard state == .loaded else { return }
            solutionPage.updateImage(to: solutionPage.image.value)
		}

		// add an scan image observer, as the scan will be used for sizing
        scanImageObserver = scanImage.state.sink { [weak self] state in
            // Logger.correction.debug("[Page|\(self?.identifier.stringIdentifier ?? "nil")] scan image state changed to \(state)")
            guard state == .loaded else { return }
            self?.scanImageChanged(self?.scanImage.image.value)
        }
    }

    func update(correctionModel: CorrectionDataModel,
                pageModel: CorrectionPageModel,
                problems: [Problem],
                emptyBox: EmptyBox?) {
        self.model = pageModel

        if pageModel.landscape {
        	self.pageSize = CGSize(width: correctionModel.pageHeight, height: correctionModel.pageWidth)
        } else {
            self.pageSize = CGSize(width: correctionModel.pageWidth, height: correctionModel.pageHeight)
        }

        self.layers.value = (1...correctionModel.passes).map({
            let layerIdentifier = CorrectionPageLayerIdentifier(pageIdentifier: identifier.identifier, layer: $0)
            return Layer(identifier: layerIdentifier.unmanaged(owner: self), kit: kit!)
        })

        self.problems.value = problems
        self.emptyBox.value = emptyBox

        self.emptyBoxObserver = emptyBox?.selected.sink(receiveValue: { [unowned self] selected in
            if selected {
                self.empty = false
            }
        })
    }

    func createSubmissions() -> [CorrectionSubmission.CorrectionImageSubmission] {
        layers.value.compactMap { ($0 as? Layer)?.createSubmission() }
    }

    func submissionDidFinish() {
        for layer in layers.value {
            if let layer = layer as? Layer {
                layer.submissionDidFinish()
            }
        }
    }

    func updateEmpty() {
        guard editable else { return }
        for problem in self.problems.value {
            if let problem = problem as? Problem {
                problem.empty = empty
            }
        }
    }

    func scanImageChanged(_ image: Image?) {
        // Logger.correction.debug("[Page|\(self.identifier.stringIdentifier)] scan image changed to \(image.debugDescription)")
        guard let image = image else {
			return
		}

        let imageSize = image.size
        assert(image.scale == 1.0, "backend provided image has a scale other than 1.0")

        for layer in layers.value {
			layer.underlyingScanSize = imageSize
        }
    }
}

extension Page: LockStatePropagation {
    func lockStateChanged(_ lockState: LockState) {
        editable = lockState.editable
        for layer in layers.value {
            if let casted = layer as? LockStatePropagation {
                casted.lockStateChanged(lockState)
            }
        }

        scanImage.lockStateChanged(lockState)
        solutionImage?.lockStateChanged(lockState)
        rawImage.lockStateChanged(lockState)
        personalizedSolutionImage.lockStateChanged(lockState)

        // Discard scan image when lock is invalidated
        if scanImage.resource.state == .loaded && lockState == .unknown {
            scanImage.invalidate()
        }
        if let sol = solutionImage,
            sol.resource.state == .loaded && lockState == .unknown {
            sol.invalidate()
        }
    }

    func invalidate() {
        for layer in layers.value {
            if let casted = layer as? LockStatePropagation {
                casted.invalidate()
            }
        }
        scanImage.invalidate()
        solutionImage?.invalidate()
        rawImage.invalidate()
        personalizedSolutionImage.invalidate()
    }
}

extension Page: PageType {
    var pageIdentifier: CorrectionPageIdentifier {
        return self.identifier.identifier
    }

    var erid: String? {
        return correction?.erid
    }

    var containingCorrection: CorrectionType? {
        return correction
    }

    var scan: LoadableImageType {
        return scanImage
    }

    var rawScan: LoadableImageType {
        return rawImage
    }

    var solution: LoadableImageType? {
        return solutionImage
    }

    var personalizedSolution: LoadableImageType? {
        return personalizedSolutionImage
    }

    var focusedProblem: Int? {
        return correction?.focusedProblem
    }

    var focusedPass: Int? {
        return correction?.focusedPass
    }

    func prefetch() {
        Logger.correction.debug("[Page|\(self.identifier.stringIdentifier)] prefetch() called")
        scanImage.load()
        personalizedSolutionImage.load()
        solutionImage?.load()
        (solutionPage.value as? SolutionPage)?.load()
        for layer in layers.value {
            if let focusedPass = focusedPass, layer.pass > focusedPass {
                continue
            }
            guard let pageModel = self.model,
                pageModel.layers.contains(layer.pass) else {
                    return
            }
            layer.image.load()
        }
    }

    var hasChanges: Bool {
        return layers.value.contains(\.hasChanges) || (emptyBox.value?.hasChanges ?? false)
    }
}

extension Page: CorrectionPageType {}
