// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import CoreGraphics
import ImageIO
import UIKit
import Logging
import Nuke

/// A remote image which can be loaded
public protocol LoadableImageType: StatefulLoadable {
    /// the image (observable)
    var image: Observable<Image?> { get }
}

/// A single page in a correction
public protocol PageType: Editable, Prefetchable {
    var pageIdentifier: CorrectionPageIdentifier { get }

    /// the page number
    var pageNumber: Int { get }
    var erid: String? { get }

    var containingCorrection: CorrectionType? { get }

    /// a loadable image for the scan of this page
    var scan: LoadableImageType { get }

    /// the raw unaltered scan of this page
    var rawScan: LoadableImageType { get }

    /// a loadable image for the solution of this page (if available)
    var solution: LoadableImageType? { get }

    var personalizedSolution: LoadableImageType? { get }

    var segmentation: PageSegmentationType? { get }

    /// the focused problem of the correction to which this page belongs to
    var focusedProblem: Int? { get }

    /// the focused pass of the correction to which this page belongs to
    var focusedPass: Int? { get }

    /// the observable layers for this page
    var layers: Observable<[LayerType]> { get }

    /// the problems on this page
    var problems: Observable<[ProblemType]> { get }

    /// the page's size
    var pageSize: CGSize { get }

    /// whether or not the page is completely empty
    var empty: Bool { get set }

    /// A page empty box, if any
    var emptyBox: Observable<EmptyBoxType?> { get }
}

public extension PageType {
    /// returns the layer corresponding to the specified pass
    /// - Parameter pass: the pass of the layer
    /// - Returns: the layer, if available
    func layer(pass: Int) -> LayerType? {
        return layers.value.first { $0.pass == pass }
    }
}
