// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import Foundation
import CoreGraphics

#if os(iOS)
import UIKit
#endif

struct ImageData {
    let width: Int
    let height: Int
    let bytesPerPixel: Int
    let data: [UInt8]
}

extension Image {
    func pixelData(of rect: CGRect) -> ImageData? {
        let width = Int(rect.size.width)
        let height = Int(rect.size.height)

        let colorSpace = CGColorSpaceCreateDeviceGray()
        let dataSize = width * height * colorSpace.numberOfComponents
        var pixelData = [UInt8](repeating: 0, count: Int(dataSize))
        let context = CGContext(data: &pixelData,
                                width: width,
                                height: height,
                                bitsPerComponent: 8,
                                bytesPerRow: colorSpace.numberOfComponents * width,
                                space: colorSpace,
                                bitmapInfo: CGImageAlphaInfo.none.rawValue)
        guard let ctx = context else { return nil }

        guard let cgImage = self.cgImage else { return nil }

        let rect = CGRect(x: -rect.origin.x, y: -rect.origin.y, width: self.size.width, height: self.size.height)

        ctx.draw(cgImage, in: rect)

        return ImageData(width: width, height: height, bytesPerPixel: colorSpace.numberOfComponents, data: pixelData)
    }
}

class PageSegmentation {
    static let queue = DispatchQueue.init(label: "segmentation", qos: .userInitiated)
    // TODO: what if the app keeps a strong ref (which it probably will)
    weak var page: Page?

    var state: Observable<ResourceObjectState> = .init(.idle)

    static let bindingEdgeInset: CGFloat = 0.07
    var imageHeight: CGFloat = 0.0
    //	private let pageScaleFactor: Scale

    var croppedTargetRect: CGRect?
    var rowIntensities: [Double]?
    var ranges: [ClosedRange<Int>]?
    var segments: Observable<[Segment]> = .init()

    var image: Image?
    var imageObserver: AnyCancellable?

    init(page: Page) {
        self.page = page
    }

    func load() {
        guard state.value == .idle else { return }
        state.value = .loading

        // check if KKK template
        if let page = self.page,
           page.kit!.exam(withId: page.pageIdentifier.exam).isKKK {
            // This is a KKK template exam, don't perform segmentation
            return
        }

        performSegmentation()
    }

    func performSegmentation() {
        Self.queue.async { [weak self] in
            self?.performSegmentationInBackground()
        }
    }

    func performSegmentationInBackground() {
        if let image = page?.scanImage.image.value {
            // image already available -> start right away
            imageAvailable(image)
        } else {
            // image not yet available -> observe
            imageObserver = page?.scanImage.state.receive(on: Self.queue).sink { [weak self] _ in
                if let image = self?.page?.scanImage.image.value {
                    self?.imageAvailable(image)
                }
            }
        }
    }

    func imageAvailable(_ image: Image) {
        self.image = image

        // discard image observer
        imageObserver = nil

        // save image height
        imageHeight = image.size.height

        // calculate the rect (without the binding edge)
        let rect = calculateCroppedTargetRect()
        croppedTargetRect = rect

        guard let rows = calculateRowIntensities(inRect: rect) else { return }
        rowIntensities = rows

        let ranges = findRanges(in: rows)
        self.ranges = ranges

        self.segments.value = convertRangesToSegments(ranges)
        state.value = .loaded
    }

    func calculateCroppedTargetRect() -> CGRect {
        guard let image = self.image else { fatalError("no image") }

        var rect = CGRect(origin: .zero, size: image.size)
        // try to get the first non-multiple choice box
        let problem = page?.problems.value.first(.credits, for: \.evaluationType)
        if let frame = problem?.normalizedFrame {
            if frame.minX > 0.5 {
                // boxes on the right side -> binding edge on the left, so adjust x offset
                rect.origin.x += rect.size.width * PageSegmentation.bindingEdgeInset
            }
            // otherwise, don't adjust.
        }

        rect.size.width -= (rect.size.width * PageSegmentation.bindingEdgeInset)

        return rect
    }

    func calculateRowIntensities(inRect rect: CGRect) -> [Double]? {
        guard let image = self.image else { fatalError("no image") }

        guard let imageData = image.pixelData(of: rect) else { return nil }

        return imageData.data.chunked(into: imageData.width).map { (row) -> Double in
            let sum = row.reduce(0) { (current, pixel) -> Double in
                current + (Double(pixel) / Double(UInt8.max))
            }
            return 1.0 - (sum / Double(row.count))
        }
    }

    let baseThreshold: Double = 0.0001
    let horizontalLineThreshold: Double = 0.25

    func findRanges(in rows: [Double]) -> [ClosedRange<Int>] {
        var ranges: [ClosedRange<Int>] = []

        var currentLower: Int?

        for (idx, value) in rows.enumerated() {
            if value > baseThreshold, currentLower == nil {
                currentLower = idx
            } else if value < baseThreshold, let lower = currentLower {
                ranges.append(lower...idx)
                currentLower = nil
            }
        }

        return ranges
    }

    func convertRangesToSegments(_ ranges: [ClosedRange<Int>]) -> [Segment] {
        var segments: [Segment] = []
        guard let problems = page?.problems.value else { return segments }
        for problem in problems {
            guard let problem = problem as? Problem else { continue }
            switch problem.evaluationType {
            case .multipleChoice:
                if let segment = createMultipleChoiceSegment(for: problem) {
                    segments.append(.multipleChoice(segment))
                }
            case .credits:
                if let segment = createTextAnswerSegment(for: problem) {
                    segments.append(.textAnswer(segment))
                }
            }
        }
        return segments
    }

    func normalize(_ range: ClosedRange<Int>) -> ClosedRange<CGFloat> {
        return (CGFloat(range.lowerBound) / imageHeight)...(CGFloat(range.upperBound) / imageHeight)
    }

    func createMultipleChoiceSegment(for problem: Problem) -> MultipleChoiceSegment? {
        let frame = problem.normalizedFrame
        let searchRange = Int(frame.minY * imageHeight)...Int(frame.maxY * imageHeight)

        if let match = matchingRange(for: searchRange) {
            var questionRange: ClosedRange<CGFloat>?
            if let idxMatch = ranges?.lastIndex(of: match), idxMatch > 0,
                let range = ranges?[idxMatch-1] {
                questionRange = normalize(range)
            }
            let overallRange = min(match.lowerBound, searchRange.lowerBound)...max(match.upperBound, searchRange.upperBound)

            return MultipleChoiceSegment(
                problem: problem.identifier,
                range: normalize(overallRange),
                questionRange: questionRange,
                optionAreaRange: normalize(match)
            )
        } else {
            return MultipleChoiceSegment(
                problem: problem.identifier,
                range: normalize(searchRange),
                questionRange: nil,
                optionAreaRange: normalize(searchRange)
            )
        }
    }

    func createTextAnswerSegment(for problem: Problem) -> TextAnswerSegment? {
        let frame = problem.normalizedFrame
        let searchRange = Int(frame.minY * imageHeight)...Int(frame.maxY * imageHeight)

        //print("Matching \(searchRange)")
        var segmentRange = searchRange
        var areaRange: ClosedRange<Int>?
        if let match = matchingRange(for: searchRange) {
            segmentRange = match

            if let rows = rowIntensities?[match].enumerated() {
                for (idx, value) in rows where value > horizontalLineThreshold {
                    areaRange = (match.lowerBound+idx)...match.upperBound
                    //print("-> Found \(areaRange!)")
                    break
                }
            }
        }

        let normalizedAreaRange: ClosedRange<CGFloat>?
        if let areaRange = areaRange {
            normalizedAreaRange = normalize(areaRange)
        } else {
            normalizedAreaRange = nil
        }

        return TextAnswerSegment(
            problem: problem.identifier,
            problemRange: normalize(searchRange),
            range: normalize(segmentRange),
            textAreaRange: normalizedAreaRange
        )
    }

    fileprivate func matchingRange(for range: ClosedRange<Int>) -> ClosedRange<Int>? {
        guard let ranges = self.ranges else { return nil }

        let match = ranges.first { (known) -> Bool in
            known.overlaps(range)
        }

        return match
    }
}

extension PageSegmentation: PageSegmentationType { }
