// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import CoreGraphics

/// represents a text answer segment
public struct TextAnswerSegment: Codable {
    /// the problem
    public let problem: ProblemIdentifier
    /// the problem's range
    public let problemRange: ClosedRange<CGFloat>

    /// the segment's range
    public var range: ClosedRange<CGFloat>

    /// the range of the text area
    public var textAreaRange: ClosedRange<CGFloat>?
}

/// represents a multiple choice segment
public struct MultipleChoiceSegment: Codable {
    /// the associated problem
    public let problem: ProblemIdentifier

    /// the range of the whole segment
    public var range: ClosedRange<CGFloat>
    /// the range of the question text
    public var questionRange: ClosedRange<CGFloat>?
    /// the range of the answer area
    public var optionAreaRange: ClosedRange<CGFloat>
}

/**
 represents a segment inferred by the `PageSegmentation`

 can currently be either text or multiple choice
 */
public enum Segment: Codable {
    enum CodingKeys: CodingKey {
        case textAnswer
        case multipleChoice
    }

    /// a text segment
    case textAnswer(TextAnswerSegment)

    /// a multiple choice segment
    case multipleChoice(MultipleChoiceSegment)

    public func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        switch self {
        case .textAnswer(let value):
            try container.encode(value, forKey: .textAnswer)
        case .multipleChoice(let value):
            try container.encode(value, forKey: .multipleChoice)
        }
    }

    public init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: CodingKeys.self)
        do {
            let value =  try container.decode(TextAnswerSegment.self, forKey: .textAnswer)
            self = .textAnswer(value)
        } catch {
            let value =  try container.decode(MultipleChoiceSegment.self, forKey: .multipleChoice)
            self = .multipleChoice(value)
        }
    }
}

/// the opaque type for the page segmentation
public protocol PageSegmentationType: StatefulLoadable {
    /// the segments which could be inferred on the page (observable)
    var segments: Observable<[Segment]> { get }
}
