// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import CoreGraphics

class CorrectionBox: CorrectionBoxType {
    let identifier: String
    let value: Double
    let normalizedFrame: CGRect

    var originalSelected: Bool
    var selected: Observable<Bool>
    var editable: Bool = true
    var editabilityWasOverridden: Bool = false

    var empty: Bool = false {
        didSet { updateEmpty() }
    }

    unowned var problemPass: ProblemPass

    var hasChanges: Bool {
        return originalSelected != selected.value
    }

    var pass: Int {
        return problemPass.pass
    }

    init(
        using model: CorrectionBoxModel,
        in correction: CorrectionDataModel,
        problemPass: ProblemPass
    ) {
        identifier = model.boxId
        value = model.value
        normalizedFrame = model.normalizedFrame(using: correction)
        originalSelected = model.isSelected
        selected = Observable(model.isSelected)

        self.problemPass = problemPass
    }

    func setSelected(_ selected: Bool, allowModificationToAll override: Bool = false) {
        guard editable || override else { return }
        guard self.selected.value != selected else { return }

        editabilityWasOverridden = editabilityWasOverridden || override

        problemPass.set(box: self, selected: selected, override: override)
    }

    /// this function is called by the problem pass to set the value
    func internalSetSelected(_ selected: Bool) {
        // check once again, so no unnecessary observer calls are made
        guard self.selected.value != selected else { return }

        self.selected.value = selected
    }

    func createSubmission() -> CorrectionSubmission.CorrectionBoxSubmission? {
        guard editable || editabilityWasOverridden, hasChanges else { return nil }

        return CorrectionSubmission.CorrectionBoxSubmission(
            boxId: identifier,
            status: selected.value ? 1 : 0)
    }

    func submissionDidFinish() {
        originalSelected = selected.value
    }

    func updateEmpty() {
        guard editable else { return }

        if empty {
            self.selected.value = false
        } else {
            self.selected.value = originalSelected
        }
    }

    func invalidate() {
        self.selected.value = false
    }

    func update(using model: CorrectionBoxModel) {
        originalSelected = model.isSelected
        self.selected.value = model.isSelected
    }
}
