// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import CoreGraphics
import Logging

/**
 **Internal Documentation**

 The representation of a problem (most of the times a subproblem) which can be assessed.

 # Structure

 A `Problem` has multiple `Pass`-es, which in turn have multiple `Box`-es.
 However, those classes are strongly coupled, as – for example – the selection of a box depends on
 the type of problem (`.credits` or `.multipleChoice`). This could be solved using delegates,
 but as we're not exposing this type directly, it can be done by using unowned references.

 */
class Problem: ProblemType {
    let identifier: ProblemIdentifier
    let page: Int

    var passes: [ProblemPassType] = []
    var normalizedFrame: CGRect = .zero

    var editable: Bool = true {
        didSet {
            for pass in passes {
                (pass as? ProblemPass)?.editable = editable
            }
        }
    }

    var empty: Bool = false {
        didSet { updateEmpty() }
    }

    let evaluationType: ProblemEvaluationType
    var editabilityWasOverridden: Bool = false

    // init from model
    init(problem: CorrectionProblemModel, in correction: CorrectionDataModel) {
        // set static members
        identifier = ProblemIdentifier(from: problem)
        page = problem.page
        evaluationType = problem.isCredit ? .credits : .multipleChoice

        // self is valid at this point, so create the individual problem passes
        let groupedBoxes = problem.groupBoxesByPass()
        var passes: [ProblemPass] = []
        var frame: CGRect?
        for pass in 1...correction.passes {
            guard let boxes = groupedBoxes[pass] else { continue }
            let credit = problem.creditForPass(pass)
            let problemPass = ProblemPass(pass: pass, credit: credit, boxes: boxes, in: correction, problem: self)

            // update frame
            if let currentFrame = frame {
                frame = currentFrame.union(problemPass.normalizedFrame)
            } else {
                frame = problemPass.normalizedFrame
            }

            passes.append(problemPass)
        }

        self.normalizedFrame = frame ?? .zero
        self.passes = passes
    }

    var hasChanges: Bool {
        guard editable || editabilityWasOverridden else { return false }
        return passes.contains(where: { ($0 as? ProblemPass)?.hasChanges ?? false })
    }

    func createSubmission() -> CorrectionSubmission.CorrectionProblemSubmission? {
        guard hasChanges else { return nil }

        var submission = CorrectionSubmission.CorrectionProblemSubmission(
            problem: identifier.problem,
            subproblem: identifier.subproblem ?? "",
            credits: [],
            comment: nil,
            crosses: []
        )

        submission = passes.reduce(into: submission) { (submission, pass) in
            (pass as? ProblemPass)?.addSubmissions(to: &submission)
        }

        // check once again
        guard !submission.crosses.isEmpty || !submission.credits.isEmpty else { return nil }

        return submission
    }

    func submissionDidFinish() {
        for pass in passes {
            if let pass = pass as? ProblemPass {
                pass.submissionDidFinish()
            }
        }
    }

    func updateEmpty() {
        for pass in passes {
            if let pass = pass as? ProblemPass {
                pass.empty = empty
            }
        }
    }

    func update(using model: CorrectionProblemModel) {
        let groupedBoxes = model.groupBoxesByPass()

        for pass in passes {
            if let pass = pass as? ProblemPass {
                let credit = model.creditForPass(pass.pass)
                guard let boxes = groupedBoxes[pass.pass] else {
                    Logger.correction.error("correction structure changed (no pass \(pass.pass)")
                    continue
                }
                pass.update(usingCredit: credit, boxes: boxes)
            }
        }
    }
}

extension Problem: LockStatePropagation {
    func lockStateChanged(_ lockState: LockState) {
        // ignore
    }

    func invalidate() {
        for pass in passes {
            if let pass = pass as? ProblemPass {
                pass.invalidate()
            }
        }
    }
}
