// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import CoreGraphics
import Logging

class ProblemPass {
    let pass: Int
    var originalCredit: Double?
    var credit: Double?

    var boxes: [CorrectionBoxType] = []

    var editable: Bool = true {
        didSet {
            for box in boxes {
                (box as? CorrectionBox)?.editable = editable
            }
        }
    }
    var empty: Bool = false {
        didSet { updateEmpty() }
    }

    var normalizedFrame: CGRect = .zero

    unowned var problem: Problem

    var hasChanges: Bool = false

    var editabilityWasOveridden: Bool = false

    init(
        pass: Int,
        credit: Double?,
        boxes: [CorrectionBoxModel],
        in correction: CorrectionDataModel,
        problem: Problem) {

        self.pass = pass
        self.originalCredit = credit
        self.credit = credit
        self.problem = problem

        let boxes = boxes.map { CorrectionBox(using: $0, in: correction, problemPass: self) }
        self.boxes = boxes

        calculateBoundingBox()
    }

    private func calculateBoundingBox() {
        // calculate 'bounding box' of all boxes
        var frame: CGRect = boxes.first?.normalizedFrame ?? .zero
        for box in boxes {
            frame = frame.union(box.normalizedFrame)
        }
        self.normalizedFrame = frame
    }

    func set(box: CorrectionBoxType, selected: Bool, override: Bool = false) {
        guard editable || override else { return }

        // update selection depending on type
        switch evaluationType {
        case .credits:
            // credits -> select/deselect the box and set the credits
            for someBox in boxes {
                (someBox as? CorrectionBox)?.internalSetSelected(someBox === box && selected)
            }
            credit = selected ? box.value : nil
        case .multipleChoice:
            // multiple choice -> just toggle the box
            box.selected.value = selected
        }

        // check if something did change
        updateHasChanges(override: override)
    }

    func updateHasChanges(override: Bool = false) {
        guard editable || override else { return }
        self.editabilityWasOveridden = editabilityWasOveridden || override
        self.problem.editabilityWasOverridden = self.problem.editabilityWasOverridden || self.editabilityWasOveridden

        if empty {
            hasChanges = true
            return
        }
        // check if credit type, and credits changed
        if .credits == evaluationType, credit != originalCredit {
            hasChanges = true
            return
        }

        // check if any of the boxes did change
        for box in boxes {
            if let box = box as? CorrectionBox, box.hasChanges {
                hasChanges = true
                return
            }
        }

        // nope. no changes
        hasChanges = false
    }

    func addSubmissions(to submission: inout CorrectionSubmission.CorrectionProblemSubmission) {
        guard editable || editabilityWasOveridden else { return }
        guard hasChanges else { return }

        if .credits == evaluationType, let credit = credit, credit != originalCredit {
            let creditSubmission = CorrectionSubmission.CorrectionCreditSubmission(credit: credit, correctionPass: pass)
            submission.credits.append(creditSubmission)
        }

        let boxSubmissions = boxes.compactMap { ($0 as? CorrectionBox)?.createSubmission() }
        submission.crosses.append(contentsOf: boxSubmissions)
    }

    func submissionDidFinish() {
        guard hasChanges else { return }
        for box in boxes {
            if let box = box as? CorrectionBox {
                box.submissionDidFinish()
            }
        }
        hasChanges = false
    }

    func updateEmpty() {
        guard editable else { return }

        hasChanges = true

        if empty {
            credit = 0.0
        } else {
            credit = originalCredit
        }
        for box in boxes {
            if let box = box as? CorrectionBox {
                box.empty = empty
            }
        }
    }

    func invalidate() {
        originalCredit = nil
        credit = nil

        for box in boxes {
            if let box = box as? CorrectionBox {
                box.invalidate()
            }
        }
    }

    func update(usingCredit credit: Double?, boxes: [CorrectionBoxModel]) {
        self.originalCredit = credit
        self.credit = credit

        for box in self.boxes {
            if let box = box as? CorrectionBox {
                if let model = boxes.first(box.identifier, for: \.boxId) {
                    box.update(using: model)
                } else {
                    Logger.correction.error("correction structure changed (no box with id \(box.identifier)")
                }
            }
        }
    }
}

extension ProblemPass: ProblemPassType {
    var evaluationType: ProblemEvaluationType {
        return problem.evaluationType
    }
}
