// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import CoreGraphics

/// the identifier of a problem (either a complete problem, or a subproblem)
public enum ProblemIdentifier: Equatable, RawRepresentable, Codable, Hashable {
    /// parse the identifier from a string formatted as `[Problem].[Subproblem]`, for example `1.a`
    /// - Parameter rawValue: the string
    public init?(rawValue: String) {
        let components = rawValue.components(separatedBy: ".")
        if components.count > 0, let first = Int(components[0]) {
            if components.count > 1 {
                let second = components[1]
                self = .subproblem(first, second)
            } else {
                self = .problem(first)
            }
        } else {
            return nil
        }
    }

    /// the raw value, for example `5.b`
    public var rawValue: String {
        if let sub = subproblem {
            return "\(problem).\(sub)"
        } else {
            return "\(problem)"
        }
    }

    public typealias RawValue = String

    /// a problem
    case problem(Int)

    /// a subproblem
    case subproblem(Int, String)

    /// the problem
    public var problem: Int {
        switch self {
        case let .problem(problem):
            return problem
        case let .subproblem(problem, _):
            return problem
        }
    }

    /// the subproblem if available
    public var subproblem: String? {
        switch self {
        case let .subproblem(_, sub):
            return sub
        default:
            return nil
        }
    }

    func matches(_ moreSpecific: ProblemIdentifier) -> Bool {
        switch self {
        case let .problem(prob):
            return moreSpecific.problem == prob
        case let .subproblem(prob, subprob):
            return moreSpecific.problem == prob && moreSpecific.subproblem == subprob
        }
    }
}

public protocol BoxType: AnyObject {
    /// the normalized frame of this box
    var normalizedFrame: CGRect { get }

    /// whether this box is selected
    var selected: Observable<Bool> { get }

    /// can be used to change the selection state.
    ///
    /// using this method will take the evaluation type into account
    /// - Parameter selected: whether to select this box
    func setSelected(_ selected: Bool, allowModificationToAll override: Bool)
}

/// the opaque type for an individual box of a problem / problem pass
public protocol CorrectionBoxType: BoxType {
    /// the value this box represents
    var value: Double { get }

    /// whether this box can be changed
    var editable: Bool { get }
}

public protocol EmptyBoxType: BoxType {
    var hasChanges: Bool { get }
}

/// the opaque type for a distinct pass of a problem
public protocol ProblemPassType: AnyObject {
    /// the pass
    var pass: Int { get }

    /// the assigned credits
    var credit: Double? { get }

    /// the evaluation type
    var evaluationType: ProblemEvaluationType { get }

    /// the individual boxes in this pass
    var boxes: [CorrectionBoxType] { get }

    /// the normalized frame of the boxes for this pass
    var normalizedFrame: CGRect { get }
}

/// the evaluation type of a problem
public enum ProblemEvaluationType {
    /// credits
    case credits
    /// multiple choice
    case multipleChoice
}

/// the opaque type for a problem
public protocol ProblemType: AnyObject {
    /// the identifier
    var identifier: ProblemIdentifier { get }

    /// the page the problem is on (might continue on the next pages)
    var page: Int { get }

    /// the evaluation type of the problem (credit or multiple choice)
    var evaluationType: ProblemEvaluationType { get }

    /// the distinct correction passes for this problem
    var passes: [ProblemPassType] { get }

    /// whether this problem can be changed
    var editable: Bool { get }

    /// the normalized frame of the boxes for this problem
    var normalizedFrame: CGRect { get }
}
