// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation

/**
 The snippet manager contains the associated snippets for its owning exam (see `ExamType`).

 All changes are automatically saved and restored.
 */
public protocol SnippetManagerType: AnyObject {
    /// all snippets for this exam
    var allSnippets: [SnippetType] { get }

    /// adds a snippet to the manager
    /// - Parameters:
    ///   - problem: the problem which is associated with this snippet
    ///   - name: a name for the snippet (can be changed later)
    ///   - image: the snippet's image
    @discardableResult
    func addSnippet(problem: ProblemIdentifier, name: String, image: Image) -> SnippetType

    /// returns the snippets for a specific problem
    /// - Parameter problem: the problem to filter by
    func snippets(forProblem problem: ProblemIdentifier) -> [SnippetType]


    /// returns the snippets excluding a specific problem
    ///
    /// - Note:
    ///     This only exists because we can't expose a `Set` of all snippets (because a set can't use a protocol as generic constraint).
    ///     So, this can be thought more of a performance-convenience-method
    /// - Parameter problem: the problem to exclude
    func snippets(excludingProblem problem: ProblemIdentifier) -> [SnippetType]

    /// removes a snippet from the manager
    /// - Parameter snippet: the snippet to remove
    func removeSnippet(_ snippet: SnippetType)

    /// removes all snippets
    func removeAll()
}
