// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

class Snippet: Hashable, Codable {
    let identifier: UUID
    let problem: ExamProblemIdentifier
    let created: Date

    var name: String {
        didSet {
            manager?.save()
        }
    }
    var image: Image?

    weak var manager: SnippetManager?

    init(problem: ExamProblemIdentifier) {
        identifier = UUID()
        self.problem = problem
        self.name = problem.problem.rawValue
        self.created = Date()
    }

    @discardableResult
    func loadImage(usingBaseURL url: URL) -> Bool {
        let imageURL = url.appendingPathComponent("\(identifier).png")
        guard let data = try? Data(contentsOf: imageURL) else { return false }
        self.image = Image(data: data)
        return self.image != nil
    }

    func saveImage(usingBaseURL url: URL) {
        guard let data = image?.pngData() else {
            Logger.snippet.error("[\(identifier.uuid)] could not create image data")
            return
        }

        let fileURL = url.appendingPathComponent("\(identifier).png")
        do {
            try data.write(to: fileURL, options: .atomicWrite)
        } catch let error {
            Logger.snippet.error("[\(identifier.uuidString)] save failed: \(error)")
        }
    }

    func deleteImage(usingBaseURL url: URL) {
        let imageURL = url.appendingPathComponent("\(identifier).png")
        do {
            try FileManager.default.removeItem(at: imageURL)
        } catch let error {
            Logger.snippet.error("[\(identifier.uuidString)] delete failed: \(error)")
        }
    }

    enum CodingKeys: String, CodingKey {
        case identifier
        case problem
        case name
        case created
    }

    func hash(into hasher: inout Hasher) {
        hasher.combine(identifier)
    }

    static func == (lhs: Snippet, rhs: Snippet) -> Bool {
        return lhs.identifier == rhs.identifier
    }
}

extension Snippet: SnippetType { }
