// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

extension SnippetManager {
    func save() {
        kit?.queue.async { [weak self] in
            self?.saveInBackground()
        }
    }

    func saveInBackground() {
        guard let databaseURL = self.databaseURL else { return }

        let encoder = JSONEncoder()
        do {
            let data = try encoder.encode(self.snippets)
            try data.write(to: databaseURL, options: [.atomicWrite])
            Logger.snippet.info("[SnippetManager|\(identifier)] saved \(snippets.count) snippets")
        } catch let error {
            Logger.snippet.info("[SnippetManager|\(identifier)] could not save snippets: \(error)")
        }
    }

    func load() {
        kit?.queue.async { [weak self] in
            self?.loadInBackground()
        }
    }

    func loadInBackground() {
        // check if already loaded
        guard snippets.isEmpty else { return }
        // ensure base url (creates directory)
        guard let baseURL = ensureBaseURL(), let databaseURL = self.databaseURL else { return }

        // try loading data
        do {
            try autoreleasepool {
                let data = try Data(contentsOf: databaseURL)
                let decoder = JSONDecoder()
                let loaded: Set<Snippet> = try decoder.decode(Set<Snippet>.self, from: data)
                for snippet in loaded {
                    snippet.loadImage(usingBaseURL: baseURL)
                    snippet.manager = self
                }
                Logger.snippet.info("[SnippetManager|\(identifier)] loaded \(loaded.count) snippets")
                DispatchQueue.main.async { [weak self] in
                    self?.snippets = loaded
                    self?.state.value = .loaded
                }
            }
        } catch let error {
            Logger.snippet.error("[SnippetManager|\(identifier)] could not load snippets: \(error)")
            DispatchQueue.main.async { [weak self] in
                self?.reset()
                self?.state.value = .loaded
            }
        }
    }

    func reset() {
        guard let databaseURL = self.databaseURL else { return }

        try? FileManager.default.removeItem(at: databaseURL)

        snippets = []
        save()
    }

    func ensureBaseURL() -> URL? {
        if let url = self.baseURL {
            return url
        }

        // swiftlint:disable identifier_name
        let fm: FileManager = FileManager.default
        guard var url = fm.documentsURL(forBackend: kit?.backendURL.host ?? "") else {
            return nil
        }

        url.appendPathComponent("\(identifier)", isDirectory: true)

        if !fm.fileExists(atPath: url.path) {
            try? fm.createDirectory(at: url, withIntermediateDirectories: true, attributes: nil)
        }

        self.baseURL = url
        return url
    }
}
