// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import os.log

extension FileManager {
    func documentsURL(forBackend backend: String) -> URL? {
        guard var baseURL = urls(for: .documentDirectory, in: .userDomainMask).first else { return nil }
        if let bundleIdentifier = Bundle.main.bundleIdentifier {
            baseURL = baseURL.appendingPathComponent(bundleIdentifier, isDirectory: true)
        }
        return baseURL.appendingPathComponent(backend, isDirectory: true)
    }
}

class SnippetManager: KitObject<ExamIdentifier> {
    var baseURL: URL?
    var databaseURL: URL? {
        ensureBaseURL()?.appendingPathComponent("snippets.json")
    }
    var snippets: Set<Snippet> = Set()
    var state: Observable<ResourceObjectState> = .init(.idle)

    required init(identifier: ExamIdentifier, kit: Kit) {
        super.init(identifier: identifier, kit: kit)

        load()
    }

    func add(_ snippet: Snippet) {
        guard let baseURL = ensureBaseURL() else { return }
        let (inserted, _) = snippets.insert(snippet)
        guard inserted else { return }

        snippet.manager = self
        snippet.saveImage(usingBaseURL: baseURL)
        save()
    }

    func remove(_ snippet: Snippet) {
        guard snippet.manager === self else { return }

        if snippets.remove(snippet) != nil {
            // remove manager
            snippet.manager = nil

            // delete image
            if let baseURL = ensureBaseURL() {
                snippet.deleteImage(usingBaseURL: baseURL)
            }

            // save database
            save()
        }
    }
}

extension SnippetManager: SnippetManagerType {
    var allSnippets: [SnippetType] {
        return Array(snippets)
    }

    func addSnippet(problem: ProblemIdentifier, name: String, image: Image) -> SnippetType {
        let snippedID = ExamProblemIdentifier(exam: identifier, problem: problem)
        let snippet = Snippet(problem: snippedID)
        snippet.image = image
        snippet.name = name

        add(snippet)

        return snippet
    }

    func removeSnippet(_ snippet: SnippetType) {
        guard let snippet = snippet as? Snippet else { return }
        remove(snippet)
    }

    func snippets(forProblem problem: ProblemIdentifier) -> [SnippetType] {
        snippets.filter { (snippet) -> Bool in
            snippet.problem.problem == problem
        }.sorted(by: \.created).reversed()
    }

    func snippets(excludingProblem problem: ProblemIdentifier) -> [SnippetType] {
        snippets.filter { (snippet) -> Bool in
            snippet.problem.problem != problem
        }.sorted(by: \.created).reversed()
    }

    func removeAll() {
        reset()
    }
}
