// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import CoreGraphics
import ImageIO
#if os(iOS)
import MobileCoreServices
#endif
import Logging

class SolutionLayer: KitObject<UnmanagedIdentifier<SolutionPageLayerIdentifier>> {
	let layerImage: SolutionLayerImage

	var underlyingScanSize: CGSize? {
		didSet {
			layerImage.underlyingScanSize = underlyingScanSize
			tryCreatingDrawing(blank: false)
		}
	}

	var drawing: Observable<Drawing?> = .init()
	var editable: Bool = false
	var layerStateObserver: AnyCancellable?

	var lastModified: Date {
        // For Debugging the conflict behavior, uncomment the following 3 lines
        // Without further modification, this will prevent saving after esolving the conflicts
//        #if DEBUG
//        return Date.distantPast
//        #endif
		return layerImage.lastModified
	}

	var hasChanges: Bool = false

	required init(identifier: UnmanagedIdentifier<SolutionPageLayerIdentifier>, kit: Kit) {
		layerImage = SolutionLayerImage(identifier: identifier.identifier, kit: kit)
		super.init(identifier: identifier, kit: kit)

        layerStateObserver = layerImage.state.sink { [weak self] state in
			if case let .error(error) = state,
			   let err = error as? NetworkError,
			   err == .notFound {
				// Layer returned 404 not found, we need to create the layer ourselves
				self?.tryCreatingDrawing(blank: true)
				return
			}
			self?.tryCreatingDrawing(blank: false)
		}
	}

	func createSubmission() -> CorrectionSubmission.CorrectionImageSubmission? {
		// ensure drawing and changes
		guard let drawing = drawing.value, drawing.elements.count > 0 else { return nil }
		// ensure image
		guard let image = drawing.getImage() else {
            Logger.submission.error("[SolutionLayer] could not create image for drawing \(identifier.stringIdentifier)")
			return nil
		}

		guard let data = CFDataCreateMutable(nil, 0) else { return nil }
		guard let destination = CGImageDestinationCreateWithData(data, kUTTypePNG, 1, nil) else { return nil }
		CGImageDestinationAddImage(destination, image, nil)
		CGImageDestinationFinalize(destination)

		let page = identifier.identifier.page
		let layer = identifier.identifier.layer

		let submission = CorrectionSubmission.CorrectionImageSubmission(page: page, layer: layer, data: data as Data)
		return submission
	}

	func submissionDidFinish() {
		hasChanges = false

		guard let drawing = drawing.value, drawing.elements.count > 0 else { return }
		// ensure image
        guard drawing.getImage() != nil else {
            Logger.submission.error("[SolutionLayer] could not create image for drawing \(identifier.stringIdentifier)")
			return
		}

        layerImage.reloadImage()
	}

	func tryCreatingDrawing(blank: Bool) {
		// don't create another one
		guard drawing.value == nil else { return }

		// check necessary components
		guard let scanSize = underlyingScanSize else {
			return
		}

		if !blank && layerImage.state.value != .loaded {
			return
		}

        let image: CGImage? = layerImage.image.value?.cgImage

		let drawing = Drawing(size: scanSize, originalImage: image)
		drawing.delegates.add(self)
		self.drawing.value = drawing
	}

	func propagateLockState(_ lockState: LockState) {
		editable = lockState.editable

	}

    func reloadLayer() {
        drawing.value = nil
        hasChanges = false
        layerImage.reloadImage()
    }
}

extension SolutionLayer: DrawingDelegate {
	func drawing(_ drawing: Drawing, didAppendElement element: Drawing.Element) {
		hasChanges = true
	}

	func drawing(_ drawing: Drawing, didRemoveLastElement element: Drawing.Element) {
		hasChanges = true
	}
}

extension SolutionLayer: LayerType {
	var pass: Int {
		return identifier.identifier.layer
	}

	var image: LoadableImageType {
		return layerImage
	}
}
