// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Combine
import Logging

#if os(iOS)
import UIKit
#endif

class SolutionPage: KitObject<SolutionPageIdentifier> {
	// var image: Observable<Image?> = Observable()
    var image: Observable<Image?> = Observable()
    var state: Observable<ResourceObjectState> = Observable(.idle)

	// MARK: - PageType Properties
	var layers: Observable<[LayerType]> = Observable()
	var problems: Observable<[ProblemType]> = Observable()

    var pageSize: CGSize = .zero

    var serverLastModified: Date = Date.distantPast {
        didSet {
            if serverLastModified > oldValue {
                // New version of layer from server
                // Update layer
                self.drawingLayer?.reloadLayer()
            }
        }
    }

	var layerStateObserver: AnyCancellable?
	var editable: Bool = false
	var empty: Bool = false
    var emptyBox: Observable<EmptyBoxType?> = Observable(nil)

	var drawingLayer: SolutionLayer? {
		layers.value.first as? SolutionLayer
	}

	var solutionLayerImage: SolutionLayerImage? {
		drawingLayer?.layerImage
	}

	var imageToMerge: CGImage?

	var isLayerImageLoaded: Bool = false

    var layerImage: Observable<Image?> = .init()

	/** If a submission was successful the page needs to be refreshed aka the layer needs to be reloaded. */
	var needsRefresh: Bool = false

	let solutionImage: SolutionImage

	private var layerSubscriber: AnyCancellable?

    lazy var resource: SolutionImageResource = createDelegatedResource()

	required init(identifier: SolutionPageIdentifier, kit: Kit) {
		solutionImage = SolutionImage(identifier: identifier, kit: kit)
		super.init(identifier: identifier, kit: kit)
		createLayers()
	}

	private func createLayers() {
		let id = SolutionPageLayerIdentifier(pageIdentifier: identifier,
											 layer: 1)
		let lay = SolutionLayer(identifier: id.unmanaged(owner: self), kit: kit!)
        layerImage = lay.layerImage.image
		self.layers.value = [lay]
        layerSubscriber = lay.layerImage.state.sink { [weak self] state in
			if state == .loaded { self?.isLayerImageLoaded = true }
		}
	}

	func loadLayerIfPresent() {
		guard let state = drawingLayer?.layerImage.state, state.value == .idle else { return }
		let solution = kit!.object(for: identifier.exam) as Solution
		if solution.model?.pagesWithLayers[identifier.page] != nil {
			drawingLayer?.layerImage.load()
		}
	}

	func updateImage(to image: Image?) {
        // Logger.solution.debug("[SolutionPage|\(self.identifier.stringIdentifier)] updated image to \(image.debugDescription)")
		guard let image = image else { return }
		let size = image.size

		for layer in layers.value {
			layer.underlyingScanSize = size
		}
	}

	func createSubmission() -> SolutionSubmissionModel? {
		guard let solutionLayer = (layers.value.last as? SolutionLayer) else { return nil }

		let submissionData: Data?

		/** This is exexuted on a background thread - kit.queue.async - so it should be ok.*/
		if let imageToMerge = imageToMerge,
		   let drawing = drawingLayer?.drawing.value,
		   let image = drawing.getImage() {

			let lhsImage = UIImage(cgImage: image)
			let rhsImage = UIImage(cgImage: imageToMerge)

			let rect = CGRect(x: 0, y: 0, width: lhsImage.size.width, height: lhsImage.size.height)
			let renderer = UIGraphicsImageRenderer(size: lhsImage.size)

			let result = renderer.pngData { _ in
				lhsImage.draw(in: rect, blendMode: .normal, alpha: 1)
				rhsImage.draw(in: rect, blendMode: .normal, alpha: 1)
			}

            submissionData = result
		} else {
			submissionData = solutionLayer.createSubmission()?.data
		}

		guard let data = submissionData else { return nil }

		let lastModDate = solutionLayer.lastModified
		let submission = SolutionSubmissionModel(page: pageNumber,
											image: data,
											date: lastModDate)
		return submission
	}

	func submissionDidFinish() {
		layers.value.forEach({ ($0 as? SolutionLayer)?.submissionDidFinish() })
		update(using: self.resource)
		imageToMerge = nil
	}

	func copy() -> SolutionPage {
		let newPage = SolutionPage(identifier: identifier, kit: kit!)
        newPage.update(using: self.resource)
        newPage.pageSize = self.pageSize
        return newPage
	}

    func triggerSolutionRefresh() {
        let solution = kit!.object(for: identifier.exam) as Solution
        solution.forceRefresh()
    }
}

extension SolutionPage: ResourceKitObjectType {
	func update(using resource: SolutionImageResource) {
        image.value = resource.createImage()
        self.pageSize = image.value?.size ?? .zero
        updateImage(to: image.value)
	}

	func createResource() -> SolutionImageResource {
		return SolutionImageResource(identifier, in: kit!)
	}
}

extension SolutionPage: LoadableImageType { }

extension SolutionPage: SolutionPageType {
	var rawScan: LoadableImageType {
		solutionImage
	}

	var pageIdentifier: CorrectionPageIdentifier {
		.init(exam: identifier.exam, erid: erid ?? "", page: pageNumber)
	}

	var pageNumber: Int { identifier.page }

	var erid: String? { identifier.exam.stringIdentifier }

	var containingCorrection: CorrectionType? { nil}

	var scan: LoadableImageType { solutionImage }

	var solution: LoadableImageType? {
		nil
	}

    var personalizedSolution: LoadableImageType? {
        return nil
    }

    var loadableLayerImage: LoadableImageType? {
        solutionLayerImage
    }

	var segmentation: PageSegmentationType? { nil }
	var focusedProblem: Int? { nil }
	var focusedPass: Int? { 1 }

	func prefetch() { }

	var hasChanges: Bool {
		return layers.value.contains(\.hasChanges)
	}

	/** Computes the CGImage of the current drawing on a background thread. */
	func getDrawingImage(onCompletion: @escaping ((CGImage?) -> Void)) {
		DispatchQueue.global(qos: .background).async { [weak self] in
			let image = self?.drawingLayer?.drawing.value?.getImage()
			onCompletion(image)
		}
	}

}

extension SolutionPage: LockStatePropagation {
		func lockStateChanged(_ lockState: LockState) { }

		func invalidate() {
            image.value = nil
            resource = createDelegatedResource()
            state.value = .idle
		}
}
