// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Foundation
import Logging

extension Solution: Submitable {

	var editable: Bool { true }

	var hasChanges: Bool {
		pages.value.contains(\.hasChanges)
	}

	func submitChanges() {
		guard hasChanges else {
            // Nothing to submit, but maybe we need to force refresh the solution?
            self.resetSubmission()
            self.forceRefresh()
            return
        }

		kit?.queue.async { [weak self] in
			self?.submitInBackground()
		}
	}

	private func submitInBackground() {
        Logger.submission.info("[Solution] attempting to submit changes")
		submissionState.value = .loading

		let pagesToSubmit = pages.value
			.compactMap { ($0 as? SolutionPage) }
			.filter { $0.hasChanges }

		let submission: [SolutionSubmissionModel] = pagesToSubmit
			.compactMap { page in
            return page.createSubmission()
		}

		guard !submission.isEmpty else {
            Logger.submission.error("[Solution] submission is empty")
			return
		}

		let submissionResource = SolutionSubmissionResource(
			identifier,
			submission: submission,
			in: kit!
        )

		self.submissionResource = submissionResource

		submissionResource.delegate = self
		submissionResource.load()
	}

	func submissionDidFinish() {
        Logger.submission.info("[Solution] submission succeeded")
		// back to no changes
		submissionState.value = .loaded
		DispatchQueue.main.async {
            self.submissionCompletion?(.success(()))
		}

		for page in pages.value {
			if let page = page as? SolutionPage {
				page.submissionDidFinish()
			}
		}

		assert(!hasChanges)

		submissionState.value = .idle

        self.forceRefresh()
	}

	func submissionDidFail(_ error: Error) {
        Logger.submission.info("[Solution] submission failed, \(error)")
		submissionState.value = .error(error)
	}
}
