// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

import Combine
import CoreGraphics
import Foundation
import Logging

final class Solution: KitObject<ExamIdentifier> {
    typealias ModelType = SolutionDescriptionModel
    typealias ResourceType = SolutionDescriptionResource

	// MARK: - Submitable

	var submissionState: Observable<ResourceObjectState> = Observable(.idle)

	var submissionCompletion: ((Result<Void, Error>) -> Void)?
	var submissionResource: SolutionSubmissionResource?

	// MARK: - Other properties

	var pages: Observable<[PageType]> = Observable()

	var model: SolutionDescriptionModel?
    lazy var resource: SolutionDescriptionResource = createDelegatedResource()
    var state: Observable<ResourceObjectState> = Observable(.idle)

	func resetSubmission() {
		submissionState.value = .idle
		pages.value.forEach { page in
			(page as? SolutionPage)?.layers.value
				.forEach { ($0 as? SolutionLayer)?.hasChanges = false }
		}
	}

	internal func prepare() {
        Logger.solution.debug("preparing solution for exam \(identifier.stringIdentifier)")
        self.load()
	}

    func forceRefresh() {
        guard self.state.value != .loading else {
            return
        }
        self.state.value = .idle
        self.resource = createDelegatedResource()
        self.load()
    }

	private func createAndLoadPagesIfNecessary(of description: SolutionDescriptionModel) {
		guard description.numberOfPages > 0 else { return }
        if pages.value.count != description.numberOfPages {
            pages.value = (1...description.numberOfPages).map { index in
                let ident = SolutionPageIdentifier(exam: identifier, page: index)
                return kit!.object(for: ident) as SolutionPage
            }
        }

        for (index, page) in pages.value.enumerated() {
            let page = (page as? SolutionPage)
            if let pageDesc = description.pagesWithLayers[index + 1] {
                page?.serverLastModified = pageDesc.date
            }
        }
	}
}

// Resource Delegate Implementation

extension Solution: ResourceDelegate {
    func resourceDidLoad(_ resource: BaseResource) {
        if resource === self.submissionResource {
            submissionDidFinish()
        } else if resource === self.resource {
            // Fugly but necessery, because we need to implement the resource delegate ourselves for the Submission
            guard resource === self.resource else { return }
            update(using: self.resource)
            state.value = .loaded
        }
    }

    func resource(_ resource: BaseResource, didFailWithError error: Error) {
        Logger.network.error("failed to load \(resource): \(error)")
        if resource === self.submissionResource {
            submissionDidFail(error)
        } else if resource === self.resource {
            state.value = .error(error)
        }
    }
}

extension Solution: ModelResourceKitObjectType {
    func createResource() -> ResourceType {
        return SolutionDescriptionResource(identifier: identifier, in: kit!)
    }

    func update(using model: ModelType) {
        self.createAndLoadPagesIfNecessary(of: model)
    }
}

extension Solution: SolutionType {

	func saveMergedPages(_ pages: [Rejection]) {
		let indexedPages: [Int: Rejection] = pages.reduce(into: [:]) { res, next in
			let (rejection, _) = next
			return res[rejection.page] = next
		}

		for (index, page) in self.pages.value.enumerated() {
			if let (rej, newPage) = indexedPages[page.pageNumber] {
				(newPage as? SolutionPage)?.solutionLayerImage?.lastModified = rej.date
                guard let newPage = newPage as? SolutionPage,
                      let oldPage = page as? SolutionPage else {
                    return
                }
                var pages = self.pages.value
                pages[index] = newPage
                kit?.replaceObject(with: oldPage.identifier, object: newPage)
				self.pages.value = pages
			}
		}

		submitChanges()
	}

	func save() {
        submitChanges()
	}

	func getCopyOfPages(_ pages: [Int]) -> [SolutionPageType] {
		let pages = self.pages.value
			.filter { pages.contains($0.pageNumber) }
			.compactMap { ($0 as? SolutionPage)?.copy() }
		return pages
	}
}
