// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#if canImport(UIKit) && !os(watchOS)
import UIKit

public extension Constrainable {
	@discardableResult
	func pinToEdges(of view: UIView? = nil,
									withInsets insets: UIEdgeInsets,
									safeConstrainable constrainable: Bool = false) -> Self {
		if #available(iOS 11.0, *), constrainable {
			return top(to: \.safeAreaLayoutGuide.topAnchor, of: view, constant: insets.top)
				.right(to: \.safeAreaLayoutGuide.rightAnchor, of: view, constant: insets.right)
				.bottom(to: \.safeAreaLayoutGuide.bottomAnchor, of: view, constant: insets.bottom)
				.left(to: \.safeAreaLayoutGuide.leftAnchor, of: view, constant: insets.left)
		}

		return top(to: \.topAnchor, of: view, constant: insets.top)
			.right(to: \.rightAnchor, of: view, constant: insets.right)
			.bottom(to: \.bottomAnchor, of: view, constant: insets.bottom)
			.left(to: \.leftAnchor, of: view, constant: insets.left)
	}

    @discardableResult
    func pinHorizontally(to view: UIView? = nil,
												 withInsets insets: UIEdgeInsets,
												 safeConstrainable constrainable: Bool = false) -> Self {
        if #available(iOS 11.0, *), constrainable {
            return right(to: \.safeAreaLayoutGuide.rightAnchor, of: view, constant: insets.right)
                .left(to: \.safeAreaLayoutGuide.leftAnchor, of: view, constant: insets.left)
        }
        return right(to: \.rightAnchor, of: view, constant: insets.right)
            .left(to: \.leftAnchor, of: view, constant: insets.left)
    }

    @discardableResult
    func pinVertically(to view: UIView? = nil,
											 withInsets insets: UIEdgeInsets,
											 safeConstrainable constrainable: Bool = false) -> Self {
        if #available(iOS 11.0, *), constrainable {
            return top(to: \.safeAreaLayoutGuide.topAnchor, of: view, constant: insets.top)
                .bottom(to: \.safeAreaLayoutGuide.bottomAnchor, of: view, constant: insets.bottom)
        }

        return top(to: \.topAnchor, of: view, constant: insets.top)
            .bottom(to: \.bottomAnchor, of: view, constant: insets.bottom)
    }

    @discardableResult
	func pinToEdges(of view: UIView? = nil,
									withInset inset: CGFloat = 0,
									safeConstrainable constrainable: Bool = false) -> Self {
		let insets = UIEdgeInsets(top: inset, left: inset, bottom: inset, right: inset)
		return pinToEdges(of: view, withInsets: insets, safeConstrainable: constrainable)
	}

	@discardableResult
	func center(in view: UIView? = nil, withMargin margin: CGFloat = 0, relation: ConstraintRelation = .equal) -> Self {
		return centerX(to: \.centerXAnchor, of: view, relation: relation, constant: margin)
			.centerY(to: \.centerYAnchor, of: view, relation: relation, constant: margin)
	}

	@discardableResult
	func size(_ size: CGSize, relation: ConstraintRelation = .equal) -> Self {
		return width(size.width, relation: relation)
			.height(size.height, relation: relation)
	}

	@discardableResult
	func aspectRatio(heightToWidth: CGFloat,
									 priority: UILayoutPriority = .required,
									 relation: ConstraintRelation = .equal) -> Self {
		addAspectRatio(heightToWidth: heightToWidth, priority: priority, relation: relation)
		return self
	}

	@discardableResult
	func addAspectRatio(heightToWidth: CGFloat,
											priority: UILayoutPriority = .required,
											relation: ConstraintRelation = .equal) -> NSLayoutConstraint {
		return target.heightAnchor
			.constraint(to: target.widthAnchor, multiplier: heightToWidth, relation: relation)
			.priority(priority)
			.activate()
	}

	@discardableResult
	func aspectRatio(widthToHeight: CGFloat,
									 priority: UILayoutPriority = .required,
									 relation: ConstraintRelation = .equal) -> Self {
		addAspectRatio(widthToHeight: widthToHeight, priority: priority, relation: relation)
		return self
	}

	@discardableResult
	func addAspectRatio(widthToHeight: CGFloat,
											priority: UILayoutPriority = .required,
											relation: ConstraintRelation = .equal) -> NSLayoutConstraint {
		return target.widthAnchor
			.constraint(to: target.heightAnchor, multiplier: widthToHeight, relation: relation)
			.priority(priority)
			.activate()
	}
}
#endif
