// Copyright 2018-2022 Chair of Network Architectures and Services
//                     Technical University of Munich
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#if canImport(UIKit) && !os(watchOS)
import UIKit

public extension UIViewController {
	/**
	adds a ChildViewController to the current ViewController.
	*/
	func add(_ child: UIViewController) {
		addChild(child)
		view.addSubview(child.view)
		child.didMove(toParent: self)
	}

	/**
	adds a child view controller to the current ViewController once with the legacy autoresizing mask.
	If there is already a child of type T this function does nothing
	*/
	func addOnceAutoresizing<T: UIViewController>(_ child: T, sizeToFitSuperview: Bool = true) {
		guard !children.contains(where: { type(of: $0) == type(of: child) }) else { return }

		add(child)

		if sizeToFitSuperview {
			child.view.frame = view.bounds
			child.view.autoresizingMask = [.flexibleWidth, .flexibleHeight]
		}
	}

	/// Adds a child view controller to the current ViewController once using constrains to fit the child to the parent if `sizeToFitSuperview` is `true`.
	/// If there is already a child of type T this function does nothing.
	/// - Parameters:
	///   - child: The view controller to add to the container
	///   - sizeToFitSuperview: Determines whether the childs edges are pinned to the container. Defaults to `true`.
	func addOnce<T: UIViewController>(_ child: T, sizeToFitSuperview: Bool = true) {
		guard !children.contains(where: { type(of: $0) == type(of: child) }) else {	return }

		add(child)

		if sizeToFitSuperview {
			child.view.usingAutoLayout().pinToEdges()
		}
	}

	/**
	removes all child view controller from a view controller.
	*/
	func clearChildren() {
		for child in children {
			child.remove()
		}
	}

	/**
	removes the current ChildViewController from the parent ViewController.
	*/
	func remove() {
		guard parent != nil else { return }
		willMove(toParent: nil)
		removeFromParent()
		view.removeFromSuperview()
	}

	var embeddedInNavigationController: UINavigationController {
		UINavigationController(rootViewController: self)
	}
}

#endif
